/*======================================================================
 * fdistill - distills floating point data to integers or byte data
 *
 * Notes:
 * As noted at the code walkthrough for this program 9/20/99, a future
 * modification that would be useful would be to specify the byte-order
 * of the output integers.
 *
 * 08/24/1999 brodzik brodzik@nsidc.org 303-492-8263
 * National Snow & Ice Data Center, University of Colorado, Boulder
 * Copyright (C) 1999-2004 University of Colorado
 *
 *======================================================================*/
#include "define.h"

static const char fdistill_c_rcsid[] = "$Header: /tmp_mnt/FILES/mjb/src/fdistill/fdistill.c,v 1.8 2004-02-08 16:42:17 brodzik Exp $";

#define ROUND(x) ((x) < 0 ? (int)((x)-.5) : (int)((x)+.5))

#define usage \
"fdistill $Revision: 1.8 $ \n"\
"usage: fdistill [-blnsuv -m factor] [infile]\n"\
"\n"\
" input : infile - input file containing floating point values\n"\
"                  (default is read from stdin)\n"\
"\n"\
" output : distilled output data is written to stdout\n"\
"\n"\
" option : b - byte data (default)\n"\
"          l - long integers (4 bytes)\n"\
"          n - no warnings about data overflow (default is warnings on)\n"\
"          s - short integers (2 bytes)\n"\
"          u - unsigned data (default is signed) \n"\
"          m factor - multiply by scale factor\n"\
"          v - verbose mode (can be repeated)\n"\
"          V - displays RCS id\n"\
"\n"

main (int argc, char *argv[]) {
  char *option;
  int verbose=0;
  float scale_factor=1.0;
  char in_filename[FILENAME_MAX];
  int data_bytes=1;
  FILE *infile;
  FILE *outfile=stdout;
  bool signed_data=TRUE;
  bool overflow_warning=TRUE;
  size_t status;
  float value, max_value, min_value;
  int1 int1_value;
  int2 int2_value;
  int4 int4_value;
  byte1 byte1_value;
  byte2 byte2_value;
  byte4 byte4_value;
  
  /* 
   * Get command line arguments
   */
  while (--argc > 0 && (*++argv)[0] == '-')
  { for (option = argv[0]+1; *option != '\0'; option++)
    { switch (*option)
      { case 'b':
	  data_bytes = 1;
	  break;
	case 'l':
	  data_bytes = 4;
	  break;
	case 'n':
	  overflow_warning = FALSE;
	  break;
	case 's':
	  data_bytes = 2;
	  break;
	case 'u':
	  signed_data = FALSE;
	  break;
	case 'v':
	  verbose++;
	  break;
	case 'V':
	  fprintf(stderr,"%s\n", fdistill_c_rcsid);
	  break;
	case 'm':
	  ++argv; --argc;
	  if (sscanf(*argv,"%f", &scale_factor) != 1) {
	    fprintf(stderr,"fdistill: can't read scale factor %s\n", *argv);
	    error_exit(usage);
	  }
	  break;
	default:
	  fprintf(stderr,"fdistill: Invalid option %c\n", *option);
	  error_exit(usage);
      } /* end switch */
    } /* end loop for each input command option */
  } /* end loop while still input arguments */

  /*
   * Read input from stdin if there are no more arguments
   */
  if (0 == argc) {
    infile = stdin;
    if (verbose >= 2) fprintf(stderr,">> Reading from stdin...\n");
  } else if (1 == argc) {
    if (1 != sscanf(*argv,"%s",in_filename)) error_exit(usage);
    infile = fopen(in_filename,"rb");
    if (NULL == infile) {
      perror(in_filename);
      exit(EXIT_FAILURE);
    }
    if (verbose >= 2) fprintf(stderr,">> Opened %s...\n",in_filename);
    ++argv; --argc;
  } else {
    fprintf(stderr,"fdistill: wrong number of input arguments\n");
    error_exit(usage);
  }

  if (verbose >= 1) {
    fprintf(stderr,"> fdistill: Beginning processing for:\n");
    fprintf(stderr,">  element size (bytes)=%d\n",data_bytes);
    fprintf(stderr,">  signed data         =%d\n",signed_data);
    fprintf(stderr,">  overflow warnings   =%d\n",overflow_warning);
    fprintf(stderr,">  scale factor        =%f\n",scale_factor);
  }

  /* Set the extrema for overflow warnings */
  if (overflow_warning) {
    switch (data_bytes * (signed_data ? -1 : 1)) { 
    case -1: 
      min_value = INT1_MIN;
      max_value = INT1_MAX;
      break;
    case -2: 
      min_value = INT2_MIN;
      max_value = INT2_MAX;
      break;
    case -4: 
      min_value = INT4_MIN;
      max_value = INT4_MAX;
      break;
    case  1: 
      min_value = BYTE1_MIN;
      max_value = BYTE1_MAX;
      break;
    case  2: 
      min_value = BYTE2_MIN;
      max_value = BYTE2_MAX;
      break;
    case  4: 
      min_value = BYTE4_MIN;
      max_value = BYTE4_MAX;
      break;
    default: assert(NEVER); /* should never execute */
    }
  }

  while (!feof(infile)) {

    status = fread(&value, sizeof(float), 1, infile);
    if (feof(infile)) break;
    if (1 != status) {
      fprintf(stderr,"fdistill: Error reading float data.\n");
      perror(in_filename);
      exit(EXIT_FAILURE);
    }
    
    value = ROUND(value*scale_factor);

    if (overflow_warning && (min_value > value || max_value < value)) {
      fprintf(stderr,"fdistill: warning, data overflow for value=%f\n",
	      value);
    }

    switch (data_bytes * (signed_data ? -1 : 1)) { 
    case -1: 
      int1_value = value; 
      status = fwrite(&int1_value,sizeof(int1),1,outfile);
      if (1 != status) {
	error_exit("fdistill: Error writing 1 byte signed int value.\n");
      }
      break;
    case -2: 
      int2_value = value; 
      status = fwrite(&int2_value,sizeof(int2),1,outfile);
      if (1 != status) {
	error_exit("fdistill: Error writing 2 byte signed int value.\n");
      }
      break;
    case -4: 
      int4_value = value; 
      status = fwrite(&int4_value,sizeof(int4),1,outfile);
      if (1 != status) {
	error_exit("fdistill: Error writing 4 byte signed int value.\n");
      }
      break;
    case  1: 
      byte1_value = value; 
      status = fwrite(&byte1_value,sizeof(byte1),1,outfile);
      if (1 != status) {
	error_exit("fdistill: Error writing 1 byte unsigned int value.\n");
      }
      break;
    case  2: 
      byte2_value = value; 
      status = fwrite(&byte2_value,sizeof(byte2),1,outfile);
      if (1 != status) {
	error_exit("fdistill: Error writing 2 byte unsigned int value.\n");
      }
      break;
    case  4: 
      byte4_value = value; 
      status = fwrite(&byte4_value,sizeof(byte4),1,outfile);
      if (1 != status) {
	error_exit("fdistill: Error writing 4 byte unsigned int value.\n");
      }
      break;
    default: assert(NEVER); /* should never execute */
    }

  } /* end while not EOF */

  if (verbose >= 2) fprintf(stderr,"> fdistill: exiting successfully\n");
  exit(EXIT_SUCCESS);

} 

