/*======================================================================
 * bootstrap - storage and retrieval for ancillary information needed
 *               to run Bootstrap sea ice algorithm
 *
 * Notes: This object represents a break from the standard pmalgorithm
 *        paradigm in two ways, 1) the Bootstrap algorithm requires
 *        ancillary information, including date and landmask, that other 
 *        pmalgorithms don't require, and 2) the algorithm itself is based
 *        on a regression of a scatter plot of tbs, which requires that
 *        the entire grid be processed, rather than running an algorithm
 *        on a single pixel at a time.
 *        Furthermore, certain assumptions about the landmask are made,
 *        It is assumed that the 
 *
 * 02/11/2000 brodzik brodzik@zamboni.colorado.edu 303-492-8263
 * National Snow & Ice Data Center, Boulder, CO
 * Copyright (C) 2000-2006 University of Colorado at Boulder
 *
 *$Id: bootstrap.c 22374 2013-01-25 17:38:07Z kbeam $
 *
 *======================================================================*/
#include "define.h"
#include "matrix.h"
#include "matrix_io.h"
#include "bootstrap.h"
#include "pmalgorithm.h"
#include "land_spillover.h"
#include "fortran.h"
#include <dirent.h>


static const char bootstrap_c_rcsid[] = "$Header: /tmp_mnt/FILES/pmalgos/bootstrap.c,v 1.3.8.1 2008-06-10 20:36:35 brodzik Exp $";


/* prototypes for FORTRAN subroutine */
extern void FORTRAN_NAME(seaice_bootstrap) (char *overlap, char *date,
                                            byte2 *inors, int *nrows,
                                            int *ncols, byte2 *t19v,
                                            byte2 *t22v, byte2 *t37h,
                                            byte2 *t37v, byte1 *landmask,
                                            byte2 *seaice, int *verbose);

extern void FORTRAN_NAME(seaice_bootstrap_v02) (int *nrows, int *ncols,
						byte2 *t37v, byte2 *t19v,
						byte2 *t37h, byte2* t22v,
						byte2 *seaice, float *landval,
						float *missval, byte1 *land, 
						int2 *imonth, int2 *iday,
						int2 *hem, char *satellite,
                                                int *verbose);


/*
 * is_bootstrap_algorithm - Return true if the algorithm id is one of
 *                          the valid bootstrap algorithms.
 *
 *  Input: algorithm - the algorithm id to test.
 *
 *  Returns: True if the algorithm id is a bootstrap algorithm.
 */
bool is_bootstrap_algorithm(pmalgo_id algorithm)
{
  bool ret_value = FALSE;

  if (algorithm == PMALGO_BOOTN_SEAICE ||
      algorithm == PMALGO_BOOTS_SEAICE ||
      is_bootstrap_v02_algorithm(algorithm)){
    ret_value = TRUE;
  }

  return ret_value;
}

/*
 * is_bootstrap_v02_algorithm - Return true if the algorithm id is one of
 *                              the valid version 2 bootstrap algorithms.
 *
 *  Input: algorithm - the algorithm id to test.
 *
 *  Returns: True if the algorithm id is a version 2 bootstrap algorithm.
 */
bool is_bootstrap_v02_algorithm(pmalgo_id algorithm)
{
  bool ret_value = FALSE;

  if (algorithm == PMALGO_BOOTN_V02_SEAICE ||
      algorithm == PMALGO_BOOTS_V02_SEAICE ||
      algorithm == PMALGO_BOOTN_V02_PUB_SEAICE ||
      algorithm == PMALGO_BOOTS_V02_PUB_SEAICE) {
    ret_value = TRUE;
  }

  return ret_value;
}

/*
 * grid_is_polarstero - Return true of the grid object is pointing to a polar
 *                      stereo (N3B or S3B) grid.  The Bootstrap V2 algorithms
 *                      require one of these two grids.  
 *
 *  Input:  grid - grid_class pointer to test.
 *
 *  Returns:  True if this grid points to polarstereo grid.
 */
bool grid_is_polarstero(grid_class* grid)
{
    bool ret_value=FALSE;

    if (strstr(grid->gpd_filename, "N3B.gpd")
        || strstr(grid->gpd_filename, "S3B.gpd")){
        ret_value=TRUE;
    }
    
    return ret_value;
}

    
/*
 * valid_v2_environment - assert that the BOOTSTRAP_V2_ANCILLARY environmental
 *                        variable is set.  If the variable is NULL or
 *                        invalid, the function sets the environmental
 *                        variable to the default value and tests again
 *                        recursively.
 *
 *
 * Returns: TRUE if the a non-null BOOTSTRAP_V2_ANCILLARY environmental
 *          variable exists and points to a valid directory.
 */
bool valid_v2_environment(void)
{
    const char* env_var_name = "BOOTSTRAP_V2_ANCILLARY";
    const char* default_dir = "/share/data/pmalgos/bootstrap_seaice";
    const char* default_file = "sst_n_01.hdf";
    const int overwrite = 1;
    char* user_boot_env = getenv(env_var_name);
    DIR *test;
    bool ret_value = TRUE;

    if (NULL == user_boot_env) {
        /* User didn't set the variable, try the default location */
        setenv(env_var_name, default_dir, overwrite);
        fprintf(stderr, "environmental variable: BOOTSTRAP_V2_ANCILLARY is NULL\n");
        fprintf(stderr, "Trying to set to default: %s\n", default_dir);
        return valid_v2_environment();
    }
    
    test = opendir( user_boot_env );
    if (NULL == test) {
        /* Invalid directory try setting to default */
        setenv(env_var_name, default_dir, overwrite);
        fprintf(stderr, "Invalid environmental variable: BOOTSTRAP_V2_ANCILLARY\n");
        fprintf(stderr, "Trying the default value %s\n", default_dir);
        return valid_v2_environment();
    } else {
        closedir(test);
        /* check this directory for the sst_n_01.hdf file */
        char test_file[FILENAME_MAX];
        FILE* fp;
        sprintf(test_file, "%s//%s",user_boot_env, default_file);
        fp = fopen(test_file, "r");
        if (NULL == fp) {
            ret_value = FALSE;
        } else {
            fclose(fp);
        }
    }


        
    return ret_value;
}


/*----------------------------------------------------------------------
 *  init_bootstrap - initialize Bootstrap algorithm ancillary information
 *
 *    input: algorithm_name - Bootstrap N or S algorithm name
 *           sat_name - satellite name (must match one of
 *                      bootstrap_sat_name)
 *           date - yyyymmdd date of input tbs
 *           bootstrap_recoding_pcent - minimum seaice threshold for bootstrap recode thresholding
 *           scale_percent_seaice - flag to turn on scaling to 0-100%
 *           land_spillover - pre-initialized land spillover class
 *           grid - grid_class to use to read landmask and weights files
 *           verbose - verbosity flag
 *
 *    returns: pointer to new bootstrap_class instance
 *             or NULL (and reason to stderr) in case of error 
 *
 *----------------------------------------------------------------------*/
bootstrap_class *init_bootstrap(const char *algorithm_name,
				const char *sat_name,
				const char *date,
				const int bootstrap_recoding_pcent,
				const bool scale_percent_seaice,
				land_spillover_class *land_spillover,
				grid_class *grid,
				int verbose) {

  int i, row, col;
  bootstrap_class *this;
  bool status;

  if (0 < verbose) {
    fprintf(stderr,"> %s\n",bootstrap_h_rcsid);
    fprintf(stderr,"> %s\n",bootstrap_c_rcsid);
  }

  /* Allocate storage for object, and initialize any non-zero defaults */
  this = (bootstrap_class *)calloc(1,sizeof(bootstrap_class));
  if (NULL == this) {
    perror("init_bootstrap: bootstrap_class object");
    return NULL;
  }

  if (strstr(algorithm_name, "_N_")) {
      this->hem = BOOTSTRAP_NORTH_HEM;
  } else if (strstr(algorithm_name, "_S_")) {
      this->hem = BOOTSTRAP_SOUTH_HEM;
  } else {
      assert(0);
  }

  /* Do version specific things  */
  if (strstr(algorithm_name, "V02"))
  {
      strcpy(this->algo_version, BOOTSTRAP_VERSION_02);
      this->scale_percent_seaice = scale_percent_seaice;
  }
  else
  {
      strcpy(this->algo_version, BOOTSTRAP_VERSION_01);
      this->scale_percent_seaice = FALSE;
  }

  // Find the bootstrap satellite name, defaulting it if not found
  sprintf(this->sat_name, "%-9s", bootstrap_sat_name[BOOTSTRAP_DEFAULT]);
  for (i = 0; i < BOOTSTRAP_NUM_SATS; i++) {
      if (0 == strcmp(bootstrap_sat_name[i],sat_name)) {
          sprintf(this->sat_name, "%-9s", sat_name);
          break;
      }
  }
    
  this->bootstrap_recoding_pcent = (0 <= bootstrap_recoding_pcent ?
                                    bootstrap_recoding_pcent :
                                    PMALGO_BOOTSTRAP_MIN_ICE_PERCENT);


  
  this->land_spillover = land_spillover;

  strncpy(this->date,date,BOOTSTRAP_DATE_LENGTH);
  this->grid = grid;

  /* 
   * Allocate an array for the landmask, and initialize all pixels to
   * PMALGO_BOOTSTRAP_LAND.  The pixels where the algorithm will be run
   * should be chosen by the user, with set_mask_bootstrap (below).  
   */
  this->landmask = (byte1 **)matrix(this->grid->rows,this->grid->cols,
                                    sizeof(**this->landmask),FALSE);
  if (NULL == this->landmask) {
    perror("init_bootstrap: error allocating matrix for landmask");
    close_bootstrap(this);
    return NULL;
  }
  
  for (row = 0; row < this->grid->rows; row++) {
      for (col = 0; col < this->grid->cols; col++) {
          this->landmask[row][col] = PMALGO_BOOTSTRAP_LAND;
      }
  }
    
  this->seaice = NULL;
  this->verbose = verbose;

  return this;

}


/*----------------------------------------------------------------------
 *  set_mask_bootstrap - use the input loci to determine mask pixels where
 *                       Bootstrap sea ice algorithm should be run
 * 
 *    input: this - bootstrap_class object
 *           loci - loci array
 *           use_value - value in loci array that mask should be set for
 *               algorithm to be run
 *
 *----------------------------------------------------------------------*/
void set_mask_bootstrap(bootstrap_class *this, 
			byte1 **loci, byte1 use_value) {

  int num, row, col;

  assert(NULL != this);
  assert(NULL != this->grid);

  num = 0;

  for (row = 0; row < this->grid->rows; row++) {
    for (col = 0; col < this->grid->cols; col++) {
        if (use_value == loci[row][col]) { 
            this->landmask[row][col] = PMALGO_BOOTSTRAP_OCEAN;
            num++;
        }
    }
  }

  if (this->verbose) {
      fprintf(stderr,"> set_mask_bootstrap: setting %d pixels with value %d",
              num,use_value);
      fprintf(stderr,"in loci to ocean in bootstrap mask\n");
  }

  return;

}

/*----------------------------------------------------------------------
 *  get_seaice_bootstrap - return the Bootstrap sea ice concentration
 *                         for the input pixel location
 * 
 *    input: this - bootstrap_class object
 *           row - row number
 *           col - column number
 *           seaice - Bootstrap sea ice concentration (%) at the input pixel 
 *
 *    result: TRUE if successful, or FALSE otherwise, with error message
 *           to stderr
 *
 *    notes:  The first time this routine is called (for any pixel location), 
 *    the FORTRAN program is called to calculate Bootstrap sea ice 
 *    concentrations for the full grid, which are then stored in array
 *    this->seaice.  On subsequent calls, the %sea ice value is returned.
 *             
 *----------------------------------------------------------------------*/
bool get_seaice_bootstrap(bootstrap_class *this, 
			  byte2 **tbs[PMALGO_NUM_CHANNELS],
			  int row, int col, float *seaice) {

  byte2 hem;
  static float landval = (float)PMALGO_BOOTSTRAP_LAND;
  static float missval = (float)PMALGO_BOOTSTRAP_MISSING_DATA;
  int2 year;
  int2 month;
  int2 day;
  float scale_factor;     

  assert(0 <= row && row < this->grid->rows);
  assert(0 <= col && col < this->grid->cols);
  assert(NULL != this);

  if (NULL == this->seaice) {

      this->seaice = (byte2 **)initialize_matrix(this->grid,
                                                 sizeof(**this->seaice),
                                                 NULL,"seaice data",
                                                 (bool)this->verbose);
      if (NULL == this->seaice) {
          fprintf(stderr,"get_seaice_bootstrap: initialize_matrix failed.\n");
          return FALSE;
      }


      hem = (byte2)this->hem;
      if (bootstrap_is_v1(this))
      {
          if (BOOTSTRAP_NORTH_HEM == (bootstrap_hemisphere)hem) {
              FORTRAN_NAME(seaice_bootstrap) (&this->sat_name[0],
                                              &this->date[0], &hem,
                                              &this->grid->rows,
                                              &this->grid->cols,
                                              &tbs[PMALGO_SSMI_19V][0][0],
                                              &tbs[PMALGO_SSMI_22V][0][0],
                                              &tbs[PMALGO_SSMI_37H][0][0],
                                              &tbs[PMALGO_SSMI_37V][0][0],
                                              &this->landmask[0][0],
                                              &this->seaice[0][0],
                                              &this->verbose);
          } else {
              FORTRAN_NAME(seaice_bootstrap) (&this->sat_name[0],
                                              &this->date[0], &hem,
                                              &this->grid->rows,&this->grid->cols,
                                              &tbs[PMALGO_SSMI_19V][0][0],
                                              &tbs[PMALGO_SSMI_22V][0][0],
                                              NULL,
                                              &tbs[PMALGO_SSMI_37V][0][0],
                                              &this->landmask[0][0],
                                              &this->seaice[0][0],
                                              &this->verbose);
          }
      }
      else if (bootstrap_is_v2(this))
      {
          sscanf(this->date, "%4hd%2hd%2hd", &year, &month, &day);
          FORTRAN_NAME(seaice_bootstrap_v02) (&this->grid->rows,&this->grid->cols,
                                              &tbs[PMALGO_SSMI_37V][0][0],
                                              &tbs[PMALGO_SSMI_19V][0][0], 
                                              &tbs[PMALGO_SSMI_37H][0][0],
                                              &tbs[PMALGO_SSMI_22V][0][0], 
                                              &this->seaice[0][0],
                                              &landval, &missval,
                                              &this->landmask[0][0], &month, &day,
                                              &hem,
                                              &this->sat_name[0],
                                              &this->verbose);

      }
  }

  /* For original bootstrap processing always scale by 10. */
  scale_factor = (float)PMALGO_BOOTSTRAP_V1_SCALE_FACTOR;

  /* For boot v2 you have scaled and unscaled. */
  if (bootstrap_is_v2(this)) {
      scale_factor =  (float)PMALGO_BOOTSTRAP_V2_NO_SCALING;
      if (this->scale_percent_seaice) {
          scale_factor = PMALGO_BOOTSTRAP_V2_SCALE_FACTOR;
      }
  }
  

  
  *seaice = (float)this->seaice[row][col]/scale_factor;

  return TRUE;

}


/* ----------------------------------------
 *
 * bootstrap_is_v2 - return true if we are doing Bootstrap Version 2
 *
 *   Input:
 *     this - Bootstrap_class object
 *
 *   Returns:
 *     True if we are V2 bootstrap
 *
 * ---------------------------------------- */
bool bootstrap_is_v2(bootstrap_class *this)
{
    bool is_v2 = FALSE;
    assert(NULL != this);

    if (streq(BOOTSTRAP_VERSION_02, this->algo_version)){
        is_v2 = TRUE;
    }
    return is_v2;
}


/* ----------------------------------------
 *
 * bootstrap_is_v1 - return true if we are doing Original Bootstrap Version 1
 *                   processing
 *
 *   Input:
 *     this - Bootstrap_class object
 *
 *   Returns:
 *     True if we are V1 bootstrap
 *
 * ---------------------------------------- */
bool bootstrap_is_v1(bootstrap_class *this)
{
    return ( ! bootstrap_is_v2(this));
}





/*----------------------------------------------------------------------
 *  bootstrap_do_spillover - return the land spillover object
 *
 *    input: this - bootstrap object
 *
 *    result: land spillover object
 *
 *----------------------------------------------------------------------
*/
land_spillover_class *bootstrap_spillover(bootstrap_class *this) {

  assert(NULL != this);

  return this->land_spillover;

}




/*-----------------------------------------------------------------------------

 * bootstrap_recode_seaice - Recode the output data for BOOTSTRAP data
 *    products.  If scaling is desired, scale the output to 0-100%.
 *
 * input: this - bootstrap_class object
 *        grid - grid object
 *        data - seaice data
 *        verbose - verbosity level
 *
 * result: scales the sea ice, including land and missing codes, by 
 *         the value of PMALGO_BOOTSTRAP_SCALE
 *--------------------------------------------------------------------------*/
bool bootstrap_recode_seaice( bootstrap_class *this,
                              grid_class *grid,
                              float **data,
                              int verbose ) {

    float ice_scale_factor;
    int row, col;
    bool scaling_ice=FALSE;

    /* This isn't needed at the moment, but we will need it in the near
     * future. */
    return TRUE;
}



/*----------------------------------------------------------------------
 *  close_bootstrap - free storage associated with this object
 * 
 *    input: this - bootstrap_class object
 *
 *    result: deallocate all memory associated with this object
 *
 *----------------------------------------------------------------------*/
void close_bootstrap(bootstrap_class *this) {

  if (NULL == this) return;
  if (NULL != this->landmask) free(this->landmask);
  if (NULL != this->seaice) free(this->seaice);
  free(this);
}

