/*======================================================================
 * goddard - storage and retrieval for ancillary information needed
 *           to run Goddard time series sea ice algorithm
 *
 * Notes: This object represents a break from the standard pmalgorithm
 *        paradigm in that the algorithm requires ancillary information, 
 *        including hemisphere and satellite platform that other 
 *        pmalgorithms don't require.
 *
 * 05/24/2006 brodzik brodzik@nsidc.org 303-492-8263
 * National Snow & Ice Data Center, Boulder, CO
 * Copyright (C) 2006 University of Colorado at Boulder
 *
 *$Id: goddard.c 22374 2013-01-25 17:38:07Z kbeam $
 *
 *======================================================================*/
#include "define.h"
#include "goddard.h"
#include "matrix.h"
#include "matrix_io.h"
#include "pmalgorithm.h"
#include "land_spillover.h"

static const char goddard_c_rcsid[] = "$Header: /tmp_mnt/FILES/pmalgos/goddard.c,v 1.2.6.1 2008-06-10 20:36:35 brodzik Exp $";

/*----------------------------------------------------------------------
 * is_goddard_algorithm - Returns true if the algorithm is any one of
 *                        the valid Goddard algorithm ids.
 *
 *   Input: algorithm - the algorithm id to test
 *
 *   Returns: True if the algorithm is a valid Goddard id, else False
 *----------------------------------------------------------------------*/
bool is_goddard_algorithm(pmalgo_id algorithm) {
  bool ret_value = FALSE;

  if (algorithm == PMALGO_GTS_SEAICE ||
      algorithm == PMALGO_GTS_PUB_SEAICE ||
      algorithm == PMALGO_GTS_MYICE) {
    ret_value = TRUE;
  }

  return ret_value;
}

/*----------------------------------------------------------------------
 *  init_goddard - validates and initializes Goddard time series 
 *                 algorithm ancillary information
 *
 *    input: spillover_open_water_pcent - spillover open water definition
 *               any values lower than this are considered open water for spillover
             goddard_recoding_pcent - goddard recoding cutoff percent seaice (any seaice
 *               values lower than this are reset to 0%)
 *           hemisphere_name - {"N" , "S"}
 *           platform id - SSM/I platform id
 *           scale_percent_seaice - flag to scale output to 0-100% range
 *                                  default range is 0-250
 *           spillover - spillover object or NULL is spillover correction
 *                       won't be performed.
 *           grid - grid information to read landmask and weights files
 *           landmask_filename - filename with land/ocean mask (0=ocean, 1=land)
 *           weights_filename - filename with weights mask, typically some annual
 *                             minimum seaice concentration
 *           verbose - verbosity flag
 *
 *    returns: pointer to new goddard_class instance
 *             or NULL (and reason to stderr) in case of error 
 *
 * For spillover correction reference, see:
 *
 * Cavalieri, D. J., C. L. Parkinson, P. Gloersen, J. C. Comiso,
 * H. J. Zwally.  1999.  Deriving long-term time series of sea ice cover
 * from satellite passive-microwave multisensor data sets.  Journal of
 * Geophysical Research, 104(C7), 15,803-15,814.
 *----------------------------------------------------------------------*/
goddard_class *init_goddard(const int goddard_recoding_pcent,
                            const char *hemisphere_name,
			    pmalgo_ssmi_platform platform,
                            const bool scale_percent_seaice,
                            land_spillover_class *land_spillover,
                            grid_class *grid,
			    int verbose) {

    int i;
    goddard_class *this;
    bool status;

    if (0 < verbose) {
	fprintf(stderr,"> %s\n",goddard_h_rcsid);
	fprintf(stderr,"> %s\n",goddard_c_rcsid);
    }

    /* Allocate storage for object, and initialize any non-zero defaults */
    this = (goddard_class *)calloc(1,sizeof(goddard_class));
    if (NULL == this) {
	perror("init_goddard: goddard_class object");
	return NULL;
    }

    if (0 == strcmp("N", hemisphere_name)) {
	this->hem = PMALGO_NORTH;
    } else if (0 == strcmp("S", hemisphere_name)) {
	this->hem = PMALGO_SOUTH;
    } else {
	perror("init_goddard: invalid hemisphere, should be N or S");
	close_goddard(this);
	return NULL;
    }
    
    this->platform = platform;
    if (PMALGO_NO_PLATFORM == this->platform) {
	perror("init_goddard: invalid platform");
	close_goddard(this);
	return NULL;
    }

    this->goddard_recoding_pcent = ( 0 <= goddard_recoding_pcent ?
                                             goddard_recoding_pcent :
                                             PMALGO_GTS_MIN_ICE_PERCENT );
        
    this->scale_percent_seaice = scale_percent_seaice;

    this->land_spillover = land_spillover;

    this->verbose = verbose;

    /* F15 is just a placeholder for now */
    if ( PMALGO_F15 == this->platform ) {
	fprintf( stderr,"init_goddard: platform F15 is not yet implemented.  Exiting.\n" );
	close_goddard(this);
	return NULL;
    }
        
    /* F17rev1 is not used for sea ice algorithms */
    if ( PMALGO_F17rev1 == this->platform ) {
	fprintf( stderr,"init_goddard: platform F17rev1 is not valid for"
                 " sea ice algorithms (use F17).  Exiting.\n" );
	close_goddard(this);
	return NULL;
    }
        
    return this;

}


/*----------------------------------------------------------------------
 *  goddard_hemisphere - return the hemisphere_id
 * 
 *    input: this - goddard_class object
 *
 *    result: hemisphere_id
 *
 *----------------------------------------------------------------------*/
pmalgo_hemisphere goddard_hemisphere(goddard_class *this) {

  assert(NULL != this);

  return this->hem;

}

/*----------------------------------------------------------------------
 *  goddard_platform - return the platform_id
 * 
 *    input: this - goddard_class object
 *
 *    result: platform_id
 *
 *----------------------------------------------------------------------*/
pmalgo_ssmi_platform goddard_platform(goddard_class *this) {

  assert(NULL != this);

  return this->platform;

}

/*----------------------------------------------------------------------
 *  goddard_scale_percent_seaice - return the scaling flag
 * 
 *    input: this - goddard_class object
 *
 *    result: scale_percent_seaice flag
 *
 *----------------------------------------------------------------------*/
bool goddard_scale_percent_seaice(goddard_class *this) {

  assert(NULL != this);

  return this->scale_percent_seaice;

}

/*----------------------------------------------------------------------
 *  goddard_spillover - return the land spillover object
 * 
 *    input: this - goddard_class object
 *
 *    result: land spillover object
 *
 *----------------------------------------------------------------------*/
land_spillover_class *goddard_spillover(goddard_class *this) {

  assert(NULL != this);

  return this->land_spillover;

}

/*----------------------------------------------------------------------
 *  goddard_recode_seaice - do the recoding to match the Goddard final product.
 *                          This includes minimum ice thresholding and some other
 *                          recoded values logic, where climatology, loci and
 *                          algorithm output values interact.
 * 
 * 
 *    input: this - goddard_class object
 *           grid - grid object
 *           data - seaice data
 *           verbose - verbosity level
 *
 *    result: performs recoding on the seaice data:
 *            1) any seaice values less than this->goddard_recoding_pcent are set to 0%
 *            2) resets any missing values (where TBs were 0) that are outside the
 *               ice_possible climatology to 0.  Nobody cares that there were zero-valued
 *               TBs there.
 *            3) any non-loci suspect seaice values are set to 0% (this is done to match
 *               output from IDL implementation we originally used for this algorithm);
 *               this preserves the pole hole from the loci mask, but resets any
 *               algorithm-generated suspect values to 0%
 *
 *    returns: TRUE for success, FALSE otherwise (an error message will be
 *             written to stderr)
 *                
 *
 *----------------------------------------------------------------------*/
bool goddard_recode_seaice( goddard_class *this,
                            grid_class *grid,
                            float **data,
                            bool use_climatology,
                            byte1 **climatology,
                            bool use_loci,
                            byte1 **loci,
                            bool fill_algo_pixels_only,
                            bool do_user_fill_value,
                            float fill_value,
                            int verbose ) {
    
    float ice_scale_factor;
    float min_seaice_threshold;
    int row, col;
    
    assert( NULL != this );
    assert( NULL != grid );
    assert( this->do_spillover );

    /*
     * If scaling is requested, process seaice values as percents
     * otherwise, process seaice values in range 0-250
     */
    ice_scale_factor = this->scale_percent_seaice ? 1.0 : 2.5;
    min_seaice_threshold = this->goddard_recoding_pcent * ice_scale_factor;

    for (row = 0; row < grid->rows; row++) {
        for (col = 0; col < grid->cols; col++) {

            /*
             * This SUSPECT_ICE check is going to reset 2 sets of
             * SUSPECT_ICE: 1) the suspect values from the algorithm itself,
             * and any SUSPECT_ICE values passed in the loci mask.
             * The first case is correct, but the second case will be added back,
             * below.
             */
            data[ row ][ col ] = ( data[ row ][ col ] < min_seaice_threshold
                || PMALGO_GTS_SUSPECT_ICE == data[ row ][ col ] )
                ? 0
                : data[ row ][ col ];

            /*
             * Apparently the GODDARD guys didn't care if TBs were zero anyplace
             * where the climatology was set to no ice, so they turned these missing
             * values off.
             */
            if ( use_climatology ) {

                if ( !climatology[ row ][ col ]
                     && fill_algo_pixels_only
                     && do_user_fill_value
                     && (byte1)fill_value == (byte1)data[ row ][ col ] ) {
                    data[ row ][ col ] = 0;
                }

                if ( !climatology[ row ][ col ]
                     && 250 >= data[ row ][ col ] ) {
                    data[ row ][ col ] = 0;
                }
                
            }

            /*
             * If the loci mask has pole hole values set to the standard SUSPECT ICE value
             * then retain them.  The suspect_ice that came out of the algorithm
             * in other locations to 0
             */
            if ( use_loci ) {

                if ( PMALGO_GTS_SUSPECT_ICE == loci[ row ][ col ] ) {
                    data[ row ][ col ] = loci[ row ][ col ];
                }

            }

        }

    }

    return TRUE;
    
}
    
/*----------------------------------------------------------------------
 *  close_goddard - free storage associated with this object
 * 
 *    input: this - goddard_class object
 *
 *    result: deallocate all memory associated with this object
 *
 *----------------------------------------------------------------------*/
void close_goddard(goddard_class *this) {

  if (NULL == this) return;

  free(this);

}

