/*  /home/powell/tasks/sun0272/granule.c

    This program produces snow depth and quality flags from the
    NASDA AMSR orbital data for one granule.

    Usage:
    DIR=/snow1/powell/sun0272/tlkt.v1.5
    cc -o $DIR/appl/pi_we granule.c -I$DIR/inc $DIR/lib/amsrtlkt.a

    The flags array (bits 7,6,5,4,3,2,1,0) has the following meaning:
    .	Bit 7		Unused
    .	Bit 6		Sea ice
    .	Bit 5		Land surface: 0=smooth, 1=rough
    .	Bit 4		Land: 0=snow possible, 1=snow impossible
    .	Bit 3		0=water, 1=land
    .	Bit 2-0		Snow class			*/

#define __GLOBAL__

#include <stdio.h>
#include <string.h>
#include <aptlkt.h>
#include <tlkt.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <math.h>

#define DEBUG

#ifdef DEBUG
#define TOP_LAT		90
#define BOT_LAT		30
#define BEG_LON		(-180)
#define END_LON		(+180)
#define ROWSbug		2*(TOP_LAT-BOT_LAT)
#define COLSbug		2*(END_LON-BEG_LON)
#endif


#define SWE_TO_SNOW	3.0

#define ERROR	-1

#define LAND_MASK_DSN	 "/snow1/powell/sun0272/AncillaryData/FlagsMask_90N-90S_0-360E"
#define FOREST_COVER_DSN "/snow1/powell/sun0272/AncillaryData/albedo_90N-90S_0-360E"

#define ROWS	360
#define COLS	720

#define V18	4
#define H18	5
#define V23	6
#define V36	8
#define H36	9

#define V89A	0
#define V89B	1
#define H89A	2
#define H89B	3

#define LAT	0
#define LON	1

#define MAX_SNOW_CLASS	7
#define NO_SNOW		0
#define UNKNOWN_SNOW	9999

/*** Bit flags for the flags array. (Bits 0,1,2 are for SnowClass) ***/
#define LAND			0x08
#define SNOW_IMPOSSIBLE		0x10
#define ROUGH			0x20
#define ICE			0x40

/*** Possible values for the SNOW output flags.  ***/
#define FLAG_NORMAL_NOSNOW	0
#define FLAG_WATER		1
#define FLAG_SNOW_IMPOSSIBLE	2
#define FLAG_ICE		3
#define FLAG_SURFACE_TOO_WARM	4
#define FLAG_FOREST		5
#define FLAG_MOUNTAIN		6
#define FLAG_RAINING		7
#define FLAG_WET_SNOW		8
#define FLAG_DRY_SNOW		9
#define FLAG_WET_SOIL		10
#define FLAG_DRY_SOIL		11
#define FLAG_OUTOFRANGE		12

/*** Prototype ***/
int flag_conv(void);

unsigned char	flags[ROWS][COLS];
float		ForestCover[ROWS][COLS];

#ifdef DEBUG
struct {
  float		snow[ROWSbug][COLSbug];
  unsigned char	flag[ROWSbug][COLSbug];
} save;
int		DryCount=0, WetCount=0;
#endif

/*==================================================================================*/

int main()
{
  double	v18, v23, v36, h36, v89;
  double	scat, sc36, sc89, scx;
  double	snow, albedo, coef, FloatSnow;
  float		longitude;
  int		FlagHandle, month, MsgSize;
  int		row, col, line, samp, SnowClass;
  char		message[512];

  static float	SNOWconst[MAX_SNOW_CLASS+1][12] = {
    /* Jan  Feb  Mar  Apr  May  Jun  Jul  Aug  Sep  Oct  Nov  Dec    */
    0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,   /* water     */
    3.6, 3.2, 2.8, 2.4, 2.4, 2.4, 2.4, 4.8, 4.8, 6.4, 5.6, 4.0,   /* tundra    */
    4.0, 3.6, 3.2, 2.8, 2.4, 2.4, 2.4, 4.8, 4.8, 6.4, 5.6, 4.8,   /* taiga     */
    0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,   /* ice       */
    4.8, 4.8, 4.8, 4.8, 4.8, 4.8, 4.8, 4.8, 4.8, 4.8, 4.8, 4.8,   /* maritime  */
    4.8, 4.8, 4.8, 4.8, 4.8, 4.8, 4.8, 4.8, 4.8, 4.8, 4.8, 4.8,   /* ephemeral */
    4.8, 4.8, 4.8, 4.8, 4.8, 4.8, 4.8, 4.8, 4.8, 4.8, 4.8, 4.8,   /* prairie   */
    4.8, 4.8, 4.8, 4.8, 4.8, 4.8, 4.8, 4.8, 4.8, 6.4, 6.0, 5.6};  /* alpine    */


  /*** Initialize tool kit ***/
  if(initlkt_(NULL, NULL) != NULL)
    trmtlkt_(NULL);

  /*** Get the month from the first line of the observation time data.  ***/
  icodeno = iTFK_OBT;
  idatano = 0;
  if(rdreal4_(&icodeno, &idatano, &iobsline, &iObsTimeNum, ObsTime) != NULL) {
    fprintf(stderr,"Error reading the observation time -- %d\n", idatano );
    trmtlkt_(NULL);
  }
  month = ObsTime[0][1] - 1;

  /*** Read flags array showing land/water/ice, level/mountain, snow possible/impossible,
       and snow class. ***/
  if( (FlagHandle = open( LAND_MASK_DSN, O_RDONLY)) == -1 ) {
    strcpy(message,"Cannot open the land mask file.\n");
    MsgSize = strlen(message);
    mestlkt_(message, &MsgSize);
    trmtlkt_(ERROR);
  }
  if( read( FlagHandle, (char *)flags[0], sizeof(flags) ) != sizeof( flags ) ) {
    strcpy(message,"Cannot read the land mask file.\n");
    MsgSize = strlen(message);
    mestlkt_(message, &MsgSize);
    trmtlkt_(ERROR);
  }
  if( close( FlagHandle )) {
    strcpy(message,"Cannot close the land mask file.\n");
    MsgSize = strlen(message);
    mestlkt_(message, &MsgSize);
    trmtlkt_(ERROR);
  }

  /*** Read forest cover array. ***/
  if( (FlagHandle = open( FOREST_COVER_DSN, O_RDONLY)) == -1 ) {

    strcpy(message,"Cannot open the forest cover file.\n");
    MsgSize = strlen(message);
    mestlkt_(message, &MsgSize);
    trmtlkt_(ERROR);
  }
  if( read( FlagHandle,
	    (char *)ForestCover[0], sizeof(ForestCover) ) != sizeof( ForestCover ) ) {
    strcpy(message,"Cannot read the forest cover file.\n");
    MsgSize = strlen(message);
    mestlkt_(message, &MsgSize);
    trmtlkt_(ERROR);
  }
  if( close( FlagHandle )) {
    strcpy(message,"Cannot close the forest cover file.\n");
    MsgSize = strlen(message);
    mestlkt_(message, &MsgSize);
    trmtlkt_(ERROR);
  }

  /*** Read AMSR brightness Temperature for 18 GHz vertical. ***/
  icodeno = iTFK_TE8;
  idatano = 5;
  if(rdreal4_(&icodeno, &idatano, &iobsline, &iobscan, bL1bbt[V18][0]) != NULL) {
    strcpy(message,"Error reading the 18 GHz vertical.\n");
    MsgSize = strlen(message);
    mestlkt_(message, &MsgSize);
    trmtlkt_(ERROR);
  }


  /*** Read AMSR brightness Temperature for 23 GHz vertical. ***/
  icodeno = iTFK_TE8;
  idatano = 7;
  if(rdreal4_(&icodeno, &idatano, &iobsline, &iobscan, bL1bbt[V23][0]) != NULL) {
    strcpy(message,"Error reading the 23 GHz vertical.\n");
    MsgSize = strlen(message);
    mestlkt_(message, &MsgSize);
    trmtlkt_(ERROR);
  }

  /*** Read AMSR brightness Temperature for 36 GHz vertical. ***/
  icodeno = iTFK_TE8;
  idatano = 9;
  if(rdreal4_(&icodeno, &idatano, &iobsline, &iobscan, bL1bbt[V36][0]) != NULL) {
    strcpy(message,"Error reading the 36 GHz vertical.\n");
    MsgSize = strlen(message);
    mestlkt_(message, &MsgSize);
    trmtlkt_(ERROR);
  }

  /*** Read AMSR brightness Temperature for 36 GHz horizontal. ***/
  icodeno = iTFK_TE8;
  idatano = 10;
  if(rdreal4_(&icodeno, &idatano, &iobsline, &iobscan, bL1bbt[H36][0]) != NULL) {
    strcpy(message,"Error reading the 36 GHz horizontal.\n");
    MsgSize = strlen(message);
    mestlkt_(message, &MsgSize);
    trmtlkt_(ERROR);
  }

  /*** Read AMSR brightness Temperature 89 GHz vertical A scan. ***/
  icodeno = iTFK_TF8;
  idatano = 1;
  if(rdreal4_(&icodeno, &idatano, &iobsline, &iobscanD, bL1bbt89[0][0]) != NULL) {
    strcpy(message,"Error reading the 89 GHz vertical A scan  band.\n");
    MsgSize = strlen(message);
    mestlkt_(message, &MsgSize);
    trmtlkt_(ERROR);
  }

  /*** Read AMSR brightness Temperature 89 GHz vertical B scan. ***/
  icodeno = iTFK_TF8;
  idatano = 2;
  if(rdreal4_(&icodeno, &idatano, &iobsline, &iobscanD, bL1bbt89[1][0]) != NULL) {
    strcpy(message,"Error reading the 89 GHz vertical B scan  band.\n");
    MsgSize = strlen(message);
    mestlkt_(message, &MsgSize);
    trmtlkt_(ERROR);
  }

  /*** Read AMSR latitude/longitude observation points except 89GHz-B.
       Latitude: +90.0 to -90.0    Longitude: -180.0 to +180.0	***/
  icodeno = iTFK_LE8;
  idatano = 0;
  if(rdreal4_(&icodeno, &idatano, &iobsline, &iobscanD, zLatLonA) != NULL) {
    strcpy(message,"Error reading the latitude/longitudes.\n");
    MsgSize = strlen(message);
    mestlkt_(message, &MsgSize);
    trmtlkt_(ERROR);
  }

  /*** Read in the surface temperatures. ***/
  icodeno = iTFK_GTT;
  idatano = 0;
  if(rdreal4_(&icodeno, &idatano, &iobsline, &iobscan, SurfATemp) != NULL) {
    strcpy(message,"Error reading the surface temperatures.\n");
    MsgSize = strlen(message);
    mestlkt_(message, &MsgSize);
    trmtlkt_(ERROR);
  }

#ifdef DEBUG
  /*** Set output structure to default values. ***/
  for( row=0; row<ROWSbug; row++ ) {
    for( col=0; col<COLSbug; col++ ) {
      save.snow[row][col] = 8888.0;
      save.flag[row][col] = 13;
    }
  }
#endif

  /*** Compute the SNOW for each pixel of the granule. ***/
  for( line=0; line<iobsline; line++ ) {
    for( samp=0; samp<iobscan; samp++ ) {

      ProcDATA[0][line][samp] = 0;
      ProcDataQ[0][line][samp][0] = 0;

      /*** Ignore this sample if it is bad. ***/
      if( (v36 = bL1bbt[V36][line][samp]) == -9999  ||
	  (h36 = bL1bbt[H36][line][samp]) == -9999  ||
	  (v18 = bL1bbt[V18][line][samp]) == -9999  ||
	  (v23 = bL1bbt[V23][line][samp]) == -9999 )
	continue;

      if( bL1bbt89[V89A][line][samp*2  ] == -9999  ||
	  bL1bbt89[V89B][line][samp*2  ] == -9999  ||
	  bL1bbt89[V89A][line][samp*2+1] == -9999  ||
	  bL1bbt89[V89B][line][samp*2+1] == -9999 )
	continue;

      snow = 0;

      /*** Get flags array indices from latitude/longitude of this sample.
	   Convert longitudes from -180_+180 to 0_360. ***/
      longitude = zLatLonA[LON][line][samp*2];
      col = (longitude>=0.0) ? longitude*2 : 720.0+longitude*2;
      row = ( 90.0 - zLatLonA[LAT][line][samp*2] ) * 2;

      /*** Check for permanent ice. */
      if( (flags[row][col] & ICE) ) {
	ProcDataQ[0][line][samp][0] = FLAG_ICE;
	ProcDATA[0][line][samp] = NO_SNOW;
	goto BottomLoop;		/*** Go to bottom loop -- there is no snow. ***/
      }

      /*** Check for water. ***/
      if( ! (flags[row][col] & LAND) ) {
	ProcDataQ[0][line][samp][0] = FLAG_WATER;
	ProcDATA[0][line][samp] = NO_SNOW;
	goto BottomLoop;		/*** Go to bottom loop -- there is no snow. ***/
      }

      /*** See if this is a mountainous region. ***/
      if( (flags[row][col] & ROUGH) ) {
	ProcDataQ[0][line][samp][0] = FLAG_MOUNTAIN;
	ProcDATA[0][line][samp] = UNKNOWN_SNOW;
	goto BottomLoop;		/*** Go to bottom loop -- there is no snow. ***/
      }

      /*** Check for heavy forest ***/
      if( ForestCover[row][col] > 0.5 ) {
	ProcDataQ[0][line][samp][0] = FLAG_FOREST;
	ProcDATA[0][line][samp] = UNKNOWN_SNOW;
	goto BottomLoop;		/*** Go to bottom loop -- there is no snow. ***/
      }

      /*** Check snow impossible flag ***/
      if( (flags[row][col] & SNOW_IMPOSSIBLE) ) {
	ProcDataQ[0][line][samp][0] = FLAG_SNOW_IMPOSSIBLE;
	ProcDATA[0][line][samp] = NO_SNOW;
	goto BottomLoop;		/*** Go to bottom loop -- there is no snow. ***/
      }

      if( v36<50 || v36>320 || h36<50 || h36>320 ||
	  v18<50 || v18>320 || v23<50 || v23>320 ) {
	ProcDataQ[0][line][samp][0] = FLAG_OUTOFRANGE;
	ProcDATA[0][line][samp] = NO_SNOW;
	goto BottomLoop;		/*** Go to bottom loop -- there is no snow. ***/
      }

      if( v36 >= 250.0  || h36 >= 240.0   ) {
	ProcDataQ[0][line][samp][0] = FLAG_NORMAL_NOSNOW;
	ProcDATA[0][line][samp] = NO_SNOW;
	goto BottomLoop;		/*** Go to bottom loop -- there is no snow. ***/
      }

      v89 = ( bL1bbt89[V89A][line][samp*2  ] + bL1bbt89[V89B][line][samp*2  ] +
	      bL1bbt89[V89A][line][samp*2+1] + bL1bbt89[V89B][line][samp*2+1] ) * 0.25;


      /*** Check the surface temperature. ***/
      if( (3.57 + 0.698*v18 + 0.333*v23 - 0.434*v36 + 0.422*v89) >= 275.0 ) {
	ProcDataQ[0][line][samp][0] = FLAG_SURFACE_TOO_WARM;
	ProcDATA[0][line][samp] = NO_SNOW;
	goto BottomLoop;		/*** Go to bottom loop -- there is no snow. ***/
      }


      /*** Come here if there is snow.  Get scattering index. ***/
      sc36 = v18-v36-3.0;
      sc89 = v23-v89-3.0;
      scx = v36-v89-1.0;
      scat = sc89;
      scat = sc36>scat ? sc36 : scat;
      scat = scx>scat ? scx : scat;

      /*** See if it's raining.  (Grody and Basist, 1996) ***/
      if( (v23 > 258.0)	        	||
	  (v23 > 254.0 && scat < 2)	||
	  (v23 > (165.0 + 0.49*v89) )  ) {
	ProcDataQ[0][line][samp][0] = FLAG_RAINING;
	ProcDATA[0][line][samp] = UNKNOWN_SNOW;
	goto BottomLoop;		/*** Go to bottom loop -- there is no snow. ***/
      }

      /*** See if it's wet snow. */
      if( (v36 - h36) > 10.0 && SurfATemp[line][samp] >= 270.0 ) {
	ProcDataQ[0][line][samp][0] = FLAG_WET_SNOW;
	ProcDATA[0][line][samp] = UNKNOWN_SNOW;
	goto BottomLoop;		/*** Go to bottom loop -- there is no snow. ***/
      }

      /*** Compute snow for wet or dry soil. ***/
      if( ( ((v36-v18)/18.0) >= -0.3 )
	  && SurfATemp[line][samp] >= 270.0
	  && SurfATemp[line][samp] <= 273.0 ) {
	ProcDataQ[0][line][samp][0] = FLAG_WET_SOIL;
	FloatSnow = (5.0/SWE_TO_SNOW) * (v18-v36);
	snow = (FloatSnow>2.5) ? FloatSnow : 0;

      } else {

	ProcDataQ[0][line][samp][0] = FLAG_DRY_SOIL;
	SnowClass = (flags[row][col] & MAX_SNOW_CLASS);
	coef = SNOWconst[SnowClass][month];
	albedo = ForestCover[row][col];
	FloatSnow = (coef/SWE_TO_SNOW) * (v18-v36-5.0) / (1.-albedo);
	snow = (FloatSnow>2.5) ? FloatSnow : 0;
      }
      ProcDATA[0][line][samp] = snow<0 ? NO_SNOW : snow;

    BottomLoop:  ;
#ifdef DEBUG
      /*** Save the data for later IDL debugging. ***/
      if( row < ROWSbug ) {
	save.snow[row][col] = ProcDATA[0][line][samp];
	save.flag[row][col] = ProcDataQ[0][line][samp][0];
      }
#endif
    }		/* End: for( sample...   */
  }		/* End: for( line...     */

#ifdef DEBUG
  /*** Write out the saved debug data. ***/
  if( write(2, (char *)save.snow, sizeof(save) ) != sizeof(save) ) {
    perror("Error writing saved debug output.");
    return(1);
  }
#endif


  /*** Tool kit terminate process ***/
  trmtlkt_(NULL);

  return(0);
}
