/*============================================================================
 * This file contains functions needed for applying the land spillover
 * correction algorithm developed by Cavalieri et al. The functions were
 * originally developed by M.J. Brodzik for the Goddard time series algorithm
 * and moved to a seperate file for use by multiple sea ice algorithms by
 * J. Smith on 2/11/2010.
 *
 * National Snow and Ice Data Center, University of Colorado, Boulder
 * Copyright 2010 Regents of the University of Colorado
 *==========================================================================*/

#include "land_spillover.h"
#include "matrix.h"
#include "matrix_io.h"
#include "morph_sub.h"

#define MORPH_DILATE 2   /* morphological dilation operator */
#define MORPH_RANK 16   /* morphological rank operator */
#define MORPH_BINIMG 512 /* values in Img will be 0 and non-zero */
#define MORPH_NZPAD_OFF 0    /* padding off */
#define MORPH_AUTOSE 0 /* */
#define MORPH_BINSE 256 /* values in Structuring element will be 0 and non-zero */
#define MORPH_SET 0  /* SorF set operation */
#define MORPH_LOWER_THRESH 1 /* lower binary threshold value */
#define MORPH_UPPER_THRESH 1 /* upper binary threshold value */
#define MORPH_TRUE 255 /* flag value for true in return from MorphSub */

typedef enum {
    SHORE=0, NEAR_SHORE, OFF_SHORE, NUM_SHORE_TYPES
} shore_type;
static const char *shore_type_name[] = { "SHORE", "NEAR SHORE", "OFF SHORE" };
static const byte1 shore_type_value[] = { 2, 4, 8 };
static const byte1 shore_ice_threshold_pcent[] = { 60, 40, 20 };
static const int shore_neighborhood[] = { 7, 5, 3 };
static const int open_water_pixel_threshold = 3;


land_spillover_class *init_land_spillover(const int spillover_open_water_pcent,
					  const int algorithm_open_water_pcent,
					  const char *landmask_filename,
					  const char *weights_filename,
					  grid_class *grid,
					  int verbose)
{

    land_spillover_class *this;


    this = (land_spillover_class *)calloc(1, sizeof(land_spillover_class));
    if (this == NULL)
    {
	perror("init_landspillover_class: land_spillover_class object");
	return NULL;
    }

    this->se_path = morph_se_path(verbose);

    this->spillover_open_water_pcent = ( 0 <= spillover_open_water_pcent ?
                                         spillover_open_water_pcent :
                                         algorithm_open_water_pcent);

    if (!make_proximity_mask(this, landmask_filename, grid, verbose))
    {
	fprintf(stderr, "\n");
    }

    if (!make_CMIN_weights(this, weights_filename, grid, verbose))
    {
	fprintf(stderr, "\n");
    }

    return this;
}


/*----------------------------------------------------------------------
 *  morph_se_path - retrieves the environment setting for MORPH_SE_PATH
 *
 *    input: verbose flag for verbose output
 *
 *    result: char * will contain value of MORPH_SE_PATH environment var
iable
 *
 *----------------------------------------------------------------------*/
static char *morph_se_path( bool verbose ) {

    char *path;
    char *spath;

    path = getenv( "MORPH_SE_PATH" );
    if ( NULL == path ) {

        if ( 0 < verbose )
            fprintf( stderr, "> morph_se_path: No environment variable MORPH_SE_PATH set, using default.\n" );
        path = strdup( "/share/data/pmalgos/spillover_kernels/" );

    }

    fprintf( stderr,"morph_se_path: Using spillover kernels at %s\n", path );

    return path;

}

/*----------------------------------------------------------------------
 *  make_proximity_mask - dilate the landmask for the Goddard spillover correction
 *
 *    input:    se_path - character string containing the MORPH_SE_PATH
 *                        environment variable.
 *              landmask_filename - filename with landmask (0=ocean/1=land) to dilate
 *              grid - grid object
 *              verbose - set > 0 for verbose output
 *
 *    output:   land_ocean_mask - dilated land/ocean mask, in which 0 indicates
 *                                ocean and 1 indicates land.
 *
 *    result: TRUE if successful, or FALSE otherwise, with error message
 *           to stderr
 *
 *    proximity_mask M will contain:
 *                  0 = water
 *                  1 = land
 *                  2 = shore
 *                  4 = near-shore
 *                  8 = off-shore
 *
 *----------------------------------------------------------------------
*/
#define OCEAN 0
#define LAND 1
static bool make_proximity_mask(land_spillover_class *this,
				const char *landmask_filename,
				grid_class *grid,
				int verbose) {

    byte1 **landmask;
    byte1 **shore;
    byte1 **near_shore;
    byte1 **off_shore;
    int row, col;
    int morph_verbose;
    int morph_dummy = 0;
    char SE_filename[MAX_STRING];

#ifdef SPILLOVER_DEBUG
    size_t status;
    char M_filename[MAX_STRING];
#endif

    if ( 0 < verbose) {
        fprintf( stderr, "> Doing landmask dilation...\n" );
        morph_verbose = 1;
    }

    /* Allocate memory for the M matrix */
    this->M = (byte1 **)matrix(
	grid->rows, grid->cols, sizeof( byte1 ), matrix_ZERO );
    if ( NULL == this->M ) {
	perror("make_proximity_mask: error allocating matrix for M");
	return FALSE;
    }

    /* Read the landmask */
    landmask = (byte1 **)initialize_matrix( grid, sizeof(**landmask),
                                            landmask_filename, "landmask input",
                                            verbose );
    if (NULL == landmask) {
      perror("make_proximity_mask: error allocating matrix for landmask");
      return FALSE;
    }

    /* Allocate memory for each shore category and the matrix M*/
    shore = (byte1 **)matrix( grid->rows, grid->cols, sizeof( byte1 ), matrix_ZERO );
    near_shore = (byte1 **)matrix( grid->rows, grid->cols, sizeof( byte1 ), matrix_ZERO );
    off_shore = (byte1 **)matrix( grid->rows, grid->cols, sizeof( byte1 ), matrix_ZERO );

    if ( NULL == shore || NULL == near_shore || NULL == off_shore ) {
      perror("make_proximity_mask: error allocating matrices for shore/near shore/off shore");
      return FALSE;
    }

    /*
     * Do dilation and combination with "not land" for each shore category
     * dilate expects to get only the data part of the matrix (not the pointers to the rows)
     */
    sprintf( SE_filename, "%s/%s", this->se_path, "nsidc_roundSE3x3.txt"
);
    MorphSub( MORPH_DILATE, *(landmask), *(shore), grid->cols, grid->rows,
              MORPH_BINIMG, MORPH_NZPAD_OFF,
              MORPH_LOWER_THRESH, MORPH_UPPER_THRESH,
              SE_filename, MORPH_AUTOSE,
              morph_dummy, morph_dummy, morph_dummy,
              MORPH_BINSE, MORPH_SET,
              morph_dummy, morph_dummy, morph_verbose);

    sprintf( SE_filename, "%s/%s", this->se_path, "nsidc_roundSE5x5.txt"
);
    MorphSub( MORPH_DILATE, *(landmask), *(near_shore), grid->cols, grid->rows,
              MORPH_BINIMG, MORPH_NZPAD_OFF,
              MORPH_LOWER_THRESH, MORPH_UPPER_THRESH,
              SE_filename, MORPH_AUTOSE,
              morph_dummy, morph_dummy, morph_dummy,
              MORPH_BINSE, MORPH_SET,
              morph_dummy, morph_dummy, morph_verbose);

    sprintf( SE_filename, "%s/%s", this->se_path, "nsidc_roundSE7x7.txt"
);
    MorphSub( MORPH_DILATE, *(landmask), *(off_shore), grid->cols, grid->rows,
              MORPH_BINIMG, MORPH_NZPAD_OFF,
              MORPH_LOWER_THRESH, MORPH_UPPER_THRESH,
              SE_filename, MORPH_AUTOSE,
              morph_dummy, morph_dummy, morph_dummy,
              MORPH_BINSE, MORPH_SET,
              morph_dummy, morph_dummy, morph_verbose);

    /*
     * Calculate the composite mask with land/shore/near shore/off shore/water classifications.
     */
    for ( row=0; row < grid->rows; row++ ) {
      for ( col=0; col < grid->cols; col++ ) {

        if ( LAND == landmask[row][col] ) {

	    this->M[row][col] = LAND;

        } else if ( 0 < shore[row][col] ) {

	    this->M[row][col] = shore_type_value[ SHORE ];

        } else if ( 0 < near_shore[row][col] && 0 == shore[row][col] ) {

	    this->M[row][col] = shore_type_value[ NEAR_SHORE ];

        } else if ( 0 < off_shore[row][col] && 0 == near_shore[row][col]
 ) {

	    this->M[row][col] = shore_type_value[ OFF_SHORE ];

        }

      }
    }

#ifdef SPILLOVER_DEBUG

    /* If debugging, write the proximity mask to a file */
    strncpy( M_filename, "spillover_M.bin", MAX_STRING );
    status = write_matrix( M_filename, (void **)this->M,
                           grid->rows, grid->cols, sizeof( byte1 ) );
    if ( 0 == status ) {
      fprintf( stderr, "make_proximity_matrix: Error writing M matrix.\n" );
    } else {
      fprintf( stderr, "SPILLOVER_DEBUG> Wrote M matrix to %s\n", M_filename );
    }

#endif

    free( landmask );
    free( shore );
    free( near_shore );
    free( off_shore );

    if ( 0 < verbose) {
        fprintf( stderr, "> Goddard landmask dilation finished.\n" );
    }

    return TRUE;

}

/*----------------------------------------------------------------------
 *  make_CMIN_weights - populate the array of CMIN weights to use for spillover correction
 *
 *    input: land_ocean_mask - dilated land/ocean mask (0: ocean, 1: land)
 *              weights_filename - filename with minimum annual ice concentrations
 *              grid - grid object
 *              verbose - non-zero for verbose output
 *
 *    output: ice_threshold - thresholded ice array (Cavalieri et al.)
 *
 *    result: TRUE if successful, or FALSE otherwise, with error message
 *           to stderr
 *
 *    Weights matrix CMIN is created according to Cavalieri et al.
 *
 *    Begin with matrix P of minimum monthly avg ice concentrations throughout
 *    a given year (Cavalieri et al. say 1984 for SMMR and 1992 for SSM/I), and then
 *    resetting off-shore, near-shore and shore pixels via:
 *    1) at off-shore pixels, any P values exceeding 20% are reduced to 20%
 *    2) at near-shore pixels, any P values exceeding 40% are reduced to 40%
 *    3) at shore pixels, any P values exceeding 60% are reduced to 60%
 *
 *----------------------------------------------------------------------
*/
static bool make_CMIN_weights(land_spillover_class *this,
		       const char *weights_filename,
		       grid_class *grid,
		       int verbose) {

    static byte1 **ice_threshold_tmp;
    int row, col;

#ifdef SPILLOVER_DEBUG
  size_t status;
  char CMIN_filename[MAX_STRING];
#endif

  assert( NULL != this->M );

  if ( 0 < verbose) {
    fprintf( stderr, "> Begin make_CMIN_weights...\n" );
  }

  /* Read the annual ice concentrations into the CMIN matrix*/
  this->CMIN = (byte1 **)initialize_matrix( grid,
					    sizeof(**(this->CMIN)),
					    weights_filename,
					    "weights input",
					    verbose );
  if ( NULL == this->CMIN ) {
    perror("make_CMIN_weights: error allocating matrix for weights");
    return FALSE;
  }

  /*
   * Populate the ice_threshold matrix according to Cavilieri et al.
   */
  for ( row=0; row < grid->rows; row++ ) {
    for ( col=0; col < grid->cols; col++ ) {

      if ( shore_type_value[ OFF_SHORE ] == this->M[row][col] ) {
          if ( shore_ice_threshold_pcent[ OFF_SHORE ] < this->CMIN[row][col] ) {
	      this->CMIN[row][col] = shore_ice_threshold_pcent[ OFF_SHORE ];
          }
      } else if ( shore_type_value[ NEAR_SHORE ] == this->M[row][col] )
{
          if ( shore_ice_threshold_pcent[ NEAR_SHORE ] < this->CMIN[row][col] ) {
              this->CMIN[row][col] = shore_ice_threshold_pcent[ NEAR_SHORE ];
          }
      } else if ( shore_type_value [ SHORE ] == this->M[row][col] ) {
          if ( shore_ice_threshold_pcent[ SHORE ] < this->CMIN[row][col]
 ) {
	      this->CMIN[row][col] = shore_ice_threshold_pcent[ SHORE ];
        }
      }

    }
  }

#ifdef SPILLOVER_DEBUG

  /* If debugging, write the CMIN mask to a file */
    strncpy( CMIN_filename, "spillover_CMIN.bin", MAX_STRING );
    status = write_matrix( CMIN_filename, (void **)this->CMIN,
                           grid->rows, grid->cols, sizeof( byte1 ) );
    if ( 0 == status ) {
      fprintf( stderr, "make_CMIN_weights: Error writing CMIN matrix.\n"
 );
    } else {
      fprintf( stderr, "SPILLOVER_DEBUG> Wrote CMIN matrix to %s\n",
	       CMIN_filename );
    }

#endif

  if ( 0 < verbose) {
    fprintf( stderr, "> End make_CMIN_weights...\n" );
  }

  return TRUE;

}


/*----------------------------------------------------------------------
 *  spillover_seaice - do the spillover correction to the input seaice grid
 *
 *    input: land_ocean_mask - dilated land/ocean mask.
 *           ice_threshold - thresholded ice array (Cavalieri et al.)
 *           scale_percent_seaice - TRUE: scale the sea ice to 0-100
 *           spillover_open_water_pcent - maximum percent sea ice concentration
 *                                        that is considered to be open water
 *                                        (typically 15 percent).
 *           scaling_factor - scaling factor to convert sea ice concentrations
 *                            from what the algorithm output to 0-100.
 *           max_ice_percent - the maximum percentage that is considered to
 *                             be a valid calculation by an algorithm.
 *           grid - grid object
 *           data - seaice data
 *           verbose - verbosity level
 *
 *    result: performs spillover correction on input seaice data
 *
 *    returns: TRUE for success, FALSE otherwise (an error message will be
 *             written to stderr)
 *
 *
 *----------------------------------------------------------------------*/
bool spillover_seaice( land_spillover_class *this,
                       bool scale_percent_seaice,
                       float spillover_open_water_pcent,
                       float scaling_factor,
                       int max_ice_percent,
                       grid_class *grid,
                       float **data,
                       int verbose ) {

    float ice_scale_factor;
    float open_water_threshold;
    float max_seaice_threshold;
    int row, col;
    char SE_filename[MAX_STRING];
    int morph_dummy = 0;
    int morph_verbose = verbose;
    byte1 **open_water;
    byte1 **adjust;
    int i;
    int count;

#ifdef SPILLOVER_DEBUG
    size_t status;
    char adjust_filename[MAX_STRING];
    char open_water_filename[MAX_STRING];
    char raw_filename[MAX_STRING];
#endif

    assert( NULL != grid );

    /*
     * If scaling is requested, this means that the range of sea ice values
     * differs from 0-100.
     */
    ice_scale_factor = scale_percent_seaice ? 1.0 : scaling_factor;
    open_water_threshold = spillover_open_water_pcent * ice_scale_factor;
    max_seaice_threshold = max_ice_percent * ice_scale_factor;

    if (verbose > 1)
	fprintf(stderr, "spillover parameters: ice_scale_factor = %f, open_water_threshold = %f, max_seaice_threshold = %f\n", ice_scale_factor, open_water_threshold, max_seaice_threshold);


    /*
     * Do the spillover correction to the input gridded seaice data:
     *
     * At any time when the NxN neighborhood of a "shore" pixel contains 3 or
     * more open-water pixels (concentration < threshold) then the calculated
     * concentration is reduced by the corresponding value in the
     * weight_mask. Wherever the adjustment results in a negative
     * concentraction, the concentration is set to zero.
     *
     * The neighborhood NxN sizes will be adjusted for type of shore pixel.
     * (XxX, XxX, XxX) for (shore, near-shore, far-shore).  This adjustment
     * matches what K. Knowles implemented in the original IDL implementation
     * of this algorithm.
     *
     * N.B. Walt thinks that Cavalieri et al. actually did this iteratively:
     * 1) do 3x3 correction of offshore, then
     * 2) do 3x3 correction of nearshore on the results from 1, then
     * 3) do 3x3 correction of shore on the results from 2.
     *
     * We may want to try this alternative some day.
     */

    /*
     * Find pixels in seaice data that are open water.
     */
    open_water = (byte1 **)matrix( grid->rows, grid->cols, sizeof( byte1 ),
				   matrix_ZERO );
    if ( NULL == open_water ) {
      perror("spillover_seaice: error allocating matrix for open_water");
      return FALSE;
    }

    for (row = 0; row < grid->rows; row++) {
        for (col = 0; col < grid->cols; col++) {
            if ( data[ row ][ col ] < open_water_threshold) {
                open_water[ row ][ col ] = 1;
            }
        }
    }

#ifdef SPILLOVER_DEBUG

        /*
         * If debugging, write the raw seaice data to a file
         */
        sprintf( raw_filename, "%s", "spillover.raw.bin" );
        status = write_matrix( raw_filename, (void **)data,
                               grid->rows, grid->cols, sizeof( float ) );
        if ( 0 == status ) {
            fprintf( stderr, "bootstrap_spillover_seaice: Error writing raw grid.\n" );
        } else {
            fprintf( stderr, "SPILLOVER_DEBUG> Wrote raw grid to %s\n", raw_filename );
        }

#endif

    adjust = (byte1 **)matrix( grid->rows, grid->cols, sizeof( byte1 ), matrix_ZERO );
    if ( NULL == adjust ) {
      perror("bootstrap_spillover_seaice: error allocating matrix for shore values");
      return FALSE;
    }

    /*
     * Loop for each successive shore type (offshore, nearshore, shore)
     * starting with the furthest away and moving to the shore pixels
     */
    for ( i = OFF_SHORE; i >= SHORE; i-- ) {

        /* Initialize the candidates for adjustment for each new shore category */
        for (row = 0; row < grid->rows; row++) {
            for (col = 0; col < grid->cols; col++) {
                adjust[ row ][ col ] = 0;
            }
        }

        /*
         * Use the rank filter function to find where there are >=3 open water
         * pixels in the NxN neighborhood (where N=3, 5, 7) for this shore
         * type; these will be the pixels that are candidates for adjustment
         */
        sprintf( SE_filename, "%s/square%1dx%1d.txt", this->se_path,
                 shore_neighborhood[ i ], shore_neighborhood[ i ] );
        MorphSub( MORPH_RANK, *open_water, *adjust, grid->cols, grid->rows,
                  MORPH_BINIMG, MORPH_NZPAD_OFF,
                  MORPH_LOWER_THRESH, MORPH_UPPER_THRESH,
                  SE_filename, MORPH_AUTOSE,
                  morph_dummy, morph_dummy, morph_dummy,
                  MORPH_BINSE, MORPH_SET,
                  open_water_pixel_threshold, morph_dummy, morph_verbose );

        /* Find and adjust the shore pixels for this type */
        count = 0;
        for (row = 0; row < grid->rows; row++) {
            for (col = 0; col < grid->cols; col++) {

                /*
                 * This shore pixel has at least the required number of open
                 * water neighbors, so it needs to be adjusted.  Use the CMIN
                 * array to adjust concentrations downward.  Only adjust
                 * seaice locations that are legitimate seaice values (not
                 * flagged values).  Reset negative adjusted values to 0.
                 */
                if ( this->M[ row ][ col ] == shore_type_value[ i ]
                     && MORPH_TRUE == adjust[ row ][ col]
                     && data[ row ][ col ] <= max_seaice_threshold ) {

                    data[ row ][ col ] = data[ row ][ col ] -
                        ( this->CMIN[ row][ col ] * ice_scale_factor );
                    if ( data[ row ][ col ] < 0.0 ) {
                        data[ row ][ col ] = 0.0;
                    }
                    count++;

                }
            }

        }

#ifdef SPILLOVER_DEBUG

        fprintf( stderr, "SPILLOVER_DEBUG> %d %s pixels adjusted.\n",
                 count, shore_type_name[ i ] );

        /*
         * If debugging, write the adjust mask to a file
         */
        sprintf( adjust_filename, "%s%1d%s",
                 "spillover_adjust.",
                 shore_type_value[ i ],
                 ".bin" );
        status = write_matrix( adjust_filename, (void **)adjust,
                               grid->rows, grid->cols, sizeof( byte1 ) );
        if ( 0 == status ) {
            fprintf( stderr, "bootstrap_spillover_seaice: Error writing adjust grid.\n" );
        } else {
            fprintf( stderr, "SPILLOVER_DEBUG> Wrote adjust grid to %s\n", adjust_filename );
        }

#endif

    }

#ifdef SPILLOVER_DEBUG

        /*
         * If debugging, write the open water to a file
         */
        sprintf( open_water_filename, "%s", "spillover.open_water.bin" );
        status = write_matrix( open_water_filename, (void **)open_water,
                               grid->rows, grid->cols, sizeof( byte1 ) );
        if ( 0 == status ) {
            fprintf( stderr, "bootstrap_spillover_seaice: Error writing open water grid.\n" );
        } else {
            fprintf( stderr, "SPILLOVER_DEBUG> Wrote open water grid to %s\n", open_water_filename );
        }

#endif

    if ( NULL != open_water ) free( open_water );
    if ( NULL != adjust ) free( adjust );

    return TRUE;

}
