/*======================================================================
 * pmalgorithm - operations to perform algorithms on passive microwave (PM)
 *               brightness temperatures (TBs)
 *
 * Notes:
 * To add a new algorithm:
 * 1) Add the algorithm id to pmalgo_id in the header file
 * 2) Add the corresponding string to pmalgo_id_name in the header file
 * 3) Add a line of descriptive info to PMALGO_ALGORITHM_DESCRIPTIONS
 *    in the header file
 * 4) Declare a new run_xxx function to Algorithm Function List (below)
 *    (the interface must be the same as all others)
 * 5) Define the new run_xxx function in the Local Functions section (at
 *    the end of this file); use of a macro is optional; Include any
 *    publication references used for the implementation in the
 *    header for the function
 * 6) Add a new section to the case statement in init_pmalgo
 * 7) If the new algorithm requires input data besides the standard input
 *    of loci and/or climatology (e.g. some algorithms need an addtional surface 
 *    classification, or a forest mask), then add the check for the optional 
 *    inputs to pmalgos.c, including a call to whatever initialization might be 
 *    required there e.g. the Bootstrap algorithms require a call to 
 *    init_bootstrap_pmalgo, and the ones that need a forest mask require a 
 *    call to init_sfc_pmalgo.
 * 8) Edit the man page file pmalgos.1, to include details of this algorithm.
 *
 * 03/22/1999 brodzik brodzik@zamboni.colorado.edu 303-492-8263
 * National Snow & Ice Data Center, Boulder, CO
 * Copyright (C) 1999-2006 University of Colorado at Boulder
 *
 *======================================================================*/
#include "define.h"
#include "matrix_io.h"
#include "pmalgorithm.h"
#include "bootstrap.h"
#include "goddard.h"
#include "snow_gb.h"
#include "surface.h"
#include "land_spillover.h"
#include "fortran.h"

static const char pmalgorithm_c_rcsid[] = "$Header: /tmp_mnt/FILES/pmalgos/pmalgorithm.c,v 1.58.2.1 2008-06-10 20:36:35 brodzik Exp $";

/* 
 * Conversion factor for input TBs, converts integer tenths of Kelvins to
 * floating point values in Kelvins
 */
#define TB_SCALE 0.1 

/* 
 * Algorithm Function List 
 * New algorithm functions must retain this calling sequence.
 */
static bool run_nsidc_m1_swe (void *private,
                              byte2 **data[PMALGO_NUM_CHANNELS], 
			      int col, int row, float *value, bool *valid_tbs);
static bool run_nsidc_m4_swe (void *private,
			      byte2 **data[PMALGO_NUM_CHANNELS], 
			      int col, int row, float *value, bool *valid_tbs);
static bool run_nsidc_s0_swe (void *private,
			      byte2 **data[PMALGO_NUM_CHANNELS], 
			      int col, int row, float *value, bool *valid_tbs);
static bool run_nsidc_s1_swe (void *private,
			      byte2 **data[PMALGO_NUM_CHANNELS], 
			      int col, int row, float *value, bool *valid_tbs);
static bool run_nsidc_s3_swe (void *private,
			      byte2 **data[PMALGO_NUM_CHANNELS], 
			      int col, int row, float *value, bool *valid_tbs);
static bool run_nsidc_s4_swe (void *private,
			      byte2 **data[PMALGO_NUM_CHANNELS], 
			      int col, int row, float *value, bool *valid_tbs);
static bool run_nsidc_s5_swe (void *private,
			      byte2 **data[PMALGO_NUM_CHANNELS], 
			      int col, int row, float *value, bool *valid_tbs);
static bool run_nsidc_s6_swe (void *private,
			      byte2 **data[PMALGO_NUM_CHANNELS], 
			      int col, int row, float *value, bool *valid_tbs);
static bool run_nsidc_s10_swe (void *private,
			       byte2 **data[PMALGO_NUM_CHANNELS], 
			       int col, int row, float *value, bool *valid_tbs);
static bool run_nsidc_s11_swe (void *private,
			       byte2 **data[PMALGO_NUM_CHANNELS], 
			       int col, int row, float *value, bool *valid_tbs);
static bool run_nsidc_s12_swe (void *private,
			       byte2 **data[PMALGO_NUM_CHANNELS], 
			       int col, int row, float *value, bool *valid_tbs);
static bool run_nsidc_a0_swe (void *private,
			      byte2 **data[PMALGO_NUM_CHANNELS], 
			      int col, int row, float *value, bool *valid_tbs);
static bool run_nsidc_a4_swe (void *private,
			      byte2 **data[PMALGO_NUM_CHANNELS], 
			      int col, int row, float *value, bool *valid_tbs);
static bool run_nsidc_a10_swe (void *private,
                               byte2 **data[PMALGO_NUM_CHANNELS], 
                               int col, int row, float *value, bool *valid_tbs);
static bool run_gw_swe (void *private,
			byte2 **data[PMALGO_NUM_CHANNELS], 
			int col, int row, float *value, bool *valid_tbs);
static bool run_gw_wet_swe (void *private,
			    byte2 **data[PMALGO_NUM_CHANNELS], 
			    int col, int row, float *value, bool *valid_tbs);
static bool run_gw_modwet_swe (void *private,
			       byte2 **data[PMALGO_NUM_CHANNELS], 
			       int col, int row, float *value, 
			       bool *valid_tbs);
static bool run_hb_dry_comb (void *private,
			     byte2 **data[PMALGO_NUM_CHANNELS], 
			     int col, int row, float *value, bool *valid_tbs);
static bool run_hb_fresh_comb (void *private,
			       byte2 **data[PMALGO_NUM_CHANNELS], 
			       int col, int row, float *value, 
			       bool *valid_tbs);
static bool run_tait1_swe (void *private,
			   byte2 **data[PMALGO_NUM_CHANNELS], 
			   int col, int row, float *value, bool *valid_tbs);
static bool run_chang_depth (void *private,
			     byte2 **data[PMALGO_NUM_CHANNELS], 
			     int col, int row, float *value, bool *valid_tbs);
static bool run_gb_snow (void *private,
			 byte2 **data[PMALGO_NUM_CHANNELS], 
			 int col, int row, float *value, bool *valid_tbs);
static bool run_gb_surface (void *private,
			    byte2 **data[PMALGO_NUM_CHANNELS], 
			    int col, int row, float *value, bool *valid_tbs);
static bool run_chang_surface (void *private,
			       byte2 **data[PMALGO_NUM_CHANNELS], 
			       int col, int row, float *value, 
			       bool *valid_tbs);
static bool run_nsidc_surface (void *private,
			       byte2 **data[PMALGO_NUM_CHANNELS], 
			       int col, int row, float *value, 
			       bool *valid_tbs);
static bool run_boot_seaice (void *private,
			     byte2 **data[PMALGO_NUM_CHANNELS], 
			     int col, int row, float *value, bool *valid_tbs);
static bool run_ntn_seaice (void *private,
			    byte2 **data[PMALGO_NUM_CHANNELS], 
			    int col, int row, float *value, bool *valid_tbs);
static bool run_nts_seaice (void *private,
			    byte2 **data[PMALGO_NUM_CHANNELS], 
			    int col, int row, float *value, bool *valid_tbs);
static bool run_ntn_myice (void *private,
			   byte2 **data[PMALGO_NUM_CHANNELS], 
			   int col, int row, float *value, bool *valid_tbs);
static bool run_gts_seaice (void *private,
			    byte2 **data[PMALGO_NUM_CHANNELS], 
			    int col, int row, float *value, bool *valid_tbs);
static bool run_gts_myice (void *private,
			   byte2 **data[PMALGO_NUM_CHANNELS], 
			   int col, int row, float *value, bool *valid_tbs);
static bool run_ssmi_xpgr (void *private,
			   byte2 **data[PMALGO_NUM_CHANNELS], 
			   int col, int row, float *value, bool *valid_tbs);
static bool run_smmr_xpgr (void *private,
			   byte2 **data[PMALGO_NUM_CHANNELS], 
			   int col, int row, float *value, bool *valid_tbs);
static bool run_nagler_depth (void *private,
			      byte2 **data[PMALGO_NUM_CHANNELS], 
			      int col, int row, float *value, bool *valid_tbs);
static bool run_nagler_csi (void *private,
			    byte2 **data[PMALGO_NUM_CHANNELS], 
			    int col, int row, float *value, bool *valid_tbs);
static bool run_za_s1_fs (void *private,
                          byte2 **data[PMALGO_NUM_CHANNELS], 
                          int col, int row, float *value, bool *valid_tbs);
static bool run_no_algorithm (void *private,
			      byte2 **data[PMALGO_NUM_CHANNELS], 
			      int col, int row, float *value, bool *valid_tbs);

/* Other local functions */
static void fetch_swe_tbs( void *private, byte2 in_t19h, byte2 in_t37h,
                           float *out_t19h, float *out_t37h );
static bool run_nasateam_ice (byte2 **data[PMALGO_NUM_CHANNELS], 
			      int col, int row, byte2 hem, float *total, 
			      float *myr, bool *valid_tbs);
static bool run_goddard_ice (byte2 **data[PMALGO_NUM_CHANNELS], 
			     int col, int row, pmalgo_hemisphere hem, 
			     pmalgo_ssmi_platform platform,
                             bool scale_percent_seaice,
			     float *total, float *myr, bool *valid_tbs);
static double altitude_correct_tb ( double tb,
                                    void *private,
                                    int2 elevation_meters,
                                    pmalgo_channel channel );

/* parameter for calling seaice_nasateam_ */
typedef enum {
    NASATEAM_NORTH_HEM = 1, NASATEAM_SOUTH_HEM
} nasateam_hemisphere;

/* prototype for FORTRAN subroutines */
extern void FORTRAN_NAME(seaice_nasateam) (byte2 *tb19h, byte2 *tb19v,
					   byte2 *tb22v, byte2 *tb37v,
					   byte2 *ihem, 
					   byte1 *total_ice, byte1 *multi_yr_ice);

extern void FORTRAN_NAME(seaice_goddard) (byte2 *tb19h, byte2 *tb19v,
					  byte2 *tb22v, byte2 *tb37v,
					  byte2 *ihem, byte2 *iplatform,
                                          byte2 *scale,
					  byte2 *total_ice, byte2 *multi_yr_ice);


/*----------------------------------------------------------------------
 *  init_pmalgo - initialize object according to a specific algorithm
 *
 *    input: algorithm_name - algorithm name (must match one of 
 *                            pmalgo_id_name)
 *           platform_name - platform name
 *           verbose - verbosity flag
 * 
 *    returns: pointer to new pmalgo_class instance
 *             or NULL (and reason to stderr) in case of error 
 *
 *----------------------------------------------------------------------*/
pmalgo_class *init_pmalgo(char *algorithm_name,
                          char *platform_name,
			  int verbose) {

    int i;
    pmalgo_id id;
    pmalgo_class *this;

    if (0 < verbose) {
        fprintf(stderr,"> %s\n",pmalgorithm_h_rcsid);
        fprintf(stderr,"> %s\n",pmalgorithm_c_rcsid);
    }

    /* Allocate storage for object, and initialize any non-zero defaults */
    this = (pmalgo_class *)calloc(1,sizeof(pmalgo_class));
    if (NULL == this) {
        perror("init_pmalgo: pmalgo_class object");
        return NULL;
    }
    this->id = PMALGO_NO_ALGORITHM;
    this->sensor = PMALGO_NO_SENSOR;

    this->platform = PMALGO_NO_PLATFORM;
    for (i = 0; i <= PMALGO_NUM_PLATFORMS; i++) {
	if (0 == strcmp(pmalgo_ssmi_platform_name[i],platform_name)) {
	    this->platform = (pmalgo_ssmi_platform)i;
	    break;
	}
    }
    if ( PMALGO_NO_PLATFORM == this->platform
         && 0 != strcmp(pmalgo_ssmi_platform_name[0],platform_name) ) {
        fprintf(stderr,"init_pmalgo: Unrecognized platform_name = %s.\n", platform_name );
        close_pmalgo(this);
        return NULL;
    }
    
    for (i = 0; i < PMALGO_NUM_CHANNELS; i++) {
        this->channel[i] = PMALGO_NO_CHANNEL;
    }

    for (id = (pmalgo_id)0; id < PMALGO_NUM_ALGORITHMS; id++) {
        if (0 == strcmp(pmalgo_id_name[id],algorithm_name)) {
            this->id = id;
            break;
        }
    }

    if ( 0 == strcmp( "NSIDC_A2_SWE", algorithm_name )
         || 0 == strcmp( "NSIDC_S2_SWE", algorithm_name ) ) {
        fprintf(stderr,"pmalgos: NSIDC_A2_SWE and NSIDC_S2_SWE are obsolete.\n");
        fprintf(stderr,"pmalgos: See NSIDC_S11_SWE and NSIDC_S12_SWE for updated versions.\n");
        close_pmalgo(this);
        return NULL;
    }

    if (PMALGO_NO_ALGORITHM == this->id) {
        fprintf(stderr,"init_pmalgo: invalid algorithm_name = %s\n",
                algorithm_name);
        close_pmalgo(this);
        return NULL;
    }

    if (PMALGO_NSIDC_M1_SWE == this->id) {
        this->sensor = PMALGO_SMMR;
        this->channel[0] = PMALGO_SMMR_18H;
        this->channel[1] = PMALGO_SMMR_37H;
        this->num_channels = 2;
        this->num_days = 1;
        this->direction = strdup(pmalgo_direction_name
                                 [PMALGO_COLD_PASS_DIRECTION_ONLY]);
        this->grid_algorithm = strdup(pmalgo_grid_algorithm_name
                                      [PMALGO_INVERSE_DISTANCE_WEIGHTING]);
        this->run = run_nsidc_m1_swe;
    } else if (PMALGO_NSIDC_M4_SWE == this->id) {
        this->sensor = PMALGO_SMMR;
        this->channel[0] = PMALGO_SMMR_18H;
        this->channel[1] = PMALGO_SMMR_37H;
        this->num_channels = 2;
        this->num_days = 1;
        this->direction = strdup(pmalgo_direction_name
                                 [PMALGO_COLD_PASS_DIRECTION_ONLY]);
        this->grid_algorithm = strdup(pmalgo_grid_algorithm_name
                                      [PMALGO_INVERSE_DISTANCE_WEIGHTING]);
        this->run = run_nsidc_m4_swe;
    } else if (PMALGO_NSIDC_S0_SWE == this->id) {
        this->sensor = PMALGO_SSMI;
        this->channel[0] = PMALGO_SSMI_19H;
        this->channel[1] = PMALGO_SSMI_37H;
        this->num_channels = 2;
        this->num_days = 1;
        this->direction = strdup(pmalgo_direction_name
                                 [PMALGO_COLD_PASS_DIRECTION_ONLY]);
        this->grid_algorithm = strdup(pmalgo_grid_algorithm_name
                                      [PMALGO_BACKUS_GILBERT]);
        this->run = run_nsidc_s0_swe;
    } else if (PMALGO_NSIDC_S1_SWE == this->id) {
        this->sensor = PMALGO_SSMI;
        this->channel[0] = PMALGO_SSMI_19H;
        this->channel[1] = PMALGO_SSMI_37H;
        this->num_channels = 2;
        this->num_days = 1;
        this->direction = strdup(pmalgo_direction_name
                                 [PMALGO_COLD_PASS_DIRECTION_ONLY]);
        this->grid_algorithm = strdup(pmalgo_grid_algorithm_name
                                      [PMALGO_BACKUS_GILBERT]);
        this->run = run_nsidc_s1_swe;
    } else if (PMALGO_NSIDC_S3_SWE == this->id) {
        this->sensor = PMALGO_SSMI;
        this->channel[0] = PMALGO_SSMI_19H;
        this->channel[1] = PMALGO_SSMI_37V;
        this->channel[2] = PMALGO_SSMI_37H;
        this->num_channels = 3;
        this->num_days = 1;
        this->direction = strdup(pmalgo_direction_name
                                 [PMALGO_COLD_PASS_DIRECTION_ONLY]);
        this->grid_algorithm = strdup(pmalgo_grid_algorithm_name
                                      [PMALGO_BACKUS_GILBERT]);
        this->run = run_nsidc_s3_swe;
    } else if (PMALGO_NSIDC_S4_SWE == this->id) {
        this->sensor = PMALGO_SSMI;
        this->channel[0] = PMALGO_SSMI_19H;
        this->channel[1] = PMALGO_SSMI_37H;
        this->num_channels = 2;
        this->num_days = 1;
        this->direction = strdup(pmalgo_direction_name
                                 [PMALGO_COLD_PASS_DIRECTION_ONLY]);
        this->grid_algorithm = strdup(pmalgo_grid_algorithm_name
                                      [PMALGO_BACKUS_GILBERT]);
        this->run = run_nsidc_s4_swe;
    } else if (PMALGO_NSIDC_S5_SWE == this->id) {
        this->sensor = PMALGO_SSMI;
        this->channel[0] = PMALGO_SSMI_19V;
        this->channel[1] = PMALGO_SSMI_19H;
        this->channel[2] = PMALGO_SSMI_22V;
        this->channel[3] = PMALGO_SSMI_37V;
        this->channel[4] = PMALGO_SSMI_37H;
        this->channel[5] = PMALGO_SSMI_85V;
        this->channel[6] = PMALGO_SSMI_85H;
        this->num_channels = 7;
        this->num_days = 1;
        this->direction = strdup(pmalgo_direction_name
                                 [PMALGO_COLD_PASS_DIRECTION_ONLY]);
        this->grid_algorithm = strdup(pmalgo_grid_algorithm_name
                                      [PMALGO_BACKUS_GILBERT]);
        this->run = run_nsidc_s5_swe;
    } else if (PMALGO_NSIDC_S6_SWE == this->id) {
        this->sensor = PMALGO_SSMI;
        this->channel[0] = PMALGO_SSMI_19V;
        this->channel[1] = PMALGO_SSMI_19H;
        this->channel[2] = PMALGO_SSMI_22V;
        this->channel[3] = PMALGO_SSMI_37V;
        this->channel[4] = PMALGO_SSMI_37H;
        this->channel[5] = PMALGO_SSMI_85V;
        this->channel[6] = PMALGO_SSMI_85H;
        this->num_channels = 7;
        this->num_days = 1;
        this->direction = strdup(pmalgo_direction_name
                                 [PMALGO_COLD_PASS_DIRECTION_ONLY]);
        this->grid_algorithm = strdup(pmalgo_grid_algorithm_name
                                      [PMALGO_BACKUS_GILBERT]);
        this->run = run_nsidc_s6_swe;
    } else if (PMALGO_NSIDC_S10_SWE == this->id) {
        this->sensor = PMALGO_SSMI;
        this->channel[0] = PMALGO_SSMI_19H;
        this->channel[1] = PMALGO_SSMI_37H;
        this->num_channels = 2;
        this->num_days = 1;
        this->direction = strdup(pmalgo_direction_name
                                 [PMALGO_COLD_PASS_DIRECTION_ONLY]);
        this->grid_algorithm = strdup(pmalgo_grid_algorithm_name
                                      [PMALGO_BACKUS_GILBERT]);
        this->run = run_nsidc_s10_swe;
    } else if (PMALGO_NSIDC_S11_SWE == this->id) {
        this->sensor = PMALGO_SSMI;
        this->channel[0] = PMALGO_SSMI_19H;
        this->channel[1] = PMALGO_SSMI_37H;
        this->num_channels = 2;
        this->num_days = 1;
        this->direction = strdup(pmalgo_direction_name
                                 [PMALGO_COLD_PASS_DIRECTION_ONLY]);
        this->grid_algorithm = strdup(pmalgo_grid_algorithm_name
                                      [PMALGO_BACKUS_GILBERT]);
        this->run = run_nsidc_s11_swe;
    } else if (PMALGO_NSIDC_S12_SWE == this->id) {
        this->sensor = PMALGO_SSMI;
        this->channel[0] = PMALGO_SSMI_19H;
        this->channel[1] = PMALGO_SSMI_37H;
        this->num_channels = 2;
        this->num_days = 1;
        this->direction = strdup(pmalgo_direction_name
                                 [PMALGO_COLD_PASS_DIRECTION_ONLY]);
        this->grid_algorithm = strdup(pmalgo_grid_algorithm_name
                                      [PMALGO_BACKUS_GILBERT]);
        this->run = run_nsidc_s12_swe;
    } else if (PMALGO_NSIDC_A0_SWE == this->id) {
        this->sensor = PMALGO_AMSR;
        this->channel[0] = PMALGO_AMSR_18H;
        this->channel[1] = PMALGO_AMSR_36H;
        this->num_channels = 2;
        this->num_days = 1;
        this->direction = strdup(pmalgo_direction_name
                                 [PMALGO_COLD_PASS_DIRECTION_ONLY]);
        this->grid_algorithm = strdup(pmalgo_grid_algorithm_name
                                      [PMALGO_BACKUS_GILBERT]);
        this->run = run_nsidc_a0_swe;
    } else if (PMALGO_NSIDC_A4_SWE == this->id) {
        this->sensor = PMALGO_AMSR;
        this->channel[0] = PMALGO_AMSR_18H;
        this->channel[1] = PMALGO_AMSR_36H;
        this->num_channels = 2;
        this->num_days = 1;
        this->direction = strdup(pmalgo_direction_name
                                 [PMALGO_COLD_PASS_DIRECTION_ONLY]);
        this->grid_algorithm = strdup(pmalgo_grid_algorithm_name
                                      [PMALGO_BACKUS_GILBERT]);
        this->run = run_nsidc_a4_swe;
    } else if (PMALGO_NSIDC_A10_SWE == this->id) {
        this->sensor = PMALGO_AMSR;
        this->channel[0] = PMALGO_AMSR_18H;
        this->channel[1] = PMALGO_AMSR_36H;
        this->num_channels = 2;
        this->num_days = 1;
        this->direction = strdup(pmalgo_direction_name
                                 [PMALGO_COLD_PASS_DIRECTION_ONLY]);
        this->grid_algorithm = strdup(pmalgo_grid_algorithm_name
                                      [PMALGO_BACKUS_GILBERT]);
        this->run = run_nsidc_a10_swe;
    } else if (PMALGO_GW_SWE == this->id) {
        this->sensor = PMALGO_SSMI;
        this->channel[0] = PMALGO_SSMI_19V;
        this->channel[1] = PMALGO_SSMI_37V;
        this->num_channels = 2;
        this->num_days = 1;
        this->direction = strdup(pmalgo_direction_name
                                 [PMALGO_COLD_PASS_DIRECTION_ONLY]);
        this->grid_algorithm = strdup(pmalgo_grid_algorithm_name
                                      [PMALGO_BACKUS_GILBERT]);
        this->run = run_gw_swe;
    } else if (PMALGO_GW_WET_SWE == this->id) {
        this->sensor = PMALGO_SSMI;
        this->channel[0] = PMALGO_SSMI_19V;
        this->channel[1] = PMALGO_SSMI_37V;
        this->channel[2] = PMALGO_SSMI_37H;
        this->num_channels = 3;
        this->num_days = 1;
        this->direction = strdup(pmalgo_direction_name
                                 [PMALGO_COLD_PASS_DIRECTION_ONLY]);
        this->grid_algorithm = strdup(pmalgo_grid_algorithm_name
                                      [PMALGO_BACKUS_GILBERT]);
        this->run = run_gw_wet_swe;
    } else if (PMALGO_GW_MODWET_SWE == this->id) {
        this->sensor = PMALGO_SSMI;
        this->channel[0] = PMALGO_SSMI_19V;
        this->channel[1] = PMALGO_SSMI_37V;
        this->channel[2] = PMALGO_SSMI_37H;
        this->num_channels = 3;
        this->num_days = 1;
        this->direction = strdup(pmalgo_direction_name[PMALGO_COLD_PASS_DIRECTION_ONLY]);
        this->grid_algorithm = strdup(pmalgo_grid_algorithm_name
                                      [PMALGO_BACKUS_GILBERT]);
        this->run = run_gw_modwet_swe;
    } else if (PMALGO_HB_DRY_COMB == this->id) {
        this->sensor = PMALGO_SSMI;
        this->channel[0] = PMALGO_SSMI_19V;
        this->channel[1] = PMALGO_SSMI_19H;
        this->channel[2] = PMALGO_SSMI_37V;
        this->channel[3] = PMALGO_SSMI_37H;
        this->num_channels = 4;
        this->num_days = 1;
        this->direction = strdup(pmalgo_direction_name[PMALGO_COLD_PASS_DIRECTION_ONLY]);
        this->grid_algorithm = strdup(pmalgo_grid_algorithm_name
                                      [PMALGO_BACKUS_GILBERT]);
        this->run = run_hb_dry_comb;
    } else if (PMALGO_HB_FRESH_COMB == this->id) {
        this->sensor = PMALGO_SSMI;
        this->channel[0] = PMALGO_SSMI_85V;
        this->channel[1] = PMALGO_SSMI_85H;
        this->channel[2] = PMALGO_SSMI_37V;
        this->channel[3] = PMALGO_SSMI_37H;
        this->num_channels = 4;
        this->num_days = 1;
        this->direction = strdup(pmalgo_direction_name[PMALGO_COLD_PASS_DIRECTION_ONLY]);
        this->grid_algorithm = strdup(pmalgo_grid_algorithm_name
                                      [PMALGO_BACKUS_GILBERT]);
        this->run = run_hb_fresh_comb;
    } else if (PMALGO_TAIT1_SWE == this->id) {
        this->sensor = PMALGO_SSMI;
        this->channel[0] = PMALGO_SSMI_19V;
        this->channel[1] = PMALGO_SSMI_37H;
        this->channel[2] = PMALGO_SSMI_85H;
        this->num_channels = 3;
        this->num_days = 1;
        this->direction = strdup(pmalgo_direction_name
                                 [PMALGO_COLD_PASS_DIRECTION_ONLY]);
        this->grid_algorithm = strdup(pmalgo_grid_algorithm_name
                                      [PMALGO_BACKUS_GILBERT]);
        this->run = run_tait1_swe;
    } else if (PMALGO_CHANG_DEPTH == this->id) {
        this->sensor = PMALGO_SSMI;
        this->channel[0] = PMALGO_SSMI_19V;
        this->channel[1] = PMALGO_SSMI_37V;
        this->channel[2] = PMALGO_SSMI_37H;
        this->num_channels = 3;
        this->num_days = 1;
        this->direction = strdup(pmalgo_direction_name
                                 [PMALGO_COLD_PASS_DIRECTION_ONLY]);
        this->grid_algorithm = strdup(pmalgo_grid_algorithm_name
                                      [PMALGO_BACKUS_GILBERT]);
        this->run = run_chang_depth;
    } else if (PMALGO_GB_SNOW == this->id) {
        this->sensor = PMALGO_SSMI;
        this->channel[0] = PMALGO_SSMI_19V;
        this->channel[1] = PMALGO_SSMI_19H;
        this->channel[2] = PMALGO_SSMI_22V;
        this->channel[3] = PMALGO_SSMI_37V;
        this->channel[4] = PMALGO_SSMI_85V;
        this->channel[5] = PMALGO_SSMI_85H;
        this->num_channels = 6;
        this->num_days = 1;
        this->direction = strdup(pmalgo_direction_name
                                 [PMALGO_COLD_PASS_DIRECTION_ONLY]);
        this->grid_algorithm = strdup(pmalgo_grid_algorithm_name
                                      [PMALGO_BACKUS_GILBERT]);
        this->run = run_gb_snow;
    } else if (PMALGO_GB_SURFACE == this->id) {
        this->sensor = PMALGO_SSMI;
        this->channel[0] = PMALGO_SSMI_19V;
        this->channel[1] = PMALGO_SSMI_19H;
        this->channel[2] = PMALGO_SSMI_22V;
        this->channel[3] = PMALGO_SSMI_37V;
        this->channel[4] = PMALGO_SSMI_85V;
        this->channel[5] = PMALGO_SSMI_85H;
        this->num_channels = 6;
        this->num_days = 1;
        this->direction = strdup(pmalgo_direction_name
                                 [PMALGO_COLD_PASS_DIRECTION_ONLY]);
        this->grid_algorithm = strdup(pmalgo_grid_algorithm_name
                                      [PMALGO_BACKUS_GILBERT]);
        this->run = run_gb_surface;
    } else if (PMALGO_CHANG_SURFACE == this->id) {
        this->sensor = PMALGO_SSMI;
        this->channel[0] = PMALGO_SSMI_19V;
        this->channel[1] = PMALGO_SSMI_22V;
        this->channel[2] = PMALGO_SSMI_37V;
        this->channel[3] = PMALGO_SSMI_37H;
        this->channel[4] = PMALGO_SSMI_85V;
        this->channel[5] = PMALGO_SSMI_85H;
        this->num_channels = 6;
        this->num_days = 1;
        this->direction = strdup(pmalgo_direction_name
                                 [PMALGO_COLD_PASS_DIRECTION_ONLY]);
        this->grid_algorithm = strdup(pmalgo_grid_algorithm_name
                                      [PMALGO_BACKUS_GILBERT]);
        this->run = run_chang_surface;
    } else if (PMALGO_NSIDC_SURFACE == this->id) {
        this->sensor = PMALGO_SSMI;
        this->channel[0] = PMALGO_SSMI_19V;
        this->channel[1] = PMALGO_SSMI_22V;
        this->channel[2] = PMALGO_SSMI_37V;
        this->channel[3] = PMALGO_SSMI_37H;
        this->channel[4] = PMALGO_SSMI_85V;
        this->channel[5] = PMALGO_SSMI_85H;
        this->num_channels = 6;
        this->num_days = 1;
        this->direction = strdup(pmalgo_direction_name
                                 [PMALGO_COLD_PASS_DIRECTION_ONLY]);
        this->grid_algorithm = strdup(pmalgo_grid_algorithm_name
                                      [PMALGO_BACKUS_GILBERT]);
        this->run = run_nsidc_surface;
    } else if (PMALGO_BOOTN_SEAICE == this->id ||
	PMALGO_BOOTS_SEAICE == this->id ) { 
        this->sensor = PMALGO_SSMI;
        this->channel[0] = PMALGO_SSMI_19V;
	if (PMALGO_BOOTN_SEAICE == this->id){
	    this->channel[1] = PMALGO_SSMI_22V;
	    this->channel[2] = PMALGO_SSMI_37V;
	    this->channel[3] = PMALGO_SSMI_37H;
	    this->num_channels = 4;
	} else {
	    this->channel[1] = PMALGO_SSMI_22V;
	    this->channel[2] = PMALGO_SSMI_37V;
	    this->num_channels = 3;
	}
        this->num_days = 1;
        this->direction = strdup(pmalgo_direction_name
                                 [PMALGO_BOTH_PASS_DIRECTIONS]);
        this->grid_algorithm = strdup(pmalgo_grid_algorithm_name
                                      [PMALGO_DITB]);
        this->run = run_boot_seaice;
    } else if (PMALGO_BOOTN_V02_SEAICE == this->id || 
	PMALGO_BOOTS_V02_SEAICE == this->id ) { 
        this->sensor = PMALGO_SSMI;
        this->channel[0] = PMALGO_SSMI_19V;
        this->channel[1] = PMALGO_SSMI_22V;
        this->channel[2] = PMALGO_SSMI_37V;
        this->channel[3] = PMALGO_SSMI_37H;
        this->num_channels = 4;
        this->num_days = 1;
        this->direction = strdup(pmalgo_direction_name
                                 [PMALGO_BOTH_PASS_DIRECTIONS]);
        this->grid_algorithm = strdup(pmalgo_grid_algorithm_name
                                      [PMALGO_DITB]);
        this->run = run_boot_seaice;
    } else if (PMALGO_BOOTN_V02_PUB_SEAICE == this->id || 
	PMALGO_BOOTS_V02_PUB_SEAICE == this->id ) {
        /* This algorithm is identical to the BOOT[N|S]_V02_SEAICE
           algorithm except that it uses the new GSFC regressions.
           We re-use as much of the existing Bootstrap algorithm
           code as possible by using a new F17 platform value.
           The new platform selects the latest regression subroutine
           in the FORTRAN code (BOOT_ICE_[N|S]SBA). */
        if (this->platform == PMALGO_F17) {
          this->platform = PMALGO_F17_final;
        }
        this->sensor = PMALGO_SSMI;
        this->channel[0] = PMALGO_SSMI_19V;
        this->channel[1] = PMALGO_SSMI_22V;
        this->channel[2] = PMALGO_SSMI_37V;
        this->channel[3] = PMALGO_SSMI_37H;
        this->num_channels = 4;
        this->num_days = 1;
        this->direction = strdup(pmalgo_direction_name
                                 [PMALGO_BOTH_PASS_DIRECTIONS]);
        this->grid_algorithm = strdup(pmalgo_grid_algorithm_name
                                      [PMALGO_DITB]);
        this->run = run_boot_seaice;
    } else if (PMALGO_NTN_SEAICE == this->id ) { 
        this->sensor = PMALGO_SSMI;
        this->channel[0] = PMALGO_SSMI_19V;
        this->channel[1] = PMALGO_SSMI_19H;
        this->channel[2] = PMALGO_SSMI_22V;
        this->channel[3] = PMALGO_SSMI_37V;
        this->num_channels = 4;
        this->num_days = 1;
        this->direction = strdup(pmalgo_direction_name
                                 [PMALGO_BOTH_PASS_DIRECTIONS]);
        this->grid_algorithm = strdup(pmalgo_grid_algorithm_name
                                      [PMALGO_DITB]);
        this->run = run_ntn_seaice;
    } else if (PMALGO_NTS_SEAICE == this->id ) {
        this->sensor = PMALGO_SSMI;
        this->channel[0] = PMALGO_SSMI_19V;
        this->channel[1] = PMALGO_SSMI_19H;
        this->channel[2] = PMALGO_SSMI_22V;
        this->channel[3] = PMALGO_SSMI_37V;
        this->num_channels = 4;
        this->num_days = 1;
        this->direction = strdup(pmalgo_direction_name
                                 [PMALGO_BOTH_PASS_DIRECTIONS]);
        this->grid_algorithm = strdup(pmalgo_grid_algorithm_name
                                      [PMALGO_DITB]);
        this->run = run_nts_seaice;
    } else if (PMALGO_NTN_MYICE == this->id ) { 
        this->sensor = PMALGO_SSMI;
        this->channel[0] = PMALGO_SSMI_19V;
        this->channel[1] = PMALGO_SSMI_19H;
        this->channel[2] = PMALGO_SSMI_22V;
        this->channel[3] = PMALGO_SSMI_37V;
        this->num_channels = 4;
        this->num_days = 1;
        this->direction = strdup(pmalgo_direction_name
                                 [PMALGO_BOTH_PASS_DIRECTIONS]);
        this->grid_algorithm = strdup(pmalgo_grid_algorithm_name
                                      [PMALGO_DITB]);
        this->run = run_ntn_myice;
    } else if (PMALGO_GTS_SEAICE == this->id ) { 
        this->sensor = PMALGO_SSMI;
        this->channel[0] = PMALGO_SSMI_19V;
        this->channel[1] = PMALGO_SSMI_19H;
        this->channel[2] = PMALGO_SSMI_22V;
        this->channel[3] = PMALGO_SSMI_37V;
        this->num_channels = 4;
        this->num_days = 1;
        this->direction = strdup(pmalgo_direction_name
                                 [PMALGO_BOTH_PASS_DIRECTIONS]);
        this->grid_algorithm = strdup(pmalgo_grid_algorithm_name
                                      [PMALGO_DITB]);
        this->run = run_gts_seaice;
    } else if (PMALGO_GTS_PUB_SEAICE == this->id ) {
        /* This algorithm is identical to the GTS_SEAICE algorithm
           except that it uses the new published tie-points and weather
           thresholds. We re-use as much of the Goddard algorithm
           code as possible by switching to a new F17 platform and using
           this new platform to index into the tie-point coefficients
           array and to select new weather thresholds in the FORTRAN
           code. */
        if (this-> platform == PMALGO_F17) {
          this->platform = PMALGO_F17_final;
        }
        this->sensor = PMALGO_SSMI;
        this->channel[0] = PMALGO_SSMI_19V;
        this->channel[1] = PMALGO_SSMI_19H;
        this->channel[2] = PMALGO_SSMI_22V;
        this->channel[3] = PMALGO_SSMI_37V;
        this->num_channels = 4;
        this->num_days = 1;
        this->direction = strdup(pmalgo_direction_name
                                 [PMALGO_BOTH_PASS_DIRECTIONS]);
        this->grid_algorithm = strdup(pmalgo_grid_algorithm_name
                                      [PMALGO_DITB]);
        this->run = run_gts_seaice;
    } else if (PMALGO_GTS_MYICE == this->id ) { 
        this->sensor = PMALGO_SSMI;
        this->channel[0] = PMALGO_SSMI_19V;
        this->channel[1] = PMALGO_SSMI_19H;
        this->channel[2] = PMALGO_SSMI_22V;
        this->channel[3] = PMALGO_SSMI_37V;
        this->num_channels = 4;
        this->num_days = 1;
        this->direction = strdup(pmalgo_direction_name
                                 [PMALGO_BOTH_PASS_DIRECTIONS]);
        this->grid_algorithm = strdup(pmalgo_grid_algorithm_name
                                      [PMALGO_DITB]);
        this->run = run_gts_myice;
    } else if (PMALGO_SMMR_XPGR == this->id) {
        this->sensor = PMALGO_SMMR;
        this->channel[0] = PMALGO_SMMR_18H;
        this->channel[1] = PMALGO_SMMR_37V;
        this->num_channels = 2;
        this->num_days = 1;
        this->direction = strdup(pmalgo_direction_name
                                 [PMALGO_WARM_PASS_DIRECTION_ONLY]);
        this->grid_algorithm = strdup(pmalgo_grid_algorithm_name
                                      [PMALGO_BACKUS_GILBERT]);
        this->run = run_smmr_xpgr;
    } else if (PMALGO_SSMI_XPGR == this->id) {
        this->sensor = PMALGO_SSMI;
        this->channel[0] = PMALGO_SSMI_19H;
        this->channel[1] = PMALGO_SSMI_37V;
        this->num_channels = 2;
        this->num_days = 1;
        this->direction = strdup(pmalgo_direction_name
                                 [PMALGO_WARM_PASS_DIRECTION_ONLY]);
        this->grid_algorithm = strdup(pmalgo_grid_algorithm_name
                                      [PMALGO_BACKUS_GILBERT]);
        this->run = run_ssmi_xpgr;
    } else if (PMALGO_NAGLER_DEPTH == this->id) {
        this->sensor = PMALGO_SSMI;
        this->channel[0] = PMALGO_SSMI_19V;
        this->channel[1] = PMALGO_SSMI_37V;
        this->channel[2] = PMALGO_SSMI_85V;
        this->num_channels = 3;
        this->num_days = 3;
        this->direction = strdup(pmalgo_direction_name
                                 [PMALGO_COLD_PASS_DIRECTION_ONLY]);
        this->grid_algorithm = strdup(pmalgo_grid_algorithm_name
                                      [PMALGO_BACKUS_GILBERT]);
        this->run = run_nagler_depth;
    } else if (PMALGO_NAGLER_CSI == this->id) {
        this->sensor = PMALGO_SSMI;
        this->channel[0] = PMALGO_SSMI_37V;
        this->channel[1] = PMALGO_SSMI_85V;
        this->num_channels = 2;
        this->num_days = 1;
        this->direction = strdup(pmalgo_direction_name
                                 [PMALGO_COLD_PASS_DIRECTION_ONLY]);
        this->grid_algorithm = strdup(pmalgo_grid_algorithm_name
                                      [PMALGO_BACKUS_GILBERT]);
        this->run = run_nagler_csi;
    } else if (PMALGO_ZA_S1_FS == this->id) {
        this->sensor = PMALGO_SSMI;
        this->channel[0] = PMALGO_SSMI_19V;
        this->channel[1] = PMALGO_SSMI_37V;
        this->num_channels = 2;
        this->num_days = 1;
        this->direction = strdup(pmalgo_direction_name
                                 [PMALGO_COLD_PASS_DIRECTION_ONLY]);
        this->grid_algorithm = strdup(pmalgo_grid_algorithm_name
                                      [PMALGO_BACKUS_GILBERT]);
        this->run = run_za_s1_fs;
    } else if (PMALGO_SMMR_ALL_CHANNELS == this->id) {
        this->sensor = PMALGO_SMMR;
        this->channel[0] = PMALGO_SMMR_06V;
        this->channel[1] = PMALGO_SMMR_06H;
        this->channel[2] = PMALGO_SMMR_10V;
        this->channel[3] = PMALGO_SMMR_10H;
        this->channel[4] = PMALGO_SMMR_18V;
        this->channel[5] = PMALGO_SMMR_18H;
        this->channel[6] = PMALGO_SMMR_21V;
        this->channel[7] = PMALGO_SMMR_21H;
        this->channel[8] = PMALGO_SMMR_37V;
        this->channel[9] = PMALGO_SMMR_37H;
        this->num_channels = 10;
        this->num_days = 0;
        this->direction = strdup(pmalgo_direction_name
                                 [PMALGO_EITHER_PASS_DIRECTION]);
        this->grid_algorithm = strdup(pmalgo_grid_algorithm_name
                                      [PMALGO_INVERSE_DISTANCE_WEIGHTING]);
        this->run = run_no_algorithm;
    } else if (PMALGO_SSMI_ALL_CHANNELS == this->id) {
        this->sensor = PMALGO_SSMI;
        this->channel[0] = PMALGO_SSMI_19V;
        this->channel[1] = PMALGO_SSMI_19H;
        this->channel[2] = PMALGO_SSMI_22V;
        this->channel[3] = PMALGO_SSMI_37V;
        this->channel[4] = PMALGO_SSMI_37H;
        this->channel[5] = PMALGO_SSMI_85V;
        this->channel[6] = PMALGO_SSMI_85H;
        this->num_channels = 7;
        this->num_days = 0;
        this->direction = strdup(pmalgo_direction_name
                                 [PMALGO_EITHER_PASS_DIRECTION]);
        this->grid_algorithm = strdup(pmalgo_grid_algorithm_name
                                      [PMALGO_BACKUS_GILBERT]);
        this->run = run_no_algorithm;
    } else {
        assert(0);
        fprintf(stderr,"init_pmalgo: invalid algorithm id = %d\n",
                this->id);
        close_pmalgo(this);
        return NULL;
    }

    this->private = NULL;

    return this;

}


/*----------------------------------------------------------------------
 *  init_bootstrap_pmalgo - initialize parameters specific to Bootstrap
 *                          seaice algorithms
 * 
 *    input: this - pmalgo_class object
 *           date - date (yyyymmdd) of input tbs
 *           spillover_open_water_pcent - open water definition for spillover correction
 *           bootstrap_recoding_pcent - minimum seaice threshold for bootstrap recode thresholding
 *           scale_percent_seaice - flag to turn on scaling to 0-100%
 *           land_spillover - land spillover class
 *           grid - grid_class to use to read landmask and weights files
 *           verbose - verbosity flag
 *
 *    result: TRUE if successful, or FALSE otherwise, with error message
 *           to stderr
 *
 *----------------------------------------------------------------------*/
bool init_bootstrap_pmalgo(pmalgo_class *this,
			   const char *date,
			   const int bootstrap_recoding_pcent,
			   const bool scale_percent_seaice,
			   land_spillover_class *land_spillover,
			   grid_class *grid,
			   int verbose) {

    assert(NULL != this);
    
    this->private = (void *)init_bootstrap(pmalgo_id_name[this->id],
                                           pmalgo_ssmi_platform_name[this->platform],
                                           date,
                                           bootstrap_recoding_pcent,
                                           scale_percent_seaice,
                                           land_spillover,
                                           grid, verbose);
    
    if (NULL != this->private) {
        return TRUE;
    } else {
        return FALSE;
    }
					 
}



/*----------------------------------------------------------------------
 *  mask_bootstrap_pmalgo - use the input loci to determine mask pixels where
 *                          Bootstrap sea ice algorithm should be run
 * 
 *    input: this - pmalgo_class object
 *           loci - loci array
 *           use_value - value in loci array that mask should be set for
 *               algorithm to be run
 *
 *----------------------------------------------------------------------*/
void mask_bootstrap_pmalgo(pmalgo_class *this,
			   byte1 **loci, byte1 use_value) {

    assert(NULL != this);
    assert(NULL != this->private);
    set_mask_bootstrap((bootstrap_class *)this->private,loci,use_value);
    return;
}



/*----------------------------------------------------------------------
 *  init_goddard_pmalgo - initialize parameters specific to Goddard time
 *                        series seaice algorithms
 * 
 *    input: this - pmalgo_class object
 *           spillover_open_water_pcent - open water definition for spillover correction
 *           goddard_recoding_pcent - minimum seaice threshold for goddard recode thresholding
 *           hemisphere_name - input hemisphere name
 *           scale_percent_seaice - flag to turn on scaling to 0-100%
 *           land_spillover - land_spillover_class object for spillover
 *                            processing. NULL if spillover processing won't
 *                            be performed.
 *      Only required for spillover processing:
 *           grid - grid_class to use to read landmask and weights files
 *           verbose - verbosity flag
 *
 *    result: TRUE if successful, or FALSE otherwise, with error message
 *           to stderr
 *
 *----------------------------------------------------------------------*/
bool init_goddard_pmalgo(pmalgo_class *this,
                         const int goddard_recoding_pcent,
			 const char *hemisphere_name,
                         const bool scale_percent_seaice,
                         land_spillover_class *land_spillover,
                         grid_class *grid,
			 int verbose) {

    assert(NULL != this);
    assert(is_goddard_algorithm(this->id));
    
    this->private = (void *)init_goddard(goddard_recoding_pcent,
                                         hemisphere_name,
                                         this->platform,
                                         scale_percent_seaice,
                                         land_spillover,
                                         grid,
                                         verbose);
    if (NULL != this->private) {
        return TRUE;
    } else {
        return FALSE;
    }
					 
}


/*----------------------------------------------------------------------
 *  init_sfc_pmalgo - initialize parameters specific to algorithms that
 *                    require ancillary surface information
 * 
 *    input: this - pmalgo_class object
 *      Only required for CHANG_DEPTH and NSIDC_S3_SWE:
 *           month - month of data acquisition
 *           surface_filename - gridded Chang or NSIDC surface classification
 *      Only required for NSIDC_S11_SWE and NSIDC_S12_SWE:
 *           elevation_filename - gridded elevation (meters)
 *      Only required for NSIDC_S11_SWE and NSIDC_S12_SWE:
 *           high_elevation_model_id - high elevation model number
 *           threshold_elevation_meters - threshold elevation for applying model
 *      Required for all but NSIDC_S6_SWE, NSIDC_S11_SWE and NSIDC_S12_SWE:
 *           landmask_filename - Chang's landmask with bitwise settings for
 *                               permanent ice, land/water, roughness, 
 *                               snow climatology, and snow category
 *      Required for all but NSIDC_S12_SWE:
 *           forest_cover_filename - average forest_cover for the pixel
 *      Required for all:
 *           grid - grid object
 *           verbose - verbosity flag
 *
 *    result: TRUE if successful, or FALSE otherwise, with error message
 *           to stderr
 *
 *----------------------------------------------------------------------*/
bool init_sfc_pmalgo(pmalgo_class *this,
		     int month,
		     const char *surface_filename,
		     const char *forest_cover_filename,
		     const char *landmask_filename,
		     const char *elevation_filename,
		     grid_class *grid,
                     byte1 high_elevation_model_id,
                     int2 threshold_elevation_meters,		         
		     int verbose) {

    assert(NULL != this);
    assert(PMALGO_CHANG_SURFACE == this->id ||
           PMALGO_NSIDC_SURFACE == this->id ||
           PMALGO_CHANG_DEPTH == this->id ||
           PMALGO_NSIDC_A4_SWE == this->id ||
           PMALGO_NSIDC_A10_SWE == this->id ||
           PMALGO_NSIDC_M4_SWE == this->id ||
           PMALGO_NSIDC_S1_SWE == this->id ||
           PMALGO_NSIDC_S3_SWE == this->id ||
           PMALGO_NSIDC_S4_SWE == this->id ||
           PMALGO_NSIDC_S5_SWE == this->id ||
           PMALGO_NSIDC_S6_SWE == this->id ||
           PMALGO_NSIDC_S10_SWE == this->id ||
           PMALGO_NSIDC_S11_SWE == this->id ||
           PMALGO_NSIDC_S12_SWE == this->id);

    this->private = (void *)init_sfc(pmalgo_id_name[this->id],
                                     this->platform,
                                     month,
                                     surface_filename,
                                     forest_cover_filename,
                                     landmask_filename,
                                     elevation_filename,
                                     grid,
                                     high_elevation_model_id,
                                     threshold_elevation_meters,
                                     verbose);
    if (NULL != this->private) {
        return TRUE;
    } else {
        return FALSE;
    }
					 
}


/*----------------------------------------------------------------------
 *  postprocessing_filter_pmalgo - Do postprocessing filtering.  This is any
 *  processing that must be performed after the entire grid of values has been
 *  calculated.  The main processing loop (that will have been performed prior
 *  to calling this function) operates on one pixel at a time, independent of
 *  the surrounding pixels. This processing will do any filtering needed,
 *  where output for a given pixel does depend on the values of the
 *  surrounding pixels.
 *
 *  Currently this step is only used for GODDARD_SEAICE and
 *  BOOTSTRAP_[N/S]_[V02_]SEAICE.  It does the spillover correction.  It also 
 *  does a bunch of special work to recode values to match the Goddard final
 *  process coding logic.
 * 
 *    input: this - pmalgo_class object
 *           grid - grid object
 *           data - pointer data matrix with algorithm output
 *           verbose - verbosity flag
 *
 *    result: TRUE if successful, or FALSE otherwise, with error message
 *           to stderr
 *
 *----------------------------------------------------------------------*/
bool postprocessing_filter_pmalgo( pmalgo_class *this,
                                   grid_class *grid,
                                   float **data,
                                   bool use_climatology,
                                   byte1 **climatology,
                                   bool use_loci,
                                   byte1 **loci,
                                   bool fill_algo_pixels_only,
                                   bool do_user_fill_value,
                                   float fill_value,
                                   int verbose ) {

    /*
     * GODDARD algorithms may need spillover processing.
     * When spillover processing is requested, also do the Goddard recoding.
     */
    if ( is_goddard_algorithm(this->id) ) {

        goddard_class* gc_this;
        gc_this = ((goddard_class *)this->private);

        if ( goddard_spillover( ( goddard_class * )this->private ) ) {

            if ( 0 < verbose ) {
                fprintf( stderr, "> Doing GODDARD spillover correction...\n" );
            }

            if ( !spillover_seaice( 
		     ((goddard_class *)this->private)->land_spillover,
		     ((goddard_class *)this->private)->scale_percent_seaice,
		     PMALGO_GTS_OPEN_WATER_PERCENT,
		     PMALGO_GTS_SCALE_FACTOR,
		     PMALGO_GTS_MAX_ICE_PERCENT,
		     grid,
		     data,
		     verbose ) ) {
                return FALSE;
            }

            if ( 0 < verbose ) {
                fprintf( stderr, "> Doing GODDARD recoding...\n" );
            }

            if ( !goddard_recode_seaice( ( goddard_class * )this->private,
                                         grid,
                                         data,
                                         use_climatology,
                                         climatology,
                                         use_loci,
                                         loci,
                                         fill_algo_pixels_only,
                                         do_user_fill_value,
                                         fill_value,
                                         verbose ) ) {
                return FALSE;
            }
            
        }

    }

    else if (is_bootstrap_algorithm(this->id))
    {
        bootstrap_class* bc_this;
        bc_this = ((bootstrap_class *)this->private);         

        if ( bootstrap_spillover( bc_this ) ) {
        
            if ( 0 < verbose ) {
                fprintf( stderr, "> Doing Bootstrap spillover correction...\n");
            }

            if ( !spillover_seaice (
                     bc_this->land_spillover,
                     bc_this->scale_percent_seaice,
                     PMALGO_BOOTSTRAP_OPEN_WATER_PERCENT,
                     PMALGO_BOOTSTRAP_V2_SCALE_FACTOR,
                     PMALGO_BOOTSTRAP_MAX_ICE_PERCENT,
                     grid,
                     data,
                     verbose ) ) {
                return FALSE;
            }
        }
        
        /* Recode all bootstrap data.  Only scale if desired. */
        if (bc_this->scale_percent_seaice) {
            if ( ! bootstrap_recode_seaice( bc_this, grid, data, verbose ) ) {
                return FALSE;   
            }
        }
        
    }

    return TRUE;

}


/*----------------------------------------------------------------------
 *  close_pmalgo - free storage associated with this object
 * 
 *    input: this - pmalgo_class object
 *
 *    result: deallocate all memory associated with this object
 *
 *----------------------------------------------------------------------*/
void close_pmalgo(pmalgo_class *this) {

    if (NULL == this) return;
    if (NULL != this->direction) free(this->direction);
    if (NULL != this->grid_algorithm) free(this->grid_algorithm);
    if (NULL != this->private) {
        if (is_bootstrap_algorithm(this->id)) {
            close_bootstrap((bootstrap_class *)this->private);
        } else if (is_goddard_algorithm(this->id)) {
            close_goddard((goddard_class *)this->private);
        }
    }
    free(this);
}



/*************************Local functions********************************/


/*----------------------------------------------------------------------
 *  run_nsidc_m1_swe - Calculate Chang SWE from SMMR Tbs
 * 
 *    input : private - not used for this algorithm
 *            data - TB matrix [channel][row][col] (TBs in tenths of Kelvins)
 *            col - column to operate on
 *            row - row to operate on
 * 
 *    output: value - algorithm result (mm SWE > 0)
 *            valid_tbs - TRUE if input TBs for this [row][col] are non-zero,
 *                        FALSE otherwise
 *
 *    result: TRUE if algorithm returns positive SWE, FALSE otherwise
 *            
 *  Reference:
 *    Chang, A. T. C., Foster, J. L., and Hall, D. K. (1987), Nimbus-7 SMMR
 *      derived global snow cover parameters. Ann. Glaciol. 9:39-44.
 *
 *    Chang, A.T.C., J.L. Foster, D.K. Hall, H.W. Powell, Y.L.Chien, 1990 
 *      (revised 7/92). "Nimbus-7 SMMR Derived Global Snow Cover and Snow 
 *      Depth Data Set,"  A PLDS User Publication.
 * *----------------------------------------------------------------------*/
#define CHANG_SWE_DENSITY 3.0
#define CHANG_SWE_COEFFICIENT 1.59
#define CHANG_DEPTH_MIN_CM 2.5
#define CHANG_SWE_MIN_MM (CHANG_SWE_DENSITY * CHANG_DEPTH_MIN_CM)
#define TB_GRADIENT(t1,t2) ((float)((t1) - (t2)))
#define CHANG_SWE(t18h,t37h) \
((CHANG_SWE_DENSITY) * (CHANG_SWE_COEFFICIENT) * (TB_GRADIENT(t18h,t37h)))

static bool run_nsidc_m1_swe (void *private,
                              byte2 **data[PMALGO_NUM_CHANNELS], 
                              int col, int row, float *value, bool *valid_tbs) {

    float swe;

    assert(NULL == private);

    *valid_tbs = FALSE;
  
    if (0 < data[PMALGO_SMMR_18H][row][col] 
        && 0 < data[PMALGO_SMMR_37H][row][col]) {
        *valid_tbs = TRUE;
        swe = CHANG_SWE(data[PMALGO_SMMR_18H][row][col] * TB_SCALE,
                        data[PMALGO_SMMR_37H][row][col] * TB_SCALE);
        if (CHANG_SWE_MIN_MM < swe) {
            *value = swe; 
            return TRUE;
        }
    }
  
    return FALSE;

}

/*----------------------------------------------------------------------
 *  run_nsidc_m4_swe - Forest-corrected Chang SWE from SMMR Tbs
 *                     (I wish I had some version of the Grody-Basist precip tests 
 *                     to use here.)
 * 
 *    input : private - pointer to sfc_class initialized with forest
 *                      cover information
 *            data - TB matrix [channel][row][col] (TBs in tenths of Kelvins)
 *            col - column to operate on
 *            row - row to operate on
 * 
 *    output: value - algorithm result (mm SWE > 0)
 *            valid_tbs - TRUE if input TBs for this [row][col] are non-zero,
 *                        FALSE otherwise
 *
 *    result: TRUE if algorithm returns positive SWE, 
 *            FALSE otherwise
 *           
 *  References: See references for run_nsidc_m1_swe.
 *              Forest correction reference is:
 *  Chang, A.T.C., J.L. Foster, and D.K. Hall. (1996). Effects of forest on
 *    the snow parameters derived from microwave measurements during the 
 *    BOREAS winter field campaign. Hydrological Process, Vol. 10, 1565-1574.
 *    
 *  Note: the user should limit input forest fractions to the range 0.0 (no
 *    forest) to 0.50 (50% forest).  For purposes of testing alternatives, 
 *    nothing will be done here to enforce this limit.  R. Armstrong, 
 *    personal communication.  This is our interpretation of the plots in
 *    Chang et al.'s paper, that only consider f for values up to 0.5.
 *    Also, note that using higher f values, e.g. 90%, would increase the
 *    SWE value by an unrealistic factor of 10.  Limiting f to 0.5 will
 *    yield a maximum correction of a factor of 2.
 *
 *----------------------------------------------------------------------*/
#define NSIDC4_SMMR(t18h,t36h,forest_cover) \
(CHANG_SWE((t18h),(t36h)) / (1. - (forest_cover)))

static bool run_nsidc_m4_swe (void *private,
			      byte2 **data[PMALGO_NUM_CHANNELS], 
			      int col, int row, float *value, bool *valid_tbs) {

    float swe;

    assert(NULL != private);

    *valid_tbs = FALSE;
  
    if (0 < data[PMALGO_SMMR_18H][row][col] 
        && 0 < data[PMALGO_SMMR_37H][row][col]) {
        *valid_tbs = TRUE;
        swe = NSIDC4_SMMR(data[PMALGO_SMMR_18H][row][col] * TB_SCALE,
                          data[PMALGO_SMMR_37H][row][col] * TB_SCALE,
                          forest_cover_sfc((sfc_class *)private,col,row));
      
        if (CHANG_SWE_MIN_MM < swe) {
            *value = swe;
            return TRUE;
        }
    }

    return FALSE;

}


/*----------------------------------------------------------------------
 *  run_nsidc_s0_swe - Calculate Chang SWE from SSM/I Tbs, without
 *                       correcting the SSM/I channels.  The only reason 
 *                       we're doing this is to verify that the correction
 *                       is appropriate in the long time series
 * 
 *    input : private - not used for this algorithm
 *            data - TB matrix [channel][row][col] (TBs in tenths of Kelvins)
 *            col - column to operate on
 *            row - row to operate on
 * 
 *    output: value - algorithm result (mm SWE > 0)
 *            valid_tbs - TRUE if input TBs for this [row][col] are non-zero,
 *                        FALSE otherwise
 *
 *    result: TRUE if algorithm returns positive SWE, FALSE otherwise
 *            
 *  Reference:
 *    Chang, A. T. C., Foster, J. L., and Hall, D. K. (1987), Nimbus-7 SMMR
 *      derived global snow cover parameters. Ann. Glaciol. 9:39-44.
 *
 *    Chang, A.T.C., J.L. Foster, D.K. Hall, H.W. Powell, Y.L.Chien, 1990 
 *      (revised 7/92). "Nimbus-7 SMMR Derived Global Snow Cover and Snow 
 *      Depth Data Set,"  A PLDS User Publication.
 * *----------------------------------------------------------------------*/
static bool run_nsidc_s0_swe (void *private,
			      byte2 **data[PMALGO_NUM_CHANNELS], 
			      int col, int row, float *value, bool *valid_tbs) {

    float swe;

    assert(NULL == private);

    *valid_tbs = FALSE;
  
    if (0 < data[PMALGO_SSMI_19H][row][col] 
        && 0 < data[PMALGO_SSMI_37H][row][col]) {
        *valid_tbs = TRUE;
        swe = CHANG_SWE(data[PMALGO_SSMI_19H][row][col] * TB_SCALE,
                        data[PMALGO_SSMI_37H][row][col] * TB_SCALE);
        if (CHANG_SWE_MIN_MM < swe) {
            *value = swe; 
            return TRUE;
        }
    }
  
    return FALSE;

}


/*----------------------------------------------------------------------
 *  run_nsidc_s1_swe - Calculate SWE with NSIDC1 algorithm (Al Chang's SMMR
 *    algorithm modified to use adjusted SSM/I Tbs
 * 
 *    input : private - pointer to sfc_class initialized with optional platform id
 *            data - TB matrix [channel][row][col] (TBs in tenths of Kelvins)
 *            col - column to operate on
 *            row - row to operate on
 * 
 *    output: value - algorithm result (mm SWE > 0) 
 *            valid_tbs - TRUE if input TBs for this [row][col] are non-zero,
 *                        FALSE otherwise
 *
 *    result: TRUE if algorithm returns positive SWE, FALSE otherwise
 *           
 *  Note:
 *    SMMR-SSM/I conversion used:
 *       SMMR_18H = (SSMI_19H - 6.0 Kelvins)
 *       SMMR_37H = (SSMI_37H - 1.0 Kelvins)
 *
 *  References:
 *    The SMMR to SSMI  conversion is based on unpublished research by 
 *       R. Armstrong, National Snow and Ice Data Center.  From e-mail 
 *       correspondence from Armstrong to M. J. Brodzik, Feb 11, 1999:
 *
 *    "The offsets used to adapt the Chang SMMR snow algorithm to SSM/I data
 *     were based on empirical studies involving comparisons with snow extent
 *     information derived from visible satellite data as well as on results
 *     obtained in comparison studies  undertaken during the overlap period 
 *     (July 7, 1987 to August 20, 1987) of the SMMR and SSM/I instruments 
 *     (Jezek at al.,1991; Rott, 1989).
 *
 *     Jezek, K.C., C. Merry, D. Cavalieri, S. Grace, J. Bedner,, D. Wilson, 
 *     and D. Lampkin, 1991. Comparison between SMMR and SSM/I passive 
 *     microwave data collected over the Antarctic ice sheet, Byrd Polar 
 *     Research Center, The Ohio State University, Technical Report Number 
 *     91-03. 
 *
 *     Rott, H. 1989. Multispectral microwave signatures of the Antarctic ice
 *     sheet. In:Microwave Radiometry and Remote Sensing Applications, Ed. P.
 *     Pampaloni, VSP, Utrecht, NL, pp.89-101."
 *     
 *    See also:
 *    Chang, A. T. C., Foster, J. L., and Hall, D. K. (1987), Nimbus-7 SMMR
 *      derived global snow cover parameters. Ann. Glaciol. 9:39-44.
 *
 *    Chang, A.T.C., J.L. Foster, D.K. Hall, H.W. Powell, Y.L.Chien, 1990 
 *      (revised 7/92). "Nimbus-7 SMMR Derived Global Snow Cover and Snow 
 *      Depth Data Set,"  A PLDS User Publication.
 *
 *----------------------------------------------------------------------*/
#define NSIDC_19H_OFFSET 6.0
#define NSIDC_37H_OFFSET 1.0
#define NSIDC1_SWE(t19h,t37h) CHANG_SWE(((t19h) - (NSIDC_19H_OFFSET)),\
((t37h) - (NSIDC_37H_OFFSET)))

static bool run_nsidc_s1_swe (void *private,
			      byte2 **data[PMALGO_NUM_CHANNELS], 
			      int col, int row, float *value, bool *valid_tbs) {

    float swe;
    float t19h, t37h;

    assert(NULL != private);

    *valid_tbs = FALSE;
  
    if (0 < data[PMALGO_SSMI_19H][row][col] 
        && 0 < data[PMALGO_SSMI_37H][row][col]) {

        *valid_tbs = TRUE;

        fetch_swe_tbs( private, data[PMALGO_SSMI_19H][row][col], data[PMALGO_SSMI_37H][row][col],
                       &t19h, &t37h );

        swe = NSIDC1_SWE( t19h, t37h );

        if (CHANG_SWE_MIN_MM < swe) {
            *value = swe;
            return TRUE;
        }
    }

    return FALSE;

}


/*----------------------------------------------------------------------
 *  run_nsidc_s5_swe - Calculate SWE with NSIDC1 algorithm (Al Chang's SMMR
 *    algorithm modified to use adjusted SSM/I Tbs, plus the Grody-Basist
 *    precip tests)
 * 
 *    input : private - pointer to sfc_class initialized with optional platform id
 *            data - TB matrix [channel][row][col] (TBs in tenths of Kelvins)
 *            col - column to operate on
 *            row - row to operate on
 * 
 *    output: value - algorithm result (mm SWE > 0) 
 *            valid_tbs - TRUE if input TBs for this [row][col] are non-zero,
 *                        FALSE otherwise
 *
 *    result: TRUE if algorithm returns positive SWE or precip, 
 *            FALSE otherwise
 *           
 *  References: See references for run_nsidc1_swe and run_gb_snow
 *
 *----------------------------------------------------------------------*/
static bool run_nsidc_s5_swe (void *private,
			      byte2 **data[PMALGO_NUM_CHANNELS], 
			      int col, int row, float *value, bool *valid_tbs) {
    float swe;
    gb_return_code return_code;
    gb_surface_condition surface_condition;

    assert(NULL != private);

    *valid_tbs = FALSE;
  
    if (0 < data[PMALGO_SSMI_19V][row][col] 
        && 0 < data[PMALGO_SSMI_19H][row][col]
        && 0 < data[PMALGO_SSMI_22V][row][col]
        && 0 < data[PMALGO_SSMI_37V][row][col]
        && 0 < data[PMALGO_SSMI_37H][row][col]
        && 0 < data[PMALGO_SSMI_85V][row][col]
        && 0 < data[PMALGO_SSMI_85H][row][col]) {
        *valid_tbs = TRUE;
        return_code = snow_gb(&surface_condition,
                              data[PMALGO_SSMI_19V][row][col] * TB_SCALE,
                              data[PMALGO_SSMI_19H][row][col] * TB_SCALE,
                              data[PMALGO_SSMI_22V][row][col] * TB_SCALE,
                              data[PMALGO_SSMI_37V][row][col] * TB_SCALE,
                              data[PMALGO_SSMI_85V][row][col] * TB_SCALE,
                              data[PMALGO_SSMI_85H][row][col] * TB_SCALE,
                              0);
        if (GB_PC1 == surface_condition) {
            *value = PMALGO_GB_PC1;
            return TRUE;
        } else if (GB_PC2 == surface_condition) {
            *value = PMALGO_GB_PC2;
            return TRUE;
        } else {
            swe = NSIDC1_SWE(data[PMALGO_SSMI_19H][row][col] * TB_SCALE,
                             data[PMALGO_SSMI_37H][row][col] * TB_SCALE);
      
            if (CHANG_SWE_MIN_MM < swe) {
                *value = swe;
                return TRUE;
            }
        }
    }

    return FALSE;

}

/*----------------------------------------------------------------------
 *  run_nsidc_s6_swe - Calculate SWE with forest-corrected NSIDC1 algorithm 
 *    (Al Chang's SMMR algorithm modified to use adjusted SSM/I Tbs, adjusted
 *    for places with forest-cover, plus the Grody-Basist precip tests)
 * 
 *    input : private - pointer to sfc_class initialized with forest
 *                      cover information
 *            data - TB matrix [channel][row][col] (TBs in tenths of Kelvins)
 *            col - column to operate on
 *            row - row to operate on
 * 
 *    output: value - algorithm result (mm SWE > 0,
 *                    PMALGO_GB_PC1, PMALGO_GB_PC2) 
 *            valid_tbs - TRUE if input TBs for this [row][col] are non-zero,
 *                        FALSE otherwise
 *
 *    result: TRUE if algorithm returns positive SWE or precip, 
 *            FALSE otherwise
 *           
 *  References: See references for run_nsidc_s1_swe and run_gb_snow.
 *              Forest correction reference is:
 *  Chang, A.T.C., J.L. Foster, and D.K. Hall. (1996). Effects of forest on
 *    the snow parameters derived from microwave measurements during the 
 *    BOREAS winter field campaign. Hydrological Process, Vol. 10, 1565-1574.
 *    
 *  Note: the user should limit input forest fractions to the range 0.0 (no
 *    forest) to 0.50 (50% forest).  For purposes of testing alternatives, 
 *    nothing will be done here to enforce this limit.  R. Armstrong, 
 *    personal communication.  This is our interpretation of the plots in
 *    Chang et al.'s paper, that only consider f for values up to 0.5.
 *    Also, note that using higher f values, e.g. 90%, would increase the
 *    SWE value by an unrealistic factor of 10.  Limiting f to 0.5 will
 *    yield a maximum correction of a factor of 2.
 *
 *----------------------------------------------------------------------*/
#define NSIDC6_SWE(t19h,t37h,forest_cover) \
(NSIDC1_SWE((t19h),(t37h)) / (1. - (forest_cover)))

static bool run_nsidc_s6_swe (void *private,
			      byte2 **data[PMALGO_NUM_CHANNELS], 
			      int col, int row, float *value, bool *valid_tbs) {

    float swe;
    gb_return_code return_code;
    gb_surface_condition surface_condition;

    assert(NULL != private);

    *valid_tbs = FALSE;
  
    if (0 < data[PMALGO_SSMI_19V][row][col] 
        && 0 < data[PMALGO_SSMI_19H][row][col]
        && 0 < data[PMALGO_SSMI_22V][row][col]
        && 0 < data[PMALGO_SSMI_37V][row][col]
        && 0 < data[PMALGO_SSMI_37H][row][col]
        && 0 < data[PMALGO_SSMI_85V][row][col]
        && 0 < data[PMALGO_SSMI_85H][row][col]) {
        *valid_tbs = TRUE;
        return_code = snow_gb(&surface_condition,
                              data[PMALGO_SSMI_19V][row][col] * TB_SCALE,
                              data[PMALGO_SSMI_19H][row][col] * TB_SCALE,
                              data[PMALGO_SSMI_22V][row][col] * TB_SCALE,
                              data[PMALGO_SSMI_37V][row][col] * TB_SCALE,
                              data[PMALGO_SSMI_85V][row][col] * TB_SCALE,
                              data[PMALGO_SSMI_85H][row][col] * TB_SCALE,
                              0);
        if (GB_PC1 == surface_condition) {
            *value = PMALGO_GB_PC1;
            return TRUE;
        } else if (GB_PC2 == surface_condition) {
            *value = PMALGO_GB_PC2;
            return TRUE;
        } else {
            swe = NSIDC6_SWE(data[PMALGO_SSMI_19H][row][col] * TB_SCALE,
                             data[PMALGO_SSMI_37H][row][col] * TB_SCALE,
                             forest_cover_sfc((sfc_class *)private,col,row));
      
            if (CHANG_SWE_MIN_MM < swe) {
                *value = swe;
                return TRUE;
            }
        }
    }

    return FALSE;

}


/*----------------------------------------------------------------------
 *  run_nsidc_a0_swe - Calculate SWE from AMSR Tbs, assuming that AMSR
 *                     frequencies are the same as SSM/I frequencies.
 *                     Eventually we should do a correction (or justify that
 *                     it isn't necessary).
 * 
 *    input : private - not used for this algorithm
 *            data - TB matrix [channel][row][col] (TBs in tenths of Kelvins)
 *            col - column to operate on
 *            row - row to operate on
 * 
 *    output: value - algorithm result (mm SWE > 0)
 *            valid_tbs - TRUE if input TBs for this [row][col] are non-zero,
 *                        FALSE otherwise
 *
 *    result: TRUE if algorithm returns positive SWE, FALSE otherwise
 *            
 *  Reference:
 *    See refs for nsidc_s1 algorithm
 * *----------------------------------------------------------------------*/
static bool run_nsidc_a0_swe (void *private,
			      byte2 **data[PMALGO_NUM_CHANNELS], 
			      int col, int row, float *value, bool *valid_tbs) {

    float swe;

    assert(NULL == private);

    *valid_tbs = FALSE;
  
    if (0 < data[PMALGO_AMSR_18H][row][col] 
        && 0 < data[PMALGO_AMSR_36H][row][col]) {
        *valid_tbs = TRUE;
        swe = NSIDC1_SWE(data[PMALGO_AMSR_18H][row][col] * TB_SCALE,
                         data[PMALGO_AMSR_36H][row][col] * TB_SCALE);
        if (CHANG_SWE_MIN_MM < swe) {
            *value = swe; 
            return TRUE;
        }
    }
  
    return FALSE;

}


/*----------------------------------------------------------------------
 *  run_nsidc_a4_swe - Calculate SWE from AMSR Tbs, assuming that
 *                     AMSR frequencies are the same as SSM/I frequencies,
 *                     using forest-corrected NSIDC1
 * 
 *    input : private - pointer to sfc_class initialized 
 *                      with forest cover information
 *            data - TB matrix [channel][row][col] (TBs in tenths of Kelvins)
 *            col - column to operate on
 *            row - row to operate on
 * 
 *    output: value - algorithm result (mm SWE > 0)
 *            valid_tbs - TRUE if input TBs for this [row][col] are non-zero,
 *                        FALSE otherwise
 *
 *    result: TRUE if TBs were non-zero, FALSE otherwise
 *            
 *  Reference: See references for run_nsidc_s4_swe. 
 *
 *----------------------------------------------------------------------*/
static bool run_nsidc_a4_swe (void *private,
			      byte2 **data[PMALGO_NUM_CHANNELS], 
			      int col, int row, float *value, bool *valid_tbs) {


    float swe;

    assert(NULL != private);

    if (0 == data[PMALGO_AMSR_18H][row][col] 
        || 0 == data[PMALGO_AMSR_36H][row][col]) {

        *valid_tbs = FALSE;
        return FALSE;

    }

    *valid_tbs = TRUE;
  
    swe = NSIDC6_SWE(data[PMALGO_AMSR_18H][row][col] * TB_SCALE,
                     data[PMALGO_AMSR_36H][row][col] * TB_SCALE,
                     forest_cover_sfc((sfc_class *)private,col,row));
      
    if (CHANG_SWE_MIN_MM < swe) {
        *value = swe;
        return TRUE;
    }

    return FALSE;

}



/*----------------------------------------------------------------------
 *  run_gw_swe - Calculate SWE with Goodison-Walker algorithm
 * 
 *    input : private - not used for this algorithm
 *            data - TB matrix [channel][row][col] (TBs in tenths of Kelvins)
 *            col - column to operate on
 *            row - row to operate on
 * 
 *    output: value - algorithm result (mm SWE > 0) 
 *            valid_tbs - TRUE if input TBs for this [row][col] are non-zero,
 *                        FALSE otherwise
 *
 *    result: TRUE if algorithm returns positive SWE, FALSE otherwise
 *
 *  Reference:
 *    Goodison, B. E., and Walker, A. E. (1994), Canadian development and 
 *      use of snow cover information from passive microwave satellite data.
 *      ESA/NASA International Workshop, pp. 245-262.
 *
 *    The actual algorithm used here was given to M. J. Brodzik by e-mail
 *    from Anne Walker, dated Mar 31 1997.
 *
 *----------------------------------------------------------------------*/
#define GOODISON_WALKER_SWE(t19v,t37v) \
        ((-2.74 * (float)((t37v) - (t19v))) - 20.7)

static bool run_gw_swe (void *private,
			byte2 **data[PMALGO_NUM_CHANNELS], 
			int col, int row, float *value, bool *valid_tbs) {
    float swe;

    assert(NULL == private);

    *valid_tbs = FALSE;

    if (0 < data[PMALGO_SSMI_19V][row][col] 
        && 0 < data[PMALGO_SSMI_37V][row][col]) {
        *valid_tbs = TRUE;
        swe = GOODISON_WALKER_SWE(data[PMALGO_SSMI_19V][row][col] * TB_SCALE,
                                  data[PMALGO_SSMI_37V][row][col] * TB_SCALE);
        if (0.0 < swe) {
            *value = swe;
            return TRUE;
        }
    } 

    return FALSE;

}


/*----------------------------------------------------------------------
 *  run_gw_wet_swe - Calculate SWE with Goodison-Walker algorithm, and
 *                   return wet snow indicator
 * 
 *    input : private - not used for this algorithm
 *            data - TB matrix [channel][row][col] (TBs in tenths of Kelvins)
 *            col - column to operate on
 *            row - row to operate on
 * 
 *    output: value - algorithm result (mm SWE > 0, or PMALGO_WET_SNOW_FLAG
 *              if wet snow)
 *            valid_tbs - TRUE if input TBs for this [row][col] are non-zero,
 *                        FALSE otherwise
 *
 *    result: TRUE if algorithm returns SWE or wet snow flag, FALSE otherwise
 *
 *  Reference:
 *    Goodison, B. E., and Walker, A. E. (1994), Canadian development and 
 *      use of snow cover information from passive microwave satellite data.
 *      ESA/NASA International Workshop, pp. 245-262.
 *
 *----------------------------------------------------------------------*/
#define WALKER_WET_SNOW(t37v,t37h) (10. < (float)((t37v) - (t37h)))

static bool run_gw_wet_swe (void *private,
			    byte2 **data[PMALGO_NUM_CHANNELS], 
			    int col, int row, float *value, bool *valid_tbs) {
  
    float swe;

    assert(NULL == private);

    *valid_tbs = FALSE;

    /* 
     * Run Goodison-Walker SWE first, and then 
     * do wet snow test for any swe <= 0
     */
    if (0 < data[PMALGO_SSMI_19V][row][col]
        && 0 < data[PMALGO_SSMI_37V][row][col]
        && 0 < data[PMALGO_SSMI_37H][row][col]) {
        *valid_tbs = TRUE;
        swe = GOODISON_WALKER_SWE(data[PMALGO_SSMI_19V][row][col] * TB_SCALE,
                                  data[PMALGO_SSMI_37V][row][col] * TB_SCALE);
        if (0.0 < swe) {
            *value = swe;
            return TRUE;
        } else if (WALKER_WET_SNOW(data[PMALGO_SSMI_37V][row][col] * TB_SCALE,
                                   data[PMALGO_SSMI_37H][row][col] * TB_SCALE)) {
            *value = PMALGO_WET_SNOW_FLAG;
            return TRUE;
        }
    }

    return FALSE;

}
  

/*----------------------------------------------------------------------
 *  run_gw_modwet_swe - Calculate SWE with Goodison-Walker algorithm, and
 *                      return NSIDC-modified wet snow indicator (includes
 *                      a test for "reasonably cold" Tbs)
 * 
 *    input : private - not used for this algorithm
 *             data - TB matrix [channel][row][col] (TBs in tenths of Kelvins)
 *            col - column to operate on
 *            row - row to operate on
 * 
 *    output: value - algorithm result (mm SWE > 0, 
 *                    or PMALGO_WET_SNOW_FLAG if wet snow) 
 *            valid_tbs - TRUE if input TBs for this [row][col] are non-zero,
 *                        FALSE otherwise
 *
 *    result: TRUE if algorithm returns SWE or wet snow flag, FALSE otherwise
 *
 *  Reference:
 *    Goodison, B. E., and Walker, A. E. (1994), Canadian development and 
 *      use of snow cover information from passive microwave satellite data.
 *      ESA/NASA International Workshop, pp. 245-262.
 *
 *----------------------------------------------------------------------*/
#define FREEZING_THRESHOLD_KELVINS 255
#define MODIFIED_WALKER_WET_SNOW(t37v,t37h) \
         (WALKER_WET_SNOW(t37v,t37h) && \
	 FREEZING_THRESHOLD_KELVINS > (t37v) && \
	 FREEZING_THRESHOLD_KELVINS > (t37h))

static bool run_gw_modwet_swe (void *private,
			       byte2 **data[PMALGO_NUM_CHANNELS], 
			       int col, int row, float *value, 
			       bool *valid_tbs) {

    float swe;

    assert(NULL == private);

    *valid_tbs = FALSE;

    /* 
     * Run Goodison-Walker SWE first, and then 
     * do the modified wet snow test for any swe <= 0
     */
    if (0 < data[PMALGO_SSMI_19V][row][col]
        && 0 < data[PMALGO_SSMI_37V][row][col]
        && 0 < data[PMALGO_SSMI_37H][row][col]) {
        *valid_tbs = TRUE;
        swe = GOODISON_WALKER_SWE(data[PMALGO_SSMI_19V][row][col] * TB_SCALE, 
                                  data[PMALGO_SSMI_37V][row][col] * TB_SCALE);
    
        if (0.0 < swe) {
            *value = swe;
            return TRUE;
        } else if (MODIFIED_WALKER_WET_SNOW(data[PMALGO_SSMI_37V][row][col] 
                                            * TB_SCALE,
                                            data[PMALGO_SSMI_37H][row][col]
                                            * TB_SCALE)) {
            *value = PMALGO_WET_SNOW_FLAG;
            return TRUE;
        }
    }

    return FALSE;

}
  

/*----------------------------------------------------------------------
 *  run_hb_dry_comb - Calculate Hilbrunner's TB COMB parameter, for dry snow
 *                    conditions
 * 
 *    input : private - not used for this algorithm
 *            data - TB matrix [channel][row][col] (TBs in tenths of Kelvins)
 *            col - column to operate on
 *            row - row to operate on
 * 
 *    output: value - algorithm COMB result (Kelvins)
 *            valid_tbs - TRUE if input TBs for this [row][col] are non-zero,
 *                        FALSE otherwise
 *
 *    result: TRUE if algorithm returns positive COMB, FALSE otherwise
 *
 *  Reference:
 *    Hiltbrunner, Daniel. Land Surface Temperature and Microwave Emissivity 
 *       from SSM/I DAta. Ph.D. diss., University of Bern.
 *    Maetzler, Christian.  E-mail to Richard L. Armstrong, 4 Jun 1999.
 *
 *----------------------------------------------------------------------*/
#define HB_TCOMBA(ch1v,ch1h,ch2v,ch2h) ((float)((ch1v)-(ch1h)+(ch2v)-(ch2h)))
#define HB_TCOMBB(ch1v,ch2v) (2.*((float)((ch1v)-(ch2v))))

static bool run_hb_dry_comb (void *private,
			     byte2 **data[PMALGO_NUM_CHANNELS], 
			     int col, int row, float *value, 
			     bool *valid_tbs) {

    float comb, B;

    assert(NULL == private);

    *valid_tbs = FALSE;

    if (0 < data[PMALGO_SSMI_19V][row][col] 
        && 0 < data[PMALGO_SSMI_19H][row][col]
        && 0 < data[PMALGO_SSMI_37V][row][col]
        && 0 < data[PMALGO_SSMI_37H][row][col]) {
        *valid_tbs = TRUE;
        B = HB_TCOMBB(data[PMALGO_SSMI_19V][row][col] * TB_SCALE,
                      data[PMALGO_SSMI_37V][row][col] * TB_SCALE);
        comb = (0.0 >= B ? 0.0 :
                B + HB_TCOMBA(data[PMALGO_SSMI_19V][row][col] * TB_SCALE,
                              data[PMALGO_SSMI_19H][row][col] * TB_SCALE,
                              data[PMALGO_SSMI_37V][row][col] * TB_SCALE,
                              data[PMALGO_SSMI_37H][row][col] * TB_SCALE));
	    
        if (0.0 < comb) {
            *value = comb;
            return TRUE;
        }
    } 

    return FALSE;

}
  
  

/*----------------------------------------------------------------------
 *  run_hb_fresh_comb - Calculate Hilbrunner's TB COMB parameter, for fresh
 *                      snow conditions
 * 
 *    input : private - not used for this algorithm
 *            data - TB matrix [channel][row][col] (TBs in tenths of Kelvins)
 *            col - column to operate on
 *            row - row to operate on
 * 
 *    output: value - algorithm COMB result (Kelvins)
 *            valid_tbs - TRUE if input TBs for this [row][col] are non-zero,
 *                        FALSE otherwise
 *
 *    result: TRUE if algorithm returns positive COMB, FALSE otherwise
 *
 *  Reference:
 *    Hiltbrunner, Daniel. Land Surface Temperature and Microwave Emissivity 
 *       from SSM/I DAta. Ph.D. diss., University of Bern.
 *    Maetzler, Christian.  E-mail to Richard L. Armstrong, 4 Jun 1999.
 *
 *----------------------------------------------------------------------*/
static bool run_hb_fresh_comb (void *private,
			       byte2 **data[PMALGO_NUM_CHANNELS], 
			       int col, int row, float *value, 
			       bool *valid_tbs) {

    float comb, B;

    assert(NULL == private);

    *valid_tbs = FALSE;

    if (0 < data[PMALGO_SSMI_85V][row][col] 
        && 0 < data[PMALGO_SSMI_85H][row][col]
        && 0 < data[PMALGO_SSMI_37V][row][col]
        && 0 < data[PMALGO_SSMI_37H][row][col]) {
        *valid_tbs = TRUE;
        B = HB_TCOMBB(data[PMALGO_SSMI_85V][row][col] * TB_SCALE,
                      data[PMALGO_SSMI_37V][row][col] * TB_SCALE);
        comb = (0.0 >= B ? 0.0 :
                B + HB_TCOMBA(data[PMALGO_SSMI_85V][row][col] * TB_SCALE,
                              data[PMALGO_SSMI_85H][row][col] * TB_SCALE,
                              data[PMALGO_SSMI_37V][row][col] * TB_SCALE,
                              data[PMALGO_SSMI_37H][row][col] * TB_SCALE));
        if (0.0 < comb) {
            *value = comb;
            return TRUE;
        }
    } 

    return FALSE;

}
  

/*----------------------------------------------------------------------
 *  run_tait1_swe - Calculate Tait's Category 1 (simplest conditions) SWE 
 *                  (non-forested, non-complex terrain, no melting snow, 
 *                  no depth hoar)
 *
 *    input : private - not used for this algorithm
 *            data - TB matrix [channel][row][col] (TBs in tenths of Kelvins)
 *            col - column to operate on
 *            row - row to operate on
 * 
 *    output: value - algorithm result (mm SWE > 0) 
 *            valid_tbs - TRUE if input TBs for this [row][col] are non-zero,
 *                        FALSE otherwise
 *
 *    result: TRUE if algorithm returns positive SWE, FALSE otherwise
 *
 *  Reference:
 *    Tait, Andrew. Estimation of snow depth and snow water equivalent using
 *       passive microwave radiation data. Ph.D. diss., University of Colorado.
 *
 *----------------------------------------------------------------------*/
#define TAIT1_SWE(t19v,t37h,t85h) (29.2 + 1.7 * ((float)((t19v) - (t37h))) \
				  - 0.9 * ((float)((t19v) - (t85h))))

static bool run_tait1_swe (void *private,
			   byte2 **data[PMALGO_NUM_CHANNELS], 
			   int col, int row, float *value, 
			   bool *valid_tbs) {

    float swe;

    assert(NULL == private);

    *valid_tbs = FALSE;

    if (0 < data[PMALGO_SSMI_19V][row][col] 
        && 0 < data[PMALGO_SSMI_37H][row][col]
        && 0 < data[PMALGO_SSMI_85H][row][col]) {
        *valid_tbs = TRUE;
        swe = TAIT1_SWE(data[PMALGO_SSMI_19V][row][col] * TB_SCALE,
                        data[PMALGO_SSMI_37H][row][col] * TB_SCALE,
                        data[PMALGO_SSMI_85H][row][col] * TB_SCALE);
        if (0.0 < swe) {
            *value = swe;
            return TRUE;
        }
    } 

    return FALSE;

}
  
/*----------------------------------------------------------------------
 *  run_chang_depth - Calculate Chang's snow depth.
 *                    This is Chang's algorithm for AMSR, adjusted for
 *                    SSM/I Tbs, and currently available ancillary data
 * 
 *    input : private - pointer to sfc_class initialized 
 *                      with ancillary surface information
 *            data - TB matrix [channel][row][col] (TBs in tenths of Kelvins)
 *            col - column to operate on
 *            row - row to operate on
 * 
 *    output: value - algorithm result (PMALGO_CHANG_NO_SNOW, cm depth > 2.5, 
 *                    PMALGO_CHANG_WET_SNOW, PMALGO_CHANG_UNKNOWN_SNOW)
 *            valid_tbs - TRUE if input TBs for this [row][col] are non-zero,
 *                        FALSE otherwise
 *
 *    result: TRUE if TBs were non-zero, FALSE otherwise
 *            
 *  Reference:
 *    E-mail and C program, granule.c, Chang to Armstrong, Armstrong sent to
 *    brodzik, Jan 4, 2001; Revised version e-mail to Brodzik, 
 *    March 19, 2001.
 * 
 *
 *----------------------------------------------------------------------*/
#define CHANG_WET_SOIL_SWE(t19v,t37v) (5.0 * ((t19v) - (t37v)))
#define CHANG_DRY_SOIL_SWE(coef,forest_cover,t19v,t37v) \
(coef * ((t19v) - (t37v) - 5.0) / (1. - (forest_cover)))
#define CHANG_WET_SOIL_DEPTH(t19v,t37v) \
(CHANG_WET_SOIL_SWE((t19v),(t37v)) / CHANG_SWE_DENSITY)
#define CHANG_DRY_SOIL_DEPTH(coef,forest_cover,t19v,t37v) \
(CHANG_DRY_SOIL_SWE((coef),(forest_cover),(t19v),(t37v)) / CHANG_SWE_DENSITY)
static bool run_chang_depth (void *private,
			     byte2 **data[PMALGO_NUM_CHANNELS], 
			     int col, int row, float *value, bool *valid_tbs) {

    sfc_type surface;
    float t19v,t37v, t37h;

    assert(NULL != private);


    if (0 == data[PMALGO_SSMI_19V][row][col] 
        || 0 == data[PMALGO_SSMI_37V][row][col]
        || 0 == data[PMALGO_SSMI_37H][row][col]) {
        *valid_tbs = FALSE;
        return FALSE;
    }

    *valid_tbs = TRUE;
  
    /* Do TB scaling once */
    t19v = data[PMALGO_SSMI_19V][row][col] * TB_SCALE;
    t37v = data[PMALGO_SSMI_37V][row][col] * TB_SCALE;
    t37h = data[PMALGO_SSMI_37H][row][col] * TB_SCALE;

    /* Retrieve the surface category */
    surface = retrieve_sfc((sfc_class *)private,col,row);
    if (SFC_WET_SOIL == surface) {
        *value = CHANG_WET_SOIL_DEPTH(t19v,t37v);
        *value = (*value > CHANG_DEPTH_MIN_CM ? *value : PMALGO_CHANG_NO_SNOW);

    } else if (SFC_DRY_SOIL == surface) {
        *value = 
            CHANG_DRY_SOIL_DEPTH(swe_coef_sfc((sfc_class *)private,col,row),
                                 forest_cover_sfc((sfc_class *)private,col,row),
                                 t19v,t37v);
        *value = (*value > CHANG_DEPTH_MIN_CM ? *value : PMALGO_CHANG_NO_SNOW);

    } else if (SFC_NORMAL_NOSNOW == surface || 
               SFC_WATER == surface || 
               SFC_SNOW_IMPOSSIBLE == surface ||
               SFC_ICE == surface ||
               SFC_TOO_WARM == surface ||
               SFC_OUTOFRANGE == surface) {
        *value = (float)PMALGO_CHANG_NO_SNOW;
    } else {
        *value = (float)PMALGO_CHANG_UNKNOWN_SNOW;
    }

    /* 
     * Use Walker's test for wet snow, 
     * only in places where SWE was attempted in the first place 
     */
    if (SFC_WET_SOIL == surface || SFC_DRY_SOIL == surface) {
        if ((float)PMALGO_CHANG_NO_SNOW == *value) {
            if (MODIFIED_WALKER_WET_SNOW(t37v,t37h)) {
                *value = (float)PMALGO_CHANG_WET_SNOW;
                return TRUE;
            }
        }
    }

    return TRUE;

}

/*----------------------------------------------------------------------
 *  run_nsidc_s3_swe - NSIDC's modifications to Chang's AMSR decision tree,
 *                     to calculate SWE from H-polarized SSM/I Tbs
 * 
 *    input : private - pointer to sfc_class initialized 
 *                      with ancillary surface information
 *            data - TB matrix [channel][row][col] (TBs in tenths of Kelvins)
 *            col - column to operate on
 *            row - row to operate on
 * 
 *    output: value - algorithm result (PMALGO_CHANG_NO_SNOW, mm SWE > 0, 
 *                    PMALGO_CHANG_WET_SNOW, PMALGO_CHANG_UNKNOWN_SNOW)
 *            valid_tbs - TRUE if input TBs for this [row][col] are non-zero,
 *                        FALSE otherwise
 *
 *    result: TRUE if TBs were non-zero, FALSE otherwise
 *            
 *  Reference:
 *    E-mail and C program, granule.c, Chang to Armstrong, Armstrong sent to
 *    brodzik, Jan 4, 2001.
 *    How to run H-pol algorithm decided verballe, Armstrong and Brodzik, 
 *    Feb. 6, 2001.
 *
 *----------------------------------------------------------------------*/
static bool run_nsidc_s3_swe (void *private,
			      byte2 **data[PMALGO_NUM_CHANNELS], 
			      int col, int row, float *value, bool *valid_tbs) {

    sfc_type surface;
    float t19h, t37v, t37h;

    assert(NULL != private);

    if (0 == data[PMALGO_SSMI_19H][row][col] 
        || 0 == data[PMALGO_SSMI_37V][row][col]
        || 0 == data[PMALGO_SSMI_37H][row][col]) {

        *valid_tbs = FALSE;
        return FALSE;

    }

    *valid_tbs = TRUE;
  
    /* Do TB scaling once */
    t19h = data[PMALGO_SSMI_19H][row][col] * TB_SCALE;
    t37v = data[PMALGO_SSMI_37V][row][col] * TB_SCALE;
    t37h = data[PMALGO_SSMI_37H][row][col] * TB_SCALE;

    /* Retrieve the surface category */
    surface = retrieve_sfc((sfc_class *)private,col,row);
    if (SFC_WET_SOIL == surface || SFC_DRY_SOIL == surface) {
        *value = 
            CHANG_DRY_SOIL_SWE(swe_coef_sfc((sfc_class *)private,col,row),
                               forest_cover_sfc((sfc_class *)private,col,row),
                               t19h,t37h);
        *value = (*value <= CHANG_SWE_MIN_MM ? PMALGO_CHANG_NO_SNOW : *value);

    } else if (SFC_NORMAL_NOSNOW == surface || 
               SFC_WATER == surface || 
               SFC_SNOW_IMPOSSIBLE == surface ||
               SFC_ICE == surface ||
               SFC_TOO_WARM == surface ||
               SFC_OUTOFRANGE == surface) {
        *value = (float)PMALGO_CHANG_NO_SNOW;
    } else {
        *value = (float)PMALGO_CHANG_UNKNOWN_SNOW;
    }

    /* 
     * Use Walker's test for wet snow, 
     * only in places where SWE was attempted in the first place 
     */
    if (SFC_WET_SOIL == surface || SFC_DRY_SOIL == surface) {
        if ((float)PMALGO_CHANG_NO_SNOW == *value) {
            if (MODIFIED_WALKER_WET_SNOW(t37v,t37h)) {
                *value = (float)PMALGO_CHANG_WET_SNOW;
                return TRUE;
            }
        }
    }

    return TRUE;

}

/*----------------------------------------------------------------------
 *  run_nsidc_s4_swe - forest-corrected NSIDC1
 * 
 *    input : private - pointer to sfc_class initialized 
 *                      with forest cover information
 *            data - TB matrix [channel][row][col] (TBs in tenths of Kelvins)
 *            col - column to operate on
 *            row - row to operate on
 * 
 *    output: value - algorithm result (mm SWE > 0)
 *            valid_tbs - TRUE if input TBs for this [row][col] are non-zero,
 *                        FALSE otherwise
 *
 *    result: TRUE if TBs were non-zero, FALSE otherwise
 *            
 *  Reference: See references for run_nsidc_s1_swe. 
 *             Forest correction reference is:
 *  Chang, A.T.C., J.L. Foster, and D.K. Hall. (1996). Effects of forest on
 *    the snow parameters derived from microwave measurements during the 
 *    BOREAS winter field campaign. Hydrological Process, Vol. 10, 1565-1574.
 *
 *  Note: the user should limit input forest fractions to the range 0.0 (no
 *    forest) to 0.50 (50% forest).  For purposes of testing alternatives, 
 *    nothing will be done here to enforce this limit.  R. Armstrong, 
 *    personal communication.  This is our interpretation of the plots in
 *    Chang et al.'s paper, that only consider f for values up to 0.5.
 *    Also, note that using higher f values, e.g. 90%, would increase the
 *    SWE value by an unrealistic factor of 10.  Limiting f to 0.5 will
 *    yield a maximum correction of a factor of 2.
 *----------------------------------------------------------------------*/
static bool run_nsidc_s4_swe (void *private,
			      byte2 **data[PMALGO_NUM_CHANNELS], 
			      int col, int row, float *value, bool *valid_tbs) {


    float swe;
    float t19h, t37h;

    assert(NULL != private);

    if (0 == data[PMALGO_SSMI_19H][row][col] 
        || 0 == data[PMALGO_SSMI_37H][row][col]) {

        *valid_tbs = FALSE;
        return FALSE;

    }

    *valid_tbs = TRUE;
  
    fetch_swe_tbs( private, data[PMALGO_SSMI_19H][row][col], data[PMALGO_SSMI_37H][row][col],
                   &t19h, &t37h );
    
    swe = NSIDC6_SWE( t19h, t37h, forest_cover_sfc((sfc_class *)private,col,row));
      
    if (CHANG_SWE_MIN_MM < swe) {
        *value = swe;
        return TRUE;
    }

    return FALSE;

}

/*----------------------------------------------------------------------
 *  run_nsidc_s10_swe - runs forest-corrected SWE algorithm on SSM/I TBs,
 *                      using a different correction on the TBs than
 *                      we've used before
 * 
 *    input : private - pointer to sfc_class initialized 
 *                      with forest cover information
 *            data - TB matrix [channel][row][col] (TBs in tenths of Kelvins)
 *            col - column to operate on
 *            row - row to operate on
 * 
 *    output: value - algorithm result (mm SWE > 0)
 *            valid_tbs - TRUE if input TBs for this [row][col] are non-zero,
 *                        FALSE otherwise
 *
 *    result: TRUE if TBs were non-zero, FALSE otherwise
 *            
 *  Reference: See references for run_nsidc_s4_swe. 
 *             Forest correction reference is:
 *  Chang, A.T.C., J.L. Foster, and D.K. Hall. (1996). Effects of forest on
 *    the snow parameters derived from microwave measurements during the 
 *    BOREAS winter field campaign. Hydrological Process, Vol. 10, 1565-1574.
 *
 *  Note: the user should limit input forest fractions to the range 0.0 (no
 *    forest) to 0.50 (50% forest).  For purposes of testing alternatives, 
 *    nothing will be done here to enforce this limit.  R. Armstrong, 
 *    personal communication.  This is our interpretation of the plots in
 *    Chang et al.'s paper, that only consider f for values up to 0.5.
 *    Also, note that using higher f values, e.g. 90%, would increase the
 *    SWE value by an unrealistic factor of 10.  Limiting f to 0.5 will
 *    yield a maximum correction of a factor of 2.
 *
 *  TB regressions are from current (Nov 2005) research by M. Savoie.
 *
 *----------------------------------------------------------------------*/
#define NSIDC_19H_REGRESSION(t19h) ((0.925 * (t19h)) + 10.110)
#define NSIDC_37H_REGRESSION(t37h) ((0.936 * (t37h)) + 10.743)
#define NSIDC10_SWE(t19h,t37h,forest_cover) \
((CHANG_SWE((NSIDC_19H_REGRESSION(t19h)),(NSIDC_37H_REGRESSION(t37h)))) / (1. - (forest_cover)))

static bool run_nsidc_s10_swe (void *private,
                               byte2 **data[PMALGO_NUM_CHANNELS], 
                               int col, int row, float *value, bool *valid_tbs) {


    float swe;
    float t19h, t37h;

    assert(NULL != private);

    if (0 == data[PMALGO_SSMI_19H][row][col] 
        || 0 == data[PMALGO_SSMI_37H][row][col]) {

        *valid_tbs = FALSE;
        return FALSE;

    }

    *valid_tbs = TRUE;
  
    fetch_swe_tbs( private, data[PMALGO_SSMI_19H][row][col], data[PMALGO_SSMI_37H][row][col],
                   &t19h, &t37h );

    swe = NSIDC10_SWE( t19h, t37h, forest_cover_sfc((sfc_class *)private,col,row));
      
    if (CHANG_SWE_MIN_MM < swe) {
        *value = swe;
        return TRUE;
    }

    return FALSE;

}


/*----------------------------------------------------------------------
 *  run_nsidc_s11_swe - runs forest-corrected SWE algorithm on SSM/I TBs,
 *                      using a different correction on the TBs than
 *                      we've used before; also, adjusts tbs for less
 *                      atmosphere at high altitudes
 * 
 *    input : private - pointer to sfc_class initialized 
 *                      with forest cover and elevation information
 *            data - TB matrix [channel][row][col] (TBs in tenths of Kelvins)
 *            col - column to operate on
 *            row - row to operate on
 * 
 *    output: value - algorithm result (mm SWE > 0)
 *            valid_tbs - TRUE if input TBs for this [row][col] are non-zero,
 *                        FALSE otherwise
 *
 *    result: TRUE if TBs were non-zero, FALSE otherwise
 *            
 *  Reference: See references for run_nsidc_s4_swe. 
 *             Forest correction reference is:
 *  Chang, A.T.C., J.L. Foster, and D.K. Hall. (1996). Effects of forest on
 *    the snow parameters derived from microwave measurements during the 
 *    BOREAS winter field campaign. Hydrological Process, Vol. 10, 1565-1574.
 *
 *  Note: the user should limit input forest fractions to the range 0.0 (no
 *    forest) to 0.50 (50% forest).  For purposes of testing alternatives, 
 *    nothing will be done here to enforce this limit.  R. Armstrong, 
 *    personal communication.  This is our interpretation of the plots in
 *    Chang et al.'s paper, that only consider f for values up to 0.5.
 *    Also, note that using higher f values, e.g. 90%, would increase the
 *    SWE value by an unrealistic factor of 10.  Limiting f to 0.5 will
 *    yield a maximum correction of a factor of 2.
 *
 *  TB regressions are from current (Nov 2005) research by M. Savoie.
 *
 *----------------------------------------------------------------------*/
#define NSIDC_19H_REGRESSION(t19h) ((0.925 * (t19h)) + 10.110)
#define NSIDC_37H_REGRESSION(t37h) ((0.936 * (t37h)) + 10.743)
#define NSIDC10_SWE(t19h,t37h,forest_cover) \
((CHANG_SWE((NSIDC_19H_REGRESSION(t19h)),(NSIDC_37H_REGRESSION(t37h)))) / (1. - (forest_cover)))

static bool run_nsidc_s11_swe (void *private,
                               byte2 **data[PMALGO_NUM_CHANNELS], 
                               int col, int row, float *value, bool *valid_tbs) {


    float swe;
    float t19h_f, t37h_f;
    double t19h, t37h;
    int2 elevation;
    pmalgo_high_elevation_model_id model_id;

    assert(NULL != private);

    if (0 == data[PMALGO_SSMI_19H][row][col] 
        || 0 == data[PMALGO_SSMI_37H][row][col]) {

        *valid_tbs = FALSE;
        return FALSE;

    }

    *valid_tbs = TRUE;

    /* Scale tbs, and do F17->F13 regression, first */
    fetch_swe_tbs( private, data[PMALGO_SSMI_19H][row][col], data[PMALGO_SSMI_37H][row][col],
                   &t19h_f, &t37h_f );
    
    /*
     * Do atmospheric correction, first, at high altitude locations, only.
     * This is a multiple linear regression of input TB and elevation,
     * and assumes a surface temperature.  The output is
     * a TB "corrected" for the thinner atmosphere at high altitudes.
     */
    elevation = elevation_sfc( (sfc_class *)private, col, row );
    if ( threshold_elevation_sfc( (sfc_class *)private ) < elevation ) {
      
        t19h = altitude_correct_tb(
            (double)t19h_f,
            private,
            elevation,
            PMALGO_SSMI_19H);
        t37h = altitude_correct_tb(
            (double)t37h_f,
            private,
            elevation,
            PMALGO_SSMI_37H);

    } else {

        t19h = (double)t19h_f;
        t37h = (double)t37h_f;

    }
      
    /*
     * Now do the empirically-based SWE algorithm
     */
    swe = (float)NSIDC10_SWE( t19h, t37h, forest_cover_sfc( (sfc_class *)private, col, row ) );
      
    if (CHANG_SWE_MIN_MM < swe) {
        *value = swe;
        return TRUE;
    }

    return FALSE;

}


/*----------------------------------------------------------------------
 *  run_nsidc_s12_swe - runs plain original nsidc1 SWE algorithm on SSM/I TBs,
 *                      adjusts tbs for less atmosphere at high altitudes
 * 
 *    input : private - pointer to sfc_class initialized 
 *                      with forest cover and elevation information
 *            data - TB matrix [channel][row][col] (TBs in tenths of Kelvins)
 *            col - column to operate on
 *            row - row to operate on
 * 
 *    output: value - algorithm result (mm SWE > 0)
 *            valid_tbs - TRUE if input TBs for this [row][col] are non-zero,
 *                        FALSE otherwise
 *
 *    result: TRUE if TBs were non-zero, FALSE otherwise
 *            
 *  Reference: See references for run_nsidc_s1_swe. 
 *
 *----------------------------------------------------------------------*/
static bool run_nsidc_s12_swe (void *private,
                               byte2 **data[PMALGO_NUM_CHANNELS], 
                               int col, int row, float *value, bool *valid_tbs) {


    float swe;
    float t19h_f, t37h_f;
    double t19h, t37h;
    int2 elevation;
    pmalgo_high_elevation_model_id model_id;

    assert(NULL != private);

    if (0 == data[PMALGO_SSMI_19H][row][col] 
        || 0 == data[PMALGO_SSMI_37H][row][col]) {

        *valid_tbs = FALSE;
        return FALSE;

    }

    *valid_tbs = TRUE;

    /* Scale tbs, and do F17->F13 regression, first */
    fetch_swe_tbs( private, data[PMALGO_SSMI_19H][row][col], data[PMALGO_SSMI_37H][row][col],
                   &t19h_f, &t37h_f );

    /*
     * Do atmospheric correction, first, at high altitude locations, only.
     * This is a multiple linear regression of input TB and elevation,
     * and assumes a surface temperature.  The output is
     * a TB "corrected" for the thinner atmosphere at high altitudes.
     */
    elevation = elevation_sfc( (sfc_class *)private, col, row );
    if ( threshold_elevation_sfc( (sfc_class *)private ) < elevation ) {
      
        t19h = altitude_correct_tb(
            (double)t19h_f,
            private,
            elevation,
            PMALGO_SSMI_19H);
        t37h = altitude_correct_tb(
            (double)t37h_f,
            private,
            elevation,
            PMALGO_SSMI_37H);

    } else {

        t19h = (double)t19h_f;
        t37h = (double)t37h_f;

    }
      
    /*
     * Now do the original nsidc1 SWE algorithm
     */
    swe = (float)NSIDC1_SWE( t19h, t37h );
      
    if (CHANG_SWE_MIN_MM < swe) {
        *value = swe;
        return TRUE;
    }

    return FALSE;

}


/*----------------------------------------------------------------------
 *  run_nsidc_a10_swe - Calculate SWE from AMSR Tbs, assuming that
 *                      AMSR frequencies are the same as SSM/I frequencies,
 *                      using NSIDC10.  This uses a different correction
 *                      on the TBs than we've used before.
 * 
 *    input : private - pointer to sfc_class initialized 
 *                      with forest cover information
 *            data - TB matrix [channel][row][col] (TBs in tenths of Kelvins)
 *            col - column to operate on
 *            row - row to operate on
 * 
 *    output: value - algorithm result (mm SWE > 0)
 *            valid_tbs - TRUE if input TBs for this [row][col] are non-zero,
 *                        FALSE otherwise
 *
 *    result: TRUE if TBs were non-zero, FALSE otherwise
 *            
 *  Reference: See references for run_nsidc_s10_swe.
 *             Uses the AMSR-E TBs as if they were SSM/I TBs, and
 *             does a regression to SMMR TBs.
 *
 *----------------------------------------------------------------------*/
static bool run_nsidc_a10_swe (void *private,
                               byte2 **data[PMALGO_NUM_CHANNELS], 
                               int col, int row, float *value, bool *valid_tbs) {


    float swe;

    assert(NULL != private);

    if (0 == data[PMALGO_AMSR_18H][row][col] 
        || 0 == data[PMALGO_AMSR_36H][row][col]) {

        *valid_tbs = FALSE;
        return FALSE;

    }

    *valid_tbs = TRUE;
  
    swe = NSIDC10_SWE(data[PMALGO_AMSR_18H][row][col] * TB_SCALE,
                      data[PMALGO_AMSR_36H][row][col] * TB_SCALE,
                      forest_cover_sfc((sfc_class *)private,col,row));
      
    if (CHANG_SWE_MIN_MM < swe) {
        *value = swe;
        return TRUE;
    }

    return FALSE;

}


/*----------------------------------------------------------------------
 *  run_gb_snow - Run Grody/Basist decision tree for snow
 *
 *    input : private - not used for this algorithm
 *            data - TB matrix [channel][row][col] (TBs in tenths of Kelvins)
 *            col - column to operate on
 *            row - row to operate on
 * 
 *    output: value - classifications for bad obs, no snow, glacier ice or snow
 *            valid_tbs - TRUE if input TBs for this [row][col] are non-zero,
 *                        FALSE otherwise
 *
 *    result: TRUE if tbs are non-zero, FALSE otherwise
 *
 *  Reference:
 *    See called C subroutine, snow_gb.c.
 *
 *----------------------------------------------------------------------*/
static bool run_gb_snow (void *private,
			 byte2 **data[PMALGO_NUM_CHANNELS], 
			 int col, int row, float *value, 
			 bool *valid_tbs) {

    gb_return_code return_code;
    gb_surface_condition surface_condition;

    assert(NULL == private);

    *valid_tbs = FALSE;

    if (0 < data[PMALGO_SSMI_19V][row][col] 
        && 0 < data[PMALGO_SSMI_19H][row][col]
        && 0 < data[PMALGO_SSMI_22V][row][col]
        && 0 < data[PMALGO_SSMI_37V][row][col]
        && 0 < data[PMALGO_SSMI_85V][row][col]
        && 0 < data[PMALGO_SSMI_85H][row][col]) {
        *valid_tbs = TRUE;
        return_code = snow_gb(&surface_condition,
                              data[PMALGO_SSMI_19V][row][col] * TB_SCALE,
                              data[PMALGO_SSMI_19H][row][col] * TB_SCALE,
                              data[PMALGO_SSMI_22V][row][col] * TB_SCALE,
                              data[PMALGO_SSMI_37V][row][col] * TB_SCALE,
                              data[PMALGO_SSMI_85V][row][col] * TB_SCALE,
                              data[PMALGO_SSMI_85H][row][col] * TB_SCALE,
                              0);
        *value = (float)return_code ;
        return TRUE;
    } 

    return FALSE;

}
  

/*----------------------------------------------------------------------
 *  run_gb_surface - Run Grody/Basist decision tree for surface classifications
 *
 *    input : private - not used for this algorithm
 *            data - TB matrix [channel][row][col] (TBs in tenths of Kelvins)
 *            col - column to operate on
 *            row - row to operate on
 * 
 *    output: value - surface classification (see snow_gb.h)
 *            valid_tbs - TRUE if input TBs for this [row][col] are non-zero,
 *                        FALSE otherwise
 *
 *    result: TRUE if tbs are non-zero, FALSE otherwise
 *
 *  Reference:
 *    See called C subroutine, snow_gb.c.
 *
 *----------------------------------------------------------------------*/
static bool run_gb_surface (void *private,
			    byte2 **data[PMALGO_NUM_CHANNELS], 
			    int col, int row, float *value, 
			    bool *valid_tbs) {

    gb_return_code return_code;
    gb_surface_condition surface_condition;

    assert(NULL == private);

    *valid_tbs = FALSE;

    if (0 < data[PMALGO_SSMI_19V][row][col] 
        && 0 < data[PMALGO_SSMI_19H][row][col]
        && 0 < data[PMALGO_SSMI_22V][row][col]
        && 0 < data[PMALGO_SSMI_37V][row][col]
        && 0 < data[PMALGO_SSMI_85V][row][col]
        && 0 < data[PMALGO_SSMI_85H][row][col]) {
        *valid_tbs = TRUE;
        return_code = snow_gb(&surface_condition,
                              data[PMALGO_SSMI_19V][row][col] * TB_SCALE,
                              data[PMALGO_SSMI_19H][row][col] * TB_SCALE,
                              data[PMALGO_SSMI_22V][row][col] * TB_SCALE,
                              data[PMALGO_SSMI_37V][row][col] * TB_SCALE,
                              data[PMALGO_SSMI_85V][row][col] * TB_SCALE,
                              data[PMALGO_SSMI_85H][row][col] * TB_SCALE,
                              0);
        *value = (float)surface_condition;
        return TRUE;
    } 

    return FALSE;

}
  
  

/*----------------------------------------------------------------------
 *  run_chang_surface - Run Chang surface classification
 *
 *    input : private - pointer to sfc_class initialized 
 *                      with ancillary surface information
 *            data - TB matrix [channel][row][col] (TBs in tenths of Kelvins)
 *            col - column to operate on
 *            row - row to operate on
 * 
 *    output: value - surface classification (see surface.h)
 *            valid_tbs - TRUE if input TBs for this [row][col] are non-zero,
 *                        FALSE otherwise
 *
 *    result: TRUE if tbs are non-zero, FALSE otherwise
 *
 *  Reference:
 *    E-mail and C program, granule.c, Chang to Armstrong, Armstrong sent to
 *    brodzik, Jan 4, 2001.
 *
 *----------------------------------------------------------------------*/
#define CHANG_MIN_TB_K 50.0
#define CHANG_MAX_TB_K 320.0
#define CHANG_SNOW_THRESHOLD_37V_K (250.0)
#define CHANG_SNOW_THRESHOLD_37H_K (240.0)
#define CHANG_SURFACE_TOO_WARM(t19v,t22v,t37v,t85v) \
((3.57 + 0.698 * (t19v) + 0.333 * (t22v) - 0.434 * (t37v) + 0.422 * (t85v)) \
>= 275.0)
#define CHANG_IS_WET_SOIL(t19v,t37v) ((((t37v) - (t19v)) / 18.0) >= -0.3)
static bool run_chang_surface (void *private,
			       byte2 **data[PMALGO_NUM_CHANNELS], 
			       int col, int row, float *value, 
			       bool *valid_tbs) {

    sfc_type surface;
    gb_surface_condition gb_surface;
    float t19v,t22v,t37v,t37h,t85v,t85h;

    assert(NULL != private);

    *valid_tbs = FALSE;
    *value = (float)SFC_NORMAL_NOSNOW;

    /*
     * Usually, an algorithm is not run at all if any of the TBs are zero, but
     * this algorithm checks later for out of range TBs, so only set
     * only set valid_tbs to FALSE if all TBs are zero, otherwise, let the
     * algorithm classify the missing data.
     * This also eliminates the problem at the edges of EASE-Grid scans, where
     * there is no 85 GHz temperature, but there are lower channel temperatures
     * for a given pixel
     */
    if (0 == data[PMALGO_SSMI_19V][row][col]  
        && 0 == data[PMALGO_SSMI_22V][row][col] 
        && 0 == data[PMALGO_SSMI_37V][row][col] 
        && 0 == data[PMALGO_SSMI_37H][row][col] 
        && 0 == data[PMALGO_SSMI_85V][row][col] 
        && 0 == data[PMALGO_SSMI_85H][row][col]) {
    
        return FALSE;

    }
    *valid_tbs = TRUE;

    /* Do TB scaling once */
    t19v = data[PMALGO_SSMI_19V][row][col] * TB_SCALE;
    t22v = data[PMALGO_SSMI_22V][row][col] * TB_SCALE;
    t37v = data[PMALGO_SSMI_37V][row][col] * TB_SCALE;
    t37h = data[PMALGO_SSMI_37H][row][col] * TB_SCALE;
    t85v = data[PMALGO_SSMI_85V][row][col] * TB_SCALE;
    t85h = data[PMALGO_SSMI_85H][row][col] * TB_SCALE;

    /* 
     * Start with (relatively) fixed surface conditions, then look at TBs
     * for changing information like precip and wet/dry soil and snow
     */
    surface = determine_nsidc_sfc((sfc_class *)private,col,row);
    assert(SFC_ICE == surface || SFC_WATER == surface || 
           SFC_MOUNTAIN == surface || SFC_FOREST == surface || 
           SFC_SNOW_IMPOSSIBLE == surface ||
           SFC_NORMAL_NOSNOW == surface);

    if (SFC_NORMAL_NOSNOW != surface) {
        *value = (float)surface;
        return TRUE;
    }

    /* ROV checks on TBS */
    if (t85v < CHANG_MIN_TB_K || CHANG_MAX_TB_K < t85v ||
        t85h < CHANG_MIN_TB_K || CHANG_MAX_TB_K < t85h ||
        t37v < CHANG_MIN_TB_K || CHANG_MAX_TB_K < t37v ||
        t37h < CHANG_MIN_TB_K || CHANG_MAX_TB_K < t37h ||
        t19v < CHANG_MIN_TB_K || CHANG_MAX_TB_K < t19v ||
        t22v < CHANG_MIN_TB_K || CHANG_MAX_TB_K < t22v ) {
        *value = (float)SFC_OUTOFRANGE;
        return TRUE;
    }
  
    /* 
     * Use Grody/Basist tests for precip only, 
     * with latest version from G&B;  Send 0.0 for the 19H TB, it's not
     * needed to do precip test
     */
    if (GB_NO_SNOW == snow_gb(&gb_surface,t19v,0.0,t22v,t37v,t85v,t85h,0)) {
        if (GB_PC1 == gb_surface || GB_PC2 == gb_surface) {
            *value = (float)SFC_RAINING;
            return TRUE;
        }
    }

    /* 
     * The order of this test and the precip tests is important: better to
     * check for precip first, and then do tests for a surface that's "too
     * warm"
     */
/*   if (CHANG_SNOW_THRESHOLD_37V_K <= t37v || */
/*       CHANG_SNOW_THRESHOLD_37H_K <= t37h) { */
/*     *value = (float)SFC_NORMAL_NOSNOW;  */
/*     return TRUE;  */
/*   } */

    /* Surface temperature check (where does this come from?) */
/*   if (CHANG_SURFACE_TOO_WARM(t19v,t22v,t37v,t85v)) { */
/*     *value = (float)SFC_TOO_WARM; */
/*     return TRUE; */
/*   } */
  
    /* 
     * Use Chang's wet soil discriminator.  Again, Chang includes a test
     * on surface temperatures, but I don't have them.
     */
    if (CHANG_IS_WET_SOIL(t19v,t37v)) {
        *value = (float)SFC_WET_SOIL;
    } else {
        *value = (float)SFC_DRY_SOIL;
    }
  
    return TRUE;

}
  

/*----------------------------------------------------------------------
 *  run_nsidc_surface - Run Chang surface classification, with NSIDC 
 *                      modifications
 *
 *    input : private - pointer to sfc_class initialized 
 *                      with ancillary surface information
 *            data - TB matrix [channel][row][col] (TBs in tenths of Kelvins)
 *            col - column to operate on
 *            row - row to operate on
 * 
 *    output: value - surface classification (see surface.h)
 *            valid_tbs - TRUE if input TBs for this [row][col] are non-zero,
 *                        FALSE otherwise
 *
 *    result: TRUE if tbs are non-zero, FALSE otherwise
 *
 *  Reference:
 *    E-mail and C program, granule.c, Chang to Armstrong, Armstrong sent to
 *    brodzik, Jan 4, 2001.
 *
 *----------------------------------------------------------------------*/
static bool run_nsidc_surface (void *private,
			       byte2 **data[PMALGO_NUM_CHANNELS], 
			       int col, int row, float *value, 
			       bool *valid_tbs) {

    sfc_type surface;
    gb_surface_condition gb_surface;
    float t19v,t22v,t37v,t37h,t85v,t85h;

    assert(NULL != private);

    *valid_tbs = FALSE;
    *value = (float)SFC_NORMAL_NOSNOW;

    /*
     * Usually, an algorithm is not run at all if any of the TBs are zero, but
     * this algorithm checks later for out of range TBs, so only set
     * only set valid_tbs to FALSE if all TBs are zero, otherwise, let the
     * algorithm classify the missing data.
     * This also eliminates the problem at the edges of EASE-Grid scans, where
     * there is no 85 GHz temperature, but there are lower channel temperatures
     * for a given pixel
     */
    if (0 == data[PMALGO_SSMI_19V][row][col]  
        && 0 == data[PMALGO_SSMI_22V][row][col] 
        && 0 == data[PMALGO_SSMI_37V][row][col] 
        && 0 == data[PMALGO_SSMI_37H][row][col] 
        && 0 == data[PMALGO_SSMI_85V][row][col]
        && 0 == data[PMALGO_SSMI_85H][row][col]) {

        return FALSE;

    }
    *valid_tbs = TRUE;

    /* Do TB scaling once */
    t19v = data[PMALGO_SSMI_19V][row][col] * TB_SCALE;
    t22v = data[PMALGO_SSMI_22V][row][col] * TB_SCALE;
    t37v = data[PMALGO_SSMI_37V][row][col] * TB_SCALE;
    t37h = data[PMALGO_SSMI_37H][row][col] * TB_SCALE;
    t85v = data[PMALGO_SSMI_85V][row][col] * TB_SCALE;
    t85h = data[PMALGO_SSMI_85H][row][col] * TB_SCALE;

    /* 
     * Start with (relatively) fixed surface conditions, then look at TBs
     * for changing information like precip and wet/dry soil and snow
     */
    surface = determine_nsidc_sfc((sfc_class *)private,col,row);
    assert(SFC_ICE == surface || SFC_WATER == surface || 
           SFC_MOUNTAIN == surface || SFC_FOREST == surface || 
           SFC_SNOW_IMPOSSIBLE == surface ||
           SFC_NORMAL_NOSNOW == surface);

    if (SFC_NORMAL_NOSNOW != surface) {
        *value = (float)surface;
        return TRUE;
    }

    /* ROV checks on TBS */
    if (t85v < CHANG_MIN_TB_K || CHANG_MAX_TB_K < t85v || 
        t85h < CHANG_MIN_TB_K || CHANG_MAX_TB_K < t85h || 
        t37v < CHANG_MIN_TB_K || CHANG_MAX_TB_K < t37v || 
        t37h < CHANG_MIN_TB_K || CHANG_MAX_TB_K < t37h || 
        t19v < CHANG_MIN_TB_K || CHANG_MAX_TB_K < t19v || 
        t22v < CHANG_MIN_TB_K || CHANG_MAX_TB_K < t22v ) { 
        *value = (float)SFC_OUTOFRANGE; 
        return TRUE; 
    } 
  
    /* 
     * Use Grody/Basist tests for precip only, 
     * with latest version from G&B;  Send 0.0 for the 19H TB, it's not
     * needed to do precip test
     */
    if (GB_NO_SNOW == snow_gb(&gb_surface,t19v,0.0,t22v,t37v,t85v,t85h,0)) {
        if (GB_PC1 == gb_surface || GB_PC2 == gb_surface) {
            *value = (float)SFC_RAINING;
            return TRUE;
        }
    }

    /* 
     * The order of this test and the precip tests is important: better to
     * check for precip first, and then do tests for a surface that's "too
     * warm" for snow 
     */
/*   if (CHANG_SNOW_THRESHOLD_37V_K <= t37v || */
/*       CHANG_SNOW_THRESHOLD_37H_K <= t37h) { */
/*     *value = (float)SFC_NORMAL_NOSNOW;  */
/*     return TRUE;  */
/*   } */

    /* Use Chang's wet soil discriminator */
    if (CHANG_IS_WET_SOIL(t19v,t37v)) {
        *value = (float)SFC_WET_SOIL;
    } else {
        *value = (float)SFC_DRY_SOIL;
    }
  
    return TRUE;

}
  
  

/*----------------------------------------------------------------------
 *  run_boot_seaice - Calculate % sea ice with Bootstrap algorithm
 * 
 *    input : private - pointer to bootstrap_class object
 *            data - TB matrix [channel][row][col] (TBs in tenths of Kelvins)
 *            col - column to operate on
 *            row - row to operate on
 * 
 *    output: value - valid algorithm result (% total sea ice, or 
 *              PMALGO_BOOTSTRAP_MISSING_DATA (for TBs < 40K), or
 *              PMALGO_BOOTSTRAP_LAND for pixels where algorithm wasn't
 *              run)
 *            valid_tbs - TRUE if input TBs for this [row][col] are non-zero,
 *                        FALSE otherwise
 *
 *    result: TRUE if algorithm returns sea ice or missing ice flag, 
 *            FALSE otherwise
 *
 *  Reference:
 *    See called FORTRAN routine, seaice_bootstrap.f.
 *
 *----------------------------------------------------------------------*/
static bool run_boot_seaice (void *private,
			     byte2 **data[PMALGO_NUM_CHANNELS], 
			     int col, int row, float *value, 
			     bool *valid_tbs) {

    float seaice;
    bootstrap_class* this;
    assert(NULL != private);
    
    this = ((bootstrap_class *)private);
    


    *valid_tbs = FALSE;
    
    if ( bootstrap_is_v1(this) ) {
        if (this->hem == BOOTSTRAP_SOUTH_HEM) {
            /* BOOTSTRAP V1, Southern Hemisphere */
            if (0 < data[PMALGO_SSMI_19V][row][col] 
                && 0 < data[PMALGO_SSMI_22V][row][col]
                && 0 < data[PMALGO_SSMI_37V][row][col]) {
                *valid_tbs = TRUE;
                if (get_seaice_bootstrap(this, data, row, col, &seaice)) {
                    *value = seaice;
                    return TRUE;
                }
            }
        } else {
            /* Boot V1, Northern Hemisphere */
            if (0 < data[PMALGO_SSMI_19V][row][col] 
                && 0 < data[PMALGO_SSMI_22V][row][col]
                && 0 < data[PMALGO_SSMI_37H][row][col]
                && 0 < data[PMALGO_SSMI_37V][row][col]) {
                *valid_tbs = TRUE;
                if (get_seaice_bootstrap(this, data, row, col, &seaice)) {
                    *value = seaice;
                    return TRUE;
                }
            }
        }
    } else if ( bootstrap_is_v2(this)) {
        /* Bootstrap V2, both hemispheres */
	    if (get_seaice_bootstrap(this, data, row, col, &seaice)) {
            *value = seaice;
            *valid_tbs = TRUE;
            return TRUE;
        }
    } else {
        /* invalid bootstrap type input. */
        assert(FALSE);
    }
    
    
        

    return FALSE;

}



/*----------------------------------------------------------------------
 *  run_ntn_seaice - Calculate % sea ice in Northern Hemisphere
 *    with "NASA Team" algorithm
 * 
 *    input : private - not used for this algorithm
 *            data - TB matrix [channel][row][col] (TBs in tenths of Kelvins)
 *            col - column to operate on
 *            row - row to operate on
 * 
 *    output: value - valid algorithm result (% total sea ice, or 
 *              PMALGO_NASATEAM_SUSPECT_ICE)
 *            valid_tbs - TRUE if input TBs for this [row][col] are non-zero,
 *                        FALSE otherwise
 *
 *    result: TRUE if algorithm returns sea ice or suspect ice flag, 
 *            FALSE otherwise
 *
 *  Reference:
 *    See called FORTRAN routine, seaice_nasateam.f.
 *
 *----------------------------------------------------------------------*/
static bool run_ntn_seaice (void *private,
			    byte2 **data[PMALGO_NUM_CHANNELS], 
			    int col, int row, float *value, bool *valid_tbs) {

    /* this is a dummy variable for the sea ice routine */ 
    static float myr;  
  
    assert(NULL == private);

    return (run_nasateam_ice (data, col, row, 
                              (byte2)NASATEAM_NORTH_HEM, value, &myr,
                              valid_tbs));

}


/*----------------------------------------------------------------------
 *  run_nts_seaice - Calculate % sea ice in Southern Hemisphere
 *    with "NASA Team" algorithm
 * 
 *    input : private - not used for this algorithm
 *            data - TB matrix [channel][row][col] (TBs in tenths of Kelvins)
 *            col - column to operate on
 *            row - row to operate on
 * 
 *    output: value - valid algorithm result (% total sea ice, or 
 *              PMALGO_NASATEAM_SUSPECT_ICE)
 *            valid_tbs - TRUE if input TBs for this [row][col] are non-zero,
 *                        FALSE otherwise
 *
 *    result: TRUE if algorithm returns sea ice or suspect ice flag, 
 *            FALSE otherwise
 *
 *  Reference:
 *    See called FORTRAN routine, seaice_nasateam.f.
 *
 *----------------------------------------------------------------------*/
static bool run_nts_seaice (void *private,
			    byte2 **data[PMALGO_NUM_CHANNELS], 
			    int col, int row, float *value, bool *valid_tbs) {

    /* this is a dummy variable for the sea ice routine */ 
    static float myr;

    assert(NULL == private);

    return (run_nasateam_ice (data, col, row, 
                              (byte2)NASATEAM_SOUTH_HEM, value, &myr,
                              valid_tbs));

}
  

/*----------------------------------------------------------------------
 *  run_ntn_myice - Calculate % multi-year sea ice in Northern Hemisphere
 *    with "NASA Team" algorithm
 * 
 *    input : private - not used for this algorithm
 *            data - TB matrix [channel][row][col] (TBs in tenths of Kelvins)
 *            col - column to operate on
 *            row - row to operate on
 * 
 *    output: value - algorithm result (% multi-year sea ice, or 
 *              PMALGO_NASATEAM_SUSPECT_ICE)
 *            valid_tbs - TRUE if input TBs for this [row][col] are non-zero,
 *                        FALSE otherwise
 *
 *    result: TRUE if algorithm returns sea ice or suspect ice flag, 
 *            FALSE otherwise
 *
 *  Reference:
 *    See called FORTRAN routine, seaice_nasateam.f.
 *
 *----------------------------------------------------------------------*/
static bool run_ntn_myice (void *private,
			   byte2 **data[PMALGO_NUM_CHANNELS], 
			   int col, int row, float *value, bool *valid_tbs) {

    /* this is a dummy variable for the sea ice routine */ 
    static float total;  
  
    assert(NULL == private);

    return (run_nasateam_ice (data, col, row, 
                              (byte2)NASATEAM_NORTH_HEM, &total, 
                              value, valid_tbs));

}


/*----------------------------------------------------------------------
 *  run_gts_seaice - Calculates total sea ice concentrations, to reproduce 
 *                   "Goddard Time Series" data; this is the NASA Team
 *                   algorithm, but uses adjusted tie points that minimize
 *                   sensor differences.
 * 
 *    input : private - pointer to goddard_class object
 *            data - TB matrix [channel][row][col] (TBs in tenths of Kelvins)
 *            col - column to operate on
 *            row - row to operate on
 * 
 *    output: value - valid algorithm result: 
 *                    0-250 (% total sea ice scaled into range 0-250)
 *                    251   ("suspect ice" is identified in seaice_goddard.f)
 *                    254   (LAND flag: any input loci values that were in
 *                           the range 1-250 will be assumed to be land 
 *                           pixels, and are reset to 254)
 *            valid_tbs - TRUE if input TBs for this [row][col] are non-zero,
 *                        FALSE otherwise
 *
 *    result: TRUE if algorithm returns sea ice or suspect ice flag, 
 *            FALSE otherwise
 *
 *  Reference:
 *    See called FORTRAN routine, seaice_goddard.f.
 *
 *----------------------------------------------------------------------*/
static bool run_gts_seaice (void *private,
			    byte2 **data[PMALGO_NUM_CHANNELS], 
			    int col, int row, float *value, bool *valid_tbs) {

    /* this is a dummy variable for the sea ice routine */ 
    static float myr;  
    static bool status;
  
    assert(NULL != private);
    status = run_goddard_ice (data, col, row, 
                              goddard_hemisphere((goddard_class *)private), 
                              goddard_platform((goddard_class *)private),
                              goddard_scale_percent_seaice((goddard_class *)private),
                              value, &myr, valid_tbs);

    return status;

}


/*----------------------------------------------------------------------
 *  run_gts_myice - Calculates multi-year sea ice concentrations, to reproduce 
 *                  "Goddard Time Series" data; this is the NASA Team
 *                  algorithm, but uses adjusted tie points that minimize
 *                  sensor differences.
 * 
 *    input : private - pointer to goddard_class object
 *            data - TB matrix [channel][row][col] (TBs in tenths of Kelvins)
 *            col - column to operate on
 *            row - row to operate on
 * 
 *    output: value - valid algorithm result: 
 *                    0-250 (% total sea ice scaled into range 0-250)
 *                    251   ("suspect ice" is identified in seaice_goddard.f)
 *                    254   (LAND flag: any input loci values that were in
 *                           the range 1-250 will be assumed to be land 
 *                           pixels, and are reset to 254)
 *            valid_tbs - TRUE if input TBs for this [row][col] are non-zero,
 *                        FALSE otherwise
 *
 *    result: TRUE if algorithm returns sea ice or suspect ice flag, 
 *            FALSE otherwise
 *
 *  Reference:
 *    See called FORTRAN routine, seaice_goddard.f.
 *
 *----------------------------------------------------------------------*/
static bool run_gts_myice (void *private,
			   byte2 **data[PMALGO_NUM_CHANNELS], 
			   int col, int row, float *value, bool *valid_tbs) {

    /* this is a dummy variable for the sea ice routine */ 
    static float total;  
    static bool status;
  
    assert(NULL != private);

    status = run_goddard_ice (data, col, row, 
                              goddard_hemisphere((goddard_class *)private),
                              goddard_platform((goddard_class *)private),
                              goddard_scale_percent_seaice((goddard_class *)private),
                              &total, value, valid_tbs);

    return status;

}


/*----------------------------------------------------------------------
 *  run_ssmi_xpgr - Calculate cross-polarization gradient ratio (XPGR)
 *                  (for Abdalati snow melt algorithm) on SSMI TBs
 * 
 *    input : private - not used for this algorithm
 *            data - TB matrix [channel][row][col] (TBs in tenths of Kelvins)
 *            col - column to operate on
 *            row - row to operate on
 * 
 *    output: value - algorithm result (unitless) 
 *            valid_tbs - TRUE if input TBs for this [row][col] are non-zero,
 *                        FALSE otherwise
 *
 *    result: TRUE when TBs are non-zero, FALSE otherwise
 *
 *  Reference:
 *    Abdalati, W., K. Steffen, 1997. "Snowmelt on the Greenland Ice Sheet as
 *      Derived from Passive Microwave Satellite Data," Journal of Climate, 
 *      Vol 10, No. 2, February 1997.
 *
 *----------------------------------------------------------------------*/
#define XPGR(t19h,t37v) (((float)(t19h) - (float)(t37v)) \
			 /((float)(t19h) + (float)(t37v)))

static bool run_ssmi_xpgr (void *private,
			   byte2 **data[PMALGO_NUM_CHANNELS], 
			   int col, int row, float *value, bool *valid_tbs) {

    *valid_tbs = FALSE;

    assert(NULL == private);

    if (0 < data[PMALGO_SSMI_19H][row][col] 
        && 0 < data[PMALGO_SSMI_37V][row][col]) {
        *valid_tbs = TRUE;
        *value = XPGR(data[PMALGO_SSMI_19H][row][col], 
                      data[PMALGO_SSMI_37V][row][col]);
        return TRUE;
    } 

    return FALSE;

}


/*----------------------------------------------------------------------
 *  run_smmr_xpgr - Calculate cross-polarization gradient ratio (XPGR)
 *                  (for Abdalati snow melt algorithm) on SMMR TBs
 *                  SMMR TBs are adjusted to SSM/I TBs per Jezek, Merry
 *                  and Cavalieri (1993).
 * 
 *    input : private - not used for this algorithm
 *            data - TB matrix [channel][row][col] (TBs in tenths of Kelvins)
 *            col - column to operate on
 *            row - row to operate on
 * 
 *    output: value - algorithm result (unitless) 
 *            valid_tbs - TRUE if input TBs for this [row][col] are non-zero,
 *                        FALSE otherwise
 *
 *    result: TRUE when TBs are non-zero, FALSE otherwise
 *
 *  Reference:
 *
 *    Abdalati, W., K. Steffen, 1997. "Snowmelt on the Greenland Ice Sheet as
 *      Derived from Passive Microwave Satellite Data," Journal of Climate, 
 *      Vol 10, No. 2, February 1997.
 *
 *    Jezek, K. C., C. J. Merry, D. J. Cavalieri, 1993. Comparison of SMMR 
 *     and SSM/I passive microwave data collected over Antarctica, Annals
 *     of Glaciology, Vol 17, pp 131-136.
 *
 *----------------------------------------------------------------------*/
#define SMMR18H_TO_SSMI19H(t18h) (((float)(t18h) - 2.62)/0.940)
#define SMMR37V_TO_SSMI37V(t37v) (((float)(t37v) - 30.20)/0.861)
#define SMMR_XPGR(t18h,t37v) \
(XPGR((SMMR18H_TO_SSMI19H(t18h)),(SMMR37V_TO_SSMI37V(t37v))))

static bool run_smmr_xpgr (void *private,
			   byte2 **data[PMALGO_NUM_CHANNELS], 
			   int col, int row, float *value, bool *valid_tbs) {

    *valid_tbs = FALSE;

    assert(NULL == private);

    if (0 < data[PMALGO_SMMR_18H][row][col] 
        && 0 < data[PMALGO_SMMR_37V][row][col]) {
        *valid_tbs = TRUE;
        *value = SMMR_XPGR(data[PMALGO_SMMR_18H][row][col] * TB_SCALE,
                           data[PMALGO_SMMR_37V][row][col] * TB_SCALE);
        return TRUE;
    } 

    return FALSE;

}


/*----------------------------------------------------------------------
 *  run_nagler_depth - Calculate snow depth with algorithm by Nagler & Rott
 * 
 *    input : private - not used for this algorithm
 *            data - TB matrix [channel][row][col] (TBs in tenths of Kelvins)
 *            col - column to operate on
 *            row - row to operate on
 * 
 *    output: value - algorithm result (cm snow depth > 5,
 *                    or NAGLER_THIN_SNOW_VALUE if thin snow) 
 *            valid_tbs - TRUE if input TBs for this [row][col] are non-zero,
 *                        FALSE otherwise
 *
 *    result: TRUE if algorithm returns snow, FALSE otherwise
 *
 *  References:
 *     Nagler, Thomas, and H. Rott 1992.  Development and intercomparisons 
 *       of snow mapping algorithms based on SSM/I data.  Proceedings 
 *       IGARSS'92, Houston, IEEE, N.p., 812-814.
 *
 *----------------------------------------------------------------------*/
#define NAGLER_THIN_SNOW_THRESHOLD 5.
#define NAGLER_THIN_SNOW_VALUE 1
#define NAGLER_SNOW_COVERED(t19v,t19v_less_t37v,t37v_less_t85v) \
     ((float)(t19v) <= 266 && \
      (((float)(t19v_less_t37v) >= 4) || ((float)(t37v_less_t85v) >= 3)))
#define NAGLER_SNOW_DEPTH_CM(t19v_less_t37v,t37v_less_t85v) \
     (-2.41 \
      + 1.2 * ((float)(t19v_less_t37v)) \
      - 0.16 * ((float)(t37v_less_t85v)))

static bool run_nagler_depth (void *private,
			      byte2 **data[PMALGO_NUM_CHANNELS], 
			      int col, int row, float *value, 
			      bool *valid_tbs) {
    float depth;
    float t19v_less_t37v, t37v_less_t85v;
  
    assert(NULL == private);

    *valid_tbs = FALSE;

    if (0 < data[PMALGO_SSMI_19V][row][col] 
        && 0 < data[PMALGO_SSMI_37V][row][col]
        && 0 < data[PMALGO_SSMI_85V][row][col]) {

        *valid_tbs = TRUE;

        t19v_less_t37v = data[PMALGO_SSMI_19V][row][col] * TB_SCALE
            - data[PMALGO_SSMI_37V][row][col] * TB_SCALE;
        t37v_less_t85v = data[PMALGO_SSMI_37V][row][col] * TB_SCALE
            - data[PMALGO_SSMI_85V][row][col] * TB_SCALE;

        if (NAGLER_SNOW_COVERED(data[PMALGO_SSMI_19V][row][col] * TB_SCALE,
                                t19v_less_t37v,t37v_less_t85v)) {
            depth = NAGLER_SNOW_DEPTH_CM(t19v_less_t37v,t37v_less_t85v);
            if (0 < depth) {
                if (NAGLER_THIN_SNOW_THRESHOLD < depth) {
                    *value = depth;
                    return TRUE;
                } else {
                    *value = NAGLER_THIN_SNOW_VALUE;
                    return TRUE;
                }
            }
        }
    }

    return FALSE;

}


/*----------------------------------------------------------------------
 *  run_nagler_csi - Calculate clear skies index by Nagler & Rott
 * 
 *    input : private - not used for this algorithm
 *            data - TB matrix [channel][row][col] (TBs in tenths of Kelvins)
 *            col - column to operate on
 *            row - row to operate on
 * 
 *    output: value - algorithm result (difference in Kelvins) 
 *            valid_tbs - TRUE if input TBs for this [row][col] are non-zero,
 *                        FALSE otherwise
 *
 *    result: FALSE if input TBs do not produce valid algorithm output,
 *            TRUE otherwise
 *
 *  References:
 *     Nagler, Thomas, and H. Rott 1992.  Development and intercomparisons 
 *       of snow mapping algorithms based on SSM/I data.  Proceedings 
 *       IGARSS'92, Houston, IEEE, N.p., 812-814.
 *
 *----------------------------------------------------------------------*/
static bool run_nagler_csi (void *private,
			    byte2 **data[PMALGO_NUM_CHANNELS], 
			    int col, int row, float *value, bool *valid_tbs) {
    float diff;
  
    assert(NULL == private);

    *valid_tbs = FALSE;

    if (0 < data[PMALGO_SSMI_37V][row][col]
        && 0 < data[PMALGO_SSMI_85V][row][col]) {

        *valid_tbs = TRUE;
        diff = data[PMALGO_SSMI_37V][row][col] * TB_SCALE
            - data[PMALGO_SSMI_85V][row][col] * TB_SCALE;
        if (0 < diff) {
            *value = diff;
            return TRUE;
        }
    }
    
    return FALSE;

}


/*----------------------------------------------------------------------
 *  run_za_s1_fs - Zhang-Armstrong frozen soil test for SSM/I, version 1.
 *  Eventually this algorithm should require a snow cover input map to
 *  determine where not to run the FS test, but for now I'm just going to
 *  run it everywhere and filter snow-covered pixels afterward.  Another
 *  option would be to leave it this way, and suggest that the user enter
 *  the snow cover map as a climatology file, with (no snow == frozen
 *  soil possible, and snow == frozen soil impossible).  This would
 *  require no new code, and take advantage of the mechanism in place for
 *  climatology-filtering.
 * 
 *    input : private - not used for this algorithm
 *            data - TB matrix [channel][row][col] (TBs in tenths of Kelvins)
 *            col - column to operate on
 *            row - row to operate on
 * 
 *    output: value - PMALGO_FROZEN_SOIL_FLAG
 *            valid_tbs - TRUE if input TBs for this [row][col] are non-zero,
 *                        FALSE otherwise
 *
 *    result: TRUE if algorithm returns frozen soil flag, FALSE otherwise
 *
 *  Reference: 
 *
 *  Zhang, T., R.L. Armstrong, and J. Smith, 2003: Investigation of the
 *  near-surface soil freeze-thaw cycle in the contiguous United States:
 *  Algorithm development and validation, Journal of Geophysical Research,
 *  Vol. 108, No.D22:8860-8874.
 *
 *----------------------------------------------------------------------*/
#define FS_FREEZING_THRESHOLD_KELVINS 258.2
#define NSIDC_FROZEN_SOIL(t19v,t37v) \
         ((0.0 > ((t37v) - (t19v))) && \
	 FS_FREEZING_THRESHOLD_KELVINS > (t37v))

static bool run_za_s1_fs (void *private,
			  byte2 **data[PMALGO_NUM_CHANNELS], 
			  int col, int row, float *value, 
			  bool *valid_tbs) {

    assert(NULL == private);

    *valid_tbs = FALSE;

    /* 
     * Only do freezing soil test for tbs > 0
     */
    if (0 < data[PMALGO_SSMI_19V][row][col]
        && 0 < data[PMALGO_SSMI_37V][row][col]) {
        *valid_tbs = TRUE;
        if (NSIDC_FROZEN_SOIL(data[PMALGO_SSMI_19V][row][col] * TB_SCALE, 
                              data[PMALGO_SSMI_37V][row][col] * TB_SCALE)) {
            *value = PMALGO_FROZEN_SOIL_FLAG;
            return TRUE;
        }
    }

    return FALSE;

}
  

/*----------------------------------------------------------------------
 *  run_no_algorithm - Degenerate function for all channels options. 
 *                     Included for completeness, it is always a
 *                     a programming error to invoke this function.
 * 
 *    input : private - not used for this algorithm
 *            data - TB matrix [channel][row][col] (TBs in tenths of Kelvins)
 *            col - column to operate on
 *            row - row to operate on
 * 
 *    output: value - placeholder for algorithm value
 *            valid_tbs - always FALSE
 *
 *    result: FALSE 
 *
 *----------------------------------------------------------------------*/
static bool run_no_algorithm (void *private,
			      byte2 **data[PMALGO_NUM_CHANNELS], 
			      int col, int row, float *value, 
			      bool *valid_tbs) {

    assert(0);
    *valid_tbs = FALSE;
    return FALSE;

}


/*----------------------------------------------------------------------
 *  run_nasateam_ice - Calculate total and multi-year % sea ice with 
 *            "NASA Team" algorithm
 * 
 *    input : data - TB matrix [channel][row][col] (TBs in tenths of Kelvins)
 *            col - column to operate on
 *            row - row to operate on
 *            hem - hemisphere indicator (1=NORTH, 2=SOUTH)
 * 
 *    output: total - algorithm result (% total sea ice, or 
 *              PMALGO_NASATEAM_SUSPECT_ICE)
 *            myr - algorithm result (% multi-year sea ice, or 
 *              PMALGO_NASATEAM_SUSPECT_ICE)
 *            valid_tbs - TRUE if input TBs for this [row][col] are non-zero,
 *                        FALSE otherwise
 *
 *    result: FALSE if input TBs do not produce valid algorithm output,
 *            TRUE otherwise
 *
 *  Reference:
 *    See called FORTRAN routine, seaice_nasateam.f.
 *
 *----------------------------------------------------------------------*/
static bool run_nasateam_ice (byte2 **data[PMALGO_NUM_CHANNELS], 
			      int col, int row, byte2 hem, 
			      float *total, float *myr, bool *valid_tbs) {
  
    static byte1 total_ice, multi_yr_ice;

    *valid_tbs = FALSE;
    if (0 < data[PMALGO_SSMI_19V][row][col]
        && 0 < data[PMALGO_SSMI_19H][row][col]
        && 0 < data[PMALGO_SSMI_22V][row][col]
        && 0 < data[PMALGO_SSMI_37V][row][col]) {

        *valid_tbs = TRUE;
        FORTRAN_NAME(seaice_nasateam) (&(data[PMALGO_SSMI_19H][row][col]),
                                       &(data[PMALGO_SSMI_19V][row][col]),
                                       &(data[PMALGO_SSMI_22V][row][col]),
                                       &(data[PMALGO_SSMI_37V][row][col]),
                                       &hem,&total_ice,&multi_yr_ice);
        assert(multi_yr_ice <= total_ice);
        if (0 < total_ice) {
            *total = total_ice;
            *myr = multi_yr_ice;
            return TRUE;
        }
    }

    return FALSE;

}


/*----------------------------------------------------------------------
 * fetch_swe_tbs - scales TBs needed for standard SWE algorithms,
 *                 and does F17->F13 regression if the ssmi sensor is F17
 *
 *    input : private - pointer to sfc_class initialized with optional platform id
 *            byte2 in_t19h, in_t37h - stored tbs, scaled to tenths of Kelvins
 *
 *    output : float *out_t19h, *out_t37h - scaled, possibly regressed TBs
 *             (Kelvins)
 *
 *    result : n/a
 *
 * F17_TO_F13 regressions are from unpublished work done by W. Meier,
 * summer, 2009; using cold-pass data from 28 Mar 2008 - 26 Mar 2009,
 * at 4 stable targets: Canada, Dome C, Salonga, and Summit.
 * Walt's regression is probably OK to use for data from Mar 2008 to Mar 2009,
 * and then beginning August 17, 2009.
 *
 * F17REV1_TO_F13 regressions are from unpublished work done by J. Smith,
 * July, 2009, using data from May 1 - August 2, 2009, excluding data from
 * may 6, 7, 8, for "all land pixels"; probably OK to use for CLASS data
 * in this time range only.  Presumably RSS data will not suffer from the
 * same (hot load thresholding) problems.
 *
 *----------------------------------------------------------------------*/
#define F17_TO_F13_19H_REGRESSION(t19h) \
  (float) ( ( ( 1.0077388 * (double)(t19h) ) - 2.0615463 ) )
#define F17_TO_F13_37H_REGRESSION(t37h) \
  (float) ( ( ( 1.0046444 * (double)(t37h) ) - 2.6514931 ) )
#define F17REV1_TO_F13_19H_REGRESSION(t19h) \
  (float) ( ( ( 0.980828291918 * (double)(t19h) ) + 6.02592105093 ) )
#define F17REV1_TO_F13_37H_REGRESSION(t37h) \
  (float) ( ( ( 0.980979712612 * (double)(t37h) ) + 7.71317671376 ) )
static void fetch_swe_tbs( void *private, byte2 in_t19h, byte2 in_t37h,
                           float *out_t19h, float *out_t37h ) {

    assert(NULL != private);

    *out_t19h = in_t19h * TB_SCALE;
    *out_t37h = in_t37h * TB_SCALE;
        
    if ( PMALGO_F17 == ssmi_platform_sfc( (sfc_class *)private ) ) {

        *out_t19h = F17_TO_F13_19H_REGRESSION( *out_t19h );
        *out_t37h = F17_TO_F13_37H_REGRESSION( *out_t37h );

    } else if ( PMALGO_F17rev1 == ssmi_platform_sfc( (sfc_class *)private ) ) {

        *out_t19h = F17REV1_TO_F13_19H_REGRESSION( *out_t19h );
        *out_t37h = F17REV1_TO_F13_37H_REGRESSION( *out_t37h );

    }

    return;
    
}


/*----------------------------------------------------------------------
 *  run_goddard_ice - Calls the FORTRAN routine that calculates total 
 *                    and multi-year sea ice concentrations,
 *                    to reproduce "Goddard Time Series" data; 
 *                    this is the NASA Team algorithm, but uses adjusted 
 *                    tie points that minimize sensor differences.
 * 
 *    input : data - TB matrix [channel][row][col] (TBs in tenths of Kelvins)
 *            col - column to operate on
 *            row - row to operate on
 *            hem - hemisphere id
 *            platform - platform id
 *            scale_percent_seaice - flag to do output scaling to 0-100%
 * 
 *    output: total - valid algorithm result for total ice concentration
 *                    0-250 (% total sea ice scaled into range 0-250)
 *                    or, if scale_percent_seaice is TRUE,
 *                    0-100 (% total sea ice)
 *                    254   (LAND flag: any input loci values that were in
 *                           the range 1-250 will be assumed to be land 
 *                           pixels, and are reset to 254)
 *            myr - valid algorithm result for multi-year ice concentration
 *                  (same range of values as for total)
 *            valid_tbs - TRUE if input TBs for this [row][col] are non-zero,
 *                        FALSE otherwise
 *
 *    result: FALSE if input TBs do not produce valid algorithm output,
 *            TRUE otherwise
 *
 *  Reference:
 *    See called FORTRAN routine, seaice_goddard.f.
 *
 *----------------------------------------------------------------------*/
static bool run_goddard_ice (byte2 **data[PMALGO_NUM_CHANNELS], 
			     int col, int row, pmalgo_hemisphere hem, 
			     pmalgo_ssmi_platform platform,
                             bool scale_percent_seaice,
			     float *total, float *myr, bool *valid_tbs) {
  
    static byte2 total_ice, multi_yr_ice;
    static byte2 ihem, iplatform;
    static byte2 scale=0;

    if (scale_percent_seaice) {
        scale = 1;
    }
    
    *valid_tbs = FALSE;
    if (0 < data[PMALGO_SSMI_19V][row][col]
        && 0 < data[PMALGO_SSMI_19H][row][col]
        && 0 < data[PMALGO_SSMI_22V][row][col]
        && 0 < data[PMALGO_SSMI_37V][row][col]) {

        *valid_tbs = TRUE;

        /* 
         * Private data for hemisphere/platform are enumerated types.
         * Use explicit casts to byte2 for pass to the FORTRAN
         */
        ihem = (byte2)hem;
        iplatform = (byte2)platform;
        FORTRAN_NAME(seaice_goddard) (&(data[PMALGO_SSMI_19H][row][col]),
                                      &(data[PMALGO_SSMI_19V][row][col]),
                                      &(data[PMALGO_SSMI_22V][row][col]),
                                      &(data[PMALGO_SSMI_37V][row][col]),
                                      &ihem,&iplatform,&scale,&total_ice,&multi_yr_ice);
        assert(multi_yr_ice <= total_ice);
        if (0 < total_ice) {
            *total = total_ice;
            *myr = multi_yr_ice;
            return TRUE;
        }
    }

    return FALSE;

}


/*----------------------------------------------------------------------
 *  altitude_correct_tb - Adjusts the observed brightness temperature
 *     for less atmosphere at high altitudes.  This correction essentially
 *     "adds more atmosphere" back into the radiative equations, adjusting
 *     the observed TB to that which would be observed through an average
 *     reference atmosphere (1 or 1.5 km in elevation).
 *
 *    input : observed_tb - observed brightness temperature, in Kelvins
 *            elevation_meters - mean elevation of surface at
 *                               observed_tb location, in meters
 *            channel - channel id
 * 
 *    output: n/a
 *
 *    result: corrected_tb - brightness temperature, adjusted to what the satellite
 *                           would have measured if the surface were near the
 *                           model's reference elevation.
 *
 *  Reference:
 *  High-altitude atmosphere corrections are from current (Feb 2007) research
 *    by M. Savoie.  See his e-mail to me, dated 2/21/2007.
 *  New corrections, models 4-8, from Matt's e-mail to me 10/22/2007.
 *
 *----------------------------------------------------------------------*/
#define METERS_PER_KM 1000
static double altitude_correct_tb ( double tb,
                                    void *private,
                                    int2 elevation_meters,
                                    pmalgo_channel channel ) {

    double b0, b1, b2;
    double corrected_tb;
    pmalgo_high_elevation_model_id model_id;

    assert( NULL != private );

    model_id = high_elevation_model_sfc( (sfc_class *)private );
    switch ( model_id ) {
        case PMALGO_HIGH_ELEVATION_MODEL_0:
            switch ( channel ) {
                case PMALGO_SSMI_19H:
                    b0 = 10.5705585;
                    b1 = 0.9572975;
                    b2 = 0.1731657;
                    break;
                case PMALGO_SSMI_37H:
                    b0 = 19.9119194;
                    b1 = 0.9187615;
                    b2 = 0.3572513;
                    break;
                default:
                    assert( 0 );
            }
            break;
        case PMALGO_HIGH_ELEVATION_MODEL_1:
            switch ( channel ) {
                case PMALGO_SSMI_19H:
                    b0 = 8.2677115;
                    b1 = 0.9657532;
                    b2 = 0.1746945;
                    break;
                case PMALGO_SSMI_37H:
                    b0 = 15.5900654;
                    b1 = 0.9346237;
                    b2 = 0.3634160;
                    break;
                default:
                    assert( 0 );
            }
            break;
        case PMALGO_HIGH_ELEVATION_MODEL_2:
            switch ( channel ) {
                case PMALGO_SSMI_19H:
                    b0 = 11.0083846;
                    b1 = 0.9557044;
                    b2 = 0.1617780;
                    break;
                case PMALGO_SSMI_37H:
                    b0 = 20.8238632;
                    b1 = 0.9154005;
                    b2 = 0.3357933;
                    break;
                default:
                    assert( 0 );
            }
            break;
        case PMALGO_HIGH_ELEVATION_MODEL_3:
            switch ( channel ) {
                case PMALGO_SSMI_19H:
                    b0 = 8.7094050;
                    b1 = 0.9641461;
                    b2 = 0.1632062;
                    break;
                case PMALGO_SSMI_37H:
                    b0 = 16.5177544;
                    b1 = 0.9312046;
                    b2 = 0.3415876;
                    break;
                default:
                    assert( 0 );
            }
            break;
        case PMALGO_HIGH_ELEVATION_MODEL_4:
            /* reference elevation is 0.0 km */
            switch ( channel ) {
                case PMALGO_SSMI_19H:
                    b0 = 12.97250;
                    b1 = 0.9318797;
                    b2 = 1.206606;
                    break;
                case PMALGO_SSMI_37H:
                    b0 = 21.01511;
                    b1 = 0.896625;
                    b2 = 1.505518;
                    break;
                default:
                    assert( 0 );
            }
            break;
        case PMALGO_HIGH_ELEVATION_MODEL_5:
            /* reference elevation is 0.5 km */
            switch ( channel ) {
                case PMALGO_SSMI_19H:
                    b0 = 10.61837;
                    b1 = 0.940172;
                    b2 = 1.217340;
                    break;
                case PMALGO_SSMI_37H:
                    b0 = 17.52656;
                    b1 = 0.9089241;
                    b2 = 1.526162;
                    break;
                default:
                    assert( 0 );
            }
            break;
        case PMALGO_HIGH_ELEVATION_MODEL_6:
            /* reference elevation is 1.0 km */
            switch ( channel ) {
                case PMALGO_SSMI_19H:
                    b0 = 8.250874;
                    b1 = 0.9485867;
                    b2 = 1.228232;
                    break;
                case PMALGO_SSMI_37H:
                    b0 = 14.00398;
                    b1 = 0.9214469;
                    b2 = 1.547181;
                    break;
                default:
                    assert( 0 );
            }
            break;
        case PMALGO_HIGH_ELEVATION_MODEL_7:
            /* reference elevation is 1.5 km */
            switch ( channel ) {
                case PMALGO_SSMI_19H:
                    b0 = 5.918678;
                    b1 = 0.9569533;
                    b2 = 1.239062;
                    break;
                case PMALGO_SSMI_37H:
                    b0 = 10.51953;
                    b1 = 0.9339397;
                    b2 = 1.568150;
                    break;
                default:
                    assert( 0 );
            }
            break;
        default:
            assert( 0 );
    }

    corrected_tb = b0 + ( b1 * tb ) + ( b2 * (double)elevation_meters / METERS_PER_KM );

#ifdef ALTITUDEDEBUG
    
    fprintf(stderr,"ALTITUDEDEBUG:altitude_correct_tb: tb [%s]           = %f\n",
            pmalgo_channel_name[channel],
            tb);
    fprintf(stderr,"ALTITUDEDEBUG:altitude_correct_tb: model_id             = %d\n",(int)model_id);
    fprintf(stderr,"ALTITUDEDEBUG:altitude_correct_tb: elevation            = %d\n",elevation_meters);
    fprintf(stderr,"ALTITUDEDEBUG:altitude_correct_tb: corrected_tb         = %f\n",corrected_tb);

#endif
    
    return corrected_tb;

}
