/*======================================================================
 * pmalgorithm - operations to perform algorithms on passive microwave (PM)
 *               brightness temperatures (TBs)
 *
 * 03/22/1999 brodzik brodzik@zamboni.colorado.edu 303-492-8263
 * National Snow & Ice Data Center, University of Colorado, Boulder
 *
 *======================================================================*/
#ifndef pmalgorithm_h_
#define pmalgorithm_h_

#include "define.h"
#include "grids.h"
#include "land_spillover.h"

static const char pmalgorithm_h_rcsid[] = "$Header: /tmp_mnt/FILES/pmalgos/pmalgorithm.h,v 1.44.4.1 2008-06-10 20:36:35 brodzik Exp $";

/* Some algorithms have special output values, listed here */
#define PMALGO_WET_SNOW_FLAG INT2_MAX
#define PMALGO_NASATEAM_SUSPECT_ICE 157
#define PMALGO_NASATEAM_LAND 168
#define PMALGO_NASATEAM_MIN_ICE_PERCENT 1
#define PMALGO_NASATEAM_MAX_ICE_PERCENT 100
#define PMALGO_BOOTSTRAP_V1_MISSING_DATA 157
#define PMALGO_BOOTSTRAP_MISSING_DATA 110
#define PMALGO_BOOTSTRAP_OCEAN 0
#define PMALGO_BOOTSTRAP_SUSPECT_ICE 157
#define PMALGO_BOOTSTRAP_V1_LAND 168
#define PMALGO_BOOTSTRAP_V2_LAND 1200
#define PMALGO_BOOTSTRAP_LAND 243
#define PMALGO_BOOTSTRAP_MIN_ICE_PERCENT 1
#define PMALGO_BOOTSTRAP_MAX_ICE_PERCENT 100
#define PMALGO_BOOTSTRAP_V2_MIN_ICE_PERCENT 10
#define PMALGO_BOOTSTRAP_V2_MAX_ICE_PERCENT 1000
#define PMALGO_BOOTSTRAP_OPEN_WATER_PERCENT 15
#define PMALGO_BOOTSTRAP_V2_NO_SCALING 1.
#define PMALGO_BOOTSTRAP_V1_SCALE_FACTOR 10.
#define PMALGO_BOOTSTRAP_V2_SCALE_FACTOR 10.
/* GTS sea ice scales percentages (0-100%) into the range 0-250 */
#define PMALGO_GTS_LAND 254
#define PMALGO_GTS_MIN_ICE 1
#define PMALGO_GTS_MAX_ICE 250
#define PMALGO_GTS_SUSPECT_ICE 251
#define PMALGO_GTS_SCALE_DEFAULT FALSE
#define PMALGO_GTS_MIN_ICE_PERCENT 0
#define PMALGO_GTS_MAX_ICE_PERCENT 100
#define PMALGO_GTS_OPEN_WATER_PERCENT 15
#define PMALGO_GTS_SCALE_FACTOR 2.5
#define PMALGO_CHANG_NO_SNOW 0
/* Try to keep these flags unique */
#define PMALGO_CHANG_WET_SNOW -1
#define PMALGO_CHANG_UNKNOWN_SNOW -2
#define PMALGO_GB_PC1 -3
#define PMALGO_GB_PC2 -4
/* Frozen soil algorithm will just return 0 or 1 */
#define PMALGO_FROZEN_SOIL_FLAG 1
/* Default value for high elevation threshold */
#define PMALGO_UNKNOWN_THRESHOLD_ELEVATION -1000

/* 
 * Passive microwave algorithms (Order is not important)
 *
 * Acronyms/abbreviations used here:
 * BOOTN - Bootstrap (Comiso) Sea Ice, Northern Hemisphere
 * BOOTS - Bootstrap (Comiso) Sea Ice, Southern Hemisphere
 * BOOTN_V02 - Upgraded Bootstrap (Comiso) Sea Ice, Northern Hemisphere 
 *             (this version was released by GSFC on 6/11/2008).
 * BOOTS_V02 - Upgraded Bootstrap (Comiso) Sea Ice, Southern Hemisphere
 *             (this version was released by GSFC on 6/11/2008).
 * CSI - Clear Skies Index
 * DRY_COMB - COMB parameter ("TB combination") used by Hiltbrunner/Maetzler
 *       for dry snow conditions
 * FRESH_COMB - COMB parameter ("TB combination") used by Hiltbrunner/Maetzler
 *       for fresh snow conditions
 * GB - (Norm) Grody and (Alan) Basist
 * GB_SURFACE - Grody/Basist multi-choice surface classification
 * GW - algorithm developed by (Barry) Goodison and (Ann) Walker
 * HB - (Daniel) Hiltbrunner
 * MODWET - algorithm returns a wet snow indicator, derivation is 
 *          slightly modified from published wet snow indicator
 * MYICE - multi-year sea ice concentration
 * NSIDC_A - NSIDC algorithms for AMSR data
 * NSIDC_M - NSIDC algorithms for SMMR data
 * NSIDC_S - NSIDC algorithms for SSM/I data
 * NTN - NASA Team, Northern Hemisphere
 * NTS - NASA Team, Southern Hemisphere
 * GTS - Goddard Time Series
 * SEAICE - total sea ice concentration
 * SWE - Snow Water Equivalent
 * TAIT1_SWE - SWE for (Andrew) Tait's category 1 (non-forested, non-complex
 *   terrain, no melting snow, no depth hoar)
 * WET - algorithm returns a wet snow indicator
 * XPGR - Cross-Polarized Gradient Ratio (used to determine melt on ice sheets)
 * NSIDC_A2_SWE - obsolete, see 
 * NSIDC_S2_SWE - obsolete, see NSIDC_S11_SWE for replacement
 */
typedef enum {
    PMALGO_NO_ALGORITHM = -1, 
    PMALGO_NSIDC_M1_SWE, PMALGO_NSIDC_M4_SWE, 
    PMALGO_NSIDC_S0_SWE, PMALGO_NSIDC_S1_SWE, 
    /* PMALGO_NSIDC_S2_SWE, obsolete */
    PMALGO_NSIDC_S3_SWE, PMALGO_NSIDC_S4_SWE,
    PMALGO_NSIDC_S5_SWE, PMALGO_NSIDC_S6_SWE,
    PMALGO_NSIDC_S10_SWE,
    PMALGO_NSIDC_S11_SWE,
    PMALGO_NSIDC_S12_SWE,
    PMALGO_NSIDC_A0_SWE,
    /* PMALGO_NSIDC_A2_SWE, obsolete */
    PMALGO_NSIDC_A4_SWE,
    PMALGO_NSIDC_A10_SWE,
    PMALGO_GW_SWE, PMALGO_GW_WET_SWE, 
    PMALGO_GW_MODWET_SWE, PMALGO_HB_DRY_COMB, PMALGO_HB_FRESH_COMB,
    PMALGO_TAIT1_SWE, PMALGO_CHANG_DEPTH, 
    PMALGO_GB_SNOW, PMALGO_GB_SURFACE,
    PMALGO_CHANG_SURFACE, PMALGO_NSIDC_SURFACE,
    PMALGO_BOOTN_SEAICE, PMALGO_BOOTS_SEAICE,
    PMALGO_BOOTN_V02_SEAICE, PMALGO_BOOTS_V02_SEAICE,
    PMALGO_BOOTN_V02_PUB_SEAICE, PMALGO_BOOTS_V02_PUB_SEAICE,
    PMALGO_NTN_SEAICE, PMALGO_NTS_SEAICE, PMALGO_NTN_MYICE,
    PMALGO_GTS_SEAICE, PMALGO_GTS_PUB_SEAICE, PMALGO_GTS_MYICE,
    PMALGO_SMMR_XPGR, PMALGO_SSMI_XPGR, 
    PMALGO_NAGLER_DEPTH, PMALGO_NAGLER_CSI,
    PMALGO_ZA_S1_FS,
    PMALGO_SMMR_ALL_CHANNELS, PMALGO_SSMI_ALL_CHANNELS,
    PMALGO_NUM_ALGORITHMS
} pmalgo_id;
static const char *pmalgo_id_name[] = {"NSIDC_M1_SWE",
				       "NSIDC_M4_SWE",
				       "NSIDC_S0_SWE",
				       "NSIDC_S1_SWE",
				       "NSIDC_S3_SWE",
				       "NSIDC_S4_SWE",
				       "NSIDC_S5_SWE",
				       "NSIDC_S6_SWE",
				       "NSIDC_S10_SWE",
				       "NSIDC_S11_SWE",
				       "NSIDC_S12_SWE",
				       "NSIDC_A0_SWE",
				       "NSIDC_A4_SWE",
				       "NSIDC_A10_SWE",
				       "GW_SWE",
				       "GW_WET_SWE",
				       "GW_MODWET_SWE",
				       "HB_DRY_COMB",
				       "HB_FRESH_COMB",
				       "TAIT1_SWE",
				       "CHANG_DEPTH",
				       "GB_SNOW",
				       "GB_SURFACE",
				       "CHANG_SURFACE",
				       "NSIDC_SURFACE",
				       "BOOTSTRAP_N_SEAICE",
				       "BOOTSTRAP_S_SEAICE",
				       "BOOTSTRAP_N_V02_SEAICE",
				       "BOOTSTRAP_S_V02_SEAICE",
				       "BOOTSTRAP_N_V02_PUB_SEAICE",
				       "BOOTSTRAP_S_V02_PUB_SEAICE",
				       "NASATEAM_N_SEAICE",
				       "NASATEAM_S_SEAICE",
				       "NASATEAM_N_MYICE",
				       "GODDARD_SEAICE",
                                       "GODDARD_PUB_SEAICE",
				       "GODDARD_MYICE",
				       "SMMR_XPGR",
				       "SSMI_XPGR",
                       "NAGLER_DEPTH",
                       "NAGLER_CSI",
				       "ZA_S1_FS",
				       "SMMR_ALL_CHANNELS",
				       "SSMI_ALL_CHANNELS"};

/* algorithm list, intended for use in usage messages of calling programs */
#define PMALGO_ALGORITHM_DESCRIPTIONS \
"           NSIDC_M1_SWE - (formerly CHANG_SWE) Chang algorithm for SMMR TBs (mm)\n"\
"           NSIDC_M4_SWE - NSIDC4_SWE, adjusted for SMMR TBs (mm)\n"\
"           NSIDC_S0_SWE - (formerly CHANG_SSMI_SWE) unadjusted Chang algorithm for SSM/I TBs (mm)\n"\
"           NSIDC_S1_SWE - (formerly NSIDC1_SWE) Chang algorithm adjusted for SSM/I TBs (mm)\n"\
"           NSIDC_S3_SWE - (formerly NSIDC3_SWE) NSIDC decision tree for SSM/I TBs (mm)\n"\
"              (requires NSIDC_SURFACE to determine where to calculate SWE)\n"\
"           NSIDC_S4_SWE - NSIDC1 + forest correction on SWE (mm)\n"\
"           NSIDC_S5_SWE - (formerly NSIDC5_SWE) NSIDC1 + Grody/Basist precip test (mm + flag)\n"\
"           NSIDC_S6_SWE - (formerly NSIDC6_SWE) NSIDC5, + forest correction on SWE (mm + flag)\n"\
"           NSIDC_S10_SWE - NSIDC4 SWE, using TB regressions SWE (mm)\n"\
"           NSIDC_S11_SWE - NSIDC10 SWE, with atmospheric correction at high altitudes (mm)\n"\
"           NSIDC_S12_SWE - NSIDC1 SWE, with atmospheric correction at high altitudes (mm)\n"\
"           NSIDC_A0_SWE - unadjusted NSIDC1 algorithm for AMSR TBs (mm)\n"\
"           NSIDC_A4_SWE - unadjusted NSIDC4 algorithm for AMSR TBs (mm)\n"\
"           NSIDC_A10_SWE - unadjusted NSIDC10 algorithm for AMSR TBs (mm)\n"\
"           GW_SWE - Goodison Walker algorithm (mm)\n"\
"           GW_WET_SWE - Goodison Walker w/wet snow flag (mm + flag)\n"\
"           GW_MODWET_SWE - Goodison Walker w/modified wet snow flag \n"\
"              (mm + flag)\n"\
"           HB_DRY_COMB - Hiltbrunner dry snow TB combination parameter (K)\n"\
"           HB_FRESH_COMB - Hiltbrunner fresh snow TB combination parameter\n"\
"              (K)\n"\
"           TAIT1_SWE - Tait Category 1 (simplest conditions) (mm)\n"\
"           CHANG_DEPTH - Chang AMSR algorithm adjusted for SSM/I (cm) (requires\n"\
"              CHANG_SURFACE to determine where to calculate snow depth)\n"\
"           GB_SNOW - Grody/Basist snow (bad obs/no snow/snow flags)\n"\
"           GB_SURFACE - Grody/Basist surface classification (various flags)\n"\
"           CHANG_SURFACE - Chang surface classification (various flags)\n"\
"           NSIDC_SURFACE - NSIDC surface classification (various flags)\n"\
"           BOOTSTRAP_N_SEAICE - Bootstrap, No. Hemisphere (%% total sea ice)\n"\
"           BOOTSTRAP_S_SEAICE - Bootstrap, So. Hemisphere (%% total sea ice)\n"\
"           BOOTSTRAP_N_V02_SEAICE - Bootstrap version 2, No. Hemisphere, NSIDC regressions (%% total sea ice)\n"\
"           BOOTSTRAP_S_V02_SEAICE - Bootstrap version 2, So. Hemisphere, NSIDC regressions (%% total sea ice)\n"\
"           BOOTSTRAP_N_V02_PUB_SEAICE - Bootstrap version 2, No. Hemisphere, GSFC regressions (%% total sea ice)\n"\
"           BOOTSTRAP_S_V02_PUB_SEAICE - Bootstrap version 2, So. Hemisphere, GSFC regressions (%% total sea ice)\n"\
"           NASATEAM_N_SEAICE - NASATeam, No. Hemisphere (%% total sea ice)\n"\
"           NASATEAM_S_SEAICE - NASATeam, So. Hemisphere (%% total sea ice)\n"\
"           NASATEAM_N_MYICE - NASATeam, No. Hemisphere (%% multi-year sea ice)\n"\
"           GODDARD_SEAICE - Goddard Time Series (%% total sea ice)\n"\
"           GODDARD_PUB_SEAICE - Goddard Time Series, Published Tie-points (%% total sea ice)\n" \
"           GODDARD_MYICE - Goddard Time Series (%% multi-year sea ice)\n"\
"           SMMR_XPGR - SMMR Cross-Polarized Gradient Ratio (unitless)\n"\
"           SSMI_XPGR - SSM/I Cross-Polarized Gradient Ratio (unitless)\n"\
"           NAGLER_DEPTH - SSM/I Nagler/Rott snow depth (cm)\n"\
"           NAGLER_CSI - SSM/I Nagler/Rott clear skies index (Kelvins)\n"\
"           ZA_S1_FS - Zhang-Armstring frozen soil algorithm for SSM/I TBs (flag)\n"\
"           SMMR_ALL_CHANNELS - All SMMR channels, no algorithm\n"\
"           SSMI_ALL_CHANNELS - All SSMI channels, no algorithm\n"

/* passive microwave sensors */
typedef enum {
    PMALGO_NO_SENSOR = -1, PMALGO_SMMR, PMALGO_SSMI, PMALGO_AMSR, PMALGO_NUM_SENSORS
} pmalgo_sensor;
static const char *pmalgo_sensor_name[] = { "SMMR", "SSMI", "AMSR" };

/* SSM/I satellite platforms */
typedef enum {
    PMALGO_NO_PLATFORM = 0,
    PMALGO_F08,
    PMALGO_F11,
    PMALGO_F13,
    PMALGO_F15_BRIDGE,
    PMALGO_F15,                 /* placeholder for now */
    PMALGO_F17,                 /* uses W. Meier tie-points */
    PMALGO_F17_final,           /* uses final tie-points */
    PMALGO_F17rev1
} pmalgo_ssmi_platform;
static const int PMALGO_NUM_PLATFORMS = 8;
static const char *pmalgo_ssmi_platform_name[] = { "NA",
                                                   "F08",
                                                   "F11",
                                                   "F13",
                                                   "F15_BRIDGE",
                                                   "F15",
                                                   "F17",
                                                   "F17_final",
                                                   "F17rev1"
};

/* passive microwave channels */
typedef enum { 
    PMALGO_NO_CHANNEL = -1, 
    PMALGO_SMMR_06V, PMALGO_SMMR_06H, PMALGO_SMMR_10V, PMALGO_SMMR_10H, 
    PMALGO_SMMR_18V, PMALGO_SMMR_18H, PMALGO_SMMR_21V, PMALGO_SMMR_21H, 
    PMALGO_SMMR_37V, PMALGO_SMMR_37H, PMALGO_SSMI_19V, PMALGO_SSMI_19H, 
    PMALGO_SSMI_22V, PMALGO_SSMI_37V, PMALGO_SSMI_37H, PMALGO_SSMI_85V, 
    PMALGO_SSMI_85H,
    PMALGO_AMSR_06V, PMALGO_AMSR_06H, PMALGO_AMSR_10V, PMALGO_AMSR_10H, 
    PMALGO_AMSR_18V, PMALGO_AMSR_18H, PMALGO_AMSR_23V, PMALGO_AMSR_23H, 
    PMALGO_AMSR_36V, PMALGO_AMSR_36H, PMALGO_AMSR_89V, PMALGO_AMSR_89H,
    PMALGO_NUM_CHANNELS
} pmalgo_channel;

/* 
 * These match filename extensions for the respective TB files 
 * that contain each channel's data 
 */
static const char *pmalgo_channel_name[] = {
    "06V", "06H", "10V", "10H", "18V", "18H",
    "21V", "21H", "37V", "37H",
    "19V", "19H", "22V", "37V", "37H", "85V", "85H",
    "06V", "06H", "10V", "10H", "18V", "18H",
    "23V", "23H", "36V", "36H", "89V", "89H",
};

/* 
 * Passive microwave orbital pass directions: the satellites are
 * sun-synchronous, so each orbit will have a "cold" and "warm" pass
 * direction
 */
typedef enum {
    PMALGO_COLD_PASS_DIRECTION_ONLY, PMALGO_WARM_PASS_DIRECTION_ONLY,
    PMALGO_EITHER_PASS_DIRECTION, PMALGO_BOTH_PASS_DIRECTIONS,
    PMALGO_NUM_PASS_DIRECTIONS
} pmalgo_direction;
static const char *pmalgo_direction_name[] = {"COLD_PASS_DIRECTION_ONLY", 
					      "WARM_PASS_DIRECTION_ONLY", 
					      "EITHER_PASS_DIRECTION",
					      "BOTH_PASS_DIRECTIONS" };


/* hemisphere (used in sea ice algorithms, to determine tie points */
typedef enum {
    PMALGO_NO_HEMISPHERE = 0, PMALGO_NORTH, PMALGO_SOUTH
} pmalgo_hemisphere;
static const char *pmalgo_hemisphere_name[] = { "NA", "N", "S" };

/* Various models for the atmospheric correction for high elevation surfaces */
typedef enum {
    PMALGO_NO_HIGH_ELEVATION_MODEL = -1,
    PMALGO_HIGH_ELEVATION_MODEL_0,
    PMALGO_HIGH_ELEVATION_MODEL_1,
    PMALGO_HIGH_ELEVATION_MODEL_2,
    PMALGO_HIGH_ELEVATION_MODEL_3,
    PMALGO_HIGH_ELEVATION_MODEL_4,
    PMALGO_HIGH_ELEVATION_MODEL_5,
    PMALGO_HIGH_ELEVATION_MODEL_6,
    PMALGO_HIGH_ELEVATION_MODEL_7,
    PMALGO_NUM_HIGH_ELEVATION_MODELS
} pmalgo_high_elevation_model_id;

/* High elevation model list, intended for use in usage messages of
   calling programs */
#define PMALGO_HIGH_ELEVATION_MODEL_DESCRIPTIONS \
"           0 : reference elevation=1   km, elevations > 3200 m\n"\
"           1 : reference elevation=1.5 km, elevations > 3200 m\n"\
"           2 : reference elevation=1   km, elevations > 4000 m\n"\
"           3 : reference elevation=1.5 km, elevations > 4000 m\n"\
"           4 : reference elevation=0.0 km, elevations > 3200 m\n"\
"           5 : reference elevation=0.5 km, elevations > 3200 m\n"\
"           6 : reference elevation=1.0 km, elevations > 3200 m\n"\
"           7 : reference elevation=1.5 km, elevations > 3200 m\n"

/* Gridding algorithm choices */
typedef enum {
    PMALGO_BACKUS_GILBERT, PMALGO_DITB, PMALGO_INVERSE_DISTANCE_WEIGHTING,
    PMALGO_NUM_GRID_ALGORITHMS
} pmalgo_grid_algorithm;
static const char *pmalgo_grid_algorithm_name[] = {
    "BACKUS_GILBERT",
    "DITB (drop-in-the-bucket)",
    "INVERSE_DISTANCE_WEIGHTING"
};

/* Structure for algorithm-specific parameters */
typedef struct {
    pmalgo_id id;            /* passive microwave algorithm id */
    pmalgo_sensor sensor;    /* PM sensor (determines channels to use) */
    pmalgo_ssmi_platform platform;    /* SSM/I platform id */
    int num_channels;        /* number of channels */
    pmalgo_channel channel[PMALGO_NUM_CHANNELS]; 
    /* array of channel ids needed */
    int num_days;            /* number of days to run algorithm */
    char *direction;         /* recommended pass direction for this sensor */
    char *grid_algorithm;    /* recommended TB gridding algorithm to use */
    bool (*run)(void *private,
                byte2 **data[PMALGO_NUM_CHANNELS], int col, int row, 
                float *value, bool *valid_tbs);
    /* runs the current algorithm on a matrix element*/
    void *private;           /* private algorithm info */
} pmalgo_class;
  
  
/* Function prototypes */
pmalgo_class *init_pmalgo(char *algorithm_name,
                          char *platform_name,
			  int verbose);
bool init_bootstrap_pmalgo(pmalgo_class *this,
			   const char *date,
			   const int bootstrap_recoding_pcent,
			   const bool scale_percent_seaice,
			   land_spillover_class *land_spillover,
			   grid_class *grid,
			   int verbose);
void mask_bootstrap_pmalgo(pmalgo_class *this,
			   byte1 **loci, byte1 use_value);
bool init_goddard_pmalgo(pmalgo_class *this,
                         const int goddard_recoding_pcent,
			 const char *hemisphere_name,
                         const bool scale_percent_seaice,
                         land_spillover_class *land_spillover,
                         grid_class *grid,
			 int verbose);
bool init_sfc_pmalgo(pmalgo_class *this,
		     int month,
		     const char *surface_filename,
		     const char *forest_cover_filename,
		     const char *landmask_filename,
		     const char *elevation_filename,
		     grid_class *grid,
                     byte1 high_elevation_model_id,
                     int2 threshold_elevation_meters,		         
		     int verbose);
bool postprocessing_filter_pmalgo( pmalgo_class *this,
                                   grid_class *grid,
                                   float **data,
                                   bool use_climatology,
                                   byte1 **climatology,
                                   bool use_loci,
                                   byte1 **loci,
                                   bool fill_algo_pixels_only,
                                   bool do_user_fill_value,
                                   float fill_value,
                                   int verbose );
void close_pmalgo(pmalgo_class *this);

#endif
