/*======================================================================
 * pmalgos - program to calculate various geophysical products from
 *           gridded passive microwave (PM) brightness temperatures (TBs).
 *
 * This file contains two main programs, compilation is controlled by the
 *    PMALGOS_INFO in the makefile in this directory.  The main programs are:
 *
 * 1) pmalgos_info is a command-line utility to return algorithm-specific
 *                 parameters needed to run a given algorithm (i.e. which
 *                 sensor the TBs should be from, which channels are needed,
 *                 etc.)
 * 2) pmalgos is the program that will execute the algorithm, given
 *                 access to the required gridded TB files for that algorithm.
 *
 * The user should beware of interaction between input fill_value, loci_values,
 * and algorithm output, and ensure that loci_values and/or fill_value will
 * not overlap the range of algorithm output.  In the following table,
 * "algorithm output" means the algorithm returned a legitimate value
 * when non-zero TBs were input, e.g. the calculation of "snow water
 * equivalent" was positive.
 *
 * In addition to the input conditions outlined below, use of a
 * climatology means that, if algorithm output is legitimate, it will
 * be saved for only those pixels where the climatology indicates that this
 * parameter is "likely," i.e. the climatology is "on".
 *
 * Input Conditions     Output
 * ----------------     ------
 * no fill_value,       Output array will be initialized to zeroes,
 * no loci_values        all pixels with non-zero TBs will be set to
 *                       algorithm output
 *
 * fill_value,          Output array will be initialized to zeroes,
 * no loci_values        all pixels with zero TBs will be set to fill_value,
 *                       all pixels with non-zero TBs will will be set to
 *                       algorithm output
 *
 * no fill_value,       Output array will be initialized to loci file values,
 * loci_values           algorithm will only be run on pixels where loci mask
 *                       matches one of input loci_values, and any of those
 *                       pixels with non-zero TBs will be set to algorithm
 *                       output
 *
 * fill_value,          Output array will be initialized to loci file values,
 * loci_values           all pixels with zero TBs will be set to fill_value,
 *                       algorithm will only be run on remaining pixels where
 *                       loci mask matches one of input loci_values, and any
 *                       of those pixels will be set to algorithm output
 *
 * An appropriate way to use both fill_value and loci_values would be
 * to run the SMMR_XPGR algorithm (whose output values range from -1. to 1.)
 * with a loci mask where ice sheet pixels are set to 10 and all other
 * pixels are set to 20, and a fill_value of 30.  The output array will
 * contain a typical "spiral pattern" of fill_values (indicating missing
 * TBs due to orbit coverage), and XPGR output for all of the ice sheet
 * pixels that had legitimate TB coverage.
 *
 * 02/04/1998 brodzik brodzik@nsidc.org 303-492-8263
 * National Snow & Ice Data Center, University of Colorado, Boulder
 *
 * Copyright (C) 1998-2011 Regents of the University of Colorado. The Regents
 * of the University of Colorado grants, and you accept, a personal,
 * nonexclusive, nontransferable license to use this software, at no charge,
 * in accordance with the terms herein, solely for internal research and
 * development, which is limited to testing, measuring, assessing, evaluating
 * the software and/or data that you generate using this software, and not for
 * any commercial purpose. You may not develop derivative works based on this
 * software, or redistribute this software.
 *
 *======================================================================*/
#include "define.h"
#include "matrix.h"
#include "matrix_io.h"
#include "maps.h"
#include "pmalgorithm.h"
#include "pmtbs.h"
#include "land_spillover.h"
#include "bootstrap.h"

static const char pmalgos_c_rcsid[] = "$Id: pmalgos.c 22410 2013-01-30 20:23:10Z kbeam $";

#define MAX_LOCI_VALUES 10
#define CLIMATOLOGY_OFF 0

/*
 * Beginning of pmalgos_info, command-line utility to return the specific
 * requirements of a given algorithm
 */
#ifdef PMALGOS_INFO

/*----------------------------------------------------------------------
 * usage_exit - Print the usage message to stderr and exit in error
 *
 *	result: Print the usage message to stderr.
 *
 *----------------------------------------------------------------------*/
static void usage_exit(void)
{
    int i;

    fprintf(stderr,"usage: pmalgos_info -v algorithm\n");
    fprintf(stderr,"\n");
    fprintf(stderr,"pmalgos_info $Revision: 1.33.4.1 $ \n");
    fprintf(stderr," input : algorithm - algorithm name, possible values:\n");
    fprintf(stderr,"%s\n",PMALGO_ALGORITHM_DESCRIPTIONS);
    fprintf(stderr,"\n");
    fprintf(stderr," output : (to STDOUT) list of parameters describing\n");
    fprintf(stderr,"          sensor, required channels, number of days, \n");
    fprintf(stderr,"          and recommended options \n");
    fprintf(stderr,"\n");
    fprintf(stderr,"          SENSOR is SMMR or SSMI\n");
    fprintf(stderr,"          CHANNELS are sensor-specific\n");
    fprintf(stderr,"          PASS_DIRECTIONS, one of:\n");
    fprintf(stderr,"          (%s",pmalgo_direction_name[0]);
    for (i = 1; i < PMALGO_NUM_PASS_DIRECTIONS; i++) {
        fprintf(stderr,", %s",pmalgo_direction_name[i]);
    }
    fprintf(stderr,")\n");
    fprintf(stderr,"          GRID_ALGORITHM, one of:\n");
    fprintf(stderr,"          (%s",pmalgo_grid_algorithm_name[0]);
    for (i = 1; i < PMALGO_NUM_GRID_ALGORITHMS; i++) {
        fprintf(stderr,", %s",pmalgo_grid_algorithm_name[i]);
    }
    fprintf(stderr,")\n");
    fprintf(stderr,"\n");
    fprintf(stderr," (the caller to pmalgos is free to ignore options, but \n");
    fprintf(stderr," should understand passive microwave algorithm may not \n");
    fprintf(stderr," be appropriate for other than recommended options).\n");
    fprintf(stderr," Sample output from pmalgos_info NSIDC1:\n");
    fprintf(stderr,"          SENSOR:          SSMI\n");
    fprintf(stderr,"          CHANNELS:        19H 37H \n");
    fprintf(stderr,"          NUMBER_OF_DAYS:  1 \n");
    fprintf(stderr,"          PASS_DIRECTIONS: COLD_PASS_DIRECTION_ONLY \n");
    fprintf(stderr,"          GRID_ALGORITHM:  BACKUS_GILBERT \n");
    fprintf(stderr,"\n");
    fprintf(stderr," option : v - verbose mode \n");
    fprintf(stderr,"\n");
    fprintf(stderr,"%s\n",pmalgos_c_rcsid);
    fprintf(stderr,"\n");

    exit(EXIT_FAILURE);

}

main (int argc, char *argv[]) {
    char *option;
    int i;
    int verbose=0;
    char algorithm_name[MAX_STRING];
    char platform_name[MAX_STRING] = "NA";
    pmalgo_class *algorithm;

    /*
     *      get command line arguments
     */
    while (--argc > 0 && (*++argv)[0] == '-') {
        for (option = argv[0]+1; *option != '\0'; option++) {
            switch (*option) {
                case 'v':
                    verbose++;
                    break;
                default:
                    fprintf (stderr, "pmalgos_info: invalid option %c\n", *option);

                    usage_exit();
            } /* end switch */
        } /* end loop for each input command option */
    } /* end loop while still input arguments */

    if (1 != argc) {
        fprintf(stderr,
                "pmalgos_info: Insufficient number of calling arguments.\n\n");
        usage_exit();
    }

    strncpy (algorithm_name, *argv, MAX_STRING);
    ++argv; --argc;

    if (0 < verbose) {
        fprintf(stderr,"> pmalgos_info: Beginning processing for:\n");
        fprintf(stderr,">  algorithm =%s\n",algorithm_name);
    }

    /* Initialize algorithm information, and print it to stdout */
    algorithm = init_pmalgo(algorithm_name, platform_name, verbose);
    if (NULL == algorithm) {
        fprintf(stderr,"pmalgos_info: error initializing algorithm %s\n",
                algorithm_name);
        exit(EXIT_FAILURE);
    } else if (0 < verbose) {
        fprintf(stderr,"> pmalgos_info: algorithm initialized OK\n");
    }

    /* Print the algorithm information to stdout */
    printf("SENSOR: %s\n",pmalgo_sensor_name[algorithm->sensor]);
    printf("CHANNELS:");
    for (i = 0; i < algorithm->num_channels; i++) {
        printf(" %s",pmalgo_channel_name[algorithm->channel[i]]);
    }
    printf("\n");
    printf("NUMBER_OF_DAYS: %d\n",algorithm->num_days);
    printf("PASS_DIRECTIONS: %s\n",algorithm->direction);
    printf("GRID_ALGORITHM: %s\n",algorithm->grid_algorithm);

    if (0 < verbose) {
        fprintf(stderr,"> pmalgos_info: exiting successfully\n");
    }

    close_pmalgo(algorithm);
    exit (EXIT_SUCCESS);

}

/*
 * Beginning of pmalgos, command-line program to perform
 * PM algorithms on gridded TBs.
 */
#else /* end of main for pmalgos_info */

/*----------------------------------------------------------------------
 * usage_exit - Print the usage message to stderr and exit in error
 *
 *	result: Print the usage message to stderr.
 *
 *----------------------------------------------------------------------*/
static void usage_exit(void)
{

    fprintf(stderr,"usage: pmalgos [-bijvz] [-c climatology_filename]\n");
    fprintf(stderr,"       [-a percent - spillover open water cutoff (%%)] \n");
    fprintf(stderr,"       [-d yyyymmdd] \n");
    fprintf(stderr,"       [-e elevation_filename] [-f fill_value [-r]]\n");
    fprintf(stderr,"       [-g chang_surface] [-h month] [-l loci_file] \n");
    fprintf(stderr,"       [-m landmask] [-n high_elevation_model_number] \n");
    fprintf(stderr,"       [-o outfile_name] \n");
    fprintf(stderr,"       [-p platform] [-q hemisphere] [-s threshold_elevation] \n");
    fprintf(stderr,"       [-t forest_cover]\n");
    fprintf(stderr,"       [-u <up to 10 loci values to use>]\n");
    fprintf(stderr,"       [-w weightmask]\n");
    fprintf(stderr,"       [-x percent goddard recoding cutoff (%%) \n");
    fprintf(stderr,"       gpd_file algorithm [tb_grids_basename]\n");
    fprintf(stderr,"\n");
    fprintf(stderr,"pmalgos $Revision: 1.33.4.1 $ \n");
    fprintf(stderr," input : gpd file  - grid parameters definition file\n");
    fprintf(stderr,"         algorithm - algorithm name, possible values:\n");
    fprintf(stderr,"%s",PMALGO_ALGORITHM_DESCRIPTIONS);
    fprintf(stderr,"         tb_grids_basename - basename (no extension)");
    fprintf(stderr," of TB data files to use\n");
    fprintf(stderr,"           (assumes channel extensions,");
    fprintf(stderr,"e.g. tb_grids_basename.19H, .19V etc,\n");
    fprintf(stderr,"           unless -b option set ");
    fprintf(stderr,"in which case they use modern extensions\n");
    fprintf(stderr,"           e.g. tb_grids_basename19h.bin, 19v.bin etc.)\n");
    fprintf(stderr,"\n");
    fprintf(stderr," output : passive microwave-derived product");
    fprintf(stderr," (floating point, same dimensions as input TB arrays\n");
    fprintf(stderr,"\n");
    fprintf(stderr," option : c filename - climatology file to use (default none)\n");
    fprintf(stderr,"          b - use modern filename extensions (19v.bin)\n" );
    fprintf(stderr,"          i - read input tb_grids_basename from stdin\n" );
    fprintf(stderr,"          f fill_value - value to set output to when");
    fprintf(stderr," input TBs are zeroes\n" );
    fprintf(stderr,"              (default is loci value for this pixel,");
    fprintf(stderr," or 0 if no loci used)\n");
    fprintf(stderr,"          o outfile_name - output file to create ");
    fprintf(stderr," (default is to stdout)\n");
    fprintf(stderr,"          l loci filename - filename containing starting ");
    fprintf(stderr," mask with\n");
    fprintf(stderr,"            land/ocean/coastline/ice_sheet classifications\n");
    fprintf(stderr,"            (required for all Bootstrap seaice algorithms)\n");
    fprintf(stderr,"            default is to operate on all pixels\n");
    fprintf(stderr,"          p platform - (F08, F11, F13, F15_BRIDGE, F17)\n");
    fprintf(stderr,"            or F17rev1 (F17rev1 for SWE algorithms only) \n");
    fprintf(stderr,"          r - (this option is only used with -f option)\n");
    fprintf(stderr,"              use fill_value only on pixels where");
    fprintf(stderr," algorithm was run\n");
    fprintf(stderr,"              (default is fill_value in any pixel");
    fprintf(stderr," with zeroes for TBs, so \n");
    fprintf(stderr,"              fill_value ends up on top of loci mask values)\n");
    fprintf(stderr,"          u value - unsigned byte values of loci mask ");
    fprintf(stderr," pixels to operate on \n" );
    fprintf(stderr,"              (may include up to 10 values)\n" );
    fprintf(stderr,"          v - verbose mode \n");
    fprintf(stderr,"***The following options are used for GODDARD_SEAICE time series\n");
    fprintf(stderr,"   seaice algorithms:\n");
    fprintf(stderr,"          p platform - (F08, F11, F13, F15_BRIDGE, F17)\n");
    fprintf(stderr,"          q hemisphere - (N or S)\n");
    fprintf(stderr,"          z - scale output to percent range 0-100%% (default is 0-250)\n");
    fprintf(stderr,"          j - do spillover_correction (default is no spillover)\n");
    fprintf(stderr,"              If spillover is on, m and w options are required:\n");
    fprintf(stderr,"              a percent - spillover open water cutoff 0-100%% (default is %d%%)\n",
            PMALGO_GTS_OPEN_WATER_PERCENT );
    fprintf(stderr,"              m landmask - bitwise spillover landmask\n");
    fprintf(stderr,"                   (byte data, 0=ocean, 1=land)\n");
    fprintf(stderr,"              w weightmask - minimum seaice concentration\n");
    fprintf(stderr,"                   to use for spillover weights \n");
    fprintf(stderr,"              x percent - goddard recoding threshold 0-100%% (default is %d%%)\n",
            PMALGO_GTS_MIN_ICE_PERCENT );
    fprintf(stderr,"***The following option may be used (optional) for the BOOTSTRAP_N_SEAICE\n");
    fprintf(stderr,"   and BOOTSTRAP_S_SEAICE algorithms:\n");
    fprintf(stderr,"          p platform - (F08, F11, F13)\n");
    fprintf(stderr,"***The following option may be used with the BOOTSTRAP_N_V02_PUB_SEAICE and\n");
    fprintf(stderr,"   BOOTSTRAP_S_V02_PUB_SEAICE algorithms to select the latest published\n");
    fprintf(stderr,"   regressions:\n");
    fprintf(stderr,"          p platform - (F17)\n");
    fprintf(stderr,"***The following option is required for Bootstrap\n");
    fprintf(stderr,"   seaice algorithms, ignored for all others:\n");
    fprintf(stderr,"          d yyyymmdd - date of TBs\n");
    fprintf(stderr,"***The following options are required for CHANG_DEPTH or\n");
    fprintf(stderr,"       NSIDC_S3_SWE\n");
    fprintf(stderr,"          h month  - month of TBs (1-12)\n");
    fprintf(stderr,"          g chang_surface - output from CHANG_ or NSIDC_\n");
    fprintf(stderr,"            SURFACE (fdistilled to byte data)\n");
    fprintf(stderr,"          m landmask - bitwise landmask (byte data)\n");
    fprintf(stderr,"          t forest_cover - forest_cover (float)\n");
    fprintf(stderr,"***The following options are required for CHANG_ or\n");
    fprintf(stderr,"       NSIDC_SURFACE\n");
    fprintf(stderr,"          m landmask - bitwise landmask (byte data)\n");
    fprintf(stderr,"          t forest_cover - forest_cover (float)\n");
    fprintf(stderr,"***The following option is required for NSIDC_A4_SWE, NSIDC_M4_SWE\n");
    fprintf(stderr,"       NSIDC_S4_SWE, NSIDC_S6_SWE, NSIDC_A10_SWE\n");
    fprintf(stderr,"       and NSIDC_S10_SWE\n");
    fprintf(stderr,"          t forest_cover - forest_cover (float)\n");
    fprintf(stderr,"***The following options are required for NSIDC_S11_SWE\n");
    fprintf(stderr,"          e elevation - elevation (meters)\n");
    fprintf(stderr,"          n high_elevation_model_number - model number, possible values:\n");
    fprintf(stderr,"%s",PMALGO_HIGH_ELEVATION_MODEL_DESCRIPTIONS);
    fprintf(stderr,"          s threshold_elevation - elevation (meters) for applying model\n");
    fprintf(stderr,"          t forest_cover - forest_cover (float)\n");
    fprintf(stderr,"***The following options are required for NSIDC_S12_SWE\n");
    fprintf(stderr,"          e elevation - elevation (meters)\n");
    fprintf(stderr,"          n high_elevation_model_number - model number, possible values:\n");
    fprintf(stderr,"%s",PMALGO_HIGH_ELEVATION_MODEL_DESCRIPTIONS);
    fprintf(stderr,"          s threshold_elevation - elevation (meters) for applying model\n");
    fprintf(stderr,"\n");
    fprintf(stderr,"%s\n",pmalgos_c_rcsid);
    fprintf(stderr,"\n");

    exit(EXIT_FAILURE);

}



/********************************************************************************
*  prefill_output                                                               
*                                                                               
*    prefill the output data field with the information from the loci mask.  In 
*    the special case of BOOTSTRAP V2, we must replace land values in the loci  
*    with the default land value for BOOTSTRAP v2: PMALGO_BOOTSTRAP_V2_LAND.    
*                                                                               
*     Input:                                                                    
*       out_data -  output matirx data                                          
*       loci - loci mask information                                            
*       pmtbs - pmtbs class to get information about the current algorithm.     
*                                                                               
*********************************************************************************/
bool prefill_output(float** out_data, byte1** loci, pmtbs_class* pmtbs,
                    bool scale_percent_seaice) 
{

    bool is_bootv2=FALSE;
    int row, col;
    grid_class* grid;
    float scale_factor=1.;
    float land_value;

    assert(NULL != pmtbs);

    grid = pmtbs->grid;
    
    /* Because BOOTSTRAP V2 uses values larger than a byte, we have to prefill
     * the output smartly, with the correct output land value and scaled if
     * desired.  */
    if(is_bootstrap_v02_algorithm(pmtbs->algorithm->id)){
        is_bootv2 = TRUE;
        if (scale_percent_seaice) {
            scale_factor = PMALGO_BOOTSTRAP_V2_MIN_ICE_PERCENT;
        }
        land_value = (float)PMALGO_BOOTSTRAP_V2_LAND / scale_factor;
    }

    

    for (row = 0; row < grid->rows; row++) {
        for (col = 0; col < grid->cols; col++) {
            out_data[row][col] = (float)loci[row][col];
            if (is_bootv2
                && loci[row][col] == PMALGO_BOOTSTRAP_LAND) {
                out_data[row][col] = land_value;
            }
        }
    }

    return TRUE;
}


/*------------------------------------------------------------------------
 * protect_invalid_loci_values
 *
 *
 *
 *	Input:
 *		loci - loci array
 *      grid - grid class to get size of loci
 *      min_ice - min value of loci to replace
 *      max_ice - max value of loci to replace
 *      land_value - value to replace loci's original value with
 *      num_loci_values - number of valid loci values
 *      loci_value - ptr to array of valid loci values
 *      errstr - char string to output for notification
 *
 *	effect: If any input loci values are between <min_ice> and <max_ice>,
 *         reset these to the output <land_value> before continuing.  Check
 *         that the user has not asked that the algorithm be run on loci
 *         values in this range, since it is the range of legitimate sea ice %
 *         and would create ambiguous results.
 *
 *------------------------------------------------------------------------*/
void protect_invalid_loci_values(byte1 **loci,
                             grid_class* grid,
                             byte1 min_ice,
                             byte1 max_ice,
                             byte1 land_value,
                             int num_loci_values,
                             byte2 *loci_value,
                             char* errstr)
{
    /* fill loci with correct values for the desired */
    int row, col, i;
    
    for (row = 0; row < grid->rows; row++) {
        for (col = 0; col < grid->cols; col++) {
            if ( min_ice <= loci[row][col] && loci[row][col] <= max_ice ){
                loci[row][col] = land_value;
            }
        }
    }
    
    for (i = 0; i < num_loci_values; i++) {
        if (min_ice <= loci_value[i] &&
            loci_value[i] <= max_ice) {
            fprintf(stderr,"pmalgos: ERROR: invalid use_loci=%d ",loci_value[i]);
            fprintf(stderr, errstr);
            exit(EXIT_FAILURE);
        }
    }
}


/*------------------------------------------------------------------------
 * check_valid_scale_switch - Warn if -z options was set for algorithms that
 *                            don't support rescaling of the data
 *
 *	scaling_seaice : boolean
 *		flag if we are scaling sea ice to 1-100% or not.
 *
 *  algorithm_name : char*
 *      algorithm name to run
 *
 *  verbose : bool
 *      whether to warn or not.
 *      
 *	effect: warn if illegal algorithm name. Otherwise no effect.
 *
 *------------------------------------------------------------------------*/
void check_valid_scale_switch (bool scaling_seaice, char* algorithm_name, bool verbose)
{

    pmalgo_id valid_algo_idx[]= {PMALGO_GTS_SEAICE,
                                 PMALGO_GTS_PUB_SEAICE,
                                 PMALGO_GTS_MYICE,
                                 PMALGO_BOOTN_V02_SEAICE,
                                 PMALGO_BOOTS_V02_SEAICE,
                                 PMALGO_BOOTN_V02_PUB_SEAICE,
                                 PMALGO_BOOTS_V02_PUB_SEAICE};

    bool is_valid = FALSE;
    int num_valid_algos = 7;
    int i;

    if (! scaling_seaice ) {
        return;
    }


    for ( i = 0; i < num_valid_algos; i++) {
        if (streq(algorithm_name, pmalgo_id_name[valid_algo_idx[i]])){
            is_valid=TRUE;
            break;
        }
    }

    if (! is_valid && 0 < verbose ) {
        fprintf(stderr,"invalid option '-z' ignoring for this algorithm");
    }
}


main (int argc, char *argv[]) {
    char *option;
    int i;
    bool do_stdin;
    bool do_stdout;
    int verbose=0;
    char algorithm_name[MAX_STRING];
    char date[MAX_STRING];
    char climatology_filename[FILENAME_MAX];
    char gpd_filename[FILENAME_MAX];
    char loci_filename[FILENAME_MAX];
    char landmask_filename[FILENAME_MAX];
    char weights_filename[FILENAME_MAX];
    char elevation_filename[FILENAME_MAX];
    char forest_cover_filename[FILENAME_MAX];
    char chang_surface_filename[FILENAME_MAX];
    char tbs_basename[FILENAME_MAX];
    char out_filename[FILENAME_MAX];
    char platform_name[MAX_STRING] = "NA";
    char hemisphere_name[MAX_STRING];
    bool use_climatology;
    bool use_loci;
    bool do_user_fill_value;
    bool do_spillover;
    bool fill_algo_pixels_only;
    bool scale_percent_seaice;
    bool do_algorithm_here, algo_likely_here;
    bool valid_tbs, algo_OK;
    byte1 **climatology;
    byte1 **loci;
    byte1 max_ice_value;
    byte1 high_elevation_model_id;
    float **out_data;
    float fill_value, algo_value;
    FILE *loci_file;
    grid_class *grid;
    pmtbs_class *pmtbs;
    land_spillover_class *land_spillover;
    int2 threshold_elevation_meters;
    byte2 loci_value[MAX_LOCI_VALUES];
    int num_loci_values=0;
    int row, col;
    int month;
    int spillover_open_water_pcent;
    int goddard_recoding_pcent;
    size_t status, bytes_per_row, nbytes;

    /*
     * Set defaults
     */
    do_stdin = FALSE;
    do_stdout = TRUE;
    loci_file = NULL;
    *date = '\0';
    *climatology_filename = '\0';
    *gpd_filename = '\0';
    *loci_filename = '\0';
    *landmask_filename = '\0';
    *weights_filename = '\0';
    *elevation_filename = '\0';
    *forest_cover_filename = '\0';
    *chang_surface_filename = '\0';
    *out_filename = '\0';
    *hemisphere_name = '\0';
    use_climatology = FALSE;
    use_loci = FALSE;
    do_user_fill_value = FALSE;
    fill_algo_pixels_only = FALSE;
    scale_percent_seaice = FALSE;
    climatology = NULL;
    high_elevation_model_id = PMALGO_NO_HIGH_ELEVATION_MODEL;
    threshold_elevation_meters = PMALGO_UNKNOWN_THRESHOLD_ELEVATION;
    loci = NULL;
    out_data = NULL;
    for (i = 0; i < MAX_LOCI_VALUES; i++) {
        loci_value[i] = BYTE2_MAX;
    }
    month = 0;
    do_spillover = FALSE;
    spillover_open_water_pcent = PMALGO_GTS_OPEN_WATER_PERCENT;
    goddard_recoding_pcent = PMALGO_GTS_MIN_ICE_PERCENT;

    /*
     *      get command line arguments
     */
    while (--argc > 0 && (*++argv)[0] == '-') {
        for (option = argv[0]+1; *option != '\0'; option++) {
            switch (*option) {
                case 'a':
                    ++argv; --argc;
                    if (1 != sscanf(*argv,"%d",&spillover_open_water_pcent)) usage_exit();
                    break;
                case 'b':
                    turn_on_new_extensions();
                    break;
                case 'c':
                    ++argv; --argc;
                    if (1 != sscanf(*argv,"%s",climatology_filename)) usage_exit();
                    use_climatology = TRUE;
                    break;
                case 'd':
                    ++argv; --argc;
                    if (1 != sscanf(*argv,"%s",date)) usage_exit();
                    break;
                case 'e':
                    ++argv; --argc;
                    if (1 != sscanf(*argv,"%s",elevation_filename)) usage_exit();
                    break;
                case 'f':
                    ++argv; --argc;
                    if (1 != sscanf(*argv,"%f",&fill_value)) usage_exit();
                    do_user_fill_value = TRUE;
                    break;
                case 'g':
                    ++argv; --argc;
                    if (1 != sscanf(*argv,"%s",chang_surface_filename)) usage_exit();
                    break;
                case 'h':
                    ++argv; --argc;
                    if (1 != sscanf(*argv,"%d",&month)) usage_exit();
                    break;
                case 'i':
                    do_stdin = TRUE;
                    break;
                case 'j':
                    do_spillover = TRUE;
                    break;
                case 'l':
                    ++argv; --argc;
                    if (1 != sscanf(*argv,"%s",loci_filename)) usage_exit();
                    loci_file = search_path_fopen(loci_filename, mapx_PATH, "r");
                    if (NULL == loci_file) {
                        fprintf(stderr,
                                "pmalgos: cannot find loci (land/ocean/coast/ice) "
                                "mask=%s in mapx_PATH\n",*argv);
                        usage_exit();
                    }
                    /*
                     * search_path_fopen always opens the file if it finds it, but
                     * I want initialize_matrix to read it later, so close it now and
                     * let initialize_matrix do its thing
                     */
                    fclose(loci_file);
                    use_loci = TRUE;
                    break;
                case 'm':
                    ++argv; --argc;
                    if (1 != sscanf(*argv,"%s",landmask_filename)) usage_exit();
                    break;
                case 'n':
                    ++argv; --argc;
                    if (1 != sscanf(*argv,"%u",&high_elevation_model_id)) usage_exit();
                    break;
                case 'o':
                    ++argv; --argc;
                    if (1 != sscanf(*argv,"%s",out_filename)) usage_exit();
                    do_stdout = FALSE;
                    break;
                case 'p':
                    ++argv; --argc;
                    if (1 != sscanf(*argv,"%s",platform_name)) usage_exit();
                    break;
                case 'q':
                    ++argv; --argc;
                    if (1 != sscanf(*argv,"%s",hemisphere_name)) usage_exit();
                    break;
                case 'r':
                    fill_algo_pixels_only = TRUE;
                    break;
                case 's':
                    ++argv; --argc;
                    if (1 != sscanf(*argv,"%d",&threshold_elevation_meters)) usage_exit();
                    break;
                case 't':
                    ++argv; --argc;
                    if (1 != sscanf(*argv,"%s",forest_cover_filename)) usage_exit();
                    break;
                case 'u':
                    ++argv; --argc;
                    if (num_loci_values >= MAX_LOCI_VALUES) {
                        fprintf (stderr,
                                 "pmalgos: too many input loci values, maximum is %d\n\n",
                                 MAX_LOCI_VALUES);
                        usage_exit();
                    }
                    if (1 != sscanf(*argv,"%hu",&loci_value[num_loci_values++])) {
                        usage_exit();
                    }
                    break;
                case 'v':
                    verbose++;
                    break;
                case 'w':
                    ++argv; --argc;
                    if (1 != sscanf(*argv,"%s",weights_filename)) usage_exit();
                    break;
                case 'x':
                    ++argv; --argc;
                    if (1 != sscanf(*argv,"%d",&goddard_recoding_pcent)) usage_exit();
                    break;
                case 'z':
                    scale_percent_seaice = TRUE;
                    break;
                default:
                    fprintf (stderr, "pmalgos: invalid option %c\n", *option);
                    usage_exit();
            } /* end switch */
        } /* end loop for each input command option */
    } /* end loop while still input arguments */

    if (2 > argc) {
        fprintf(stderr,"pmalgos: Insufficient number of calling arguments.\n\n");
        usage_exit();
    }

    if (1 != sscanf(*argv,"%s",gpd_filename)) usage_exit();
    ++argv; --argc;
    grid = init_grid(gpd_filename);
    if (NULL == grid) {
        usage_exit();
    }

    if (1 != sscanf(*argv,"%s",algorithm_name)) usage_exit();
    ++argv; --argc;

    /* Get the input basename from stdin or the argument list */
    if (do_stdin) {
        if (0 < argc) {
            fprintf(stderr,"pmalgos: WARNING, ignoring input basename "
                    "in favor of -i switch.\n");
        }

        if (NULL == fgets(tbs_basename, FILENAME_MAX, stdin)) {
            error_exit("pmalgos: error reading tbs_basename from stdin.\n");
        }
        /*
         * fgets returns a final newline, so strip it off
         * (this is different from gets which does this for you)
         */
        *(tbs_basename+strlen(tbs_basename)-1)='\0';

    } else {
        if (1 != argc) {
            fprintf(stderr,"pmalgos: Exactly one input basename is required,");
            fprintf(stderr," unless -i switch is used.\n");
            usage_exit();
        }

        if (FILENAME_MAX < strlen(*argv)) {
            fprintf(stderr,
                    "pmalgos: input tbs_basename exceeds allowable length of %d\n",
                    FILENAME_MAX);
            exit(EXIT_FAILURE);
        }
        strncpy (tbs_basename, *argv, FILENAME_MAX-1);
    }

    /* Check valid inputs */
    check_valid_scale_switch(scale_percent_seaice, algorithm_name, (bool)verbose);


    /* Load input TB data */
    pmtbs = init_pmtbs(grid, algorithm_name, platform_name, tbs_basename, (bool)verbose);
    if (NULL == pmtbs) exit(EXIT_FAILURE);

    /* If using a climatology, allocate memory and read it */
    if (use_climatology) {
        climatology = (byte1 **)initialize_matrix(grid,
                                                  sizeof(**climatology),
                                                  climatology_filename,
                                                  "climatology",
                                                  (bool)verbose);
        if (NULL == climatology) exit(EXIT_FAILURE);
    }

    /* If using an input loci allocate memory and read it */
    if (use_loci) {
        loci = (byte1 **)initialize_matrix(grid,
                                           sizeof(**loci),
                                           loci_filename,
                                           "loci (land/ocean/coast/ice) mask",
                                           (bool)verbose);
        if (NULL == loci) exit(EXIT_FAILURE);
    }

    /* Check for required inputs when spillover correction is wanted */

    if ( do_spillover ) {

        if ( '\0' == *landmask_filename ) {
            fprintf(stderr,"pmalgos: algorithm %s requires -m landmask input\n",
                    pmalgo_id_name[pmtbs->algorithm->id]);
            usage_exit();
        }

        if ( '\0' == *weights_filename ) {
            fprintf(stderr,"pmalgos: algorithm %s requires -w weightmask input\n",
                    pmalgo_id_name[pmtbs->algorithm->id]);
            usage_exit();
        }

        if (is_bootstrap_algorithm(pmtbs->algorithm->id)) {

            land_spillover = init_land_spillover(spillover_open_water_pcent,
                                                 PMALGO_BOOTSTRAP_OPEN_WATER_PERCENT,
                                                 landmask_filename,
                                                 weights_filename,
                                                 grid,
                                                 verbose);
        } else if (is_goddard_algorithm(pmtbs->algorithm->id)) {

            land_spillover = init_land_spillover(spillover_open_water_pcent,
                                                 PMALGO_GTS_OPEN_WATER_PERCENT,
                                                 landmask_filename,
                                                 weights_filename,
                                                 grid,
                                                 verbose);
        }
        if (land_spillover == NULL) {
            exit(EXIT_FAILURE);
        } else if (0 < verbose) {
            fprintf(stderr,"> pmalgos: initialized land spillover parameters\n");
        }

    } else {
        land_spillover = NULL;
    }

    /*
     * Some algorithms require certain inputs that are
     * optional or not required at all for the others.
     *
     */
    if (is_bootstrap_algorithm(pmtbs->algorithm->id)) {

        if ('\0' == *date) {
            fprintf(stderr,"pmalgos: algorithm %s requires -d date input\n",
                    pmalgo_id_name[pmtbs->algorithm->id]);
            usage_exit();
        }
        if (!use_loci || 0 == num_loci_values) {
            fprintf(stderr,"pmalgos: algorithm %s requires -l loci_mask and\n",
                    pmalgo_id_name[pmtbs->algorithm->id]);
            fprintf(stderr,"         -u loci value input\n");
            usage_exit();
        }

        /* Checks for valid BootV2 inputs */
        if (is_bootstrap_v02_algorithm(pmtbs->algorithm->id)) {
            
            if  (! valid_v2_environment() ){
                fprintf(stderr,"pmalgos: algorithm %s requires the BOOTSTRAP_V2_ANCILLARY\n",
                        pmalgo_id_name[pmtbs->algorithm->id]);
                fprintf(stderr,"         environmental variable to point to the directory containing\n");
                fprintf(stderr,"         the required monthly sea surface temperature hdf files.\n");
                usage_exit();
            }
            if (! grid_is_polarstero(grid)) {
                fprintf(stderr,"pmalgos: algorithm %s requires a polar-stereo grid (N3B or S3B)\n",
                        pmalgo_id_name[pmtbs->algorithm->id]);
                usage_exit();
            }
        }
        
                    

        if (!init_bootstrap_pmalgo(pmtbs->algorithm,date,
                                   goddard_recoding_pcent,
                                   scale_percent_seaice,
                                   land_spillover,
                                   grid, verbose)) {
            exit(EXIT_FAILURE);
        } else if (0 < verbose) {
            fprintf(stderr,"> pmalgos: initialized bootstrap parameters.\n");
        }

        /*
         * This is for backwards compatibility with old ps processing: If any
         * input loci values are between PMALGO_BOOTSTRAP_MIN_ICE_PERCENT and
         * PMALGO_BOOTSTRAP_MAX_ICE_PERCENT, reset these to the output
         * PMALGO_BOOTSTRAP_LAND value before continuing.  Check that the user
         * has not asked that the algorithm be run on loci values in this range,
         * since it is the range of legitimate sea ice % and would create
         * ambiguous results.
         */
        if (PMALGO_BOOTN_SEAICE == pmtbs->algorithm->id ||
            PMALGO_BOOTS_SEAICE == pmtbs->algorithm->id)
        {

            /* Set loci coast and land to algo_specific land value */
            /* fill loci correct values this algorithm */
            protect_invalid_loci_values( loci, grid,
                                         PMALGO_BOOTSTRAP_MIN_ICE_PERCENT,
                                         PMALGO_BOOTSTRAP_MAX_ICE_PERCENT,
                                         PMALGO_BOOTSTRAP_V1_LAND,
                                         num_loci_values, loci_value,
                                         "for bootstrap processing. \n");

        }
        else if (is_bootstrap_v02_algorithm(pmtbs->algorithm->id))
        {
            
            protect_invalid_loci_values( loci, grid,
                                         PMALGO_BOOTSTRAP_MIN_ICE_PERCENT,
                                         PMALGO_BOOTSTRAP_MAX_ICE_PERCENT,
                                         PMALGO_BOOTSTRAP_LAND,
                                         num_loci_values,loci_value,
                                         "for bootstrap V2 processing. \n");
            
        }
        
        /*
         * Set the loci mask pixels where the algorithm should be run
         */
        for (i = 0; i < num_loci_values; i++) {
            mask_bootstrap_pmalgo(pmtbs->algorithm,loci,(byte1)loci_value[i]);
        }


    } else if (PMALGO_NTN_SEAICE == pmtbs->algorithm->id ||
               PMALGO_NTN_MYICE == pmtbs->algorithm->id ||
               PMALGO_NTS_SEAICE == pmtbs->algorithm->id) {

        if (use_loci) {

            /*
             * When using a loci mask with this algorithm, make the output
             * backwards compatible with old ps processing: If any
             * input loci values are between PMALGO_NASATEAM_MIN_ICE_PERCENT and
             * PMALGO_NASATEAM_MAX_ICE_PERCENT, reset these to the output
             * PMALGO_NASATEAM_LAND value before continuing.  Check that the user
             * has not asked that the algorithm be run on loci values in this range,
             * since it is the range of legitimate sea ice % and would create
             * ambiguous results.
             */
            protect_invalid_loci_values( loci, grid,
                                         PMALGO_NASATEAM_MIN_ICE_PERCENT,
                                         PMALGO_NASATEAM_MAX_ICE_PERCENT, 
                                         PMALGO_NASATEAM_LAND,
                                         num_loci_values, loci_value,
                                         "for Nasateam processing. \n");

         } 

    } else if (is_goddard_algorithm(pmtbs->algorithm->id)) {

        if (use_loci) {

            max_ice_value = PMALGO_GTS_MAX_ICE;
            if (scale_percent_seaice) {
                max_ice_value = PMALGO_GTS_MAX_ICE_PERCENT;
            }

            protect_invalid_loci_values( loci, grid,
                                         PMALGO_GTS_MIN_ICE, max_ice_value,
                                         PMALGO_GTS_LAND,
                                         num_loci_values,loci_value,
                                         "for Goddard Time Series processing. \n");
        }

        if (!init_goddard_pmalgo(pmtbs->algorithm,
                                 goddard_recoding_pcent,
                                 hemisphere_name,
                                 scale_percent_seaice,
                                 land_spillover,
                                 pmtbs->grid,
                                 verbose)) {
            exit(EXIT_FAILURE);
        } else if (0 < verbose) {
            fprintf(stderr,"> pmalgos: initialized goddard parameters.\n");
        }

    } else if (PMALGO_CHANG_SURFACE == pmtbs->algorithm->id ||
               PMALGO_NSIDC_SURFACE == pmtbs->algorithm->id) {

        if (!init_sfc_pmalgo(pmtbs->algorithm,
                             0,NULL,
                             forest_cover_filename,
                             landmask_filename,
                             NULL,
                             grid,
                             high_elevation_model_id,
                             threshold_elevation_meters,
                             verbose)) {
            exit(EXIT_FAILURE);
        } else if (0 < verbose) {
            fprintf(stderr,"> pmalgos: initialized surface algorithm ancillary data.\n");
        }

    } else if (PMALGO_CHANG_DEPTH == pmtbs->algorithm->id ||
               PMALGO_NSIDC_S3_SWE == pmtbs->algorithm->id) {

        if (!init_sfc_pmalgo(pmtbs->algorithm,
                             month,
                             chang_surface_filename,
                             forest_cover_filename,
                             landmask_filename,
                             NULL,
                             grid,
                             high_elevation_model_id,
                             threshold_elevation_meters,
                             verbose)) {
            exit(EXIT_FAILURE);
        } else if (0 < verbose) {
            fprintf(stderr,"> pmalgos: initialized chang_depth/nsidc_s3_swe ");
            fprintf(stderr,"ancillary data.\n");
        }

    } else if (PMALGO_NSIDC_A4_SWE == pmtbs->algorithm->id ||
               PMALGO_NSIDC_M4_SWE == pmtbs->algorithm->id ||
               PMALGO_NSIDC_S4_SWE == pmtbs->algorithm->id ||
               PMALGO_NSIDC_S6_SWE == pmtbs->algorithm->id ||
               PMALGO_NSIDC_A10_SWE == pmtbs->algorithm->id ||
               PMALGO_NSIDC_S10_SWE == pmtbs->algorithm->id)  {

        if (!init_sfc_pmalgo(pmtbs->algorithm,
                             0,NULL,
                             forest_cover_filename,
                             NULL,
                             NULL,
                             grid,
                             high_elevation_model_id,
                             threshold_elevation_meters,
                             verbose)) {
            exit(EXIT_FAILURE);
        } else if (0 < verbose) {
            fprintf(stderr,"> pmalgos: initialized nsidc_s6, a/s10 or a/m/s4_swe ancillary data.\n");
        }

    } else if (PMALGO_NSIDC_S11_SWE == pmtbs->algorithm->id) {

        if (!init_sfc_pmalgo(pmtbs->algorithm,
                             0,NULL,
                             forest_cover_filename,
                             NULL,
                             elevation_filename,
                             grid,
                             high_elevation_model_id,
                             threshold_elevation_meters,
                             verbose)) {
            exit(EXIT_FAILURE);
        } else if (0 < verbose) {
            fprintf(stderr,"> pmalgos: initialized nsidc_s11 ancillary data.\n");
        }

    } else if (PMALGO_NSIDC_S12_SWE == pmtbs->algorithm->id) {

        if (!init_sfc_pmalgo(pmtbs->algorithm,
                             0,NULL,
                             NULL,
                             NULL,
                             elevation_filename,
                             grid,
                             high_elevation_model_id,
                             threshold_elevation_meters,
                             verbose)) {
            exit(EXIT_FAILURE);
        } else if (0 < verbose) {
            fprintf(stderr,"> pmalgos: initialized nsidc_s12 ancillary data.\n");
        }

    } else if (PMALGO_NSIDC_S1_SWE == pmtbs->algorithm->id ||
               PMALGO_NSIDC_S5_SWE == pmtbs->algorithm->id) {

        if (!init_sfc_pmalgo(pmtbs->algorithm,
                             0,NULL,
                             NULL,
                             NULL,
                             NULL,
                             grid,
                             high_elevation_model_id,
                             threshold_elevation_meters,
                             verbose)) {
            exit(EXIT_FAILURE);
        } else if (0 < verbose) {
            fprintf(stderr,"> pmalgos: initialized nsidc_s1/s5 ancillary data.\n");
        }

    }

    /* Allocate memory for output array */
    out_data = (float **)matrix(grid->rows, grid->cols, sizeof(**out_data),
                                TRUE);
    if (NULL == out_data) {
        exit(EXIT_FAILURE);
    } else if (0 < verbose) {
        fprintf(stderr,"> pmalgos: allocated output matrix.\n");
    }



    /*
     * If using a loci mask, initialize the output array to it,
     * otherwise output array default value will be zero
     */
    if (use_loci) {
        prefill_output(out_data, loci, pmtbs, scale_percent_seaice);
    }


    /***** verbose output ********/
    if (0 < verbose) {
        fprintf(stderr,"> pmalgos: Beginning processing for:\n");
        fprintf(stderr,">  tb_grids_basename   =%s\n",tbs_basename);
        fprintf(stderr,">  gpd                 =%s\n",gpd_filename);
        fprintf(stderr,">  algorithm           =%s\n",algorithm_name);
        fprintf(stderr,">  use_climatology     =%d\n",use_climatology);
        if (use_climatology) {
            fprintf(stderr,">  -c climatology file =%s\n",climatology_filename);
        } else {
            fprintf(stderr,">  -c climatology file =n/a\n");
        }
        fprintf(stderr,">  date                =%s\n",date);
        if (do_user_fill_value) {
            fprintf(stderr,">  -f fill_value       =%f\n",fill_value);
        } else {
            fprintf(stderr,">  Using default fill_value 0.\n");
        }
        if (use_loci) {
            fprintf(stderr,">  -l loci file        =%s\n",loci_filename);
            fprintf(stderr,">  -u operating on loci mask values of\n");
            for (i = 0; i < num_loci_values; i++) {
                fprintf(stderr,">  %d\n",loci_value[i]);
            }
        } else {
            fprintf(stderr,">  No loci file.  Operating on every pixel.\n");
        }
        if (PMALGO_CHANG_SURFACE == pmtbs->algorithm->id ||
            PMALGO_NSIDC_SURFACE == pmtbs->algorithm->id) {
            fprintf(stderr,">  -m landmask file =%s\n",landmask_filename);
            fprintf(stderr,">  -t forest_cover file =%s\n",forest_cover_filename);
        }
        if (PMALGO_CHANG_DEPTH == pmtbs->algorithm->id ||
            PMALGO_NSIDC_S3_SWE == pmtbs->algorithm->id) {
            fprintf(stderr,">  -h month         =%d\n",month);
            fprintf(stderr,">  -g surface file  =%s\n",chang_surface_filename);
            fprintf(stderr,">  -m landmask file =%s\n",landmask_filename);
            fprintf(stderr,">  -p platform                     =%s\n",platform_name);
            fprintf(stderr,">  -t forest_cover file =%s\n",forest_cover_filename);
        }
        if (PMALGO_NSIDC_S4_SWE == pmtbs->algorithm->id ||
            PMALGO_NSIDC_S6_SWE == pmtbs->algorithm->id ||
            PMALGO_NSIDC_S10_SWE == pmtbs->algorithm->id) {
            fprintf(stderr,">  -p platform                     =%s\n",platform_name);
            fprintf(stderr,">  -t forest_cover file =%s\n",forest_cover_filename);
        }
        if (PMALGO_NSIDC_A4_SWE == pmtbs->algorithm->id ||
            PMALGO_NSIDC_A10_SWE == pmtbs->algorithm->id) {
            fprintf(stderr,">  -t forest_cover file =%s\n",forest_cover_filename);
        }
        if (PMALGO_NSIDC_S1_SWE == pmtbs->algorithm->id ||
            PMALGO_NSIDC_S5_SWE == pmtbs->algorithm->id) {
            fprintf(stderr,">  -p platform                     =%s\n",platform_name);
        }
        if (PMALGO_NSIDC_S11_SWE == pmtbs->algorithm->id) {
            fprintf(stderr,">  -e elevation file =%s\n",elevation_filename);
            fprintf(stderr,">  -n high_elevation_model_number =%u\n",high_elevation_model_id);
            fprintf(stderr,">  -p platform                     =%s\n",platform_name);
            fprintf(stderr,">  -s threshold_elevation (meters) =%d\n",threshold_elevation_meters);
            fprintf(stderr,">  -t forest_cover file =%s\n",forest_cover_filename);
        }
        if (PMALGO_NSIDC_S12_SWE == pmtbs->algorithm->id) {
            fprintf(stderr,">  -e elevation file =%s\n",elevation_filename);
            fprintf(stderr,">  -n high_elevation_model_number =%u\n",high_elevation_model_id);
            fprintf(stderr,">  -p platform                     =%s\n",platform_name);
            fprintf(stderr,">  -s threshold_elevation (meters) =%d\n",threshold_elevation_meters);
        }
        if (is_goddard_algorithm(pmtbs->algorithm->id)) {
            fprintf(stderr,">  -a spillover open water cutoff  =%d%%\n",spillover_open_water_pcent);
            fprintf(stderr,">  -x goddard recoding cutoff      =%d%%\n",goddard_recoding_pcent);
            fprintf(stderr,">  -p platform                     =%s\n",platform_name);
            fprintf(stderr,">  -q hemisphere                   =%s\n",hemisphere_name);
            fprintf(stderr,">  -z (do scale output to 0-100%%)  =%d\n",scale_percent_seaice);
            fprintf(stderr,">  -j (do spillover)               =%d\n",do_spillover);
            fprintf(stderr,">  -m spillover landmask file      =%s\n",landmask_filename);
            fprintf(stderr,">  -w spillover weights file       =%s\n",weights_filename);
        }
        if (do_stdout) {
            fprintf(stderr,"> Output will be written to stdout\n");
        } else {
            fprintf(stderr,"> Output will be written to %s\n",out_filename);
        }
    } /***** END verbose output ********/



    /*
     * For each pixel at [row][col]:
     *
     *   Run the algorithm on this pixel location
     *   If the run function returns TRUE (so the input TBs were non-zero):
     *     Decide whether to use algorithm output for this pixel:
     *       If not using a loci mask, every pixel is OK to use algorithm output
     *       If using a loci mask, only pixels with loci values chosen by
     *         the user are OK to use algorithm output
     *     If it's OK to use algorithm output for this pixel, and
     *     the user didn't enter a climatology or entered a climatology
     *       and it's ON here:
     *       set the output value to the algorithm output
     *   Else the run function returns FALSE (so input TBs were zeroes):
     *     If the user entered a fill_value, set the output to the fill_value
     */
    for (row = 0; row < grid->rows; row++) {
        for (col = 0; col < grid->cols; col++) {


            /*
             * Decide whether to use algorithm output for this surface type (loci)
             */
            if (!use_loci) {
                do_algorithm_here = TRUE;
            } else {
                do_algorithm_here = FALSE;
                for (i = 0; i < num_loci_values; i++) {
                    if (loci_value[i] == loci[row][col]) {
                        do_algorithm_here = TRUE;
                        break;
                    }
                }
            }

            algo_OK = pmtbs->algorithm->run(pmtbs->algorithm->private,
                                            pmtbs->data,col,row,
                                            &algo_value,&valid_tbs);


            if (valid_tbs && algo_OK) {

                algo_likely_here = TRUE;
                if (use_climatology && CLIMATOLOGY_OFF == climatology[row][col]) {
                    algo_likely_here = FALSE;
                }

                /*
                 * If the algorithm output can be used for this surface
                 * type (loci value) and it's climatologically likely, use
                 * the output value
                 */
                if (do_algorithm_here && algo_likely_here) {
                    out_data[row][col] = algo_value;
                }

                if (is_goddard_algorithm(pmtbs->algorithm->id)) {

                    if ( do_algorithm_here && do_spillover ) {
                        /* overrides the case where climatology is off but you
                         * want to take the spillover value anyway.  We're
                         * guessing this was done to match goddard
                         * regression. 2011-03-09 MHS&MJB */
                        out_data[row][col] = algo_value;
                    }
                }

            } else if (!valid_tbs) {

                /*
                 * The input TBs were zeroes.
                 * If using a fill_value, set output to it.
                 */
                if (do_user_fill_value) {
                    if (!fill_algo_pixels_only || do_algorithm_here) {
                        out_data[row][col] = fill_value;
                    }
                }
            } /* end else run function returned FALSE, so input TBs were zeroes */
        } /* end loop for each column */
    } /* end loop for each row */

    /*
     * Do postprocessing filtering.
     * This is any processing that must be performed after the entire
     * grid of values has been calculated.  The main processing loop
     * operates on one pixel at a time, independent of the rest.
     * This processing will do any additional filtering needed, where output for
     * a given pixel depends on the values of the surrounding pixels.
     */
    if ( !postprocessing_filter_pmalgo(
             pmtbs->algorithm, pmtbs->grid, out_data,
             use_climatology, climatology,
             use_loci, loci,
             fill_algo_pixels_only,
             do_user_fill_value, fill_value,
             verbose ) ) {
        exit( EXIT_FAILURE );
    }

    if (do_stdout) {
        bytes_per_row = grid->cols * sizeof(**out_data);
        for (row = 0; row < grid->rows; row++) {
            nbytes = fwrite(out_data[row], 1, bytes_per_row, stdout);
            if (nbytes != bytes_per_row) {
                perror("stdout");
                exit(EXIT_FAILURE);
            }
        }
        fclose(stdout);
        if (0 < verbose)
            fprintf(stderr,"> pmalgos: wrote output data to stdout.\n");
    } else {
        status = write_matrix (out_filename, (void **)out_data, grid->rows,
                               grid->cols, sizeof(**out_data));
        if (0 == status) {
            fprintf(stderr,"pmalgos: write_matrix failed for %s\n",out_filename);
            exit(EXIT_FAILURE);
        } else if (0 < verbose) {
            fprintf(stderr,"> pmalgos: wrote output data to %s\n",
                    out_filename);
        }
    } /* end else write data to an external file */

    if (0 < verbose) fprintf(stderr,"> pmalgos: exiting successfully\n");

    if (NULL != out_data) free(out_data);
    if (NULL != loci) free(loci);
    if (NULL != climatology) free(climatology);
    close_pmtbs(pmtbs);

    exit(EXIT_SUCCESS);

}

#endif /* end of main for pmalgos */
