/*======================================================================
 * pmtbcomb - program to combine multiple days of gridded passive 
 *            microwave (PM) brightness temperatures (TBs).
 *
 * 08/02/1998 brodzik brodzik@zamboni.colorado.edu 303-492-8263
 * National Snow & Ice Data Center, University of Colorado, Boulder
 *
 * $Revision: 1.4 $
 *
 *======================================================================*/
#include "define.h"
#include "matrix.h"
#include "matrix_io.h"
#include "maps.h"
#include "grids.h"
#include "pmtbs.h"

static const char pmtbcomb_c_rcsid[] = "$Header: /tmp_mnt/FILES/pmalgos/pmtbcomb.c,v 1.4 2002-08-14 20:23:54 brodzik Exp $";

typedef enum {
  PMTC_NO_COMBINATION = -1, 
  PMTC_CLEAREST_SKY, PMTC_MEAN, PMTC_MAX, PMTC_MIN,
  PMTC_NUM_COMBINATIONS
} pmtc_combination_id;
static const char *pmtc_combination_id_name[] = {"CLEAREST_SKY",
						 "MEAN",
						 "MAX",
						 "MIN"};

/* combo algorithm list */
#define PMTC_COMBINATION_DESCRIPTIONS \
"           CLEAREST_SKY - choose TBs for the day with clearest\n"\
"                  atmosphere, maximum of (37V - 85V) > 0\n"\
"           MEAN - choose the (non-missing) mean in each channel \n"\
"                  for all input days\n"\
"           MAX - choose the (non-missing) max in each channel \n"\
"                  for all input days\n"\
"           MIN - choose the (non-missing) min in each channel \n"\
"                  for all input days\n"

#define CLEAR_SKY_INDEX_NOT_SET -1
#define CLEAR_SKY_DAY_NOT_SET -1

#define usage \
"pmtbcomb $Revision: 1.4 $ \n"\
"usage: pmtbcomb [-v] gpd_file geo_algorithm comb_algorithm \n"\
"       outfile_basename <tb_grids_basename...>\n"\
"\n"\
" input : gpd_file  - grid parameters definition file\n"\
"         geo_algorithm - geophysical algorithm (determines which channels\n"\
"           to combine and save, one of:\n"\
PMALGO_ALGORITHM_DESCRIPTIONS \
"         comb_algorithm - TB combination algorithm, possible values:\n"\
PMTC_COMBINATION_DESCRIPTIONS \
"         outfile_basename - basename of output TB files to create\n"\
"         tb_grids_basename - list of basenames of TB files to use\n"\
"           (assumes channel extensions, e.g. tb_grids_basename.19H, .19V etc)\n"\
"\n"\
" output : one set of combined passive microwave tbs from all input files\n"\
"\n"\
" option : v - verbose mode \n"\
"\n"

static bool run_clearest_sky_tb(pmtbs_class *out_pmtbs, int argc, char *argv[],
			       int verbose);
static bool run_mean_tb(pmtbs_class *out_pmtbs, int argc, char *argv[], 
			int verbose);
static bool run_max_tb(pmtbs_class *out_pmtbs, int argc, char *argv[], 
		       int verbose);
static bool run_min_tb(pmtbs_class *out_pmtbs, int argc, char *argv[],
		       int verbose);

main (int argc, char *argv[]) {
  char *option;
  int verbose=0;
  char geo_algorithm_name[MAX_STRING];
  char comb_algorithm_name[MAX_STRING];
  char gpd_filename[FILENAME_MAX];
  char out_filename[FILENAME_MAX];
  char *platform_name;
  grid_class *grid;
  pmtbs_class *out_pmtbs;
  pmtc_combination_id id, comb_id;
  bool status;
  
  /*
   * Set defaults
   */
  comb_id = PMTC_NO_COMBINATION;
  platform_name = strdup( pmalgo_ssmi_platform_name[PMALGO_NO_PLATFORM] );

  /* 
   *      get command line arguments
   */
  while (--argc > 0 && (*++argv)[0] == '-') {
    for (option = argv[0]+1; *option != '\0'; option++) {
      switch (*option) {
      case 'v':
	verbose++;
	break;
      default:
	fprintf(stderr, "pmtbcomb: invalid option %c\n", *option);
	fprintf(stderr,usage);
	exit(EXIT_FAILURE);
      } /* end switch */
    } /* end loop for each input command option */
  } /* end loop while still input arguments */

  if (4 > argc) {
    fprintf(stderr,"pmtbcomb: Insufficient number of calling arguments.\n\n");
    fprintf(stderr,usage);
    exit(EXIT_FAILURE);
  }

  if (1 != sscanf(*argv,"%s",gpd_filename)) error_exit(usage);
  ++argv; --argc;
  grid = init_grid(gpd_filename);
  if (NULL == grid) {
    error_exit(usage);
  }

  if (1 != sscanf(*argv,"%s",geo_algorithm_name)) error_exit(usage);
  ++argv; --argc;
  if (verbose >= 2) 
    fprintf(stderr,">> Initializing arrays for output TBs...\n");
  out_pmtbs = init_pmtbs(grid, geo_algorithm_name, platform_name,
                         NULL, (bool)verbose);
  if (NULL == out_pmtbs) exit(EXIT_FAILURE);

  if (1 != sscanf(*argv,"%s",comb_algorithm_name)) error_exit(usage);
  ++argv; --argc;
  for (id = (pmtc_combination_id)0; id < PMTC_NUM_COMBINATIONS; id++) {
    if (0 == strcmp(pmtc_combination_id_name[id],comb_algorithm_name)) {
      comb_id = id;
      break;
    }
  }
  if (PMTC_NO_COMBINATION == comb_id) {
    fprintf(stderr,"pmtbcomb: invalid comb_algorithm_name = %s\n",
	    comb_algorithm_name);
    exit(EXIT_FAILURE);
  }

  if (1 != sscanf(*argv,"%s",out_filename)) error_exit(usage);
  ++argv; --argc;

  if (verbose >= 1) {
    fprintf(stderr,"> pmtbcomb: Beginning processing for:\n");
    fprintf(stderr,">  gpd                 =%s\n",gpd_filename);
    fprintf(stderr,">  geo_algorithm       =%s\n",geo_algorithm_name);
    fprintf(stderr,">  comb_algorithm      =%s\n",
	    pmtc_combination_id_name[comb_id]);
    fprintf(stderr,">  out_filename        =%s\n",out_filename);
    fprintf(stderr,">  %s\n",pmtbcomb_c_rcsid);
  }

  switch (comb_id) {
  case PMTC_CLEAREST_SKY:
    status = run_clearest_sky_tb(out_pmtbs, argc, argv, verbose);
    if (!status) exit(EXIT_FAILURE);
    break;
  case PMTC_MEAN:
    status = run_mean_tb(out_pmtbs, argc, argv, verbose);
    if (!status) exit(EXIT_FAILURE);
    break;
  case PMTC_MAX:
    status = run_max_tb(out_pmtbs, argc, argv, verbose);
    if (!status) exit(EXIT_FAILURE);
    break;
  case PMTC_MIN:
    status = run_min_tb(out_pmtbs, argc, argv, verbose);
    if (!status) exit(EXIT_FAILURE);
    break;
  default:
    assert(NEVER);
  }

  /* Save the output TBs */
  status = write_pmtbs(out_pmtbs,out_filename,(bool)verbose);
  if (!status) exit(EXIT_FAILURE);
  if (verbose >= 2)
    fprintf(stderr,">> pmtbcomb: wrote output data to %s\n",
	    out_filename);

  if (verbose >= 2) fprintf(stderr,">> pmtbcomb: exiting successfully\n");

  close_pmtbs(out_pmtbs);
  exit(EXIT_SUCCESS);

} 



/*----------------------------------------------------------------------
 *  run_clearest_sky_tb - finds pixels with the clearest atmosphere, 
 *                        according to Nagler/Rott criteria
 * 
 *     input : out_pmtbs - pmtbs object, determines grid and geo_algorithm
 *             argc - current argument count
 *             *argv[] - current argument list
 *             verbose - verbosity flag
 * 
 *    output: out_pmtbs - pmtbs object for output TBs
 *    result: TRUE if successful, or FALSE otherwise
 *
 *----------------------------------------------------------------------*/
static bool run_clearest_sky_tb (pmtbs_class *out_pmtbs,	int argc, 
                                 char *argv[], int verbose) {
  int row, col, chan, day;
  pmtbs_class *in_pmtbs;
  char *algorithm_name;
  char *platform_name;
  char tbs_basename[FILENAME_MAX];
  float **clear_sky_index;
  int **clear_sky_day;
  float algo_value;
  int start_argc;
  char **start_argv;
  bool valid_tbs, algo_OK;
  size_t status;

  algorithm_name = strdup( pmalgo_id_name[PMALGO_NAGLER_CSI] );
  platform_name = strdup( pmalgo_ssmi_platform_name[PMALGO_NO_PLATFORM] );

  /*
   * To determine TBs with clearest atmosphere:
   *   Initialize a temporary array for saving the maximum clear skies index
   *   Initialize a temporary array for saving the day with clearest skies
   *   Loop for each input basename
   *     Initialize a pmtbs object for calculating the Nagler-Rott clear
   *       sky index (CSI)
   *     Loop for each col/row:
   *       Test for and retain new maximum CSI, and the day index when
   *         it occurs
   *     Close the input pmtbs TB object
   *   Loop for each input basename
   *     Initialize a pmtbs object for the channels needed by out_pmtbs
   *     Loop for each col/row:
   *       If today was the day with clearest skies in this pixel,
   *         save today's TBs for the required channels
   *     Close the input pmtbs TB object
   */     
  clear_sky_index = 
    (float **)initialize_matrix(out_pmtbs->grid, sizeof(**clear_sky_index), 
				NULL, "maximum clear sky index", 
				(bool)verbose);
  if (NULL == clear_sky_index) return FALSE;

  clear_sky_day = 
    (int **)initialize_matrix(out_pmtbs->grid, sizeof(**clear_sky_day), NULL,
			      "day of clearest sky", (bool)verbose);
  if (NULL == clear_sky_day) return FALSE;

  for (row = 0; row < out_pmtbs->grid->rows; row++) {
    for (col = 0; col < out_pmtbs->grid->cols; col++) {
      clear_sky_index[row][col] = CLEAR_SKY_INDEX_NOT_SET;
      clear_sky_day[row][col] = CLEAR_SKY_DAY_NOT_SET;
    }
  }
  
  /* 
   * Loop for each "day" (tbs_basename, which usually, but not necessarily
   * represents TBs for one day)
   */
  day = 0;
  start_argc = argc;
  start_argv = argv;
  if (verbose >= 2) fprintf(stderr,">> Looking for maximum CSI...\n");
  while (0 < argc) {

    if (1 != sscanf(*argv,"%s",tbs_basename)) {
      fprintf(stderr,"pmtbcomb: error reading input tbs_basename\n");
      fprintf(stderr,usage);
      return FALSE;
    }
    ++argv; --argc;

    /* Load input TB data to run clear skies index for this day */
    if (verbose >= 2) 
      fprintf(stderr, 
	      ">> Initializing next day's TB arrays for CSI...\n");
    in_pmtbs = init_pmtbs(out_pmtbs->grid, algorithm_name, platform_name,
                          tbs_basename, (bool)verbose);
    if (NULL == in_pmtbs) return FALSE;

    for (row = 0; row < out_pmtbs->grid->rows; row++) {
      for (col = 0; col < out_pmtbs->grid->cols; col++) {
	algo_OK = in_pmtbs->algorithm->run(in_pmtbs->algorithm->private,
					   in_pmtbs->data, col, row,
					   &algo_value, &valid_tbs);
	if (algo_OK && algo_value > clear_sky_index[row][col]) {
	  clear_sky_index[row][col] = algo_value;
	  clear_sky_day[row][col] = day;
	} 
      } /* end loop for each column */
    } /* end loop for each row */

    /* Close input TBs for this day */
    close_pmtbs(in_pmtbs);

    day++;

  } /* end loop for each "day" (tbs_basename) */

  free(algorithm_name);
  algorithm_name = strdup(pmalgo_id_name[out_pmtbs->algorithm->id]);
  day = 0;
  argc = start_argc;
  argv = start_argv;
  if (verbose >= 2) 
    fprintf(stderr,">> Saving TBs for days with clearest atmosphere...\n");

  /* 
   * Loop again for each "day" (see note in first loop, above)
   * this time, read the TB channels that will be used for output
   */
  while (0 < argc) {

    if (1 != sscanf(*argv,"%s",tbs_basename)) { 
      fprintf(stderr,"pmtbcomb: error reading input tbs_basename\n"); 
      fprintf(stderr,usage); 
      return FALSE; 
    } 
    ++argv; --argc; 

    /* Load input TB data */
    if (verbose >= 2) 
      fprintf(stderr, ">> Initializing next day's TB arrays...\n");
    in_pmtbs = init_pmtbs(out_pmtbs->grid, algorithm_name, platform_name,
                          tbs_basename, (bool)verbose);
    if (NULL == in_pmtbs) return FALSE;

    /* Copy TBs for clear sky days to output */
    for (row = 0; row < out_pmtbs->grid->rows; row++) { 
      for (col = 0; col < out_pmtbs->grid->cols; col++) { 
 	for (chan = 0; chan < in_pmtbs->algorithm->num_channels; chan++) { 
 	  if (day == clear_sky_day[row][col]) { 
              out_pmtbs->data[in_pmtbs->algorithm->channel[chan]][row][col] = 
                in_pmtbs->data[in_pmtbs->algorithm->channel[chan]][row][col]; 
 	  } /* end if this is the day with the clearest skies */
 	} /* end loop for each channel */
      } /* end loop for each column */ 
    } /* end loop for each row */ 

    /* Close input TBs for this day */
    close_pmtbs(in_pmtbs);

    day++;

  } /* end loop for each "day" (tbs_basename) */

#ifdef DEBUG_PICK

  /* Save the clear_sky_day array */
  status = write_matrix ("pmtbcomb.pick", (void **)clear_sky_day,
			 out_pmtbs->grid->rows, out_pmtbs->grid->cols,
			 sizeof(**clear_sky_day));
  if (0 == status) {
    fprintf(stderr,"pmtbcomb:  DEBUG error writing pmtbcomb.pick\n");
  } else {
    fprintf(stderr,"pmtbcomb:  DEBUG wrote day index to pmtbcomb.pick\n");
  }

#endif

  if (NULL != clear_sky_index) free(clear_sky_index);
  if (NULL != clear_sky_day) free(clear_sky_day);
  return TRUE;

}

/*----------------------------------------------------------------------
 *  run_mean_tb - calculates mean of non-missing TBs
 * 
 *     input : out_pmtbs - pmtbs object, determines grid and geo_algorithm
 *             argc - current argument count
 *             *argv[] - current argument list
 *             verbose - verbosity flag
 * 
 *    output: out_pmtbs - pmtbs object for output TBs
 *    result: TRUE if successful, or FALSE otherwise
 *
 *----------------------------------------------------------------------*/
static bool run_mean_tb (pmtbs_class *out_pmtbs, int argc, char *argv[], 
			 int verbose) {

  int row, col, chan;
  pmtbs_class *count_pmtbs;
  pmtbs_class *in_pmtbs;
  char tbs_basename[FILENAME_MAX];
  char buffer[MAX_STRING];
  char *algorithm_name;
  char *platform_name;
  float **sum_tbs[PMALGO_NUM_CHANNELS];
  byte2 count;
  float sum;

  algorithm_name = strdup(pmalgo_id_name[out_pmtbs->algorithm->id]);
  platform_name = strdup( pmalgo_ssmi_platform_name[PMALGO_NO_PLATFORM] );
  /*
   * To calculate MEAN:
   *   Initialize a pmtbs object with arrays for counting the number of
   *     legitimate TBs for channels needed for the out_pmtbs
   *   Initialize float arrays for summing legitimate TBs for channels 
   *     needed by out_pmtbs
   *   Loop for each input basename
   *     Initialize a pmtbs object for the TBs in these input basenames,
   *       for the channels needed
   *     Loop for each col/row:
   *       Sum every non-missing element in channels needed
   *       Increment the count of non-missing elements
   *     Close then input pmtbs TB object
   *   Loop for each col/row:
   *     Calculate the averages
   *   Close the count pmtbs object
   */
  if (verbose >= 2) 
    fprintf(stderr, ">> Initializing arrays to count non-missing TBs...\n");
  count_pmtbs = init_pmtbs(out_pmtbs->grid, algorithm_name, platform_name,
                           NULL, (bool)verbose);
  if (NULL == count_pmtbs) return(FALSE);

  for (chan = 0; chan < out_pmtbs->algorithm->num_channels; chan++) {
    sprintf(buffer,"%s TBs sum",
	    pmalgo_channel_name[out_pmtbs->algorithm->channel[chan]]); 
    sum_tbs[out_pmtbs->algorithm->channel[chan]] = 
      (float **)initialize_matrix(out_pmtbs->grid, sizeof(float), NULL, 
				  buffer, (bool)verbose);
    if (NULL == sum_tbs[out_pmtbs->algorithm->channel[chan]]) {
      fprintf(stderr,"run_mean_tb: initialize_matrix failed for channel %s\n",
	      pmalgo_channel_name[out_pmtbs->algorithm->channel[chan]]);
      return(FALSE);
    }
  }
  
  /* 
   * Loop for each "day" (tbs_basename, which usually, but not necessarily,
   * represents TBs for one day)
   */
  if (verbose >= 2) fprintf(stderr,">> summing non-missing TBs...\n");
  while (0 < argc) {

    if (1 != sscanf(*argv,"%s",tbs_basename)) {
      fprintf(stderr,"pmtbcomb: error reading input tbs_basename\n");
      fprintf(stderr,usage);
      return FALSE;
    }
    ++argv; --argc;

    /* Load input TB data for this day */
    if (verbose >= 2) 
      fprintf(stderr, ">> Initializing next day's TB arrays...\n");
    in_pmtbs = init_pmtbs(out_pmtbs->grid, algorithm_name, platform_name,
                          tbs_basename, (bool)verbose);
    if (NULL == in_pmtbs) return FALSE;

    for (row = 0; row < out_pmtbs->grid->rows; row++) {
      for (col = 0; col < out_pmtbs->grid->cols; col++) {
	for (chan = 0; chan < out_pmtbs->algorithm->num_channels; chan++) {
	  if (0 < 
	      in_pmtbs->data[in_pmtbs->algorithm->channel[chan]][row][col]) {
	    sum_tbs[out_pmtbs->algorithm->channel[chan]][row][col] +=
	      (float)in_pmtbs->
	      data[in_pmtbs->algorithm->channel[chan]][row][col];
	    count_pmtbs->
	      data[count_pmtbs->algorithm->channel[chan]][row][col] += 1;
	  } /* end of a non-missing TB */
	} /* end loop for each channel */
      } /* end loop for each column */
    } /* end loop for each row */
    close_pmtbs(in_pmtbs);
  } /* end loop for each "day" (tbs_basename) */

  /* 
   * Calculate average TB for each pixel
   */
  for (row = 0; row < out_pmtbs->grid->rows; row++) {
    for (col = 0; col < out_pmtbs->grid->cols; col++) {
      for (chan = 0; chan < out_pmtbs->algorithm->num_channels; chan++) {
	count = 
	  count_pmtbs->data[out_pmtbs->algorithm->channel[chan]][row][col];
	if (0 < count) {
	  sum = sum_tbs[out_pmtbs->algorithm->channel[chan]][row][col];
	  out_pmtbs->data[out_pmtbs->algorithm->channel[chan]][row][col] = 
	    (byte2)((sum / (float)count) + 0.5);
	}
      } /* end loop for each channel */
    } /* end loop for each column */
  } /* end loop for each row */

  /* Free memory used in this subroutine */
  close_pmtbs(count_pmtbs);
  for (chan = 0; chan < out_pmtbs->algorithm->num_channels; chan++) {
    if (NULL != sum_tbs[out_pmtbs->algorithm->channel[chan]]) 
      free(sum_tbs[out_pmtbs->algorithm->channel[chan]]);
  }
  
  return TRUE;

}

/*----------------------------------------------------------------------
 *  run_max_tb - determines maximum of non-missing TBs
 * 
 *     input : out_pmtbs - pmtbs object, determines grid and geo_algorithm
 *             argc - current argument count
 *             *argv[] - current argument list
 *             verbose - verbosity flag
 * 
 *    output: out_pmtbs - pmtbs object for output TBs
 *    result: TRUE if successful, or FALSE otherwise
 *
 *----------------------------------------------------------------------*/
static bool run_max_tb (pmtbs_class *out_pmtbs,	int argc, char *argv[],	
			int verbose) {

  int row, col, chan;
  pmtbs_class *in_pmtbs;
  char tbs_basename[FILENAME_MAX];
  char *algorithm_name;
  char *platform_name;
  byte2 tb, max_tb;

  /*
   * To calculate MAX:
   *   Assume the output TB arrays have been initialized to zeroes.
   *   Loop for each input basename
   *     Initialize a pmtbs object for the TBs in these input basenames,
   *       for the channels needed by out_pmtbs
   *     Loop for each col/row:
   *       Test for and retain new max
   *     Close the input pmtbs TB object
   */
  algorithm_name = strdup(pmalgo_id_name[out_pmtbs->algorithm->id]);
  platform_name = strdup( pmalgo_ssmi_platform_name[PMALGO_NO_PLATFORM] );
    
  /* 
   * Loop for each "day" (tbs_basename, which usually, but not necessarily,
   * represents TBs for one day)
   */
  if (verbose >= 2) fprintf(stderr,">> Finding maximum TBs...\n");
  while (0 < argc) {

    if (1 != sscanf(*argv,"%s",tbs_basename)) {
      fprintf(stderr,"pmtbcomb: error reading input tbs_basename\n");
      fprintf(stderr,usage);
      return FALSE;
    }
    ++argv; --argc;

    /* Load input TB data for this day */
    if (verbose >= 2) 
      fprintf(stderr, ">> Initializing next day's TB arrays...\n");
    in_pmtbs = init_pmtbs(out_pmtbs->grid, algorithm_name, platform_name,
                          tbs_basename, (bool)verbose);
    if (NULL == in_pmtbs) return FALSE;

    for (row = 0; row < out_pmtbs->grid->rows; row++) {
      for (col = 0; col < out_pmtbs->grid->cols; col++) {
	for (chan = 0; chan < out_pmtbs->algorithm->num_channels; chan++) {
	  tb = in_pmtbs->data[in_pmtbs->algorithm->channel[chan]][row][col];
	  max_tb = 
	    out_pmtbs->data[out_pmtbs->algorithm->channel[chan]][row][col];
	  if (tb > max_tb) {
	    out_pmtbs->data[out_pmtbs->algorithm->channel[chan]][row][col]
	      = tb;
	  } /* end if this is a new maximum */
	} /* end loop for each channel */
      } /* end loop for each column */
    } /* end loop for each row */

    close_pmtbs(in_pmtbs);

  } /* end loop for each "day" (tbs_basename) */

  return TRUE;

}

/*----------------------------------------------------------------------
 *  run_min_tb - determines minimum of non-missing TBs
 * 
 *     input : out_pmtbs - pmtbs object, determines grid and geo_algorithm
 *             argc - current argument count
 *             *argv[] - current argument list
 *             verbose - verbosity flag
 * 
 *    output: out_pmtbs - pmtbs object for output TBs
 *    result: TRUE if successful, or FALSE otherwise
 *
 *----------------------------------------------------------------------*/
static bool run_min_tb (pmtbs_class *out_pmtbs,	int argc, char *argv[],	
			int verbose) {

  int row, col, chan;
  pmtbs_class *in_pmtbs;
  char tbs_basename[FILENAME_MAX];
  char *algorithm_name;
  char *platform_name;
  byte2 tb, min_tb;

  /*
   * To calculate MIN:
   *   Assume the output TB arrays have been initialized to zeroes.
   *   Loop for each input basename
   *     Initialize a pmtbs object for the TBs in these input basenames,
   *       for the channels needed by out_pmtbs
   *     Loop for each col/row:
   *       Test for and retain new min
   *     Close the input pmtbs TB object
   */
  algorithm_name = strdup(pmalgo_id_name[out_pmtbs->algorithm->id]);
  platform_name = strdup( pmalgo_ssmi_platform_name[PMALGO_NO_PLATFORM] );

  /* 
   * Loop for each "day" (tbs_basename, which usually, but not necessarily,
   * represents TBs for one day)
   */
  if (verbose >= 2) fprintf(stderr,">> Finding minimum TBs...\n");
  while (0 < argc) {

    if (1 != sscanf(*argv,"%s",tbs_basename)) {
      fprintf(stderr,"pmtbcomb: error reading input tbs_basename\n");
      fprintf(stderr,usage);
      return FALSE;
    }
    ++argv; --argc;

    /* 
     * Load input TB data for this day, for the channels required 
     * by out_pmtbs 
     */
    if (verbose >= 2) 
      fprintf(stderr, ">> Initializing next day's TB arrays...\n");
    in_pmtbs = init_pmtbs(out_pmtbs->grid, algorithm_name, platform_name,
                          tbs_basename, (bool)verbose);
    if (NULL == in_pmtbs) return FALSE;

    for (row = 0; row < out_pmtbs->grid->rows; row++) {
      for (col = 0; col < out_pmtbs->grid->cols; col++) {
	for (chan = 0; chan < out_pmtbs->algorithm->num_channels; chan++) {
	  tb = in_pmtbs->data[in_pmtbs->algorithm->channel[chan]][row][col];
	  min_tb = 
	    out_pmtbs->data[out_pmtbs->algorithm->channel[chan]][row][col];
	  if (0 != tb) {
	    if (0 == min_tb || tb < min_tb) {
	      out_pmtbs->data[out_pmtbs->algorithm->channel[chan]][row][col] 
		= tb;
	    } /* end if this is a new minimum */
	  } /* if the new value is non-missing */
	} /* end loop for each channel */
      } /* end loop for each column */
    } /* end loop for each row */

    close_pmtbs(in_pmtbs);

  } /* end loop for each "day" (tbs_basename) */

  return TRUE;

}

