/*======================================================================
 * pmtbs - operations to store and operate on multi-channel passive
 *         microwave TB array objects
 *
 * 09/09/1997 brodzik brodzik@zamboni.colorado.edu 303-492-8263
 * National Snow & Ice Data Center, University of Colorado, Boulder
 *
 *======================================================================*/
#include "define.h"
#include "matrix_io.h"
#include "pmtbs.h"
#include "pmalgorithm.h"

static const char pmtbs_c_rcsid[] = "$Header: /tmp_mnt/FILES/pmalgos/pmtbs.c,v 1.19 2000-06-14 16:09:35 brodzik Exp $";

static bool channel_needed(pmalgo_class *algorithm, pmalgo_channel channel);
static bool load_channel(pmtbs_class *this, int channel_idx,
                         bool verbose);


/* Static variable to keep track of the type of TB files used */
static bool using_new_extensions=FALSE;

/* 
 * turn_on_new_extensions - use modern tb names extensions, e.g. '19v.bin', rather
 * than the original '19V'
 *
 *  Effect: set program to use the new channel extensions
 */
void turn_on_new_extensions( ) 
{
    using_new_extensions = TRUE;
}

/* 
 * turn_off_new_extensions - Use traditional tb name extensions, e.g. '19V'
 * rather than the new moodern '19v.bin'.
 *
 *  Effect: set program to use the original channel extensions
 */
void turn_off_new_extensions( ) 
{
    using_new_extensions = FALSE;
}

/* 
 * new_channels - boolean function to query the current state of extensions.
 * Whether the program is using modern "19v.bin" or traditional '19v'
 * extensions.  
 *
 *  Returns: TRUE if modern extensions are in effect.
 */
bool new_channels()
{
    return using_new_extensions;
}

/*
 * pmalgo_extension_format - returns the proper format for building the
 *                           filename based on the current state of
 *                           using_new_extensions.  The old way included a
 *                           period, modern extensions don't need this
 *                           separator.
 *  Input - channel : index into channel information arrays.
 *
 *  Returns: the proper separator for the current types of filenames.
 */
char* pmalgo_extension_format(int channel)
{
    char* format = "%s.%s";
    if ( new_channels( ) ) {
        format = "%s%s";
    }
    return format;
}

/*
 * pmalgo_channel_extension - return the proper filename extension based on
 *                            user input preference.
 *
 * input - channel : index into channel information arrays
 * 
 * Previously pmalgos required the TB files to have extensions that match the
 * channel_name (e.g. 19V, 22V). Updates to NSIDC's data sets have created the
 * need to create extensions of 19v.bin, 22v.bin, etc.  This routine
 * encapsulates the difference between old and new style file name extensions.
 *
 *  Returns: the proper filename extension for the channel
 */
char* pmalgo_channel_extension(int channel)
{
    static char extension[FILENAME_MAX];
    sprintf(extension, "%s",pmalgo_channel_name[channel]);

    if (new_channels()) {
        sprintf(extension, "%s%s", pmalgo_channel_name[channel], ".bin");
        /* Lowercase the polarization letter [HV] */
        extension[2] = tolower(extension[2]);
    }
    return extension;
    
}



/*----------------------------------------------------------------------
 *  init_pmtbs - initialize multi-channel object according to a specific
 *               algorithm
 *
 *    input: grid - grid object
 *           algorithm_name - algorithm name string
 *           platform_name - SSM/I platform name string
 *           tbs_basename - path and basename of TB grids to use
 *             (assumes sensor-specific channel extensions, 
 *             e.g. for SSM/I, tbs_basename.19H, .19V etc, and
 *             for SMMR, tbs_basename.18H, .21V etc)
 *             if NULL, TB grid arrays for this algorithm are allocated
 *             and initialized to zeroes
 *           verbose - verbosity flag
 * 
 *    returns: pointer to new pmtbs_class instance or NULL 
 *             (and reason to stderr) in case of error
 *
 *----------------------------------------------------------------------*/
pmtbs_class *init_pmtbs(grid_class *grid,
			char *algorithm_name,
                        char *platform_name,
			char *tbs_basename,
			bool verbose) {

  pmtbs_class *this;
  int channel_idx;

  assert(NULL != grid);

  if (verbose) {
    fprintf(stderr,"> %s\n",pmtbs_h_rcsid);
    fprintf(stderr,"> %s\n",pmtbs_c_rcsid);
  }

  /* Allocate storage for new pmtbs object */
  this = (pmtbs_class *)calloc(1,sizeof(pmtbs_class));
  if (NULL == this) {
    perror("init_pmtbs: pmtbs_class object allocation failed.");
    goto cleanup_error_return;
  }
  this->grid = grid;

  if (NULL != tbs_basename) {
    this->tbs_basename = strdup(tbs_basename);
    if (NULL == this->tbs_basename) {
      perror("init_pmtbs: tbs_basename string duplication failed.");
      goto cleanup_error_return;
    }
  }
   
  this->algorithm = init_pmalgo(algorithm_name, platform_name, (int)verbose);
  if (NULL == this->algorithm) {
    goto cleanup_error_return;
  }
  
  /* Allocate & initialize only the tb arrays required for this algorithm */
  for (channel_idx = 0; 
       channel_idx < this->algorithm->num_channels; channel_idx++) {
    if (!load_channel(this,channel_idx,verbose)) goto cleanup_error_return;
  } /* loop for each required channel */

  return this;

  cleanup_error_return:

  if (NULL != this) {
    close_pmtbs(this);
  }
  return NULL;

}




/*----------------------------------------------------------------------
 *  write_pmtbs - write the contents of this pmtbs object to external files
 * 
 *    input: this - pmtbs_class object
 *           tbs_basename - path and basename of TB grid files to create
 *             (will create sensor-specific channel extensions, 
 *             e.g. for SSM/I, tbs_basename.19H, .19V etc, and
 *             for SMMR, tbs_basename.18H, .21V etc)
 *           verbose - verbosity flag
 *
 *    returns: TRUE on success or FALSE (and reason to stderr) otherwise
 *
 *----------------------------------------------------------------------*/
bool write_pmtbs(pmtbs_class *this, char *tbs_basename, bool verbose) {

  int channel;
  char filename[MAX_STRING];
  size_t status;

  if (NULL == this) {
    fprintf(stderr,"write_pmtbs: input pmtbs_class is NULL\n");
    return FALSE;
  }

  if (NULL == tbs_basename) {
    fprintf(stderr,"write_pmtbs: input tbs_basename is NULL\n");
    return FALSE;
  }

  /*
   * For each non-NULL matrix pointer, save the contents to
   * an external file whose extension is the channel name
   */
  for (channel = 0; channel < PMALGO_NUM_CHANNELS; channel++) {
      if (NULL != this->data[channel]) {
          char* ext_format;
          ext_format = pmalgo_extension_format(channel);
          sprintf(filename,ext_format,tbs_basename,pmalgo_channel_extension(channel));

          status = write_matrix (filename, (void **)this->data[channel], 
                                 this->grid->rows, this->grid->cols, 
                                 sizeof(**this->data[channel]));
          if (0 == status) {
              fprintf(stderr,"write_pmtbs: write_matrix failed for %s\n",filename);
              return FALSE;
          } else if (verbose) {
              fprintf(stderr,"> write_pmtbs: wrote output data to %s\n",filename);
          }
      } /* end if there are data for this channel */
  } /* end loop for each channel */

  return TRUE;

}



/*----------------------------------------------------------------------
 *  close_pmtbs - free storage associated with this object
 * 
 *    input: this - pmtbs_class object
 *
 *    result: deallocate all memory associated with this object
 *
 *----------------------------------------------------------------------*/
void close_pmtbs(pmtbs_class *this) {

  int i;

  if (NULL == this) return;
  for (i = 0; i < PMALGO_NUM_CHANNELS; i++) {
    if (NULL != this->data[i]) free(this->data[i]);
  }
  if (NULL != this->algorithm) close_pmalgo(this->algorithm);
  if (NULL != this->tbs_basename) free(this->tbs_basename);
  free(this);
}


/*************************Local functions********************************/

/*----------------------------------------------------------------------
 *  channel_needed - see whether this channel is needed for this algorithm
 * 
 *    input: algorithm - pmalgo_class object
 *           channel - channel id
 *
 *    returns: TRUE if channel is needed by this algorithm, FALSE otherwise
 *
 *----------------------------------------------------------------------*/
static bool channel_needed(pmalgo_class *algorithm, pmalgo_channel channel) {

  int i;
  
  for (i = 0; i < algorithm->num_channels; i++) {
    if (channel == algorithm->channel[i]) return TRUE;
  }

  return FALSE;

}


/*----------------------------------------------------------------------
 *  load_channel - allocate memory for and initialize a channel array
 * 
 *    input: this - pmtbs_class object
 *           channel - channel to load
 *           verbose - verbosity flag
 *
 *    returns: TRUE on success, FALSE (with message to stderr) otherwise
 *
 *----------------------------------------------------------------------*/
static bool load_channel(pmtbs_class *this, int channel_idx,
                         bool verbose) {

    char *filename;

    if (NULL == this->tbs_basename) {
        filename = NULL;
    } else {
        filename = (char *)calloc(1,FILENAME_MAX);
        if (NULL == filename) {
            fprintf(stderr,"load_channel: allocating memory for filename string\n");
            return FALSE;
        }
        char* ext_format;
        ext_format = pmalgo_extension_format(channel_idx);
        sprintf(filename,ext_format,this->tbs_basename,
                pmalgo_channel_extension(this->algorithm->channel[channel_idx]));
    }
    
    this->data[this->algorithm->channel[channel_idx]] = 
        (byte2 **) 
        initialize_matrix (this->grid, 
                           sizeof(**this->data[this->algorithm->channel[channel_idx]]),
                           filename,
                           pmalgo_channel_name[this->algorithm->channel[channel_idx]],
                           verbose);

    if (NULL == this->data[this->algorithm->channel[channel_idx]]) {
        fprintf(stderr,"load_channel: initialize_matrix failed for channel %s\n",
                pmalgo_channel_name[this->algorithm->channel[channel_idx]]);
        if (NULL != filename) free(filename);
        return FALSE;
    }

    if (NULL != filename) free(filename);
    return TRUE;

}
