       SUBROUTINE SEAICE_BOOTSTRAP(OVERLAP,DATE,INORS,NROWS,NCOLS,
     + T19V,T22V,T37H,T37V,LANDMASK,SEAICE,VERBOSE)
C
C  PROGRAM TO DETERMINE SEA ICE CONCENTRATION IN POLAR  REGIONS.
C
C                         PROGRAM ICCONC
C
C                   DEVELOPED FOR J. C. COMISO    NASA
C                  BY S.E.SCHWEINFURTH S.A.R.  9-25-85
C
C  THIS PROGRAM CALCULATES SSM/I ICE CONCENTRATION FROM 4 PASSIVE
C  MICROWAVE CHANNELS (19V,22V,37V, AND 37H) FOR THE ARCTIC 
C  AND 3 CHANNELS (19V,22V AND 37V) FOR THE ANTARCTIC.
C
C
C  INPUT ARGUMENTS:
C      OVERLAP = DEFAULT - USES PARAMETERS FOR DEFAULT SATELLITE,
C                          DETERMINED BY INPUT DATE
C                F08 - FORCES F08 PARAMETERS
C                F11 - FORCES F11 PARAMETERS
C                F13 - FORCES F13 PARAMETERS
C      DATE = YYYYMMDD - DATE OF INPUT DATA
C      INORS = 1 - USE NORTHERN HEMISPHERE PARAMETERS
C              2 - USE SOUTHERN HEMISPHERE PARAMETERS
C      NROWS = NUMBER OF ROWS IN INPUT/OUTPUT ARRAYS
C      NCOLS = NUMBER OF COLS IN INPUT/OUTPUT ARRAYS
C      T37V = INTEGER ARRAY OF 37V TBs (TENTHS OF KELVINS)
C      T37H = INTEGER ARRAY OF 37H TBs (TENTHS OF KELVINS)
C             (ONLY USED FOR NORTHERN HEMISPHERE)
C      T19V = INTEGER ARRAY OF 19V TBs (TENTHS OF KELVINS)
C      T22V = INTEGER ARRAY OF 22V TBs (TENTHS OF KELVINS)
C      LANDMASK = BYTE MASK WITH 
C                 0 : OCEAN (PIXELS WHERE SEAICE WILL BE CALCULATED)
C                 NONZERO : NON-OCEAN PIXELS (NO SEAICE WILL BE 
C                     CALCULATED HERE, OUTPUT WILL BE SET TO 168
C      SEAICE = INTEGER ARRAY OF OUTPUT SEA ICE CONCENTRATIONS
C               (TENTHS OF PERCENT),
C               NON-OCEAN (LAND/COASTLINE) PIXELS ARE SET TO 168
C               OCEAN PIXELS WITH TBS LESS THAN 40K ARE SET TO 157
C
C  VARIABLES:
C      WTP(1) = WATER TIE POINT OF FIRST CHANNEL, 37V
C      WTP(2) = WATER TIE POINT OF SECOND CHANNEL, 37H
C      WTP(3) = WATER TIE POINT OF THIRD CHANNEL, 19V
C      ITP(1) = ICE TIE POINT OF FIRST CHANNEL, 37V
C      ITP(2) = ICE TIE POINT OF SECOND CHANNEL, 37H
C      ITP(3) = ICE TIE POINT OF THIRD CHANNEL, 19V
C      VH37(1) = INTERCEPT FOR CONSOLIDATED ICE LINE, VH37
C      VH37(2) = SLOPE FOR CONSOLIDATED ICE LINE, VH37
C      V1937(1) = INTERCEPT FOR CONSOLIDATED ICE LINE, V1937
C      V1937(2) = SLOPE FOR CONSOLIDATED ICE LINE, V1937
C      WINTRC,WSLOPE = INTERCEPT AND SLOPE FOR EXCLUSION OF WATER
C                      DATA (BASED ON V1937 DATA)
C      WXLIMT = PARAMETER FOR EXCLUSION OF WATER DATA (BASED ON V19 and V22)
C      RADT = RADIAL DISTANCE BETWEEN WATER AND ICE TIE POINTS 
C      TTPI,TTPS = INTERCEPT AND SLOPE OF LINE BETWEEN
C                  WATER AND ICE TIE POINTS (V1937)               
C      WTRK = LOWEST PERCENTAGE FOR WATER DATA (SET AT 0-7.99%).
C      TICEK = ADDITIONAL FACTOR USED IN THE ADJUSTMENT OF BRIGHTNESS
C              TEMPERATURES FROM THEIR SAVED VALUES TO ACTUAL VALUES.
C      NORS = NP OR SP INDEX ( 1 FOR NP, 2 FOR SP)
C      SEASON = SEASON INDEX ( 1 FOR WARM SEASON, 2 FOR COLD SEASON)
C      PL = NUMBER OF LAND PIXELS
C      P0 = NUMBER OF BAD OR MISSING DATA PIXELS
C      PW = NUMBER OF WATER PIXELS
C      PWH = NUMBER OF WEATHER FILTERED PIXELS
C      RADCNT = NUMBER OF PIXELS WHOSE ICE CONCENTRATION WAS CALCULATED
C               VIA THE V1937 RADIUS METHOD.
C      ADCNT = NUMBER OF PIXELS WHOSE ICE CONCENTRATION WAS CALCULATED
C              VIA AN ADDITIONAL METHOD (VH37 INTERSECTION METHOD). 
C      NORMCNT = NUMBER OF PIXELS WHOSE ICE CONCENTRATION WAS CALCULATED
C                VIA THE V1937 INTERSECTION METHOD.     
C
C 07/28/1999 brodzik brodzik@kryos.colorado.edu 303-492-8263
C National Snow & Ice Data Center, University of Colorado, Boulder
C
C $Revision: 1.23.8.1 $
C
C----------DECLARE VARIABLES
C
       PARAMETER (MAX_REGRESS_VALS=520000)

       DIMENSION WTP(3),ITP(3),VH37(2),V1937(2)
       INTEGER*2 SAT,NORS,SEASON
       CHARACTER*2 MONTH,DAY
       CHARACTER*4 YEAR
       CHARACTER*1 POLE
       DOUBLE PRECISION  INTRCA,SLOPEB,SIGMAYY,SIGMAA,
     > SIGMAB,R
       DOUBLE PRECISION X(MAX_REGRESS_VALS),Y(MAX_REGRESS_VALS),
     > SIGMAY(MAX_REGRESS_VALS)
       REAL*4 WTP,ITP,VH37,V1937,WINTRC,WSLOPE,WXLIMT
     > ,RADT,TTPI,TTPS,WTRK,SLOPE,XINT1,YINT1,OFFSET
     > ,XINT2,YINT2
       INTEGER*4 I
     > ,PW,PL,P0,PWH,RADCNT,ADCNT,NORMCNT,ICNT
       INTEGER*2 INORS
       CHARACTER*8  DATE
       CHARACTER*9  OVERLAP
       INTEGER*4 NROWS,NCOLS,NELEMENTS
       INTEGER*2 T19V(*)
       INTEGER*2 T22V(*)
       INTEGER*2 T37H(*)
       INTEGER*2 T37V(*)
       BYTE LANDMASK(*)
       INTEGER*2 SEAICE(*)
       INTEGER*4 VERBOSE

C
C  SET UNIVERSAL PARAMETERS
C
       WTRK=0.0
       TICEK=1.0
       NORS=INORS
       NELEMENTS = NROWS * NCOLS
       IF (0 .LT. VERBOSE) THEN
         WRITE(0,*)
     1        '> $Header: /tmp_mnt/FILES/pmalgos/seaice_bootstrap.f,v 1.23.8.1 2008-06-10 20:36:35 brodzik Exp $'
C'
         WRITE(0,*)'> DATE = ',DATE
         WRITE(0,*)'> NORS = ',NORS
         WRITE(0,*)'> NROWS = ',NROWS
         WRITE(0,*)'> NCOLS = ',NCOLS
         WRITE(0,*)'> NELEMENTS = ',NELEMENTS
       ENDIF

C
       IF (NORS .EQ. 1) THEN
          POLE='N'
       ELSE IF (NORS .EQ. 2) THEN
          POLE='S'
       ELSE
          WRITE(0,*)'SEAICE_BOOTSTRAP ERROR: Invalid hemisphere =',NORS
          RETURN
       ENDIF
c
C
C  BEGIN PROCESSING LOOP
C  IF USED, ZERO OUT THE X AND Y ARRAYS
C
       IF (NORS .EQ. 1) THEN
         DO I=1,MAX_REGRESS_VALS
           X(I)=0.0
           Y(I)=0.0
           SIGMAY(I)=0.0
         END DO
       ENDIF
C
C  EXTRACT YEAR,MONTH AND DAYFROM THE DATASET NAME. IF THE DATA USED IS  
C  AVERAGED DATA (EG: DAILY EXTENTION = .v37, WEEKLY AVERAGED = .v37wk)
C  THEN THE DAY IS AUTOMATICALLY SET TO 1.  
C
       YEAR = DATE(1:4)
       MONTH= DATE(5:6)
       DAY= DATE(7:8)
       READ(YEAR ,'(I4)') IYEAR 
       READ(MONTH,'(I2)') IMONTH
       READ(DAY,'(I2)') IDAY
       IF (0 .LT. VERBOSE) THEN 
         WRITE(0,*)'> ',POLE,IYEAR,IMONTH,IDAY
       ENDIF
C
C Do gross range of value checking on input date
C
       IF (IYEAR .LT. 1987) THEN
         WRITE(0,*) 'SEAICE_BOOTSTRAP ERROR: Input year ', iyear , 
     1   ' out of range for SSM/I.'
         RETURN
       ENDIF
       IF (IMONTH .LT. 1 .OR. IMONTH .GT. 12) THEN
         WRITE(0,*) 'SEAICE_BOOTSTRAP ERROR: Input month ', imonth , 
     1   ' out of range.'
         RETURN
       ENDIF
       IF (IDAY .LT. 1 .OR. IDAY .GT. 31) THEN
         WRITE(0,*) 'SEAICE_BOOTSTRAP ERROR: Input day ', iday , 
     1   ' out of range.'
         RETURN
       ENDIF
C
C  SET SATELLITE AND SEASON BASED ON DATE
C
       IF (OVERLAP .EQ. 'DEFAULT') THEN
         IF (IYEAR .LT. 1987 .OR. (IYEAR .EQ. 1987 .AND. IMONTH .LE. 7))
     1       THEN
           WRITE(0,*) 'SEAICE_BOOTSTRAP ERROR: Input year ', iyear , 
     1     ' and month ', imonth, ' out of range for SSM/I.'
           RETURN
         ENDIF
         IF((IYEAR .LT. 1991) .OR. 
     1        (IYEAR .EQ. 1991 .AND. IMONTH .LT. 12) .OR.
     1        (IYEAR .EQ. 1991 .AND. IMONTH .EQ. 12 .AND. IDAY .LT. 3))
     1      THEN
           SAT = 8
         ELSE IF ((IYEAR .LT. 1995) .OR.
     1          (IYEAR .EQ. 1995 .AND. IMONTH .LT. 5) .OR.
     1          (IYEAR .EQ. 1995 .AND. IMONTH .EQ. 5 .AND. IDAY .LT. 3))
     1          THEN
           SAT = 11
         ELSE 
           SAT = 13
         ENDIF
       ELSE IF (OVERLAP .EQ. 'F08') THEN
         SAT = 8
       ELSE IF (OVERLAP .EQ. 'F11') THEN
         SAT = 11
       ELSE IF (OVERLAP .EQ. 'F13') THEN
         SAT = 13
       ELSE 
         WRITE(0,*) 'SEAICE_BOOTSTRAP ERROR: Argument ', OVERLAP, 
     1   ' unknown'
         RETURN
       ENDIF
       IF(NORS .EQ. 1 .AND. IMONTH .GE. 6 .AND. IMONTH .LE. 10)
     > SEASON=1
       IF(NORS .EQ. 1 .AND. IMONTH .EQ. 10 .AND. IDAY .GE. 16)
     > SEASON=2
       IF(NORS .EQ. 1 .AND. (IMONTH .GE. 11 .OR. IMONTH .LE. 5))   
     > SEASON=2
       IF(NORS .EQ. 2 .AND. (IMONTH .GE. 11 .OR. IMONTH .LE. 1))    
     > SEASON=1
       IF(NORS .EQ. 2 .AND. (IMONTH .GE.  2 .AND. IMONTH .LE. 10))    
     > SEASON=2
       IF (0 .LT. VERBOSE) THEN
         WRITE(0,*) '> NORS=',NORS,' SEASON=',SEASON
       ENDIF
C
C  SET PARAMETERS ACCORDING TO POLE AND SEASON OR TIME OF YEAR.
C  FOR REFERENCE, SEE COMISO, JGR(1986)
C  ARTICLE.
C
C
       IF (NORS .EQ. 1) THEN
         IF (SEASON.EQ.1) THEN
           WTP(1)= 203.0
           WTP(2)= 130.0
           WTP(3)= 181.0
           ITP(1)= 267.0
           ITP(2)= 259.0
           ITP(3)= 270.0
           VH37(1)= -12.00
           VH37(2)= 1.0000
           V1937(1)= 145.000
           V1937(2)= 0.467 
           WINTRC= 72.260
           WSLOPE= 0.5800
           WXLIMT= 23
         ELSE
           WTP(1)= 202.0
           WTP(2)= 130.0
           WTP(3)= 179.0
           ITP(1)= 258.0
           ITP(2)= 249.0
           ITP(3)= 263.0
           VH37(1)= -12.000
           VH37(2)= 1.0000
           V1937(1)= 117.000
           V1937(2)= 0.5533
           WINTRC= 78.000
           wslope=0.5667
           WXLIMT= 14
         ENDIF
C        
         IF(IMONTH .EQ. 5 .AND. IDAY .GE. 15) THEN
           V1937(1)=119
           V1937(2)=0.56
         ENDIF
C        
         IF(IMONTH .EQ. 6) THEN
           VH37(1)= -12.0000
           VH37(2)=1.0000          
           V1937(1)=119
           V1937(2)=0.56
           IF(IDAY .LE.23) THEN
             WINTRC= 78.000
             WSLOPE=0.5667
           ENDIF
         ENDIF
C        
         IF(IMONTH .EQ. 7) THEN
           V1937(1)=119
           V1937(2)=0.56
           IF (IDAY .LE. 18) THEN
             VH37(1)= -70.06
             VH37(2)=1.2226
           ELSE
             VH37(1)= -25.00
             VH37(2)=1.0333
           ENDIF
         ENDIF
C        
         IF(IMONTH .EQ. 8) THEN
            V1937(1)=119
            V1937(2)=0.56
           IF (IDAY .LE.  4) THEN
             VH37(1)= -25.00
             VH37(2)=1.0333
           ELSE
             VH37(1)= -14.0000
             VH37(2)=0.9933
           ENDIF
         ENDIF
C    
         IF(IMONTH .EQ. 9) THEN
           IF (IDAY .LE. 12) THEN
             V1937(1)=119
             V1937(2)=0.56
           ELSE
             V1937(1)=103
             V1937(2)=0.6067
           ENDIF
           IF (IDAY .LE. 20) THEN
             VH37(1)= -14.0000
             VH37(2)= 0.9933
           ELSE
             VH37(1)= -12.0000
             VH37(2)= 1.0000
             WINTRC= 78.000
             WSLOPE= 0.5667 
           ENDIF 
         ENDIF
C        
         IF(IMONTH.EQ.10 ) THEN
           V1937(1)= 103
           V1937(2)= 0.6067 
           WINTRC= 78.000
           WSLOPE= 0.5667
         ENDIF
C        
         IF(IMONTH.EQ.11 .AND. IDAY .LE. 15) THEN
           V1937(1)= 103
           V1937(2)= 0.6067 
         ENDIF
       ENDIF
C
       IF(NORS .EQ. 2) THEN
         IF (SEASON.EQ.1) THEN
           WTP(1)= 203.0           
           WTP(3)= 180.0
           ITP(1)= 262.0
           ITP(3)= 256.0
           V1937(1)= 139.0
           V1937(2)= 0.4733
           VH37(1)= 0.0
           VH37(2)= 0.0
           WINTRC= 93.000 
           WSLOPE= 0.4933                           
           WXLIMT= 16
         ELSE     
           WTP(1)= 203.0           
           WTP(3)= 180.0
           ITP(1)= 262.0
           ITP(3)= 256.0
           V1937(1)= 139.0
           V1937(2)= 0.4733
           VH37(1)= 0.0
           VH37(2)= 0.0
           WINTRC= 93.000 
           WSLOPE= 0.4933      
           WXLIMT= 16
         ENDIF
         IF(IMONTH .EQ. 2 .OR. IMONTH .EQ. 3) THEN
           V1937(1)= 102
           V1937(2)= 0.6200
         ENDIF
         IF(IMONTH .EQ. 2 .AND. IDAY .LE. 7) THEN
           V1937(1)= 120.500
           V1937(2)= 0.5467
         ENDIF
         IF(IMONTH .EQ. 4 .AND.IDAY .LE. 7) THEN
           V1937(1)= 120.500
           V1937(2)= 0.5467
         ENDIF
       ENDIF
C
C  RESET COUNTS AND REWIND LAND MASK FILE
C
       PL=0        
       P0=0
       PW=0
       PWH=0
       RADCNT=0
       ADCNT=0
       NORMCNT=0      
C
C
C NEED TO CREATE 2D ARRAYS OF 37V AND 37H IN ORDER TO APPLY
C REGRESSION ANALYSIS FOR THE NEW VH37(1) AND VH37(2) VALUES.
C CURRENTLY THESE ARE JUST COMPUTED FOR THE SUMMER SEASON AND JUST FOR THE
C NP.  WE USE ALL THE POINTS WHICH ARE ABOVE THE LINE (Y=1.0X-22) FOR THE
C REGRESSION ANALYSIS.
C NOTE: 2 IS ADDED TO THE NEW INTERCEPT
C
C A. ALLEGRINO 4/26/95
C
C
       IF(NORS .EQ.1 .AND. SEASON .EQ. 1) THEN         
        ICNT=1
        DO I=1,NELEMENTS
           IF (LANDMASK(I).eq. 0) THEN
             TB1= (FLOAT (T37V(I)))/10.0
             TB2= (FLOAT (T37H(I)))/10.0
             IF(TB1 .GE. 40 .AND. TB2 .GE. 40 .AND. 
     >          TB1 .LT. 280 .AND. TB2 .LT. 280) THEN    
               IF(11 .EQ. SAT) THEN
                 TB1= (1.0 - 0.00335) *TB1 +0.01
                 TB2= (1.0 - 0.00165) *TB2 +0.03
               ENDIF
c added the F13 Wentz coefficients J. Stroeve 1/7/98
C
C Note the slope here has more digits than the one
C used later, at lines 596-7.  I am not sure whether this
C is significant.  --MJB
C
               IF(13 .EQ. SAT) THEN
                 TB1= (1.0 - 0.00335131)*TB1 - 0.49
                 TB2= (1.0 - 0.0016533)*TB2 - 0.05
               ENDIF
               C37H=TB1 -22
               IF(TB2 .GT. C37H) THEN
                 X(ICNT)=DBLE (TB1)
                 Y(ICNT)=DBLE (TB2)
                 ICNT=ICNT+1
                 IF(ICNT .GT. MAX_REGRESS_VALS) THEN
                   WRITE(0,*) 'SEAICE_BOOTSTRAP WARNING:'
                   WRITE(0,*) 'Number of elements in regression arrays'
                   WRITE(0,*) 'exceeds allocated space.'
                   WRITE(0,*) 'Algorithm results may not be reliable.'
                   WRITE(0,*) 'Resizing MAX_REGRESS_VALS may be needed.'
                   GO TO 8500
                 ENDIF
               ENDIF
             ENDIF 
           ENDIF
        END DO
8500    CONTINUE
        CALL LINFIT(X,Y,SIGMAY,ICNT,0,INTRCA,SIGMAA,SLOPEB,SIGMAB,R,
     >   SIGMAYY)
        IF(SLOPEB .LE. 1.2226) THEN
          VH37(1)= INTRCA +2
          VH37(2)= SLOPEB
          IF (0 .LT. VERBOSE) THEN 
            WRITE(0,*) '> used LINFIT:  ',VH37(1),VH37(2)
          ENDIF
        ENDIF    
       ENDIF        
C
C  DEFINE WATER-ICE LINE PARAMETERS: (1's are for VH37 
C                                     2's are for V1937)
C  AND RADIAL DISTANCE
C     
       TTPS=(ITP(3)-WTP(3))/(ITP(1)-WTP(1))
       TTPI=ITP(3)-TTPS*ITP(1)
       IF(NORS .EQ. 1) THEN 
          IF (SEASON .EQ. 1) THEN
            XINT1=266
            YINT1=270
          ELSE
            XINT1=261
            YINT1=263
          ENDIF
          SLOPE=(YINT1-WTP(3))/(XINT1-WTP(1))
          OFFSET= WTP(3) - (WTP(1)*SLOPE)
          XINT2=(OFFSET - V1937(1))/(V1937(2) - SLOPE)
          YINT2=(V1937(2) * XINT2) + V1937(1)
          RADT=SQRT( (XINT2-WTP(1))**2 + (YINT2-WTP(3))**2) 
       ELSE
          RADT=SQRT( (ITP(1)-WTP(1))**2 + (ITP(3)-WTP(3))**2)
       ENDIF 
C
C  PRINT THE VARIOUS STARTING PARAMETERS USED IN THE ALGORITHM
C
       IF (0 .LT. VERBOSE) THEN
         WRITE(0,2) WTP
         WRITE(0,222) ITP
         WRITE(0,23) (VH37(J),J=1,2)
         WRITE(0,25) (V1937(J),J=1,2)
         WRITE(0,24) WINTRC,WSLOPE
         WRITE(0,111) RADT
         WRITE(0,29) TTPI,TTPS
         WRITE(0,27) WTRK
 2       FORMAT('> WATER TIEPOINTS ARE = ',3F8.2)
 222     FORMAT('> FIRST YEAR ICE TIE PTS = ',3F8.2)
 23      FORMAT('> THE VH37 CONSOLIDATED ICE LINE IS Y= ',F8.3,' + X*',
     >        F8.3)
 25      FORMAT('> THE V1937 CONSOLIDATED ICE LINE IS Y= ',F8.3,' + X*',
     >        F8.3)
 24      FORMAT('> THE WATER EQUATION IS Y= ',F8.3,' + X*',F8.3)
 111     FORMAT('> THE RADIUS IS ',F8.3)
 29      FORMAT('> THE LINE CONNECTING V1937 WATER TIEPTS W/' //
     >         ' ICE TIEPTS IS:Y= ',F8.3,' + X*',F8.3)
 27      FORMAT('> WATER CUT OFF AT ',F4.1,'%')
       ENDIF
C
C  BEGIN PROCESSING DATA BY READING IN A RECORD OF MASK AND THREE TBs
C
       DO 200 I = 1,NELEMENTS
C
C  TEMPERATURES LESS THAN 40K (400=40*10 (WHERE 10 IS OUR
C  TEMPERATURE SCALING FACTOR)) ARE TAGGED AS BAD DATA.  SKIP THESE AND ALL
C  LAND POINTS.
C
          IF(T37V(I).LT.400 .OR. T19V(I).LT.400 .OR.
     >    T22V(I).LT.400 .OR. LANDMASK(I).ne. 0) GOTO 50
          IF (NORS .EQ. 1) THEN
            IF (T37H(I).LT.400) THEN
              GOTO 50
            ENDIF
          ENDIF
C         
          TB1=T37V(I)/(10.0*TICEK)
          IF (NORS .EQ. 1) THEN
            TB2=T37H(I)/(10.0*TICEK)
          ELSE 
            TB2=0.0
          ENDIF
          TB3=T19V(I)/(10.0*TICEK)
          TB4=T22V(I)/(10.0*TICEK)
C         
C***********************************************************************
C   F11 to F8 transformation provided by Frank Wentz thru Mike Martino 
C   A. Allegrino 3/5/96
C***********************************************************************
C
          IF(11 .EQ. SAT) THEN
           TB1= (1.0 - 0.00335) *TB1 +0.01
           TB2= (1.0 - 0.00165) *TB2 +0.03
           TB3= (1.0 - 0.00221) *TB3 -0.44
           TB4= (1.0 - 0.00161) *TB4 -0.30
          ENDIF
C***********************************************************************
C   F13 to F8 transformation provided by Frank Wentz
C***********************************************************************
C
          IF(13 .EQ. SAT) THEN
           TB1= (1.0 - 0.00335) *TB1 -0.49
           TB2= (1.0 - 0.00165) *TB2 -0.05
           TB3= (1.0 - 0.00221) *TB3 -0.04
           TB4= (1.0 - 0.00161) *TB4 -0.37
          ENDIF
C
C  CALL SUBROUTINE TO CALCULATE ICE CONCENTRATIONS.
C
           CALL ICCALC(NORS,VH37,V1937,WTP,TB1,TB2,
     >     TB3,TB4,WSLOPE,WINTRC,RADT,TTPS,TTPI,WTRK,ITBD,
     >     WXLIMT,PW,PWH,RADCNT,ADCNT,NORMCNT)
           SEAICE(I)=ITBD
c put in a statement to check for ice values over 100% and less than 0%.
c (Julienne Stroeve 5/9/97)
           IF (SEAICE(I) .GT. 1000) THEN
             SEAICE(I) = 1000
           ENDIF
           IF (SEAICE(I) .LT.   0) THEN
             SEAICE(I) = 0
           endif
           GO TO 100
C
C  COME HERE IF THERE IS BAD OR MISSING DATA, OR IF PIXEL IS LAND
c  we changed this so that if there is bad or missing data, the pixel
c  values is set to (-99) 157  (Julienne Stroeve 11/21/96)
C  IF THE PIXEL IS LAND, THE PIXEL VALUE IS SET TO -800 AND THE LAND
C  PIXEL IS COUNTED.
c  we changed this so that if the pixel is land it is set to -88 (168)
c  Julienne Stroeve 11/21/96
c  these are the same changes than Nick installed before.
C
   50      CONTINUE
C
C  COUNT THE NUMBER OF BAD OR MISSING PIXELS
C
           IF(LANDMASK(I).eq.0) THEN
              SEAICE(I) = 1570
              P0=P0+1
           ELSE
              SEAICE(I) = 1680
              PL=PL+1
           ENDIF
  100      CONTINUE

  200  CONTINUE
       IF (0 .LT. VERBOSE) THEN
         WRITE(0,210) PL,P0
         WRITE(0,213) PW,PWH
         WRITE(0,214) RADCNT,ADCNT,NORMCNT
 210     FORMAT(' # OF LAND PIXELS',I7,' BAD OR MISSING DATA PIXELS',I7)
 213     format(' # OF WATER FILTERED PIXELS',I7,' WEATHER PIX',I7)
 214     format(' # OF PIXELS BELOW WATER-ICE LINE',I7,' THAT USED VH37'
     1        ,I7,' THAT USED V1937',I7)

         WRITE(0,*) 'NORMAL END OF PROGRAM.'
       ENDIF

       RETURN

       END
C***************************************************************
C                          SUBROUTINE ICCALC
C
C                    DEVELOPED FOR J. C. COMISO     NASA
C                     BY S.E.SCHWEINFURTH  S.A.R.  9-25-85
C
C  MODIFIED TO ACCOUNT FOR SSMI DATA BY A.ALLEGRINO 2/89.
C
C  THIS SUBROUTINE CALCULATES ICE CONCENTRATION BY
C  FINDING THE RATIO OF THE DISTANCES OF (WATER VALUE TO
C  DATA VALUE) TO (WATER VALUE TO POLYNOMIAL VALUE).
C  THE POLYNOMIAL VALUE IS FOUND BY INTERSECTING THE LINE
C  THROUGH WATER AND DATA POINTS WITH THE POLYNOMIAL.
C
C
C
C      VARIABLES:
C          TB1= 37V BRIGHTNESS TEMPERATURE
C          TB2= 37H BRIGHTNESS TEMPERATURE
C          TB3= 19V BRIGHTNESS TEMPERATURE
C          TB4= 22V BRIGHTNESS TEMPERATURE
C          WTP1= 37V WATER TIE POINT
C          WTP3= 37H WATER TIE POINT
C          WTP2= 19V WATER TIE POINT
C          ET1= RESULT FROM WATER EQUATION (THE CALCULATED 19V
C               BRIGHTNESS TEMPERATURE USED TO TEST FOR WATER PIXEL
C               (ALSO USED TO DETERMINE IF THE 19V DATA WOULD BE USED
C                FOR NP ICE CONCENTRATION CALCULATION)
C          EINT1= X-COORDINATE OF THE POINT OF INTERSECTION
C          EINT2= Y-COORDINATE OF THE POINT OF INTERSECTION
C          TTPS= SLOPE OF THE LINE BETWEEN THE CONSOLIDATED ICE TIE
C                POINT AND THE WATER TIE POINT (V1937)
C          TTPI= INTERCEPT OF THE LINE BETWEEN THE CONSOLIDATED ICE
C                AND THE WATER TIE POINTS (V1937)
C
C
       SUBROUTINE ICCALC(NORS,VH37,V1937,WTP,TB1,TB2,
     > TB3,TB4,WSLOPE,WINTRC,RADT,TTPS,TTPI,WTRK,ITBD,
     > WXLIMT,PW,PWH,RADCNT,ADCNT,NORMCNT)
C
C------DECLARE VARIABLES
C
       INTEGER * 4 PW,PWH,RADCNT,ADCNT,NORMCNT
       REAL*4 VH37(2),WTP(3),TB1,TB2,TB3,TB4,ICEC,ET1
     > ,ET19V,WSLOPE,WINTRC,V1937(2),ET3
     > ,RADT,TTPS,TTPI,WTRK
       INTEGER * 2 NORS
C
C  SET UP EQUATIONS AND INITIALIZE VARIABLES.
C  DATA BELOW THE OPEN WATER LINE (FILTER) IS ASSUMED TO BE LESS THAN
C  8%.
C
       WTP1=WTP(1)
       WTP2=WTP(2)
       WTP3=WTP(3)
C
C  THIS IS A NEW WATER FILTER USING THE 19V VS 22V SCATERPLOT
C  TB4 IS THE 22V TEMPERATURE
C
C  A. ALLEGRINO 5/25/95
C
       ET1 = WSLOPE * TB4 + WINTRC
       IF(ET1.GT.TB3)THEN
        ICEC=WTRK/100.
        PW=PW+1
        GOTO 75
       ENDIF
C
C  APPLY ATMOSPHERIC FILTER. THE (22V-19V) TEMPERATURE DIFFERENCE IS
C  COMPARED WITH A CERTAIN VALUE . IF THE DIFFERENCE > THIS VALUE THEN THE PIXEL
C  IS SAID TO CONTAIN WEATHER AND IS CONVERTED TO THE VALUE FOR OPEN
C  OCEAN. THE VALUE CHANGES WITH HEMISPHERE AND SEASON, AND WAS DERIVED EMPIRICALLY.
C
       ACTDIFF=TB4-TB3
       IF(ACTDIFF .GT. WXLIMT)THEN
        ICEC=WTRK/100.
        PWH=PWH+1
        GOTO 75
       ENDIF
C
C  THERE ARE THREE OPTIONS CONTAINED IN THE FOLLOWING IF
C  STATEMENT.  THE FIRST FILTERS OUT THOSE POINTS BELOW
C  THE VH1937 WATER-ICE LINE AND USES THE V1937 PLOT WITH RADT 
C  TO CALCULATE CONCENTRATIONS.  THE SECOND OPTION IS FOR THOSE 
C  POINTS ABOVE A SET CRITERIA (AD-5) IN VH37 SPACE (NP ONLY) AND
C  USES VH37 SPACE. THE THIRD (DEFALUT) USES V1937 SPACE.
C  
       ET19V=TTPI+TTPS*TB1
       ET3=( VH37(1)-5 )+( VH37(2)*TB1 )
       IF (TB3 .LE. ET19V) THEN
         RADCNT=RADCNT+1
         R1=SQRT( (TB1-WTP1)**2 + (TB3-WTP3)**2 )
         ICEC=R1/RADT
       ELSE IF (NORS .EQ. 1 .AND. TB2 .GT. ET3) THEN
         ADCNT=ADCNT+1
         ED=TB1-WTP1
         IF(ED .EQ. 0.0)THEN
           EINT1=WTP1
         ELSE
           BET=(TB2-WTP2)/ED
           ALP=TB2-(BET*TB1)
           SLDVH=VH37(2)-BET
           IF(SLDVH .EQ. 0.0) THEN
             ICEC=WTRK/100.
             GOTO 75
           ELSE
             EINT1=(ALP-VH37(1))/SLDVH
           ENDIF
         ENDIF
         EINT2=VH37(1)+(VH37(2)*EINT1)      
         RINT=SQRT( (EINT1-WTP1)**2 + (EINT2-WTP2)**2 )
         IF (RINT .EQ. 0)THEN
            ICEC=WTRK/100.
         ELSE
            R1=SQRT( (TB1-WTP1)**2 + (TB2-WTP2)**2 ) 
            ICEC=R1/RINT
         ENDIF 
       ELSE
         NORMCNT=NORMCNT+1
         ED=TB1-WTP1
         IF(ED .EQ. 0.0)THEN
           EINT1=WTP1
         ELSE
           BET=(TB3-WTP3)/ED
           ALP=TB3-(BET*TB1)
           SLDVH=V1937(2)-BET
           IF(SLDVH .EQ. 0.0) THEN
             ICEC=WTRK/100.
             GOTO 75
           ELSE
             EINT1=(ALP-V1937(1))/SLDVH
           ENDIF
         ENDIF
         EINT2=V1937(1)+(V1937(2)*EINT1)      
         RINT=SQRT( (EINT1-WTP1)**2 + (EINT2-WTP3)**2 )
         IF (RINT .EQ. 0)THEN
            ICEC=WTRK/100.
         ELSE
            R1=SQRT( (TB1-WTP1)**2 + (TB3-WTP3)**2 ) 
            ICEC=R1/RINT
         ENDIF
       ENDIF                  
C
C  SET ICE CONCENTRATIONS LIMITS FOR NASA COLOR PROGRAM.
C  ICE CONCENTRATION UP TO 108% ARE STORED BECAUSE OF UNCERTAINTIES
C  IN THE EMISSIVITY OF CONSOLIDATED ICE.
C
75     ICEC = ICEC * 100.0
       IF(ICEC.LT.WTRK) ICEC = WTRK
       IF(ICEC.GT.108.0) ICEC = 108.0
C
C  WE PACK THE ICE CONCENTRATIONS IN THE FOLLOWING MANNER.
C  THIS IS JUST A SCALING TO BE ABLE TO USE NASA COLOR SOFTWARE
C  CONVENIENTLY.  AN ALTERNATIVE WOULD BE TO MULTIPLY ICEC BY 10
C  BEFORE STORING IT AS INTEGER.
C
       ITBD=INT((ICEC*10)+0.5)
       RETURN
       END

c********************************************************************
c
c  This linfit routine is from Bevingtons Statistical book      
c
      SUBROUTINE LINFIT(X,Y,SIGMAY,NPTS,MODE,A,SIGMAA,B,SIGMAB,R,
     >sigmayy)       
      DOUBLE PRECISION SUM,SUMX,SUMY,SUMX2,SUMXY,SUMY2             
      DOUBLE PRECISION XI,YI,WEIGHT,DELTA,VARNCE             
      double precision x,y,sigmay,a,sigmaa,b,sigmab,r,sigmayy
      DIMENSION X(1),Y(1),SIGMAY(1)                                     
C                                                                   
C                                                                      
      SUM = 0.                                                      
      SUMX = 0.                                                        
      SUMY = 0.                                                        
      SUMX2 = 0.                                                       
      SUMXY = 0.                                                       
      SUMY2 = 0.                                                       
      DO 50 I=1,NPTS                                                 
      XI = X(I)                                                       
      YI = Y(I)                                                       
      IF(MODE) 31,36,38                                               
   31 IF(YI) 34,36,32                                                  
   32 WEIGHT = 1.0/YI                                                 
      GO TO 41                                                        
   34 WEIGHT = 1.0/(-YI)                                              
       GO TO 41                                                       
   36 WEIGHT = 1.0                                                    
      GO TO 41                                                        
   38 WEIGHT = 1.0/SIGMAY(I)**2                                       
   41 SUM = SUM + WEIGHT                                              
      SUMX = SUMX + WEIGHT*XI                                         
      SUMY = SUMY + WEIGHT*YI                                         
      SUMX2 = SUMX2 + WEIGHT*XI*XI                                    
      SUMXY = SUMXY + WEIGHT*XI*YI                                    
      SUMY2 = SUMY2 + WEIGHT*YI*YI                                    
   50 CONTINUE                                                         
      DELTA = SUM*SUMX2 - SUMX*SUMX                                   
      A = (SUMX2*SUMY - SUMX*SUMXY)/DELTA                             
      B = (SUMXY*SUM - SUMX*SUMY)/DELTA                               
      IF(MODE) 62,64,62                                               
   62 VARNCE = 1                                                      
      GO TO 67                                                        
   64 C = NPTS - 2                                                    
      VARNCE = (SUMY2 + A*A*SUM + B*B*SUMX2                           
     1 - 2.0*(A*SUMY + B*SUMXY - A*B*SUMX))/C                         
      sigmayy=dsqrt(varnce)
   67 SIGMAA = DSQRT(VARNCE*SUMX2/DELTA)                              
      SIGMAB = DSQRT(VARNCE*SUM/DELTA)                                
      R = (SUM*SUMXY - SUMX*SUMY)/                                    
     1 DSQRT(DELTA*(SUM*SUMY2 - SUMY*SUMY))                           
      RETURN                                                           
      END
