/*======================================================================
 * snow_gb - calculate passive microwave-derived snow/surface 
 *           classification of Norm Grody and Alan Basist.
 *
 * 06/12/2000 brodzik brodzik@zamboni.colorado.edu 303-492-8263
 * National Snow & Ice Data Center, University of Colorado, Boulder
 *
 * References:
 * 
 * Grody, Norman C. and Alan N. Basist, "Global Identification of 
 *    Snowcover Using SSM/I Measurements," IEEE Transactions on 
 *    Geoscience and Remote Sensing, 34(1), January 1996, pp. 237-249.
 * 
 * For explicit algorithm details, see also:
 * 1st version of algorithm:
 * Basist, Alan, e-mail correspondence to Richard Armstrong, Wed, 
 *    2 Feb 2000, 14:47:30 -0700 (MST).  This e-mail was saved as v1.1.
 * 2nd version of algorithm:
 * Basist, Alan, e-mail correspondence to M. J. Brodzik, Fri, 
 *    4 Aug 2000, 13:45:51.  This e-mail was saved as v1.1.1.1.
 *
 *======================================================================*/
#include "define.h"
#include "snow_gb.h"

static const char snow_gb_c_rcsid[] = "$Header: /tmp_mnt/FILES/pmalgos/snow_gb.c,v 1.8 2004-06-28 18:44:32 brodzik Exp $";

/* Adjustments to convert brightness temperatures to antenna temperatures */
#define ADJUST_19GHZ 7
#define ADJUST_22GHZ 6
#define ADJUST_37GHZ 4
#define ADJUST_85GHZ 3

gb_return_code snow_gb(gb_surface_condition *surface_condition, 
		       byte2 tb19v, byte2 tb19h, 
		       byte2 tb22v, 
		       byte2 tb37v, 
		       byte2 tb85v, byte2 tb85h,
		       int verbose) {

  
  /* Antenna temperatures */
  long int P1, P2, P3, P4, P6, P7;

  /* Cross-polarization differences */
  long int PD19, PD85;

  /* Cross-channel differences */
  long int F14;
  long int F31, F34, F36;
  long int F46;
  
  long int snow, scat;
  float tt;

  if (2 < verbose) {
    fprintf(stderr,"%s\n",snow_gb_h_rcsid);
    fprintf(stderr,"%s\n",snow_gb_c_rcsid);
  }

  /* Initialize surface to unknown */
  *surface_condition = GB_UNK;

  /* 
   * The algorithm expects antenna temperature input.  Convert input 
   * brightness temperatures to antenna temperatures.  Convert to signed
   * long integers to avoid any problems with signs and overflow.
   */
  P1 = (long)tb19v - ADJUST_19GHZ;
  P2 = (long)tb19h - ADJUST_19GHZ;
  P3 = (long)tb22v - ADJUST_22GHZ;
  P4 = (long)tb37v - ADJUST_37GHZ;
  /* P5 omitted on purpose, this algorithm doesn't use 37h */
  P6 = (long)tb85v - ADJUST_85GHZ;
  P7 = (long)tb85h - ADJUST_85GHZ;
  
  /* polarization at 19Ghz */
  PD19 = P1 - P2;
  PD85 = P6 - P7;
  F14 = P1 - P4;
  F31 = P3 - P1;
  F34 = P3 - P4;
  F36 = P3 - P6;
  F46 = P4 - P6;

  /* Subsurface ICE */
  if(P3<=210 || (P3<=229 && P6<=240 && F36 < 0))
  {
    *surface_condition = GB_GIC;
    return(GB_GLACIER_ICE);
  }

  scat = F36;
  if (F14 > scat) scat = F14;
  if (F46 > scat) scat = F46;
  snow = 0;
  if (scat >= 1.0 && P3 < 257)
  {
    /* turn on snow, but look for exceptions */
    snow = 1;

    /* remove precip */
    tt = 165.0 + 0.49 * P6;
    if(P3 >= 258 || (float)P3 >= tt)
    {
      *surface_condition = GB_PC1;
      return(GB_NO_SNOW);
    }
    if(P3 >= 254 && scat <= 2.0 && PD85 >= 3)
    {
      *surface_condition = GB_PC2;
      return(GB_NO_SNOW);
    }
    /* F31 threshold changed from 0 to 3, per validation results
     * and phone conversation with Alan Basist, 8/14/00. */
    if(PD19 >= 18 && F14 <= 10 && F46 <= 5 && F31 <= 3)
    {
      *surface_condition = GB_CDS;
      return(GB_NO_SNOW);
    }
    /* bring back frozen ground */
    if(PD19 >= 12 && scat <= 2 && F14 <= 2)
    {
      *surface_condition = GB_FZG;
      return(GB_NO_SNOW);
    }
    /* something strange in the desert */
    if(F34 > 17 || P6 > P4)
    {
      *surface_condition = GB_MX1;
      return(GB_SNOW);
    }
  }
  /* if the code gets to this point and snow == 1 then
       you have SNOW! */
  if (snow == 1)
  {
    *surface_condition = GB_SNW;
    return(GB_SNOW);
  }

  return(GB_NO_SNOW);

}


