/*======================================================================
 * ssmiice - program to calculate SSM/I-derived ice
 *
 * Notes:
 *   Refer to ssmitbs.h for legitimate set of ssmi_algorithm_names.
 *   Error Processing:
 *     1) Returns EXIT_FAILURE, an error message to stderr and the
 *        usage message to stderr in any of the following cases:
 *        a) -c is used with a non-existent climatology filename
 *        b) -i is used with an invalid sea ice algorithm name
 *        c) -u is used with greater than MAX_USE_LOCI_VALUES numerals
 *        e) an invalid option letter is used
 *        f) the number of required inputs is incorrect
 *     2) Returns EXIT_FAILURE, and an error message to stderr
 *        in any of the following cases:
 *        a) init_grid fails for the input .gpd
 *        b) error reading the loci mask file
 *        c) error reading the climatolotdy file
 *        d) init_ssmitbs fails
 *        e) error writing the ice data
 *
 * 08/25/1997 brodzik brodzik@zamboni.colorado.edu 303-492-8263
 * National Snow & Ice Data Center, University of Colorado, Boulder
 *
 *$Log: not supported by cvs2svn $
 * Revision 1.4  1997/11/07  20:17:52  brodzik
 * Fixed usage message (it had wrong option letter for alternate seaice
 * algorithm).
 *
 * Revision 1.3  1997/10/15  16:54:32  brodzik
 * Added #include "matrix_io.h"
 * Added complete memory management cleanup.
 *
 * Revision 1.2  1997/10/08  18:31:45  brodzik
 * Fixed order of conditional for using climatology and checking
 * loci values.
 *
 * Revision 1.1  1997/10/02  20:57:01  brodzik
 * Initial revision
 *
 *
 *======================================================================*/
#include "define.h"
#include "matrix_io.h"
#include "nise_constants.h"
#include "ssmitbs.h"

static const char ssmiice_c_rcsid[] = "$Header: /tmp_mnt/FILES/pmalgos/ssmiice.c,v 1.5 1997-11-14 16:21:00 brodzik Exp $";

#define DEFAULT_SEAICE_ALGORITHM NASATEAM_NORTH
#define MAX_USE_LOCI_VALUES 10

#define usage \
"ssmiice $Revision: 1.5 $ \n"\
"usage: ssmiice [-v] [-c climatology_filename] \n"\
"       [-i alternate_ice_algorithm] \n"\
"       [-u loci_value_to_use]\n"\
"       grid.gpd loci_mask tb_grids_basename ice_filename\n"\
"\n"\
" input : gpd file  - grid parameters definition file\n"\
"         loci_mask - filename containing starting mask with \n"\
"           land/ocean/coastline/permanent_ice classifications\n"\
"           (key to expected classifications is in nise_constants.h)\n"\
"         tb_grids_basename - path and basename of TB grids to use\n"\
"           (assumes channel extensions, e.g. tb_grids_basename.19H, .19V etc)\n"\
"         ice_filename - ssmiice output file to create\n"\
"\n"\
" output : SSM/I-derived ice grid \n"\
"\n"\
" option : c - climatology file to use (default is no climatology)\n" \
"          i - alternate seaice algorithm designator, possible values:\n"\
"               NASATEAM_NORTH (default)\n" \
"               NASATEAM_SOUTH\n"\
"          v - verbose mode (can be repeated)\n"\
"          u - integer value of loci mask pixels to operate on \n" \
"              (may be repeated up to 10 values)\n" \
"\n"

main (int argc, char *argv[])
{
  char *option;
  int i;
  bool verbose=FALSE;
  bool very_verbose=FALSE;
  char climatology_filename[FILENAME_MAX];
  bool use_climatology;
  char gpd_filename[FILENAME_MAX];
  char mask_filename[FILENAME_MAX];
  char tbs_basename[FILENAME_MAX];
  char ice_filename[FILENAME_MAX];
  ssmi_algorithms seaice_algo;
  grid_class *grid;
  ssmitbs_class *ssmitbs;
  byte1 **data;
  byte1 **climatology;
  byte2 use_loci_value[MAX_USE_LOCI_VALUES];
  int num_use_loci_values=0;
  int row, col;
  byte1 ice;
  
  /*
   * Set defaults
   */
  *climatology_filename = 0;
  use_climatology = FALSE;
  seaice_algo=DEFAULT_SEAICE_ALGORITHM;
  for (i=0; i<MAX_USE_LOCI_VALUES; i++) {
    use_loci_value[i]=BYTE2_MAX;
  }

  /* 
   *      get command line arguments
   */
  while (--argc > 0 && (*++argv)[0] == '-') {
    for (option = argv[0]+1; *option != '\0'; option++) {
      switch (*option) {
        case 'c':
	  ++argv; --argc;
	  strcpy(climatology_filename,*argv);
	  if (NULL == climatology_filename) error_exit(usage);
	  use_climatology = TRUE;
	  break;
        case 'i':
	  ++argv; --argc;
	  if (!is_seaice_algorithm(*argv,&seaice_algo)) {
	    fprintf (stderr, "ssmiice: invalid ice algorithm %s\n\n", *argv);
	    fprintf (stderr,usage);
	    exit(EXIT_FAILURE);
	  }
	  break;
        case 'u':
	  ++argv; --argc;
	  if (num_use_loci_values >= MAX_USE_LOCI_VALUES) {
	    fprintf (stderr, "ssmiice: exceeded maximum use_loci_values of %d\n\n",
		     MAX_USE_LOCI_VALUES);
	    fprintf (stderr,usage);
	    exit(EXIT_FAILURE);
	  }
	  use_loci_value[num_use_loci_values++] = strtol(*argv,NULL,10);
	  break;
	case 'v':
	  if (verbose) very_verbose = TRUE;
	  verbose = TRUE;
	  break;
	default:
	  fprintf (stderr, "ssmiice: invalid option %c\n", *option);
	  fprintf (stderr,usage);
	  exit(EXIT_FAILURE);
      } /* end switch */
    } /* end loop for each input command option */
  } /* end loop while still input arguments */

  if (4 != argc) 
  {
    fprintf(stderr,"ssmiice: Insufficient number of calling arguments.\n\n");
    fprintf (stderr,usage);
    exit(EXIT_FAILURE);
  }

  strcpy (gpd_filename, *argv);
  ++argv; --argc;
  strcpy (mask_filename, *argv);
  ++argv; --argc;
  strcpy (tbs_basename, *argv);
  ++argv; --argc;
  strcpy (ice_filename, *argv);

  if (verbose) 
  {
    fprintf(stderr,"> ssmiice: Beginning processing for:\n");
    fprintf(stderr,">  gpd                 =%s\n",gpd_filename);
    fprintf(stderr,">  loci mask           =%s\n",mask_filename);
    fprintf(stderr,">  tb_grids_basename   =%s\n",tbs_basename);
    fprintf(stderr,">  ice file            =%s\n",ice_filename);
    fprintf(stderr,">  use_climatology     =%d\n",use_climatology);
    fprintf(stderr,">  -c climatology file =%s\n",climatology_filename);
    fprintf(stderr,">  -i seaice algorithm =%s\n",
	    ssmi_algorithm_name[seaice_algo]);
    fprintf(stderr,">  -u operate on loci mask values of\n");
    for (i=0; i<num_use_loci_values; i++) {
      fprintf(stderr,">  %d\n",use_loci_value[i]);
    }
  }

  /*
   * Initialize grid (used for array dimensions throughout)
   */
  grid = init_grid(gpd_filename);
  if (NULL == grid) {
    fprintf(stderr,"ssmiice: init_grid failed for gpd=%s\n",gpd_filename);
    goto cleanup_error_exit;
  } else if (verbose) {
    fprintf(stderr,"> ssmiice: initialized grid for gpd=%s\n",gpd_filename);
  }
    
  /*
   * Load input data 
   *  Initialize output data with land/ocean/mixed/corners mask
   *  climatology (if using one)
   *  and initialize tbs for specified ice algorithm
   */
  data = (byte1 **)initialize_matrix
    (grid, sizeof(**data), mask_filename,
     "loci (land/ocean/coast/ice) mask", verbose);
  if (NULL == data) {
    fprintf(stderr,"ssmiice: bad read on loci mask=%s\n",mask_filename);
    goto cleanup_error_exit;
  } else if (verbose) {
    fprintf(stderr,"> ssmiice: initialized data with loci mask=%s\n",
	    mask_filename);
  }

  if (use_climatology) {
    climatology = (byte1 **)initialize_matrix
      (grid, sizeof(**climatology),climatology_filename,"climatology",verbose);
    if (NULL == climatology) {
      fprintf(stderr,"ssmiice: bad read on climatology=%s\n",
	      climatology_filename);
      goto cleanup_error_exit;
    } else if (verbose) {
      fprintf(stderr,"> ssmiice: initialized climatology with=%s\n",
	      climatology_filename);
    }
  }
  
  ssmitbs = init_ssmitbs(gpd_filename,
			 ssmi_algorithm_name[seaice_algo],
			 tbs_basename,
			 verbose);
  if (NULL == ssmitbs) {
    fprintf(stderr,"ssmiice: init_ssmitbs failed.\n");
    goto cleanup_error_exit;
  } else if (verbose) {
    fprintf(stderr,"> ssmiice: initialized ssmitbs from %s\n",
	    tbs_basename);
  }

  /*
   * Calculate ice (on requested loci mask values only)
   * according to input algorithm and climatology
   */
  for (row=0; row<grid->rows; row++) {
    for (col=0; col<grid->cols; col++) {
      if (!use_climatology || SEAICE_LIKELY == climatology[row][col]) {
	for (i=0; i<num_use_loci_values; i++) {
	  if (use_loci_value[i] == data[row][col]) {

	    /* 
	     * Determine pixel seaice concentration,
	     * break out of i loop 
	     */
	    ice = (byte1) 
	      transform_element_ssmitbs(ssmitbs,seaice_algo,col,row);
	    data[row][col] = 
	      (NASATEAM_SUSPECT_ICE_VALUE == ice ? NISE_SUSPECT_ICE_VALUE : 
	       (0 < ice ? ice : data[row][col]));
	    break;

	  } /* end if loci mask value should be operated on */
	} /* end loop for each requested loci value */
      } /* end if no climatology masking, or we are masking && SEAICE_LIKELY */
    } /* end loop for each column */
  } /* end loop for each row */

  /*
   * Save ice data to external file
   */
  if (0 == write_matrix (ice_filename,
			 (void **) data,
			 grid->rows,
			 grid->cols,
			 sizeof(**data))) {
    fprintf(stderr,"ssmiice: write_matrix failed for %s\n",
	    ice_filename);
    goto cleanup_error_exit;
  } else if (verbose) {
    fprintf(stderr,"> ssmiice: wrote ice data to %s\n",
	    ice_filename);
  }

  if (verbose) {
    fprintf(stderr,"> ssmiice: exiting successfully\n");
  }

  close_ssmitbs(ssmitbs);
  if (NULL != climatology) free(climatology);
  free(data);
  close_grid(grid);
  exit (EXIT_SUCCESS);

  cleanup_error_exit:
  
  /* Free any allocated memory for matrices here */
  fprintf(stderr,"ssmiice: exiting in error\n");
  if (NULL != grid) close_grid(grid);
  if (NULL != data) free(data);
  if (NULL != climatology) free(climatology);
  if (NULL != ssmitbs) close_ssmitbs(ssmitbs);
  exit (EXIT_FAILURE);

} /* end of main */


