/*======================================================================
 * ssmisnow - program to calculate SSM/I-derived snow
 *
 * Notes:
 *   Refer to ssmitbs.h for legitimate set of ssmi_algorithm_names.
 *   Error Processing:
 *     1) Returns EXIT_FAILURE, an error message to stderr and the
 *        usage message to stderr in any of the following cases:
 *        a) -c is used with a non-existent climatology filename
 *        b) -d is used with an invalid snow algorithm name
 *        c) -u is used with greater than MAX_USE_LOCI_VALUES numerals
 *        d) -w is used with an invalid wet snow algorithm name
 *        e) an invalid option letter is used
 *        f) the number of required inputs is incorrect
 *     2) Returns EXIT_FAILURE, and an error message to stderr
 *        in any of the following cases:
 *        a) init_grid fails for the input .gpd
 *        b) error reading the loci mask file
 *        c) error reading the climatolotdy file
 *        d) init_ssmitbs fails
 *        e) error writing the ice data
 *
 * 08/25/1997 brodzik brodzik@zamboni.colorado.edu 303-492-8263
 * National Snow & Ice Data Center, University of Colorado, Boulder
 *
 *$Log: not supported by cvs2svn $
 * Revision 1.2  1997/10/15  16:53:31  brodzik
 * Changed ALL_CHANNELS ssmitbs initialization to LOWER_CHANNELS.
 * Added memory management cleanup in case of successful completion.
 *
 * Revision 1.1  1997/10/02  20:56:15  brodzik
 * Initial revision
 *
 *
 *======================================================================*/
#include "define.h"
#include "matrix_io.h"
#include "nise_constants.h"
#include "ssmitbs.h"

static const char ssmisnow_c_rcsid[] = "$Header: /tmp_mnt/FILES/pmalgos/ssmisnow.c,v 1.3 1997-11-14 16:20:34 brodzik Exp $";

#define DEFAULT_SNOW_ALGORITHM GOODISON_WALKER
#define DEFAULT_WET_SNOW_ALGORITHM WALKER
#define MAX_USE_LOCI_VALUES 10

#define usage \
"ssmisnow $Revision: 1.3 $ \n"\
"usage: ssmisnow [-v] [-c climatology_filename] \n"\
"       [-d alternate_snow_algo] [-w alternate_wet_snow_algo] \n"\
"       [-u loci_value_to_use]\n"\
"       grid.gpd loci_mask tb_grids_basename snow_filename\n"\
"\n"\
" input : gpd file  - grid parameters definition file\n"\
"         loci_mask - filename containing starting mask with \n"\
"           land/ocean/coastline/permanent_ice classifications\n"\
"           (key to expected classifications is in nise_constants.h)\n"\
"         tb_grids_basename - path and basename of TB grids to use\n"\
"           (assumes channel extensions, e.g. tb_grids_basename.19H, .19V etc)\n"\
"         snow_filename - ssmisnow output file to create\n"\
"\n"\
" output : SSM/I-derived snow grid \n"\
"\n"\
" option : c - climatology file to use (default is no climatology)\n" \
"          d - alternate snow algorithm designator, possible values:\n"\
"               GOODISON_WALKER (default)\n" \
"               NSIDC1\n"\
"          v - verbose mode (can be repeated)\n"\
"          w - alternate wet snow algorithm designator, possible values:\n" \
"               WALKER (default)\n" \
"               MODIFIED_WALKER\n" \
"          u - integer value of loci mask pixels to operate on \n" \
"              (may be repeated up to 10 values)\n" \
"\n"

main (int argc, char *argv[])
{
  char *option;
  int i;
  bool verbose=FALSE;
  bool very_verbose=FALSE;
  char climatology_filename[FILENAME_MAX];
  bool use_climatology;
  char gpd_filename[FILENAME_MAX];
  char mask_filename[FILENAME_MAX];
  char tbs_basename[FILENAME_MAX];
  char snow_filename[FILENAME_MAX];
  ssmi_algorithms dry_snow_algo, wet_snow_algo;
  grid_class *grid;
  ssmitbs_class *ssmitbs;
  byte1 **data;
  byte1 **climatology;
  byte2 use_loci_value[MAX_USE_LOCI_VALUES];
  int num_use_loci_values=0;
  int row, col;
  float swe;
  bool is_wet;
  
  /*
   * Set defaults
   */
  *climatology_filename = 0;
  use_climatology = FALSE;
  dry_snow_algo=DEFAULT_SNOW_ALGORITHM;
  wet_snow_algo=DEFAULT_WET_SNOW_ALGORITHM;
  for (i=0; i<MAX_USE_LOCI_VALUES; i++) {
    use_loci_value[i]=BYTE2_MAX;
  }

  /* 
   *      get command line arguments
   */
  while (--argc > 0 && (*++argv)[0] == '-') {
    for (option = argv[0]+1; *option != '\0'; option++) {
      switch (*option) {
        case 'c':
	  ++argv; --argc;
	  strcpy(climatology_filename,*argv);
	  if (NULL == climatology_filename) error_exit(usage);
	  use_climatology = TRUE;
	  break;
        case 'd':
	  ++argv; --argc;
	  if (!is_dry_snow_algorithm(*argv,&dry_snow_algo)) {
	    fprintf (stderr, "ssmisnow: invalid snow algorithm %s\n\n", *argv);
	    fprintf (stderr,usage);
	    exit(EXIT_FAILURE);
	  }
	  break;
        case 'u':
	  ++argv; --argc;
	  if (num_use_loci_values >= MAX_USE_LOCI_VALUES) {
	    fprintf (stderr, "ssmisnow: exceeded maximum use_loci_values of %d\n\n",
		     MAX_USE_LOCI_VALUES);
	    fprintf (stderr,usage);
	    exit(EXIT_FAILURE);
	  }
	  use_loci_value[num_use_loci_values++] = strtol(*argv,NULL,10);
	  break;
	case 'v':
	  if (verbose) very_verbose = TRUE;
	  verbose = TRUE;
	  break;
        case 'w':
	  ++argv; --argc;
	  if (!is_wet_snow_algorithm(*argv,&wet_snow_algo)) {
	    fprintf (stderr, "ssmisnow: invalid wet snow algorithm %s\n\n", 
		     *argv);
	    fprintf (stderr,usage);
	    exit(EXIT_FAILURE);
	  }
	  break;
	default:
	  fprintf (stderr, "ssmisnow: invalid option %c\n", *option);
	  fprintf (stderr,usage);
	  exit(EXIT_FAILURE);
      } /* end switch */
    } /* end loop for each input command option */
  } /* end loop while still input arguments */

  if (4 != argc) 
  {
    fprintf(stderr,"ssmisnow: Insufficient number of calling arguments.\n\n");
    fprintf (stderr,usage);
    exit(EXIT_FAILURE);
  }

  strcpy (gpd_filename, *argv);
  ++argv; --argc;
  strcpy (mask_filename, *argv);
  ++argv; --argc;
  strcpy (tbs_basename, *argv);
  ++argv; --argc;
  strcpy (snow_filename, *argv);

  if (verbose) 
  {
    fprintf(stderr,"> ssmisnow: Beginning processing for:\n");
    fprintf(stderr,">  gpd                 =%s\n",gpd_filename);
    fprintf(stderr,">  loci mask           =%s\n",mask_filename);
    fprintf(stderr,">  tb_grids_basename   =%s\n",tbs_basename);
    fprintf(stderr,">  snow file           =%s\n",snow_filename);
    fprintf(stderr,">  use_climatology     =%d\n",use_climatology);
    fprintf(stderr,">  -c climatology file =%s\n",climatology_filename);
    fprintf(stderr,">  -d dry snow algorithm=%s\n",
	    ssmi_algorithm_name[dry_snow_algo]);
    fprintf(stderr,">  -w wet snow algorithm=%s\n",
	    ssmi_algorithm_name[wet_snow_algo]);
    fprintf(stderr,">  -u operate on loci mask values of\n");
    for (i=0; i<num_use_loci_values; i++) {
      fprintf(stderr,">  %d\n",use_loci_value[i]);
    }
  }

  /*
   * Initialize grid (used for array dimensions throughout)
   */
  grid = init_grid(gpd_filename);
  if (NULL == grid) {
    fprintf(stderr,"ssmisnow: init_grid failed for gpd=%s\n",gpd_filename);
    goto cleanup_error_exit;
  } else if (verbose) {
    fprintf(stderr,"> ssmisnow: initialized grid for gpd=%s\n",gpd_filename);
  }
    
  /*
   * Load input data 
   *  Initialize output data with land/ocean/mixed/corners mask
   *  climatology (if using one)
   *  and initialize tbs for LOWER_CHANNELS.
   */
  data = (byte1 **)initialize_matrix
    (grid, sizeof(**data), mask_filename,
     "loci (land/ocean/coast/ice) mask", verbose);
  if (NULL == data) {
    fprintf(stderr,"ssmisnow: bad read on loci mask=%s\n",mask_filename);
    goto cleanup_error_exit;
  } else if (verbose) {
    fprintf(stderr,"> ssmisnow: initialized data with loci mask=%s\n",
	    mask_filename);
  }

  if (use_climatology) {
    climatology = (byte1 **)initialize_matrix
      (grid, sizeof(**climatology),climatology_filename,"climatology",verbose);
    if (NULL == climatology) {
      fprintf(stderr,"ssmisnow: bad read on climatology=%s\n",
	      climatology_filename);
      goto cleanup_error_exit;
    } else if (verbose) {
      fprintf(stderr,"> ssmisnow: initialized climatology with=%s\n",
	      climatology_filename);
    }
  }
  
  ssmitbs = init_ssmitbs(gpd_filename,
			 ssmi_algorithm_name[LOWER_CHANNELS],
			 tbs_basename,
			 verbose);
  if (NULL == ssmitbs) {
    fprintf(stderr,"ssmisnow: init_ssmitbs failed.\n");
    goto cleanup_error_exit;
  } else if (verbose) {
    fprintf(stderr,"> ssmisnow: initialized ssmitbs from %s\n",
	    tbs_basename);
  }

  /*
   * Calculate snow (on requested loci mask values only)
   * according to input algorithms and climatology
   */
  for (row=0; row<grid->rows; row++) {
    for (col=0; col<grid->cols; col++) {
      if (!use_climatology || SNOW_LIKELY == climatology[row][col]) {
	for (i=0; i<num_use_loci_values; i++) {
	  if (use_loci_value[i] == data[row][col]) {

	    /* 
	     * Call dry snow, call wet snow, determine pixel
	     * classification, break out of i loop 
	     */
	    swe = transform_element_ssmitbs(ssmitbs,dry_snow_algo,col,row);
	    if (very_verbose && 0 < swe) 
	      fprintf (stderr,"ssmisnow: row=%d col=%d, swe=%f\n",row,col,swe);
	    is_wet = (bool) 
	      transform_element_ssmitbs(ssmitbs,wet_snow_algo,col,row);
	    data[row][col] = 
	      (0 < swe ? NISE_DRY_SNOW_VALUE : 
	       (is_wet ? NISE_WET_SNOW_VALUE : data[row][col]));
	    break;

	  } /* end if loci mask value should be operated on */
	} /* end loop for each requested loci value */
      } /* end if no climatology masking, or we are masking, && SNOW_LIKELY */
    } /* end loop for each column */
  } /* end loop for each row */

  /*
   * Save snow data to external file
   */
  if (0 == write_matrix (snow_filename,
			 (void **) data,
			 grid->rows,
			 grid->cols,
			 sizeof(**data))) {
    fprintf(stderr,"ssmisnow: write_matrix failed for %s\n",
	    snow_filename);
    goto cleanup_error_exit;
  } else if (verbose) {
    fprintf(stderr,"> ssmisnow: wrote snow data to %s\n",
	    snow_filename);
  }

  if (verbose) {
    fprintf(stderr,"> ssmisnow: exiting successfully\n");
  }

  close_ssmitbs(ssmitbs);
  if (NULL != climatology) free(climatology);
  free(data);
  close_grid(grid);
  exit (EXIT_SUCCESS);

  cleanup_error_exit:
  
  /* Clean up and exit_failure */
  fprintf(stderr,"ssmisnow: exiting in error\n");
  if (NULL != grid) close_grid(grid);
  if (NULL != data) free(data);
  if (NULL != climatology) free(climatology);
  if (NULL != ssmitbs) close_ssmitbs(ssmitbs);
  exit (EXIT_FAILURE);

} /* end of main */


