/*======================================================================
 * surface - storage and retrieval for ancillary surface information needed
 *           to run Chang surface classification & SWE algorithms,
 *           and forest-corrected algorithms
 *
 * Notes: The default pmalgorithm paradigm is that no geographic or
 *        temporal information is required.
 *        However, some algorithms do require addition information, so
 *        this object is a way to store and retrieve that additional
 *        information.
 *        Examples include:
 *        - The surface classification algorithms require
 *        an estimate of forest cover, and Chang's landmask.
 *        - The snow depth and SWE algorithms require
 *        SSM/I platform id,
 *        an estimate of forest cover, Chang's landmask, 
 *        surface classification, and the month of acquisition.
 *        - NSIDC_A4_SWE, NSIDC_S4_SWE and NSIDC_S6_SWE require a 
 *        fractional forest cover.
 * 
 * 01/19/2001 brodzik brodzik@zamboni.colorado.edu 303-492-8263
 * National Snow & Ice Data Center, University of Colorado, Boulder
 *
 *======================================================================*/
#include "define.h"
#include "matrix.h"
#include "matrix_io.h"
#include "surface.h"
#include "pmalgorithm.h"

static const char surface_c_rcsid[] = "$Header: /tmp_mnt/FILES/pmalgos/surface.c,v 1.19.4.1 2008-06-10 20:36:35 brodzik Exp $";

#define NUM_SNOW_CLASS	8
#define MIN_MONTH 1
#define MAX_MONTH 12
#define NUM_MONTHS 12

/* 
 * The Chang landmask is a bit mask such that:
 * Bit	Used for
 * 7	Unused
 * 6	Permanent Ice
 * 5	Land surface: 0=smooth, 1=rough
 * 4    Land: 0=snow possible, 1=snow impossible
 * 3    0=water, 1=land
 * 2-0  Snow classification (I think these are from M. Sturm?)
 */
#define IS_ICE(x) ((x) & 0x40)
#define IS_ROUGH(x) ((x) & 0x20)
#define IS_SNOW_IMPOSSIBLE(x) ((x) & 0x10)
#define IS_LAND(x) ((x) & 0x08)
#define SNOW_CLASS(x) ((x) & 0x07)

/*
 * Coefficient for Chang snow depth calculation, function of 
 * time of year and prevalent snow type at a given pixel.
 * This acts as a correction for grain size??
 */
static float SNOWconst[NUM_SNOW_CLASS][NUM_MONTHS] = {
/*Jan  Feb  Mar  Apr  May  Jun  Jul  Aug  Sep  Oct  Nov  Dec    */
    0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,   /* water    0*/
    3.6, 3.2, 2.8, 2.4, 2.4, 2.4, 2.4, 4.8, 4.8, 6.4, 5.6, 4.0,   /* tundra   1*/
    4.0, 3.6, 3.2, 2.8, 2.4, 2.4, 2.4, 4.8, 4.8, 6.4, 5.6, 4.8,   /* taiga    2*/
    0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,   /* ice      3*/
    4.8, 4.8, 4.8, 4.8, 4.8, 4.8, 4.8, 4.8, 4.8, 4.8, 4.8, 4.8,   /* maritime 4*/
    4.8, 4.8, 4.8, 4.8, 4.8, 4.8, 4.8, 4.8, 4.8, 4.8, 4.8, 4.8,   /* ephemeral5*/
    4.8, 4.8, 4.8, 4.8, 4.8, 4.8, 4.8, 4.8, 4.8, 4.8, 4.8, 4.8,   /* prairie  6*/
    4.8, 4.8, 4.8, 4.8, 4.8, 4.8, 4.8, 4.8, 4.8, 6.4, 6.0, 5.6 }; /* alpine   7*/

/* Default values */
#define MAX_FOREST_COVER (0.5)

/*----------------------------------------------------------------------
 *  init_sfc - initialize ancillary surface information
 *
 *    input: algorithm_name
 *      Only required for the SSM/I SWE algos:
 *           platform - SSM/I sensor platform id 
 *      Only required for CHANG_DEPTH and NSIDC_S3_SWE algos:
 *           month - month of data acquisition
 *           surface_filename - gridded surface classification
 *      Required for all but NSIDC_M/S6_SWE and NSIDC_A/S4_SWE::
 *           landmask_filename - landmask with bitwise settings for
 *                               land/water, roughness, snow climatology,
 *                               and snow category
 *      Required for all algorithms:
 *           forest_cover_filename - forest_cover
 *           grid - grid object
 *           verbose - verbosity flag
 *
 *    returns: pointer to new sfc_class instance
 *             or NULL (and reason to stderr) in case of error 
 *
 *----------------------------------------------------------------------*/
sfc_class *init_sfc(const char *algorithm_name,
                    pmalgo_ssmi_platform platform,
		    int month,		          /* CHANG_DEPTH/NSIDC_S3_SWE only*/
		    const char *surface_filename, /* CHANG_DEPTH/NSIDC_S3_SWE only*/
		    const char *forest_cover_filename, /* all but NSIDC_S12_SWE */
		    const char *landmask_filename, /* all but NSIDC_M/S6_SWE */
                                                   /* and NSIDC_A/S4_SWE */
		    const char *elevation_filename, /* NSIDC_S11_SWE and S12 only */
		    grid_class *grid,
		    byte1 high_elevation_model_id, /* NSIDC_S11_SWE and S12 only */
		    int2 threshold_elevation_meters, /* NSIDC_S11_SWE and S12 only */
		    int verbose) {

    sfc_class *this;

    if (0 < verbose) {
        fprintf(stderr,"> %s\n",surface_h_rcsid);
        fprintf(stderr,"> %s\n",surface_c_rcsid);
    }

    /* Allocate storage for object */
    this = (sfc_class *)calloc(1,sizeof(sfc_class));
    if (NULL == this) {
        perror("init_sfc: sfc_class object");
        return NULL;
    }

    if (0 == strcmp("CHANG_SURFACE",algorithm_name)) {
        this->id = PMALGO_CHANG_SURFACE;
    } else if (0 == strcmp("NSIDC_SURFACE",algorithm_name)) {
        this->id = PMALGO_NSIDC_SURFACE;
    } else if (0 == strcmp("CHANG_DEPTH",algorithm_name)) {
        this->id = PMALGO_CHANG_DEPTH;
    } else if (0 == strcmp("NSIDC_A4_SWE",algorithm_name)) {
        this->id = PMALGO_NSIDC_A4_SWE;
    } else if (0 == strcmp("NSIDC_A10_SWE",algorithm_name)) {
        this->id = PMALGO_NSIDC_A10_SWE;
    } else if (0 == strcmp("NSIDC_M4_SWE",algorithm_name)) {
        this->id = PMALGO_NSIDC_M4_SWE;
    } else if (0 == strcmp("NSIDC_S1_SWE",algorithm_name)) {
        this->id = PMALGO_NSIDC_S1_SWE;
    } else if (0 == strcmp("NSIDC_S3_SWE",algorithm_name)) {
        this->id = PMALGO_NSIDC_S3_SWE;
    } else if (0 == strcmp("NSIDC_S4_SWE",algorithm_name)) {
        this->id = PMALGO_NSIDC_S4_SWE;
    } else if (0 == strcmp("NSIDC_S5_SWE",algorithm_name)) {
        this->id = PMALGO_NSIDC_S5_SWE;
    } else if (0 == strcmp("NSIDC_S6_SWE",algorithm_name)) {
        this->id = PMALGO_NSIDC_S6_SWE;
    } else if (0 == strcmp("NSIDC_S10_SWE",algorithm_name)) {
        this->id = PMALGO_NSIDC_S10_SWE;
    } else if (0 == strcmp("NSIDC_S11_SWE",algorithm_name)) {
        this->id = PMALGO_NSIDC_S11_SWE;
    } else if (0 == strcmp("NSIDC_S12_SWE",algorithm_name)) {
        this->id = PMALGO_NSIDC_S12_SWE;
    } else {
        assert(0);
    }

    if (PMALGO_CHANG_SURFACE == this->id) {

        /* 
         * Set defaults for now, but I think eventually should
         * be optional input arguments
         */
        this->forest_cover_threshold = MAX_FOREST_COVER;

    } else if (PMALGO_NSIDC_SURFACE == this->id ||
               PMALGO_NSIDC_A4_SWE == this->id ||
               PMALGO_NSIDC_A10_SWE == this->id ||
               PMALGO_NSIDC_M4_SWE == this->id ||
               PMALGO_NSIDC_S4_SWE == this->id ||
               PMALGO_NSIDC_S6_SWE == this->id ||
               PMALGO_NSIDC_S10_SWE == this->id ||
               PMALGO_NSIDC_S11_SWE == this->id) {

        /* 
         * Set defaults for now, but I think eventually should
         * be optional input arguments
         */
        this->forest_cover_threshold = 0.0;

    } else if (PMALGO_CHANG_DEPTH == this->id ||
               PMALGO_NSIDC_S3_SWE == this->id) {

        if (MIN_MONTH <= month && month <= MAX_MONTH) {
            this->month = month;
        } else {
            fprintf(stderr,"init_sfc: input month =%d out of range\n",month);
            close_sfc(this);
            return NULL;
        }
    
        this->surface = 
            (byte1 **)initialize_matrix(grid,sizeof(**this->surface),
                                        surface_filename,"surface input",
                                        verbose);
        if (NULL == this->surface) {
            close_sfc(this);
            return NULL;
        }

    }

    if (PMALGO_CHANG_SURFACE == this->id || 
        PMALGO_NSIDC_SURFACE == this->id ||
        PMALGO_CHANG_DEPTH == this->id || 
        PMALGO_NSIDC_A4_SWE == this->id || 
        PMALGO_NSIDC_A10_SWE == this->id || 
        PMALGO_NSIDC_M4_SWE == this->id || 
        PMALGO_NSIDC_S3_SWE == this->id || 
        PMALGO_NSIDC_S4_SWE == this->id || 
        PMALGO_NSIDC_S6_SWE == this->id ||
        PMALGO_NSIDC_S10_SWE == this->id ||
        PMALGO_NSIDC_S11_SWE == this->id) {

        this->forest_cover = 
            (float **)initialize_matrix(grid,sizeof(**this->forest_cover),
                                        forest_cover_filename,"forest_cover",
                                        verbose);
        if (NULL == this->forest_cover) {
            close_sfc(this);
            return NULL;
        }
    }

    if (PMALGO_CHANG_SURFACE == this->id || 
        PMALGO_NSIDC_SURFACE == this->id ||
        PMALGO_CHANG_DEPTH == this->id || 
        PMALGO_NSIDC_S3_SWE == this->id) {

        this->landmask = 
            (byte1 **)initialize_matrix(grid,sizeof(**this->landmask),
                                        landmask_filename,"Chang landmask",
                                        verbose);
        if (NULL == this->landmask) {
            close_sfc(this);
            return NULL;
        }
    }

    if (PMALGO_NSIDC_S11_SWE == this->id ||
        PMALGO_NSIDC_S12_SWE == this->id) {

        this->elevation = 
            (int2 **)initialize_matrix(grid,sizeof(**this->elevation),
                                       elevation_filename,"Elevation",
                                       verbose);
        if (NULL == this->elevation) {
            close_sfc(this);
            return NULL;
        }
    }

    if (PMALGO_NSIDC_S11_SWE == this->id ||
        PMALGO_NSIDC_S12_SWE == this->id) {

        if ( high_elevation_model_id < (byte1)PMALGO_NUM_HIGH_ELEVATION_MODELS ) {
            this->high_elevation_model_id =
                (pmalgo_high_elevation_model_id)high_elevation_model_id;
        } else {
            fprintf(stderr,"init_sfc: input high_elevation_model_id=%d out of range\n",
                    high_elevation_model_id);
            close_sfc(this);
            return NULL;
        }
      
        this->threshold_elevation_meters = threshold_elevation_meters;

    }

    if (PMALGO_NSIDC_S1_SWE == this->id ||
        PMALGO_NSIDC_S3_SWE == this->id ||
        PMALGO_NSIDC_S4_SWE == this->id ||
        PMALGO_NSIDC_S5_SWE == this->id ||
        PMALGO_NSIDC_S6_SWE == this->id ||
        PMALGO_NSIDC_S10_SWE == this->id ||
        PMALGO_NSIDC_S11_SWE == this->id ||
        PMALGO_NSIDC_S12_SWE == this->id) {

        this->platform = platform;

        if ( PMALGO_NSIDC_S3_SWE == this->id ||
             PMALGO_NSIDC_S5_SWE == this->id ||
             PMALGO_NSIDC_S6_SWE == this->id ) {

            if ( PMALGO_F17 == this->platform ) {
        
                fprintf( stderr, "init_sfc: WARNING: platform option set to F17, but" );
                fprintf( stderr, " we have not defined F17->F13 regressions for all required channels.\n");
                fprintf( stderr, "Algorithm %s will be run assuming F13 TBs.\n",
                         pmalgo_id_name[this->id] );

            }

        }
        
    }

    this->grid = grid;

    return this;

}


/*----------------------------------------------------------------------
 *  close_sfc - free storage associated with this object
 * 
 *    input: this - sfc_class object
 *
 *    result: deallocate all memory associated with this object
 *
 *----------------------------------------------------------------------*/
void close_sfc(sfc_class *this) {

    if (NULL == this) return;
    if (NULL != this->surface) free(this->surface);
    if (NULL != this->forest_cover) free(this->forest_cover);
    if (NULL != this->landmask) free(this->landmask);
    if (NULL != this->elevation) free(this->elevation);
    free(this);
}


/*----------------------------------------------------------------------
 *  is_mountain_sfc - boolean for mountainous surface at input pixel
 * 
 *    input: this - sfc_class object
 *           row - row number
 *           col - column number
 *
 *    result: TRUE iff this pixel is "rough"
 *
 *----------------------------------------------------------------------*/
bool is_mountain_sfc(sfc_class *this, int col, int row) {

    assert(0 <= row && row < this->grid->rows);
    assert(0 <= col && col < this->grid->cols);
    assert(NULL != this);
    assert(PMALGO_CHANG_SURFACE == this->id || 
           PMALGO_NSIDC_SURFACE == this->id ||
           PMALGO_CHANG_DEPTH == this->id || 
           PMALGO_NSIDC_S3_SWE == this->id);

    return (IS_ROUGH(this->landmask[row][col]) ? TRUE : FALSE);

}


/*----------------------------------------------------------------------
 *  forest_cover_sfc - return the forest_cover at input pixel 
 * 
 *    input: this - sfc_class object
 *           row - row number
 *           col - column number
 *
 *    result: surface condition
 *
 *----------------------------------------------------------------------*/
float forest_cover_sfc(sfc_class *this, int col, int row) {

    assert(0 <= row && row < this->grid->rows);
    assert(0 <= col && col < this->grid->cols);
    assert(NULL != this);
    assert(PMALGO_CHANG_DEPTH == this->id || 
           PMALGO_NSIDC_A4_SWE == this->id ||
           PMALGO_NSIDC_A10_SWE == this->id ||
           PMALGO_NSIDC_M4_SWE == this->id ||
           PMALGO_NSIDC_S3_SWE == this->id ||
           PMALGO_NSIDC_S4_SWE == this->id ||
           PMALGO_NSIDC_S6_SWE == this->id ||
           PMALGO_NSIDC_S10_SWE == this->id ||
           PMALGO_NSIDC_S11_SWE == this->id);

    return (this->forest_cover[row][col]);

}


/*----------------------------------------------------------------------
 *  elevation_sfc - return the elevation at input pixel 
 * 
 *    input: this - sfc_class object
 *           row - row number
 *           col - column number
 *
 *    result: elevation (meters)
 *
 *----------------------------------------------------------------------*/
int2 elevation_sfc(sfc_class *this, int col, int row) {

    assert(0 <= row && row < this->grid->rows);
    assert(0 <= col && col < this->grid->cols);
    assert(NULL != this);
    assert(PMALGO_NSIDC_S11_SWE == this->id ||
           PMALGO_NSIDC_S12_SWE == this->id);

    return (this->elevation[row][col]);

}


/*----------------------------------------------------------------------
 *  determine_chang_sfc - Only for CHANG_SURFACE:
 *              return the surface condition at input pixel 
 *              location, according to Chang's decision tree
 * 
 *    input: this - sfc_class object
 *           row - row number
 *           col - column number
 *
 *    result: surface condition that can be determined from forest_cover 
 *            and current land mask information
 *
 *----------------------------------------------------------------------*/
sfc_type determine_chang_sfc(sfc_class *this, int col, int row) {

    assert(0 <= row && row < this->grid->rows);
    assert(0 <= col && col < this->grid->cols);
    assert(NULL != this);
    assert(PMALGO_CHANG_SURFACE == this->id);

    if (IS_ICE(this->landmask[row][col])) {
        return SFC_ICE;
    } else if (!IS_LAND(this->landmask[row][col])) {
        return SFC_WATER;
    } else if (IS_ROUGH(this->landmask[row][col])) { 
        return SFC_MOUNTAIN;
    } else if (this->forest_cover_threshold < this->forest_cover[row][col]) {
        return SFC_FOREST;
    } else if (IS_SNOW_IMPOSSIBLE(this->landmask[row][col])) { 
        return SFC_SNOW_IMPOSSIBLE;
    } else { 
        return SFC_NORMAL_NOSNOW;
    }

}


/*----------------------------------------------------------------------
 *  determine_nsidc_sfc - Only for NSIDC_SURFACE:
 *              return the surface condition at input pixel 
 *              location, according to NSIDC's modified version of 
 *              Chang's decision tree
 * 
 *    input: this - sfc_class object
 *           row - row number
 *           col - column number
 *
 *    result: surface condition that can be determined from forest_cover
 *            and current land mask information
 *
 *----------------------------------------------------------------------*/
sfc_type determine_nsidc_sfc(sfc_class *this, int col, int row) {

    assert(0 <= row && row < this->grid->rows);
    assert(0 <= col && col < this->grid->cols);
    assert(NULL != this);
    assert(PMALGO_NSIDC_SURFACE == this->id || 
           PMALGO_CHANG_SURFACE == this->id);

    if (IS_ICE(this->landmask[row][col])) {
        return SFC_ICE;
    } else if (!IS_LAND(this->landmask[row][col])) {
        return SFC_WATER;
    } else if (IS_SNOW_IMPOSSIBLE(this->landmask[row][col])) {
        return SFC_SNOW_IMPOSSIBLE;
    } else if (IS_ROUGH(this->landmask[row][col])) {
        return SFC_MOUNTAIN;
    } else if (this->forest_cover_threshold < this->forest_cover[row][col]) {
        return SFC_FOREST;
    } else {
        return SFC_NORMAL_NOSNOW;
    }

}

/*----------------------------------------------------------------------
 *  retrieve_sfc - Only for CHANG_DEPTH or NSIDC_S3_SWE:
 *                 return the surface value at input pixel 
 * 
 *    input: this - sfc_class object
 *           row - row number
 *           col - column number
 *
 *    result: surface condition
 *
 *----------------------------------------------------------------------*/
sfc_type retrieve_sfc(sfc_class *this, int col, int row) {

    assert(0 <= row && row < this->grid->rows);
    assert(0 <= col && col < this->grid->cols);
    assert(NULL != this);
    assert(PMALGO_CHANG_DEPTH == this->id || PMALGO_NSIDC_S3_SWE == this->id);

    return (sfc_type)(this->surface[row][col]);

}


/*----------------------------------------------------------------------
 *  swe_coef_sfc - return the SWE coefficient at input pixel, for the 
 *               month and type of snow expected
 * 
 *    input: this - sfc_class object
 *           row - row number
 *           col - column number
 *
 *    result: surface condition
 *
 *----------------------------------------------------------------------*/
float swe_coef_sfc(sfc_class *this, int col, int row) {

    float value;

    assert(0 <= row && row < this->grid->rows);
    assert(0 <= col && col < this->grid->cols);
    assert(NULL != this);
    assert(PMALGO_CHANG_DEPTH == this->id ||
           PMALGO_NSIDC_S3_SWE == this->id);

    value = SNOWconst[SNOW_CLASS(this->landmask[row][col])][(this->month)-1];
    return (value);

}

/*----------------------------------------------------------------------
 *  ssmi_platform_sfc - return the stored ssmi_platform id
 * 
 *    input: this - sfc_class object
 *
 *    result: ssmi platform id
 *
 *----------------------------------------------------------------------*/
pmalgo_ssmi_platform ssmi_platform_sfc(sfc_class *this) {

    assert(NULL != this);
    assert(PMALGO_NSIDC_S1_SWE == this->id ||
           PMALGO_NSIDC_S3_SWE == this->id ||
           PMALGO_NSIDC_S4_SWE == this->id ||
           PMALGO_NSIDC_S5_SWE == this->id ||
           PMALGO_NSIDC_S6_SWE == this->id ||
           PMALGO_NSIDC_S10_SWE == this->id ||
           PMALGO_NSIDC_S11_SWE == this->id ||
           PMALGO_NSIDC_S12_SWE == this->id);

    return (this->platform);

}

/*----------------------------------------------------------------------
 *  high_elevation_model_sfc - For NSIDC_S11_SWE and NSIDC_S12_SWE:
 *                 return the requested model ID
 * 
 *    input: this - sfc_class object
 *
 *    result: high elevation model id
 *
 *----------------------------------------------------------------------*/
pmalgo_high_elevation_model_id high_elevation_model_sfc(sfc_class *this) {

    assert(NULL != this);
    assert(PMALGO_NSIDC_S11_SWE == this->id ||
           PMALGO_NSIDC_S12_SWE == this->id);

    return (this->high_elevation_model_id);

}

/*----------------------------------------------------------------------
 *  threshold_elevation_sfc - For NSIDC_S11_SWE NSIDC_S12_SWE:
 *                 return the requested threshold elevation (model will
 *                 be applied to pixels above this elevation)
 * 
 *    input: this - sfc_class object
 *
 *    result: threshold elevation
 *
 *----------------------------------------------------------------------*/
int2 threshold_elevation_sfc(sfc_class *this) {

    assert(NULL != this);
    assert(PMALGO_NSIDC_S11_SWE == this->id ||
           PMALGO_NSIDC_S12_SWE == this->id);

    return (this->threshold_elevation_meters);

}

