/*======================================================================
 * surface - storage and retrieval for ancillary surface information needed
 *           to run the Chang surface classification & SWE algorithms,
 *           and algorithms requiring forest cover and/or elevation
 *
 * 01/19/2001 brodzik brodzik@zamboni.colorado.edu 303-492-8263
 * National Snow & Ice Data Center, University of Colorado, Boulder
 *
 *======================================================================*/
#ifndef sfc_h_
#define sfc_h_

#include "define.h"
#include "pmalgorithm.h"

static const char surface_h_rcsid[] = "$Header: /tmp_mnt/FILES/pmalgos/surface.h,v 1.5.12.1 2008-06-10 20:36:35 brodzik Exp $";

/* Chang surface classifications */
typedef enum {
    /* No snow categories: */
    SFC_NORMAL_NOSNOW, SFC_WATER, SFC_SNOW_IMPOSSIBLE, SFC_ICE, SFC_TOO_WARM,
    /* Unknown snow categories */
    SFC_FOREST, SFC_MOUNTAIN, SFC_RAINING, SFC_WET_SNOW,
    /* Categories for which SWE can be calculated */
    SFC_DRY_SNOW, SFC_WET_SOIL, SFC_DRY_SOIL,
    /* TBs are out of range, this is a no snow category */
    SFC_OUTOFRANGE,
    SFC_NUM_SURFACE_CONDITIONS
} sfc_type;

#define SFC_ELEVATION_MISSING -500

/*
 * Structure for ancillary algorithm information
 * A given algorithm may use any combination of these, but
 * need not use all of them.
 */
typedef struct {
    float forest_cover_threshold;  /* max allowed forest for snow calculation */
    int month;                     /* month of input data */
    pmalgo_high_elevation_model_id high_elevation_model_id;
                                   /* high elevation model id */
    int2 threshold_elevation_meters; /* threshold (meters) for high elevation model*/
    byte1 **surface;               /* surface classifications*/
    int2 **elevation;              /* mean elevation (m) */
    byte1 **landmask;              /* bitwise landmask */
    pmalgo_id id;                  /* algorithm id */
    pmalgo_ssmi_platform platform; /* SSM/I sensor platform id */
    float **forest_cover;          /* pixel forest cover */
    grid_class *grid;              /* grid parameters */
} sfc_class;

/* Function prototypes */
sfc_class *init_sfc(const char *algorithm_name,
                    pmalgo_ssmi_platform platform,
		    int month,		         
		    const char *surface_filename,
		    const char *forest_cover_filename,
		    const char *landmask_filename,
		    const char *elevation_filename,
		    grid_class *grid,
		    byte1 high_elevation_model_id,
		    int2 threshold_elevation_meters,		         
		    int verbose);
void close_sfc(sfc_class *this);

bool is_mountain_sfc(sfc_class *this, int col, int row);
float forest_cover_sfc(sfc_class *this, int col, int row);
int2 elevation_sfc(sfc_class *this, int col, int row);
sfc_type determine_chang_sfc(sfc_class *this, int col, int row);
sfc_type determine_nsidc_sfc(sfc_class *this, int col, int row);
sfc_type retrieve_sfc(sfc_class *this, int col, int row);
float swe_coef_sfc(sfc_class *this, int col, int row);
pmalgo_ssmi_platform ssmi_platform_sfc(sfc_class *this);
pmalgo_high_elevation_model_id high_elevation_model_sfc(sfc_class *this);
int2 threshold_elevation_sfc(sfc_class *this);

#endif
