#!/usr/sbin/perl -w
#
# This perl script will test the C program pmalgos
#
# 17-Feb-2000 M. J. Brodzik brodzik@nsidc.org 303-492-8263
# National Snow & Ice Data Center, University of Colorado, Boulder
# Copyright (C) 2000-2004 University of Colorado
#-----------------------------------------------------------------------
use Env;
use lib ("${PERL_UTIL_PATH}");
use NSIDC_Unit_Test_Util;
use Getopt::Std;
use File::Basename;
use vars qw( %args );

# force autoflush after every write or print
$| = 1;
my $usage = '

test_pmalgos.pl [-a] [-t test_binary] [casenum]

Input:
    casenum - case number to run

Options:
    a - do all cases in sequence
    t test_binary - pmalgos binary to test, default is
      ${HOME}/src/pmalgos/pmalgos

';

undef %args;
if ( !getopts( 'at:', \%args ) ) {
	printf STDERR $usage;
	exit;
}

if ( !$args{a} ) {
	$args{a} = 0;
} else {
	$testcase = 1;
}

if ( !$args{t} ) {
	$args{t} = "./pmalgos";
}

if ( -1 != $#ARGV ) {
	$testcase = shift;
}

$topdir        = "/share/data/pmalgos_test_data";
$clims_dir     = "/share/data/pmalgos/";
$ancillary_dir = "${topdir}/ancillary";
$pmalgos       = $args{t};
$comparedir    = "${topdir}/compare";
$tbdir         = "${topdir}/tbs";


# Assume that test_pmalgos_nasateam.pl and test_pmalgos_bootstra.pl
# are in same directory as calling script
my $scriptname;
my $testdir;
( $scriptname, $testdir ) = fileparse($0);
$test_pmalgos_nasateam  = "${testdir}test_pmalgos_nasateam.pl -t $args{t}";
$test_pmalgos_bootstrap = "${testdir}test_pmalgos_bootstrap.pl -t $args{t}";

print "Beginning of $scriptname:\n";
print "Testing pmalgos binary: $pmalgos\n";
print "with main script      : $0\n";
print "and nasateam script   : $test_pmalgos_nasateam\n";

$elev_file = "${clims_dir}/elevation/Nlelev.pc";

# Note:  Add ancillary to environment setting for PATHMPP
# for {loci_no_path} values to work
$ENV{"PATHMPP"} = "${ancillary_dir}:" . $ENV{"PATHMPP"};
@hem_info = (
			  {
				 gpd          => "Nl.gpd",
				 loci         => "${ancillary_dir}/Nl.loci",
				 loci_no_path => "Nl.loci",
				 tb_base      => "${tbdir}/Nl911227-db",
			  },
			  {
				 gpd          => "Sl.gpd",
				 loci         => "${ancillary_dir}/Sl.loci",
				 loci_no_path => "Sl.loci",
				 tb_base      => "${tbdir}/Sl911227-db",
			  },
			  {
				 gpd          => "Nl.gpd",
				 loci         => "${ancillary_dir}/Nl.loci",
				 loci_no_path => "Nl.loci",
				 tb_base      => "${tbdir}/EASE-SMMR-NL1979002D",
			  },
			  {
				 gpd          => "Nl.gpd",
				 loci         => "${ancillary_dir}/NLGRNLND.mask",
				 loci_no_path => "Nl.loci",
				 tb_base      => "${tbdir}/Nl911227D-nn",
			  },
			  {
				 gpd          => "Nl.gpd",
				 loci         => "",
				 loci_no_path => "",
				 tb_base      => "${tbdir}/EASE-F13-NL1998365D",
			  },
			  {
				 gpd          => "Nl.gpd",
				 loci         => "",
				 loci_no_path => "",
				 tb_base      => "${tbdir}/EASE-F13-NL1999031D",
				 forest       => "${ancillary_dir}/Nligbp01.dat.max50.float",
			  },
			  {
				 gpd          => "Nl.gpd",
				 loci         => "${ancillary_dir}/Nl.loci",
				 loci_no_path => "Nl.loci",
				 tb_base      => "${tbdir}/ID2-AMSRE-X1-NL2003031D",
				 forest       => "${clims_dir}/landcover/lsb/Nl.gl_g.sds01.v4.forest.dat.max50.float",
			  },
			  {
				 gpd          => "Nl.gpd",
				 loci         => "",
				 loci_no_path => "",
				 tb_base      => "${tbdir}/EASE-SMMR-NL1979002D",
				 forest       => "${ancillary_dir}/Nligbp01.dat.max50.float",
			  },
			  {
				 gpd          => "Nl.gpd",
				 loci         => "",
				 loci_no_path => "",
				 tb_base      => "${tbdir}/EASE-F13-NL1998365D",
				 forest       => "${clims_dir}/landcover/lsb/Nl.gl_g.sds01.v4.forest.dat.max50.float",
			  },
);

$loci_land_value = 0;

my $description;

if ( $args{a} || 1 == $testcase ) {

	$description = "\nCase 1: no arguments\n" . "Expected result: error message, usage message, and EXIT_FAILURE\n";
	`$pmalgos`;
	check_for_failure( $description, $? );
	if ( $args{a} ) {
		$testcase++;
	}

}
if ( $args{a} || 2 == $testcase ) {

	$description = "\nCase 2: invalid algorithm\n" . "Expected result: error message, EXIT_FAILURE\n";
	`$pmalgos -v -l $hem_info[0]->{loci} $hem_info[0]->{gpd} NASATEAM $hem_info[0]->{tb_base}`;
	check_for_failure( $description, $? );
	if ( $args{a} ) {
		$testcase++;
	}

}
if ( $args{a} || 3 == $testcase ) {

	$description = "\nCase 3: too many use loci values\n" . "Expected result: error message, usage message, and EXIT_FAILURE\n";
	`$pmalgos -v -u 0 -u 1 -u 2 -u 3 -u 4 -u 5 -u 6 -u 7 -u 8 -u 9 -u 10 -l $hem_info[0]->{loci} $hem_info[0]->{gpd} NASATEAM_NORTH $hem_info[0]->{tb_base}`;
	check_for_failure( $description, $? );
	if ( $args{a} ) {
		$testcase++;
	}

}
if ( $args{a} || 4 == $testcase ) {

	$description = "\nCase 4: invalid gpd\n" . "Expected result: error message and EXIT_FAILURE\n";
	`$pmalgos unknown.gpd NASATEAM_NORTH $hem_info[0]->{tb_base}`;
	check_for_failure( $description, $? );
	if ( $args{a} ) {
		$testcase++;
	}

}
if ( $args{a} || ( 5 <= $testcase && $testcase <= 7 ) ) {

	print "CASES 5-7 have been replaced by test_pmalgos_nasateam...\n";
	@cmd_out = `$test_pmalgos_nasateam 2>&1`;
	if ( 0 != $? ) {
		print "@cmd_out";
		die "\nERROR: test_pmalgos_nasateam return was not EXIT_SUCCESS, prog_ret = " . ( $? >> 8 ) . ", wait_ret = " . ( $? & 255 ) . "\n";
	}
	print "@cmd_out";

	print "...and test_pmalgos_bootstrap...\n";
	@cmd_out = `$test_pmalgos_bootstrap 2>&1`;
	if ( 0 != $? ) {
		print "@cmd_out";
		die "\nERROR: test_pmalgos_bootstrap return was not EXIT_SUCCESS, prog_ret = " . ( $? >> 8 ) . ", wait_ret = " . ( $? & 255 ) . "\n";
	}
	print "@cmd_out";

	print "CASE 5-7 SUCCESS\n";
	if ( $args{a} ) {
		$testcase = 8;
	}

}
if ( $args{a} || 8 == $testcase ) {

	$description
		= "\nCase 8: legitimate run, NSIDC_M1_SWE, (formerly CHANG_SWE)\n"
		. "default path Nl.loci, climatology, fill_value\n"
		. "Expected result: successful completion, and output file to\n"
		. "                 compare to IDL output\n";
	$outfile     = "Nl.test_pmalgos.case" . $testcase;
	$clim_name   = "${clims_dir}/snow_clims/n/1966-2003/Nl.snow_climatology.01";
	$comparefile = sprintf( "%s/Nl.test_pmalgos.case%d", $comparedir, $testcase );
	`$pmalgos -v -o $outfile -f -1. -u $loci_land_value -l $hem_info[2]->{loci_no_path} -c $clim_name $hem_info[2]->{gpd} NSIDC_M1_SWE $hem_info[2]->{tb_base}`;
	check_for_success( $description, $?, $outfile, $comparefile );
	if ( $args{a} ) {
		$testcase++;
	}

}
if ( $args{a} || 9 == $testcase ) {

	$description
		= "\nCase 9: legitimate run, NSIDC_S1_SWE (formerly NSIDC1_SWE),\n"
		. "default path Nl.loci, climatology\n"
		. "Expected result: successful completion, and output file to\n"
		. "compare to previously generated output\n";
	$outfile     = "Nl.test_pmalgos.case" . $testcase;
	$comparefile = sprintf( "%s/Nl.test_pmalgos.case%d", $comparedir, $testcase );
	$clim_name   = "${clims_dir}/snow_clims/n/1978-1995/Nl.snow_climatology.12";
	`$pmalgos -v -o $outfile -u $loci_land_value -l $hem_info[0]->{loci_no_path} -c $clim_name $hem_info[0]->{gpd} NSIDC_S1_SWE $hem_info[0]->{tb_base}`;
	check_for_success( $description, $?, $outfile, $comparefile );
	if ( $args{a} ) {
		$testcase++;
	}

}
if ( $args{a} || 10 == $testcase ) {

	$description
		= "\nCase 10: legitimate run, XPGR, greenland mask for loci \n"
		. "Expected result: successful completion, and output file to\n"
		. "                 compare to IDL output\n";
	$outfile = "Nl.test_pmalgos.case" . $testcase;
	$comparefile = sprintf( "%s/Nl.test_pmalgos.case%d", $comparedir, $testcase );
	`$pmalgos -v -o $outfile -u 1 -l $hem_info[3]->{loci} $hem_info[3]->{gpd} SSMI_XPGR $hem_info[3]->{tb_base}`;
	check_for_success( $description, $?, $outfile, $comparefile );
	if ( $args{a} ) {
		$testcase++;
	}

}
if ( $args{a} || 11 == $testcase ) {

	$description = "\nCase 11: error condition, include 2 basenames in input list\n" . "Expected result: error message, usage message, and EXIT_FAILURE\n";
	`$pmalgos -v -u 1 -l $hem_info[3]->{loci} $hem_info[3]->{gpd} SSMI_XPGR $hem_info[3]->{tb_base} $hem_info[2]->{tb_base}`;
	check_for_failure( $description, $? );
	if ( $args{a} ) {
		$testcase++;
	}

}
if ( $args{a} || 12 == $testcase ) {

	$description
		= "\nCase 12: error condition, include no basenames in input list\n" . "Expected result: error message, usage message, and EXIT_FAILURE\n";
	`$pmalgos -v -u 1 -l $hem_info[3]->{loci} $hem_info[3]->{gpd} SSMI_XPGR`;
	check_for_failure( $description, $? );
	if ( $args{a} ) {
		$testcase++;
	}

}
if ( $args{a} || 13 == $testcase ) {

	$description
		= "\nCase 13: legitimate run, NAGLER_DEPTH, default path Nl.loci, \n"
		. "no climatology\n"
		. "Expected result: successful completion, and output file to\n"
		. "compare to IDL output\n";
	$outfile = "Nl.test_pmalgos.case" . $testcase;
	$comparefile = sprintf( "%s/Nl.test_pmalgos.case%d", $comparedir, $testcase );
	`$pmalgos -v -o $outfile -u $loci_land_value -l $hem_info[0]->{loci_no_path} $hem_info[0]->{gpd} NAGLER_DEPTH $hem_info[0]->{tb_base}`;
	check_for_success( $description, $?, $outfile, $comparefile );
	if ( $args{a} ) {
		$testcase++;
	}

}
if ( $args{a} || 14 == $testcase ) {

	$description
		= "\nCase 14: legitimate run, GB_SNOW, no loci or climatology\n"
		. "Expected result: successful completion, and output file to\n"
		. "compare to IDL output\n";
	$outfile = "Nl.test_pmalgos.case" . $testcase;
	$comparefile = sprintf( "%s/Nl.test_pmalgos.case%d", $comparedir, $testcase );
	`$pmalgos -v -f -2 -o $outfile $hem_info[4]->{gpd} GB_SNOW $hem_info[4]->{tb_base}`;
	check_for_success( $description, $?, $outfile, $comparefile );
	if ( $args{a} ) {
		$testcase++;
	}

}
if ( $args{a} || 15 == $testcase ) {

	$description
		= "\nCase 15: legitimate run, GB_SURFACE, no loci or climatology\n"
		. "Expected result: successful completion, and output file to\n"
		. "compare to IDL output\n";
	$outfile = "Nl.test_pmalgos.case" . $testcase;
	$comparefile = sprintf( "%s/Nl.test_pmalgos.case%d", $comparedir, $testcase );
	`$pmalgos -v -f -2 -o $outfile $hem_info[4]->{gpd} GB_SURFACE $hem_info[4]->{tb_base}`;
	check_for_success( $description, $?, $outfile, $comparefile );
	if ( $args{a} ) {
		$testcase++;
	}

}
if ( $args{a} || 16 == $testcase ) {

	$description
		= "\nCase 16: legitimate run, NSIDC_S5_SWE (formerly NSIDC5_SWE),\n"
		. "no loci or climatology\n"
		. "Expected result: successful completion, and output file to\n"
		. "                 compare to IDL output\n";
	$outfile     = "Nl.test_pmalgos.case" . $testcase;
	$comparefile = sprintf( "%s/Nl.test_pmalgos.case%d", $comparedir, $testcase );
	$cmd         = "$pmalgos -v -f -5 -o $outfile $hem_info[5]->{gpd} NSIDC_S5_SWE $hem_info[5]->{tb_base}";
	print "Calling: $cmd\n";
	`$cmd`;
	check_for_success( $description, $?, $outfile, $comparefile );
	if ( $args{a} ) {
		$testcase++;
	}

}
if ( $args{a} || 17 == $testcase ) {

	$description
		= "\nCase 17: error condition, NSIDC_S6_SWE (formerly NSIDC6_SWE) \n"
		. "with no forest cover file\n"
		. "Expected result: error message and EXIT_FAILURE\n";
	`$pmalgos -v $hem_info[5]->{gpd} NSIDC_S6_SWE $hem_info[5]->{tb_base}`;
	check_for_failure( $description, $? );
	if ( $args{a} ) {
		$testcase++;
	}

}
if ( $args{a} || 18 == $testcase ) {

	$description
		= "\nCase 18: legitimate run, NSIDC_S6_SWE (formerly NSIDC6_SWE),\n"
		. "50%% threshold forest cover file\n"
		. "Expected result: successful completion, and output file to\n"
		. "compare to IDL output\n";
	$outfile = "Nl.test_pmalgos.case" . $testcase;
	$comparefile = sprintf( "%s/Nl.test_pmalgos.case%d", $comparedir, $testcase );
	`$pmalgos -v -f -5 -t $hem_info[5]->{forest} -o $outfile $hem_info[5]->{gpd} NSIDC_S6_SWE $hem_info[5]->{tb_base}`;
	check_for_success( $description, $?, $outfile, $comparefile );
	if ( $args{a} ) {
		$testcase++;
	}


}
if ( $args{a} || 19 == $testcase ) {

	$description
		= "\nCase 19: legitimate run, NSIDC_M0_SWE (formerly CHANG_SSMI_SWE)\n"
		. "Expected result: successful completion, and output file to\n"
		. "compare to IDL output\n";
	$outfile = "Nl.test_pmalgos.case" . $testcase;
	$comparefile = sprintf( "%s/Nl.test_pmalgos.case%d", $comparedir, $testcase );
	`$pmalgos -v -f -1 -o $outfile $hem_info[4]->{gpd} NSIDC_S0_SWE $hem_info[4]->{tb_base}`;
	check_for_success( $description, $?, $outfile, $comparefile );
	if ( $args{a} ) {
		$testcase++;
	}

}
if ( $args{a} || 20 == $testcase ) {

	$description
		= "\nCase 20: legitimate run, NSIDC_A0_SWE (NSIDC1 on AMSR TBs)\n"
		. "Expected result: successful completion, and output file to\n"
		. "compare to IDL output\n";
	$outfile = "Nl.test_pmalgos.case" . $testcase;
	$comparefile = sprintf( "%s/Nl.test_pmalgos.case%d", $comparedir, $testcase );
	`$pmalgos -v -f -1 -o $outfile $hem_info[6]->{gpd} NSIDC_A0_SWE $hem_info[6]->{tb_base}`;
	check_for_success( $description, $?, $outfile, $comparefile );
	if ( $args{a} ) {
		$testcase++;
	}


}
if ( $args{a} || 21 == $testcase ) {

	$description = "\nCase 21: error condition, NSIDC_M4_SWE \n" . "with no forest cover file\n" . "Expected result: error message and EXIT_FAILURE\n";
	`$pmalgos -v $hem_info[7]->{gpd} NSIDC_M4_SWE $hem_info[7]->{tb_base}`;
	check_for_failure( $description, $? );
	if ( $args{a} ) {
		$testcase++;
	}

}
if ( $args{a} || 22 == $testcase ) {

	$description
		= "\nCase 22: legitimate run, NSIDC_M4_SWE,\n"
		. "50%% threshold forest cover file\n"
		. "Expected result: successful completion, and output file to\n"
		. "compare to IDL output\n";
	$outfile = "Nl.test_pmalgos.case" . $testcase;
	$comparefile = sprintf( "%s/Nl.test_pmalgos.case%d", $comparedir, $testcase );
	`$pmalgos -v -f -1 -t $hem_info[7]->{forest} -o $outfile $hem_info[7]->{gpd} NSIDC_M4_SWE $hem_info[7]->{tb_base}`;
	check_for_success( $description, $?, $outfile, $comparefile );
	if ( $args{a} ) {
		$testcase++;
	}

}
if ( $args{a} || 23 == $testcase ) {

	$description
		= "\nCase 23: legitimate run, ZA_S1_FS, no loci or climatology\n"
		. "Expected result: successful completion, and output file to\n"
		. "compare to IDL output\n";
	$outfile = "Nl.test_pmalgos.case" . $testcase;
	$comparefile = sprintf( "%s/Nl.test_pmalgos.case%d", $comparedir, $testcase );
	`$pmalgos -v -o $outfile $hem_info[4]->{gpd} ZA_S1_FS $hem_info[4]->{tb_base}`;
	check_for_success( $description, $?, $outfile, $comparefile );
	if ( $args{a} ) {
		$testcase++;
	}

}
if ( $args{a} || 24 == $testcase ) {

	$description
		= "\nCase 24: error condition, NSIDC_S11_SWE \n" . "with no elevation or forest file\n" . "Expected result: error message and EXIT_FAILURE\n";
	`$pmalgos -v $hem_info[8]->{gpd} NSIDC_S11_SWE $hem_info[8]->{tb_base}`;
	check_for_failure( $description, $? );
	if ( $args{a} ) {
		$testcase++;
	}

}
if ( $args{a} || 25 == $testcase ) {

	$description
		= "\nCase 25: error condition, NSIDC_S11_SWE \n"
		. "with elevation file but no forest_file\n"
		. "Expected result: error message and EXIT_FAILURE\n";
	`$pmalgos -v -e $elev_file $hem_info[8]->{gpd} NSIDC_S11_SWE $hem_info[8]->{tb_base}`;
	check_for_failure( $description, $? );
	if ( $args{a} ) {
		$testcase++;
	}

}
if ( $args{a} || 26 == $testcase ) {

	$description
		= "\nCase 26: error condition, NSIDC_S11_SWE \n"
		. "with forest file but no elevation file\n"
		. "Expected result: error message and EXIT_FAILURE\n";
	`$pmalgos -v -t $hem_info[8]->{forest} $hem_info[8]->{gpd} NSIDC_S11_SWE $hem_info[8]->{tb_base}`;
	check_for_failure( $description, $? );
	if ( $args{a} ) {
		$testcase++;
	}

}
if ( $args{a} || 27 == $testcase ) {

	$description
		= "\nCase 27: successful condition, NSIDC_S11_SWE \n"
		. "with both forest file and elevation file\n"
		. "Expected result: successful completion, and output file to\n"
		. "compare to IDL output\n";
	$outfile = "Nl.test_pmalgos.case" . $testcase;
	$comparefile = sprintf( "%s/Nl.test_pmalgos.case%d", $comparedir, $testcase );
	`$pmalgos -v -f -1 -n 2 -s 4000 -e $elev_file -t $hem_info[8]->{forest} -o $outfile $hem_info[8]->{gpd} NSIDC_S11_SWE $hem_info[8]->{tb_base}`;
	check_for_success( $description, $?, $outfile, $comparefile );
	if ( $args{a} ) {
		$testcase++;
	}


}
if ( $args{a} || 28 == $testcase ) {

	$description
		= "\nCase 28: legitimate run, NSIDC_S10_SWE,\n"
		. "50%% threshold forest cover file\n"
		. "Expected result: successful completion, and output file to\n"
		. "compare to IDL output\n";
	$outfile = "Nl.test_pmalgos.case" . $testcase;
	$comparefile = sprintf( "%s/Nl.test_pmalgos.case%d", $comparedir, $testcase );
	`$pmalgos -v -f -1 -t $hem_info[8]->{forest} -o $outfile $hem_info[8]->{gpd} NSIDC_S10_SWE $hem_info[8]->{tb_base}`;
	check_for_success( $description, $?, $outfile, $comparefile );

	if ( $args{a} ) {
		$testcase++;
	}

}
if ( $args{a} || 29 == $testcase ) {

	$description
		= "\nCase 29: legitimate run, NSIDC_A10_SWE,\n"
		. "50%% threshold forest cover file\n"
		. "Expected result: successful completion, and output file to\n"
		. "compare to IDL output\n";
	$outfile = "Nl.test_pmalgos.case" . $testcase;
	$comparefile = sprintf( "%s/Nl.test_pmalgos.case%d", $comparedir, $testcase );
	`$pmalgos -v -f -1 -t $hem_info[6]->{forest} -o $outfile $hem_info[6]->{gpd} NSIDC_A10_SWE $hem_info[6]->{tb_base}`;
	check_for_success( $description, $?, $outfile, $comparefile );
	if ( $args{a} ) {
		$testcase++;
	}


}
if ( $args{a} || 30 == $testcase ) {

	$description
		= "\nCase 30: successful condition, NSIDC_S11_SWE \n"
		. "model 0, at elevation 3200\n"
		. "Expected result: successful completion, and output file to\n"
		. "compare to IDL output\n";
	$outfile = "Nl.test_pmalgos.case" . $testcase;
	$comparefile = sprintf( "%s/Nl.test_pmalgos.case%d", $comparedir, $testcase );
	`$pmalgos -v -f -1 -n 0 -s 3200 -e $elev_file -t $hem_info[8]->{forest} -o $outfile $hem_info[8]->{gpd} NSIDC_S11_SWE $hem_info[8]->{tb_base}`;
	check_for_success( $description, $?, $outfile, $comparefile );
	if ( $args{a} ) {
		$testcase++;
	}

}
if ( $args{a} || 31 == $testcase ) {

	$description
		= "\nCase 31: successful condition, NSIDC_S11_SWE \n"
		. "model 0, at elevation 4000\n"
		. "Expected result: successful completion, and output file to\n"
		. "compare to IDL output\n";
	$outfile = "Nl.test_pmalgos.case" . $testcase;
	$comparefile = sprintf( "%s/Nl.test_pmalgos.case%d", $comparedir, $testcase );
	`$pmalgos -v -f -1 -n 0 -s 4000 -e $elev_file -t $hem_info[8]->{forest} -o $outfile $hem_info[8]->{gpd} NSIDC_S11_SWE $hem_info[8]->{tb_base}`;
	check_for_success( $description, $?, $outfile, $comparefile );
	if ( $args{a} ) {
		$testcase++;
	}

}
if ( $args{a} || 32 == $testcase ) {

	$description = "\nCase 32: error condition, obsolete algorithm NSIDC_S2_SWE \n" . "Expected result: error message and EXIT_FAILURE\n";
	`$pmalgos -v -e $elev_file $hem_info[8]->{gpd} NSIDC_S2_SWE $hem_info[8]->{tb_base}`;
	check_for_failure( $description, $? );
	if ( $args{a} ) {
		$testcase++;
	}

}
if ( $args{a} || 33 == $testcase ) {

	$description = "\nCase 33: error condition, obsolete algorithm NSIDC_A2_SWE \n" . "Expected result: error message and EXIT_FAILURE\n";
	`$pmalgos -v -e $elev_file $hem_info[8]->{gpd} NSIDC_A2_SWE $hem_info[8]->{tb_base}`;
	check_for_failure( $description, $? );
	if ( $args{a} ) {
		$testcase++;
	}

}
if ( $args{a} || 34 == $testcase ) {

	$description
		= "\nCase 34: error condition, NSIDC_S12_SWE \n" . "         with no elevation file\n" . "Expected result: error message and EXIT_FAILURE\n";
	`$pmalgos -v $hem_info[8]->{gpd} NSIDC_S12_SWE $hem_info[8]->{tb_base}`;
	check_for_failure( $description, $? );
	if ( $args{a} ) {
		$testcase++;
	}

}
if ( $args{a} || 35 == $testcase ) {

	$description
		= "\nCase 35: error condition, NSIDC_S12_SWE \n"
		. "with elevation file but no elevation model\n"
		. "Expected result: error message and EXIT_FAILURE\n";
	`$pmalgos -v -e $elev_file $hem_info[8]->{gpd} NSIDC_S12_SWE $hem_info[8]->{tb_base}`;
	check_for_failure( $description, $? );
	if ( $args{a} ) {
		$testcase++;
	}

}
if ( $args{a} || 36 == $testcase ) {

	$description
		= "\nCase 36: successful condition, NSIDC_S12_SWE \n" . "Expected result: successful completion, and output file to\n" . "compare to IDL output\n";
	$outfile = "Nl.test_pmalgos.case" . $testcase;
	$comparefile = sprintf( "%s/Nl.test_pmalgos.case%d", $comparedir, $testcase );
	`$pmalgos -v -f -1 -n 5 -s 3200 -e $elev_file -o $outfile $hem_info[8]->{gpd} NSIDC_S12_SWE $hem_info[8]->{tb_base}`;
	check_for_success( $description, $?, $outfile, $comparefile );
	if ( $args{a} ) {
		$testcase++;
	}

}

if ( $args{a} || 37 == $testcase ) {

	$description = "\nCase 37: error condition, GODDARD_SEAICE for F15 (not implemented) \n" . "Expected result: error message and EXIT_FAILURE\n";
	`$pmalgos -v -q N -p F15 N3B.gpd GODDARD_SEAICE ${topdir}/bootstrap/tbs/N/N3B.990531`;
	check_for_failure( $description, $? );
	if ( $args{a} ) {
		$testcase++;
	}

}
if ( $args{a} || 38 == $testcase ) {

	$description
		= "\nCase 38: legitimate run, NSIDC_S1_SWE\n"
		. "default path Nl.loci, climatology\n"
		. "Expected result: successful completion, and output file to\n"
		. "compare to previously generated output\n";
	$outfile = sprintf( "Nl.test_pmalgos.case%2d", $testcase );
	$comparefile = sprintf( "%s/Nl.test_pmalgos.case%2d", $comparedir, $testcase );
	$clim_name = "${clims_dir}/snow_clims/n/1978-1995/Nl.snow_climatology.12";
	`$pmalgos -v -p F17 -o $outfile -u $loci_land_value -l $hem_info[0]->{loci_no_path} -c $clim_name $hem_info[0]->{gpd} NSIDC_S1_SWE $hem_info[0]->{tb_base}`;
	check_for_success( $description, $?, $outfile, $comparefile );
	if ( $args{a} ) {
		$testcase++;
	}

}
if ( $args{a} || 39 == $testcase ) {

	$description = "\nCase 39: warning condition, NSIDC_S5_SWE with -p F17 (not implemented warning)\n"
		. "Expected results:  should match output from no -p switch, with warning message\n";
	$outfile     = "Nl.test_pmalgos.case" . $testcase;
	$comparefile = sprintf( "%s/Nl.test_pmalgos.case16", $comparedir );
	$cmd         = "$pmalgos -v -f -5 -p F17 -o $outfile $hem_info[5]->{gpd} NSIDC_S5_SWE $hem_info[5]->{tb_base}";
	print "Calling: $cmd\n";
	`$cmd`;
	check_for_success( $description, $?, $outfile, $comparefile );
	if ( $args{a} ) {
		$testcase++;
	}


}
if ( $args{a} || 40 == $testcase ) {

	$description
		= "\nCase 40: legitimate run, NSIDC_S10_SWE,\n"
		. "50%% threshold forest cover file, with -p F17\n"
		. "Expected result: successful completion, and output file to\n"
		. "compare to IDL output\n";
	$outfile = "Nl.test_pmalgos.case" . $testcase;
	$comparefile = sprintf( "%s/Nl.test_pmalgos.case%d", $comparedir, $testcase );
	`$pmalgos -v -f -1 -p F17 -t $hem_info[8]->{forest} -o $outfile $hem_info[8]->{gpd} NSIDC_S10_SWE $hem_info[8]->{tb_base}`;
	check_for_success( $description, $?, $outfile, $comparefile );

	if ( $args{a} ) {
		$testcase++;
	}

}
if ( $args{a} || 41 == $testcase ) {

	$description
		= "\nCase 41: successful condition, NSIDC_S11_SWE \n"
		. "with both forest file and elevation file, and -p F17\n"
		. "Expected result: successful completion, and output file to\n"
		. "compare to IDL output\n";
	$outfile = "Nl.test_pmalgos.case" . $testcase;
	$comparefile = sprintf( "%s/Nl.test_pmalgos.case%d", $comparedir, $testcase );
	`$pmalgos -v -f -1 -p F17 -n 2 -s 4000 -e $elev_file -t $hem_info[8]->{forest} -o $outfile $hem_info[8]->{gpd} NSIDC_S11_SWE $hem_info[8]->{tb_base}`;
	check_for_success( $description, $?, $outfile, $comparefile );
	if ( $args{a} ) {
		$testcase++;
	}


}
if ( $args{a} || 42 == $testcase ) {

	$description
		= "\nCase 42: successful condition, NSIDC_S12_SWE with -p F17\n"
		. "Expected result: successful completion, and output file to\n"
		. "compare to IDL output\n";
	$outfile = "Nl.test_pmalgos.case" . $testcase;
	$comparefile = sprintf( "%s/Nl.test_pmalgos.case%d", $comparedir, $testcase );
	`$pmalgos -v -f -1 -p F17 -n 5 -s 3200 -e $elev_file -o $outfile $hem_info[8]->{gpd} NSIDC_S12_SWE $hem_info[8]->{tb_base}`;
	check_for_success( $description, $?, $outfile, $comparefile );
	if ( $args{a} ) {
		$testcase++;
	}

}
if ( $args{a} || 43 == $testcase ) {

	$description
		= "\nCase 43: legitimate run, NSIDC_S1_SWE for sensor F17rev1\n"
		. "default path Nl.loci, climatology\n"
		. "Expected result: successful completion, and output file to\n"
		. "compare to previously generated output\n";
	$outfile = sprintf( "Nl.test_pmalgos.case%2d", $testcase );
	$comparefile = sprintf( "%s/Nl.test_pmalgos.case%2d", $comparedir, $testcase );
	$clim_name = "${clims_dir}/snow_clims/n/1978-1995/Nl.snow_climatology.12";
	`$pmalgos -v -p F17rev1 -o $outfile -u $loci_land_value -l $hem_info[0]->{loci_no_path} -c $clim_name $hem_info[0]->{gpd} NSIDC_S1_SWE $hem_info[0]->{tb_base}`;
	check_for_success( $description, $?, $outfile, $comparefile );
	if ( $args{a} ) {
		$testcase++;
	}

}

# end while testcase number is defined


1;
