#!/usr/local/bin/perl -w
#
# This perl script will test the C program pmalgos for the Bootstrap algorithm
#
# 2010-08-09 M. J. Brodzik 303-492-8263 brodzik@nsidc.org
# National Snow & Ice Data Center, Boulder, CO
# Copyright (C) 2010 University of Colorado
#
#$Id: test_pmalgos_bootstrap.pl 22406 2013-01-30 16:10:18Z savoie $
#
#-----------------------------------------------------------------------
use Env;
use lib ("${PERL_UTIL_PATH}");
use NSIDC_Unit_Test_Util;
use Getopt::Std;
use File::Temp qw/ tempdir /;
use File::Basename qw/ basename /;
use File::Spec::Functions;

sub tmp_test_file_name($$$);


# force autoflush after every write or print
$| = 1;
my $usage = '

test_pmalgos_bootstrap.pl [-t test_binary]

Input: n/a

Options:
    t test_binary - pmalgos binary to test, default is
      ./pmalgos

';

undef %args;
if ( !getopts( 't:', \%args ) ) {
	printf STDERR $usage;
	exit;
}

if ( !$args{t} ) {
	$args{t} = "./pmalgos";
}

$tmp_test_dir    = tempdir();
$topdir          = "/share/data/pmalgos_test_data";
$map_dir         = ${topdir} . "/maps";
$ancillary_dir   = ${topdir} . "/ancillary";
$fill_value      = 157;
$pmalgos         = "$args{t} ";
$pmalgos_options = "-f $fill_value -r -u 0 ";

# Use a call to the unix function hostname to determine where we are running.
# Do not use $ENV{HOSTNAME}, it returns unreliable results, depending on the
# user's shell.
$fdistill = "$ENV{FDISTILL_DIR}/fdistill -ub ";

$tbsdir  = "${topdir}/bootstrap/tbs";
$testdir = "${topdir}/";

# Note assumptions in code about filenames, here, the hemisphere and date will
# be parsed from these filenames.  Dates are assumed to be pre-2000.
@basename_list = (
				   $tbsdir . "/N/N3B.980718",
				   $tbsdir . "/N/N3B.980719",
				   $tbsdir . "/N/N3B.980804",
				   $tbsdir . "/N/N3B.980805",
				   $tbsdir . "/N/N3B.980912",
				   $tbsdir . "/N/N3B.980913",
				   $tbsdir . "/N/N3B.980920",
				   $tbsdir . "/N/N3B.980921",
				   $tbsdir . "/N/N3B.981001",
				   $tbsdir . "/N/N3B.981015",
				   $tbsdir . "/N/N3B.981016",
				   $tbsdir . "/N/N3B.981115",
				   $tbsdir . "/N/N3B.981116",
				   $tbsdir . "/N/N3B.990531",
				   $tbsdir . "/N/N3B.990601",
				   $tbsdir . "/N/N3B.990623",
				   $tbsdir . "/N/N3B.990624",
				   $tbsdir . "/S/S3B.981031",
				   $tbsdir . "/S/S3B.981101",
				   $tbsdir . "/S/S3B.990131",
				   $tbsdir . "/S/S3B.990201",
				   $tbsdir . "/S/S3B.990207",
				   $tbsdir . "/S/S3B.990208",
				   $tbsdir . "/S/S3B.990301",
				   $tbsdir . "/S/S3B.990407",
				   $tbsdir . "/S/S3B.990408",
				   $tbsdir . "/N/N3B.900101",
				   $tbsdir . "/N/N3B.900701",
				   $tbsdir . "/N/N3B.940101",
				   $tbsdir . "/N/N3B.940701",
				   $tbsdir . "/S/S3B.940101",
				   $tbsdir . "/S/S3B.940701"
);

%landmasks = (
			   N => $ancillary_dir . "/N3B.bootstrap.landmask.dat",
			   S => $ancillary_dir . "/S3B.bootstrap.landmask.dat",
);

# Note:  Add map_dir to environment setting for PATHMPP
# for {loci_no_path} values to work
$ENV{"PATHMPP"} = $ENV{"PATHMPP"} . ":${map_dir}";

my $hem;
my $yymmdd;
my $yyyymmdd;
my $algo_base = "bootstrap";
my $description;

printf STDERR "\n\n*************************Begin test_pmalgos_bootstrap tests ***************************\n\n\n";

# Test Bootstrap seaice algorithm for each basename, compare output to what was previously produced
foreach $base (@basename_list) {

	# Parse the basename to figure out the hemisphere and date
	@parts = split( '3B\.', $base );
	$hem = chop( $parts[0] );
	if ( $base !~ /(\d{6})/ ) {
		die "ERROR: error parsing date from input basename $base.";
	}
	$yymmdd = $1;

	$algo     = uc($algo_base) . "_${hem}_SEAICE";
	$yyyymmdd = '19' . $yymmdd;

	$test_file = tmp_test_file_name( $tmp_test_dir, $base, ".${algo}.tot" );
	$compare_file = $testdir . $algo_base . '/compare_ice/' . $hem . '/' . $yymmdd . '.tot';

	$description = "\ntest_pmalgos_bootstrap ${algo} for basename $base\n" . "using pmalgos at $pmalgos\n";

	$cmd = "$pmalgos $pmalgos_options -d $yyyymmdd -l $landmasks{$hem}";
	$cmd = "$cmd ${hem}3B.gpd $algo $base";
	print "Calling: $cmd | $fdistill > $test_file\n";
	`$cmd | $fdistill > $test_file`;
	check_for_success( $description, $?, $test_file, $compare_file );

}

# Bootstrap_v2 tests
printf STDERR "\n\n\n************************* bootstrap_v2 tests ***************************\n\n";

# Set environment variable to run Bootstrap_V2
# this is the the location the sst files that are needed by the processing
$ancillary_dir = "/share/data/pmalgos/bootstrap_seaice";
$ENV{"BOOTSTRAP_V2_ANCILLARY"} = "${ancillary_dir}";

$tbsdir  = "${topdir}/tbs/bootstrap";
$testdir = "${topdir}/bootstrap/compare_ice/bootstrap_v2";
@basename_list = (
				   $tbsdir . "/tb_f13_n3b20070430_v2_n",
				   $tbsdir . "/tb_f13_s3b20070430_v2_s",
				   $tbsdir . "/ssmi-f13-s3b20070430",
				   $tbsdir . "/ssmi-f13-n3b20070430",
				   $tbsdir . "/ssmi-f13-s3b20070430",
				   $tbsdir . "/ssmi-f13-n3b20070430",
				   $tbsdir . "/ssmi-f13-s3b20070430",
				   $tbsdir . "/ssmi-f13-n3b20070430",
				   $tbsdir . "/ssmi-f13-s3b20070430",
				   $tbsdir . "/ssmi-f13-n3b20070430",
				   $tbsdir . "/ssmi-f13-s3b20070430",
				   $tbsdir . "/ssmi-f13-n3b20070430",
);

@bv02_compare_list = (
   $testdir . "/bootstrap.bv02.new.apr30.n3b",
   $testdir . "/bootstrap.bv02.new.apr30.s3b",
   $testdir . "/bootstrap.bv02.new.apr30.s3b",
   $testdir . "/bootstrap.bv02.new.apr30.n3b",
   $testdir . "/bootstrap.bv02.new.apr30.clim.s3b",
   $testdir . "/bootstrap.bv02.new.apr30.clim.n3b",
   $testdir . "/bootstrap.bv02.new.apr30.scaling.s3b",
   $testdir . "/bootstrap.bv02.new.apr30.scaling.n3b",
   $testdir . "/bootstrap.bv02.new.apr30.spillover.s3b",
   $testdir . "/bootstrap.bv02.new.apr30.spillover.n3b",
   $testdir . "/bootstrap.bv02.new.apr30.spilloverscaling.s3b",
   $testdir . "/bootstrap.bv02.new.apr30.spilloverscaling.n3b",

);

# "published" regressions e.g. the best available
@bv02_pub_compare_list = (
   $testdir . "/bootstrap.bv02.pub.new.apr30.n3b",
   $testdir . "/bootstrap.bv02.pub.new.apr30.s3b",
   $testdir . "/bootstrap.bv02.pub.new.apr30.s3b",
   $testdir . "/bootstrap.bv02.pub.new.apr30.n3b",
   $testdir . "/bootstrap.bv02.pub.new.apr30.clim.s3b",
   $testdir . "/bootstrap.bv02.pub.new.apr30.clim.n3b",
   $testdir . "/bootstrap.bv02.pub.new.apr30.scaling.s3b",
   $testdir . "/bootstrap.bv02.pub.new.apr30.scaling.n3b",
   $testdir . "/bootstrap.bv02.pub.new.apr30.spillover.s3b",
   $testdir . "/bootstrap.bv02.pub.new.apr30.spillover.n3b",
   $testdir . "/bootstrap.bv02.pub.new.apr30.spilloverscaling.s3b",
   $testdir . "/bootstrap.bv02.pub.new.apr30.spilloverscaling.n3b",

);

@options_list = (
				  "-b -u OCEAN_VAL -d YYYYMMDD -l LOCI ",
				  "-b -u OCEAN_VAL -d YYYYMMDD -l LOCI ",
				  "-u OCEAN_VAL -d YYYYMMDD -l LOCI ",
				  "-u OCEAN_VAL -d YYYYMMDD -l LOCI ",
				  "-u OCEAN_VAL -d YYYYMMDD -l LOCI -c CLIMMASK",
				  "-u OCEAN_VAL -d YYYYMMDD -l LOCI -c CLIMMASK",
				  "-u OCEAN_VAL -d YYYYMMDD -l LOCI -z ",
				  "-u OCEAN_VAL -d YYYYMMDD -l LOCI -z ",
				  "-u OCEAN_VAL -d YYYYMMDD -l LOCI -j -m LANDMASK -w WEIGHTMASK ",
				  "-u OCEAN_VAL -d YYYYMMDD -l LOCI -j -m LANDMASK -w WEIGHTMASK ",
				  "-u OCEAN_VAL -d YYYYMMDD -l LOCI -j -m LANDMASK -w WEIGHTMASK -z ",
				  "-u OCEAN_VAL -d YYYYMMDD -l LOCI -j -m LANDMASK -w WEIGHTMASK -z ",
);

%loci = (
		  N => $ancillary_dir . "/N3B.bootstrap_v02.landmask.dat",
		  S => $ancillary_dir . "/S3B.bootstrap_v02.landmask.dat",
);

%landmask = (
			  N => $ancillary_dir . "/N3B.bootstrap_v02.landocean.dat",
			  S => $ancillary_dir . "/S3B.bootstrap_v02.landocean.dat",
);

%weightmask = (
	N => $ancillary_dir . "/cmin_bootstrap_v2_1992.N11.byte",
	S => $ancillary_dir . "/cmin_bootstrap_v2_1992.S11.byte",

	# N => $ancillary_dir . "/N3B.cmin_goddard_1992.N11.byte",
	# S => $ancillary_dir . "/S3B.cmin_goddard_1992.S11.byte",
);

%climmask = (
			  N => $ancillary_dir . '/iceclim.bootv2.N04_v1.byte',
			  S => $ancillary_dir . '/iceclim.bootv2.S04_v2.byte',
);

my $ocean_val = 0;
my $i;


# Test Bootstrap V2 seaice algorithm, compare output to Jeff's output
my @algorithm_stems = ( "_V02_SEAICE", "_V02_PUB_SEAICE" );
my @compare_lists = ( \@bv02_compare_list, \@bv02_pub_compare_list );

for ( my $num_algorithms = 0; $num_algorithms <= $#algorithm_stems; $num_algorithms++ ) {
	my $algorithm_stem = $algorithm_stems[$num_algorithms];
	my @compare_list   = @{ $compare_lists[$num_algorithms] };

	for ( $i = 0; $i <= $#basename_list; $i++ ) {

		$base            = $basename_list[$i];
		$compare_file    = $compare_list[$i];
		$pmalgos_options = "-p F17 " . $options_list[$i];

		# Parse the basename to figure out the hemisphere and date
		if ( $base !~ /(.)3b(\d{8})/ ) {
			die "ERROR: error parsing date from input basename $base.";
		}
		$hem      = uc($1);
		$yyyymmdd = $2;

		$algo = "BOOTSTRAP_" . $hem . $algorithm_stem;
		$test_file = tmp_test_file_name( $tmp_test_dir, $base, ".${algo}.tot" );

		$pmalgos_options =~ s/OCEAN_VAL/$ocean_val/;
		$pmalgos_options =~ s/YYYYMMDD/$yyyymmdd/;
		$pmalgos_options =~ s/LOCI/$loci{ $hem }/;
		$pmalgos_options =~ s/LANDMASK/$landmask{ $hem }/;
		$pmalgos_options =~ s/WEIGHTMASK/$weightmask{ $hem }/;
		$pmalgos_options =~ s/CLIMMASK/$climmask{$hem}/;

		$description = "\ntest_pmalgos_bootstrap ${algo} for basename $base\n" . " and options=" . $pmalgos_options . "\n" . "using pmalgos at $pmalgos\n";

		$cmd = "$pmalgos $pmalgos_options ${hem}3B.gpd $algo $base";
		print "Calling: $cmd > $test_file\n";
		`$cmd > $test_file`;
		check_for_success( $description, $?, $test_file, $compare_file );

	}
}




# Test Bootstrap error conditions try no date, no loci, no environment
# variable BOOTSTRAP_V2_ANCILLARY (currently not setting this exits
# successfully, with "STOP 1 statement executed" and an empty sea ice
# file).
@basename_list = (
				   $tbsdir . "/ssmi-f13-s3b20070430",      # no date
				   $tbsdir . "/ssmi-f13-n3b20070430",      # no loci
				   $tbsdir . "/ssmi-f13-s3b20070430",      # no env variable
				   "${topdir}/tbs/EASE-F13-NL1999031D",    # non-N/S3B grid
);

@options_list = (
				  "-u OCEAN_VAL -l LOCI ",
				  "-u OCEAN_VAL -d YYYYMMDD ",
				  "-u OCEAN_VAL -d YYYYMMDD -l LOCI -j -m LANDMASK -w WEIGHTMASK ",
				  "-u OCEAN_VAL -d YYYYMMDD -l LOCI ",
);

@description_list = ( "ERROR condition, no date", "ERROR condition, no loci", "ERROR condition, no env variable", "ERROR condition, non ps grid", );



# -----------------------------------------
# Test Bootstrap V2 seaice error conditions
# -----------------------------------------
for ( $i = 0; $i <= $#basename_list; $i++ ) {

	$base            = $basename_list[$i];
	$pmalgos_options = $options_list[$i];
	$description     = "\ntest_pmalgos_bootstrap for " . $description_list[$i] . "\n" . "using pmalgos at $pmalgos\n";

	if ( 3 ne $i ) {

		# Parse the basename to figure out the hemisphere and date
		if ( $base !~ /-(.)3b(\d{8})/ ) {
			die "ERROR: error parsing date from input basename $base.";
		}
		$hem      = uc($1);
		$yyyymmdd = $2;

		$algo = "BOOTSTRAP_" . $hem . "_V02_SEAICE";
		$test_file = tmp_test_file_name( $tmp_test_dir, $base, ".${algo}.tot" );

		$pmalgos_options =~ s/OCEAN_VAL/$ocean_val/;
		$pmalgos_options =~ s/YYYYMMDD/$yyyymmdd/;
		$pmalgos_options =~ s/LOCI/$loci{ $hem }/;
		$pmalgos_options =~ s/LANDMASK/$landmask{ $hem }/;
		$pmalgos_options =~ s/WEIGHTMASK/$weightmask{ $hem }/;
		$gpd = "${hem}3B.gpd";

	} else {

		$hem      = "N";
		$yyyymmdd = "19990131";

		$algo      = "BOOTSTRAP_" . $hem . "_V02_SEAICE";
		$test_file = tmp_test_file_name( $tmp_test_dir, $base, ".${algo}.tot" );
		$easeloci  = "/share/data/pmalgos/landcover/Nl_lo.721x721.bin";

		$pmalgos_options =~ s/OCEAN_VAL/$ocean_val/;
		$pmalgos_options =~ s/YYYYMMDD/$yyyymmdd/;
		$pmalgos_options =~ s/LOCI/$easeloci/;
		$gpd = "Nl.gpd";
	}

	if ( 2 eq $i ) {
		$ENV{"BOOTSTRAP_V2_ANCILLARY"} = ".";
	} else {
		$ENV{"BOOTSTRAP_V2_ANCILLARY"} = "${ancillary_dir}";
	}

	$cmd = "$pmalgos $pmalgos_options $gpd $algo $base";
	print "Calling: $cmd > $test_file\n";
	`$cmd > $test_file`;
	check_for_failure( $description, $? );


}

printf STDERR "\n\n\n*************************End Bootstrap v2 tests ***************************\n\n";
printf STDERR "\n\n\n*************************End test_pmalgos_bootstrap tests ***************************\n\n";


1;

sub tmp_test_file_name($$$) {
	my $tmp_dir           = shift;
	my $compare_file_base = shift;
	my $extension         = shift;
	my $base_id           = basename($compare_file_base);
	return catfile( $tmp_test_dir, $base_id . $extension );
}
