#!/usr/sbin/perl -w
#
# This perl script will test the C program pmtbcomb
#
# 2009-07-01 M. J. Brodzik brodzik@nsidc.org 303-492-8263
# National Snow & Ice Data Center, University of Colorado, Boulder 
# Copyright (C) 2009 University of Colorado
#-----------------------------------------------------------------------
use Env;
use Getopt::Std;
use File::Basename;
use vars qw( %args );

# force autoflush after every write or print
$| = 1;
my $usage = '

test_pmtbcomb.pl [-a] [-t test_binary] [casenum]

Input:
    casenum - case number to run

Options:
    a - do all cases in sequence
    t test_binary - pmalgos binary to test, default is
      ${HOME}/src/pmalgos/pmtbcomb

';

undef %args;
if ( !getopts('at:', \%args) ) {
    printf STDERR $usage;
    exit;
}

if ( !$args{a} ) {
    $args{a} = 0;
} else {
    $testcase = 1;
}

if ( !$args{t} ) {
    $args{t} = "${HOME}/src/pmalgos/pmtbcomb";
}

if (-1 != $#ARGV) {
    $testcase = shift;
}

$topdir = "/share/data/pmalgos_test_data";
$ancillary_dir = "${topdir}/maps";
$pmtbcomb = $args{t};
$comparedir = "${topdir}/pmtbcomb";
$tbdir = "${topdir}/tbs";

print "Beginning of $0:\n";
print "Testing pmtbcomb binary: $pmtbcomb\n";

# Note:  Add ancillary to environment setting for PATHMPP
# for {loci_no_path} values to work
$ENV{"PATHMPP"} = "${ancillary_dir}:" . $ENV{"PATHMPP"};

@hem_info = 
  (
   {
    gpd => "Nl.gpd",
    tb_base1 => "${tbdir}/EASE-F13-NL1998305D",
    tb_base2 => "${tbdir}/EASE-F13-NL1998365D",
   },
   {
    gpd => "Nl.gpd",
    tb_base1 => "${tbdir}/EASE-F13-NL1999031D",
    tb_base2 => "${tbdir}/EASE-F13-NL1999354D",
   },
  );


if ( $args{a} || 1 == $testcase) {

    print "Case 1: no arguments\n";
    print "Expected result: error message, usage message, and EXIT_FAILURE\n";
    `$pmtbcomb`;
    &check_for_failure( $testcase, $? );
    if ( $args{a} ) {
        $testcase++;
    }

}
if ( $args{a} || 2 == $testcase ) {

    print "Case 2: invalid gpd\n";
    print "Expected result: error message and EXIT_FAILURE\n";
    $outfile = "Nl.test_pmtbcomb.case" . $testcase;
    `$pmtbcomb unknown.gpd NSIDC_S1_SWE MEAN $outfile $hem_info[0]->{tb_base1}`;
    &check_for_failure( $testcase, $? );
    if ( $args{a} ) {
        $testcase++;
    }

}

if ( $args{a} || 3 == $testcase ) {

    print "Case 3: invalid geo_algorithm\n";
    print "Expected result: error message and EXIT_FAILURE\n";
    $outfile = "Nl.test_pmtbcomb.case" . $testcase;
    `$pmtbcomb $hem_info[0]->{gpd} BOGUS_GEO_ALGO MEAN $outfile $hem_info[0]->{tb_base1}`;
    &check_for_failure( $testcase, $? );
    if ( $args{a} ) {
        $testcase++;
    }

}

if ( $args{a} || 4 == $testcase ) {

    print "Case 4: invalid comb_algorithm\n";
    print "Expected result: error message and EXIT_FAILURE\n";
    $outfile = "Nl.test_pmtbcomb.case" . $testcase;
    `$pmtbcomb $hem_info[0]->{gpd} NSIDC_S1_SWE BOGUS_COMB_ALGO $outfile $hem_info[0]->{tb_base1}`;
    &check_for_failure( $testcase, $? );
    if ( $args{a} ) {
        $testcase++;
    }

}

if ( $args{a} || 5 == $testcase ) {

     print "\n\nCase 5: legitimate run, clear_skies\n";
     print "Expected result: successful completion, and output file to\n";
     print "                 compare to previously-generated output\n";
     $outfile = "Nl.test_pmtbcomb.case" . $testcase;
     $comparefile = sprintf("%s/Nl.test_pmtbcomb.case%d",
                            $comparedir,$testcase);
    `$pmtbcomb $hem_info[0]->{gpd} NAGLER_DEPTH CLEAREST_SKY $outfile $hem_info[0]->{tb_base1} $hem_info[0]->{tb_base2}`;

     @chans = ( '.19V', '.37V', '.85V' );
     foreach $chan (@chans) {
         
         &check_for_success( $testcase,
                             $?,
                             $outfile . $chan,
                             $comparefile . $chan );

     }

     if ( $args{a} ) {
         $testcase++;
     }

}
if ( $args{a} || 6 == $testcase ) {

     print "\n\nCase 6: legitimate run, MEAN\n";
     print "Expected result: successful completion, and output file to\n";
     print "                 compare to previously-generated output\n";
     $outfile = "Nl.test_pmtbcomb.case" . $testcase;
     $comparefile = sprintf("%s/Nl.test_pmtbcomb.case%d",
                            $comparedir,$testcase);
    `$pmtbcomb $hem_info[1]->{gpd} NSIDC_S1_SWE MEAN $outfile $hem_info[1]->{tb_base1} $hem_info[1]->{tb_base2}`;

     @chans = ( '.19H', '.37H' );
     foreach $chan (@chans) {
         
         &check_for_success( $testcase,
                             $?,
                             $outfile . $chan,
                             $comparefile . $chan );

     }

     if ( $args{a} ) {
         $testcase++;
     }

}
if ( $args{a} || 7 == $testcase ) {

     print "\n\nCase 7: legitimate run, MAX\n";
     print "Expected result: successful completion, and output file to\n";
     print "                 compare to previously-generated output\n";
     $outfile = "Nl.test_pmtbcomb.case" . $testcase;
     $comparefile = sprintf("%s/Nl.test_pmtbcomb.case%d",
                            $comparedir,$testcase);
    `$pmtbcomb $hem_info[1]->{gpd} NSIDC_S1_SWE MAX $outfile $hem_info[1]->{tb_base1} $hem_info[1]->{tb_base2}`;

     @chans = ( '.19H', '.37H' );
     foreach $chan (@chans) {
         
         &check_for_success( $testcase,
                             $?,
                             $outfile . $chan,
                             $comparefile . $chan );

     }

     if ( $args{a} ) {
         $testcase++;
     }

}
if ( $args{a} || 8 == $testcase ) {

     print "\n\nCase 8: legitimate run, MIN\n";
     print "Expected result: successful completion, and output file to\n";
     print "                 compare to previously-generated output\n";
     $outfile = "Nl.test_pmtbcomb.case" . $testcase;
     $comparefile = sprintf("%s/Nl.test_pmtbcomb.case%d",
                            $comparedir,$testcase);
    `$pmtbcomb $hem_info[1]->{gpd} NSIDC_S1_SWE MIN $outfile $hem_info[1]->{tb_base1} $hem_info[1]->{tb_base2}`;

     @chans = ( '.19H', '.37H' );
     foreach $chan (@chans) {
         
         &check_for_success( $testcase,
                             $?,
                             $outfile . $chan,
                             $comparefile . $chan );

     }

     if ( $args{a} ) {
         $testcase++;
     }

}
# end while testcase number is defined

# Examines the program return code for an expected SUCCESS case.
# If it was successful, compare output to previously created compare file.
# Dies if either status is bad, or comparison fails.
#
# arguments:
#   case number : test case number
#   status : return value from pmalgos call ($?)
#   outfile : output file from this test case
#   comparefile : compare file for this test case
sub check_for_success {

    my $testcase = shift;
    my $status = shift;
    my $outfile = shift;
    my $comparefile = shift;

    print "CASE ". $testcase . "......";
    
    if (0 != $status) {
        die "\nERROR: return was not EXIT_SUCCESS, prog_ret = " . 
            ($status >> 8) . ", wait_ret = " . ($status & 255) . "\n";
    }

    print "SUCCESS: returned EXIT_SUCCESS, checking output...";
    `diff -s $outfile $comparefile`;
    if (0 != $?) {
        die "ERROR: return was not EXIT_SUCCESS, prog_ret = " . 
            ($? >> 8) . ", wait_ret = " . ($? & 255) . "\n";
    } else {
        print "OK.\n\n";
        unlink $outfile;
    }
    print "CASE $testcase test    file was $outfile\n";
    print "CASE $testcase archive file was $comparefile\n";

    return;
    
}

# Examines the program return code for an expected FAILURE case.
# Dies if status is actually success.
#
# arguments:
#   case number : test case number
#   status : return value from pmalgos call ($?)
sub check_for_failure {

    my $testcase = shift;
    my $status = shift;

    print "CASE ". $testcase . "......";
    
    if (0 == ($status >> 8)) {
        die "ERROR: return was not EXIT_FAILURE\n";
    }

    print "OK.\n\n";

    return;
    
}

1;



