/*======================================================================
 * test_pmtbs : test driver for pmtbs
 * 
 * 2000-02-11 brodzik brodzik@nsidc.org 303-492-8263
 * National Snow & Ice Data Center, Boulder CO
 * Copyright 2000-2009 University of Colorado at Boulder
 *
 *$Id: test_pmtbs.c 15309 2009-07-01 16:23:06Z brodzik $
 * 
 *======================================================================*/
#include "define.h"
#include "matrix_io.h"
#include "pmtbs.h"
#include "pmalgorithm.h"

#define usage_message "usage: test_pmtbs case_num\n"

#define ident_message "test_pmtbs $Revision: 1.8 $"
#define NORTH_SMMR_BASEFILE "/projects/EASE/pmalgos_test/EASE-SMMR-NL1979002D"
#define NORTH_SSMI_BASEFILE "/projects/EASE/pmalgos_test/Nl911227D-nn"
#define NORTH_SSMI_BASEFILE2 "/projects/EASE/pmalgos_test/EASE-F13-NL1998365D"
#define SOUTH_SSMI_BASEFILE "/projects/EASE/pmalgos_test/Sl911227D-nn"
#define NON_EXISTENT_FILE "/usr/local/bogus.dat"
#define NORTH_GPD_FILE "Nl.gpd"
#define SOUTH_GPD_FILE "Sl.gpd"
#define COMPARE_DIR "/usr2/projects/nise/idl/compare_data/"
#define SMMR_NUM_CHANNELS 10
#define SSMI_NUM_CHANNELS 7

static void init_pmtbs_for(pmalgo_id id, bool null_basename);
static void process_algorithm(pmalgo_id id);

main (int argc, char *argv[])
{
  static const char test_pmtbs_c_RCSID[]="$Id: test_pmtbs.c 15309 2009-07-01 16:23:06Z brodzik $";
  char file_name[FILENAME_MAX];
  pmtbs_class *pmtbs;
  int row, col;
  int i;
  int case_num;
  pmalgo_id id;
  pmalgo_sensor sensor;
  grid_class *grid;

  fprintf(stderr,"%s\n", ident_message);

  if (--argc != 1) {
    fprintf(stderr,"Invalid number of input arguments\n");
    fprintf(stderr,usage_message);
    exit(EXIT_FAILURE);
  }
  case_num = strtol(*++argv,NULL,10);

  switch (case_num) {
  case 1:
    /* 
     * Case 1: Error condition
     *   Call with NULL pointer to gpd
     *   Expected result:
     *     assertion will be thrown in init_grid
     */
    strcpy(file_name,NON_EXISTENT_FILE);
    fprintf(stderr,"\n==>Case 1: Error Condition: NULL grid object\n");
    pmtbs = init_pmtbs(NULL,
		       "NSIDC1_SWE",
                       "",
		       NORTH_SSMI_BASEFILE,
		       TRUE);
    fprintf(stderr,"ERROR:  init_pmtbs should have thrown an assertion\n");
    exit(EXIT_FAILURE);

    break;
  case 2:

    /* 
     * Case 2: Error condition
     *   Call with invalid algorithm name
     *   Expected result:
     *     "unrecognized algorithm" error to stderr
     *     returned pointer is NULL
     */
    grid = init_grid(NORTH_GPD_FILE);
    fprintf(stderr,"\n==>Case 2: Error Condition: invalid algorithm name\n");
    pmtbs = init_pmtbs(grid,
		       "BOGUS_ALGORITHM",
                       "",
		       NORTH_SSMI_BASEFILE,
		       TRUE);
    if (NULL == pmtbs) {
      fprintf(stderr,"SUCCESS:  init_pmtbs returned NULL pointer\n\n\n");
    } else {
      fprintf(stderr,"ERROR:  init_pmtbs did not return NULL pointer\n");
      exit(EXIT_FAILURE);
    }

    break;
  case 3:

    /* 
     * Case 3: Legitimate condition: Each valid algorithm
     *   Call init_pmtbs for each valid algorithm, and check returned
     *     object
     *   Expected result:
     *     returned pointer is non-NULL
     *     initializes appropriate TB grids
     */
    fprintf(stderr,"\n==>Case 3: Legitimate Condition: each valid algorithm\n");
    for (id=(pmalgo_id)0; id<PMALGO_NUM_ALGORITHMS; id++) {
      fprintf(stderr,"\n==>Subcase: for algorithm %s\n",pmalgo_id_name[id]);
	  init_pmtbs_for(id,FALSE);
    } /* loop for each algorithm */

    break;
  case 4:

    /* 
     * Case 4: Initialize and invoke each algorithm
     *   Call init_pmtbs and use pmtbs->algorithm->run
     *   for each valid algorithm, and check result against
     *     results produced ahead of time in IDL
     *   Expected result:
     *     Results should match within +/- 1 integer unit.
     */
    fprintf(stderr,"\n==>Case 4: Legitimate Condition: invoke each valid algorithm\n");
    for (id=(pmalgo_id)0; id<PMALGO_NUM_ALGORITHMS; id++) {
      fprintf(stderr,"\n==>Subcase: for algorithm %s\n",pmalgo_id_name[id]);
      process_algorithm(id);

    } /* loop for each algorithm */
    break;

  case 5:

    /* 
     * Case 5: Legitimate condition: Valid algorithm, NULL tbs_basename
     *   Call init_pmtbs for a valid algorithm and NULL tbs_basename, 
     *   and check returned object
     *   Expected result:
     *     returned pointer is non-NULL
     *     initializes appropriate TB grids
     */
    fprintf(stderr,"\n==>Case 5: Legitimate Condition: NULL tbs_basename\n");
    init_pmtbs_for(PMALGO_NAGLER_DEPTH,TRUE);

    break;

  default:
    fprintf(stderr,"Invalid case %d\n",case_num);
    exit(EXIT_FAILURE);
  }

  fprintf(stderr,"End of %s\n", ident_message);
  exit(EXIT_SUCCESS);

}

/*----------------------------------------------------------------------
 *  init_pmtbs_for - initialize and test a pmtbs object
 * 
 *     input : id - algorithm to test
 *             null_basename - set basename to NULL, otherwise use a
 *               legitimate one
 * 
 *    effect: all errors are fatal
 *----------------------------------------------------------------------*/
static void init_pmtbs_for(pmalgo_id id, bool null_basename) {

  int num_chans[PMALGO_NUM_ALGORITHMS] = { 2, 2, 2, 3, 3, 4, 4, 4, 2, 2, 3,
					   2, 10, 7 };

  pmalgo_sensor expected_sensor;
  char **expected_channels;
  char gpd_file[FILENAME_MAX];
  char base_file[FILENAME_MAX];
  pmtbs_class *pmtbs;
  grid_class *grid;
  char *id_name;
  int i;
  int expected_num_channels;
  bool ok=TRUE;
  bool expected_channel[PMALGO_NUM_CHANNELS];

  for (i = 0; i < PMALGO_NUM_CHANNELS; i++) expected_channel[i] = FALSE;

  switch (id) {
  case PMALGO_NSIDC_M1_SWE:
    expected_sensor = PMALGO_SMMR;
    expected_channel[PMALGO_SMMR_18H] = TRUE;
    expected_channel[PMALGO_SMMR_37H] = TRUE;
    break;
  case PMALGO_NSIDC_S1_SWE:
    expected_sensor = PMALGO_SSMI;
    expected_channel[PMALGO_SSMI_19H] = TRUE;
    expected_channel[PMALGO_SSMI_37H] = TRUE;
    break;
  case PMALGO_GW_SWE:
    expected_sensor = PMALGO_SSMI;
    expected_channel[PMALGO_SSMI_19V] = TRUE;
    expected_channel[PMALGO_SSMI_37V] = TRUE;
    break;
  case PMALGO_GW_WET_SWE:
    expected_sensor = PMALGO_SSMI;
    expected_channel[PMALGO_SSMI_19V] = TRUE;
    expected_channel[PMALGO_SSMI_37V] = TRUE;
    expected_channel[PMALGO_SSMI_37H] = TRUE;
    break;
  case PMALGO_GW_MODWET_SWE:
    expected_sensor = PMALGO_SSMI;
    expected_channel[PMALGO_SSMI_19V] = TRUE;
    expected_channel[PMALGO_SSMI_37V] = TRUE;
    expected_channel[PMALGO_SSMI_37H] = TRUE;
    break;
  case PMALGO_NTN_SEAICE:
    expected_sensor = PMALGO_SSMI;
    expected_channel[PMALGO_SSMI_19V] = TRUE;
    expected_channel[PMALGO_SSMI_19H] = TRUE;
    expected_channel[PMALGO_SSMI_22V] = TRUE;
    expected_channel[PMALGO_SSMI_37V] = TRUE;
    break;
  case PMALGO_NTS_SEAICE:
    expected_sensor = PMALGO_SSMI;
    expected_channel[PMALGO_SSMI_19V] = TRUE;
    expected_channel[PMALGO_SSMI_19H] = TRUE;
    expected_channel[PMALGO_SSMI_22V] = TRUE;
    expected_channel[PMALGO_SSMI_37V] = TRUE;
    break;
  case PMALGO_NTN_MYICE:
    expected_sensor = PMALGO_SSMI;
    expected_channel[PMALGO_SSMI_19V] = TRUE;
    expected_channel[PMALGO_SSMI_19H] = TRUE;
    expected_channel[PMALGO_SSMI_22V] = TRUE;
    expected_channel[PMALGO_SSMI_37V] = TRUE;
    break;
  case PMALGO_SMMR_XPGR:
    expected_sensor = PMALGO_SMMR;
    expected_channel[PMALGO_SMMR_18H] = TRUE;
    expected_channel[PMALGO_SMMR_37V] = TRUE;
    break;
  case PMALGO_SSMI_XPGR:
    expected_sensor = PMALGO_SSMI;
    expected_channel[PMALGO_SSMI_19H] = TRUE;
    expected_channel[PMALGO_SSMI_37V] = TRUE;
    break;
  case PMALGO_NAGLER_DEPTH:
    expected_sensor = PMALGO_SSMI;
    expected_channel[PMALGO_SSMI_19V] = TRUE;
    expected_channel[PMALGO_SSMI_37V] = TRUE;
    expected_channel[PMALGO_SSMI_85V] = TRUE;
    break;
  case PMALGO_NAGLER_CSI:
    expected_sensor = PMALGO_SSMI;
    expected_channel[PMALGO_SSMI_37V] = TRUE;
    expected_channel[PMALGO_SSMI_85V] = TRUE;
    break;
  case PMALGO_SMMR_ALL_CHANNELS:
    expected_sensor = PMALGO_SMMR;
    expected_channel[PMALGO_SMMR_06V] = TRUE;
    expected_channel[PMALGO_SMMR_06H] = TRUE;
    expected_channel[PMALGO_SMMR_10V] = TRUE;
    expected_channel[PMALGO_SMMR_10H] = TRUE;
    expected_channel[PMALGO_SMMR_18V] = TRUE;
    expected_channel[PMALGO_SMMR_18H] = TRUE;
    expected_channel[PMALGO_SMMR_21V] = TRUE;
    expected_channel[PMALGO_SMMR_21H] = TRUE;
    expected_channel[PMALGO_SMMR_37V] = TRUE;
    expected_channel[PMALGO_SMMR_37H] = TRUE;
    break;
  case PMALGO_SSMI_ALL_CHANNELS:
    expected_sensor = PMALGO_SSMI;
    expected_channel[PMALGO_SSMI_19V] = TRUE;
    expected_channel[PMALGO_SSMI_19H] = TRUE;
    expected_channel[PMALGO_SSMI_22V] = TRUE;
    expected_channel[PMALGO_SSMI_37V] = TRUE;
    expected_channel[PMALGO_SSMI_37H] = TRUE;
    expected_channel[PMALGO_SSMI_85V] = TRUE;
    expected_channel[PMALGO_SSMI_85H] = TRUE;
    break;
  default:
    assert(0);
  }

  expected_num_channels = num_chans[id];

  if (PMALGO_SMMR == expected_sensor) {
    sprintf(gpd_file,"%s",NORTH_GPD_FILE);
    sprintf(base_file,"%s",NORTH_SMMR_BASEFILE);
  } else if (PMALGO_NTS_SEAICE == id) {
    sprintf(gpd_file,"%s",SOUTH_GPD_FILE);
    sprintf(base_file,"%s",SOUTH_SSMI_BASEFILE);
  } else {
    sprintf(gpd_file,"%s",NORTH_GPD_FILE);
    sprintf(base_file,"%s",NORTH_SSMI_BASEFILE);
  }

  if (null_basename) {
    *base_file = NULL;
  }

  grid = init_grid(gpd_file);

  id_name = strdup(pmalgo_id_name[id]);
  pmtbs = init_pmtbs(grid,
		     id_name,
                     "",
		     base_file,
		     TRUE);
  free(id_name);
  if (NULL != pmtbs) {
    
    /* Check the pieces of the returned class */
    if (NULL != strstr(pmtbs->grid->gpd_filename,gpd_file)) {
      fprintf(stderr,"  correct gpd=%s\n",gpd_file);
    } else {
      ok = FALSE;
      fprintf(stderr,"  ERROR in gpd, expected=%s, got=%s\n",
	      gpd_file,pmtbs->grid->gpd_filename);
    }

    if (NULL == base_file) {
      if (NULL == pmtbs->tbs_basename) {
	fprintf(stderr,"  correct NULL base_file\n");
      } else {
	ok = FALSE;
	fprintf(stderr,"  ERROR, object's base filename should be NULL\n");
      }
    } else if (NULL != base_file) {
      if (0 == strcmp(pmtbs->tbs_basename,base_file)) {
	fprintf(stderr,"  correct base_file=%s\n",base_file);
      } else {
	ok = FALSE;
	fprintf(stderr,"  ERROR in base filename: expected=%s, got=%s\n",
		base_file,pmtbs->tbs_basename);
      }
    }

    if (pmtbs->algorithm->id == id) {
      fprintf(stderr,"  correct algorithm=%s\n",
	      pmalgo_id_name[id]);
    } else {
      ok = FALSE;
      fprintf(stderr,"  ERROR in algorithm: expected=%s, got=%s\n",
	      pmalgo_id_name[id],
	      pmalgo_id_name[pmtbs->algorithm->id]);
    }

    if (expected_sensor == pmtbs->algorithm->sensor) {
      fprintf(stderr,"  correct sensor=%s\n",
	      pmalgo_sensor_name[expected_sensor]);
    } else {
      ok = FALSE;
      fprintf(stderr,"  ERROR in sensor: expected=%s, got=%s\n",
	      pmalgo_sensor_name[expected_sensor],
	      pmalgo_sensor_name[pmtbs->algorithm->sensor]);
    }

    if (expected_num_channels == pmtbs->algorithm->num_channels) {
      fprintf(stderr,"  correct num_channels=%d\n",pmtbs->algorithm->num_channels);

      for (i = 0; i < pmtbs->algorithm->num_channels; i++) {
	if (expected_channel[pmtbs->algorithm->channel[i]]) {
	  fprintf(stderr,"  correct channel=%s\n",
		  pmalgo_channel_name[pmtbs->algorithm->channel[i]]);
	} else {
	  ok = FALSE;
	  fprintf(stderr,"  ERROR channel list: didn't expect but got =%s\n",
		  pmalgo_channel_name[pmtbs->algorithm->channel[i]]);
	}
      } /* end loop for each channel in pmalgorithm list of channels */

      for (i = 0; i < PMALGO_NUM_CHANNELS; i++) {
	if (NULL != pmtbs->data[i]) {
	  if (expected_channel[i]) {
	    fprintf(stderr,"  correct non-null data pointer for channel=%s\n",
		    pmalgo_channel_name[i]);
	  } else {
	    ok = FALSE;
	    fprintf(stderr,"  ERROR, found NULL data pointer for channel=%s\n",
		    pmalgo_channel_name[i]);
	  }
	} else { /* the pointer for this channel's data is NULL */
	  if (!expected_channel[i]) {
	    fprintf(stderr,"  correct NULL data pointer for channel=%s\n",
		    pmalgo_channel_name[i]);
	  } else {
	    ok = FALSE;
	    fprintf(stderr,"  ERROR, found non-null data pointer for channel=%s\n",
		    pmalgo_channel_name[i]);
	  }
	}
      }
    } else {
      ok = FALSE;
      fprintf(stderr,"  ERROR in num_channels: expected=%d, got=%d\n",
	      expected_num_channels,pmtbs->algorithm->num_channels);
    }

    fprintf(stderr,"num_days       = %d\n",pmtbs->algorithm->num_days);
    fprintf(stderr,"pass_direction = %s\n",pmtbs->algorithm->direction);
    fprintf(stderr,"grid_algorithm = %s\n",pmtbs->algorithm->grid_algorithm);

    if (ok) {
      fprintf(stderr,"SUCCESS:  init_pmtbs returned correct TB grids object\n");
    } else {
      fprintf(stderr,"ERROR:  init_pmtbs returned incorrect object.\n");
      close_pmtbs(pmtbs);
      exit(EXIT_FAILURE);
    }
  } else {

    fprintf(stderr,"ERROR:  init_pmtbs returned NULL pointer\n");
    exit(EXIT_FAILURE);

  } /* end else (it returned the NULL pointer) */

  close_pmtbs(pmtbs);

  return;
}

/*----------------------------------------------------------------------
 *  process_algorithm - initialize and process a pmtbs object
 * 
 *     input : id - algorithm to test
 * 
 *    effect: any error is considered fatal
 *----------------------------------------------------------------------*/
static void process_algorithm(pmalgo_id id) {

  char gpd_file[FILENAME_MAX];
  char out_file[FILENAME_MAX];
  char compare_file[FILENAME_MAX];
  char base_file[FILENAME_MAX];
  char *id_name;
  pmtbs_class *pmtbs;
  grid_class *grid;
  int row,col;
  long int num;
  float **data;
  bool valid_tbs;

  if (PMALGO_SMMR_ALL_CHANNELS == id || PMALGO_SSMI_ALL_CHANNELS == id) {
    return;
  }
  
  if (PMALGO_NSIDC_M1_SWE == id || PMALGO_SMMR_XPGR == id) {
    sprintf(gpd_file,"%s",NORTH_GPD_FILE);
    sprintf(base_file,"%s",NORTH_SMMR_BASEFILE);
  } else if (PMALGO_NAGLER_DEPTH == id || PMALGO_NAGLER_CSI == id) {
    sprintf(gpd_file,"%s",NORTH_GPD_FILE);
    sprintf(base_file,"%s",NORTH_SSMI_BASEFILE2);
  } else {
    if (PMALGO_NTS_SEAICE == id) {
      sprintf(gpd_file,"%s",SOUTH_GPD_FILE);
      sprintf(base_file,"%s",SOUTH_SSMI_BASEFILE);
    } else {
      sprintf(gpd_file,"%s",NORTH_GPD_FILE);
      sprintf(base_file,"%s",NORTH_SSMI_BASEFILE);
    }
  }
  
  grid = init_grid(gpd_file);
      
  /* Initialize filenames, pmtbs object, and output matrix */
  sprintf(out_file,"test/test_pmtbs.%s.test_data",
	  pmalgo_id_name[id]);

  id_name = strdup(pmalgo_id_name[id]);
  pmtbs = init_pmtbs(grid,
		     id_name,
                     "",
		     base_file,
		     TRUE);
  free(id_name);
  if (NULL == pmtbs) {
    fprintf(stderr,"ERROR:  init_pmtbs returned NULL pointer\n");
    goto error_exit;
  }

  data = (float **)initialize_matrix(pmtbs->grid,
				     sizeof(**data),
				     NULL,
				     "algorithm output",
				     TRUE);
  if (NULL == data) {
    fprintf(stderr,"ERROR:  initialize_matrix returned NULL pointer\n");
    goto error_exit;
  }  

  sprintf(compare_file,"%s%s.compare_data",COMPARE_DIR,
	  pmalgo_id_name[id]);

  /* Call transform for each matrix cell */
  num = 0;
  for (row = 0; row < pmtbs->grid->rows; row++) {
    for (col = 0; col < pmtbs->grid->cols; col++) {
      if (pmtbs->algorithm->run(pmtbs->algorithm->private,
                                pmtbs->data,col,row,
                                &(data[row][col]),
				&valid_tbs)) {
	num++;
      }
    }
  }
  fprintf(stderr,"Done running %s, calculated values for %ld pixels.\n",
	  pmalgo_id_name[id],num);

  /* Save the matrix for visual comparison to expected results */
  if (0 == write_matrix (out_file, (void **)data,
			 pmtbs->grid->rows,
			 pmtbs->grid->cols,
			 sizeof(**data))) {
    fprintf(stderr,"ERROR: Error writing result matrix to %s\n",out_file);
    goto error_exit;
  } else { 
    fprintf(stderr,"process_algorithm: wrote test matrix to :%s\n",
	   out_file);
    fprintf(stderr,"PARTIAL_SUCCESS:  complete success requires visual inspection\n");
    fprintf(stderr,">>Compare test file %s \n>>compare file %s\n\n\n",out_file,compare_file);
  }

  free(data);
  close_pmtbs(pmtbs);

  return; 

  error_exit:

  if (NULL != data) free(data);
  if (NULL != pmtbs) close_pmtbs(pmtbs);
  exit(EXIT_FAILURE);

} 

  

