################################################################################
#@***h* Sea_Ice_CDR/cdr_daily_seaice_test.py
#
# NAME
#  cdr_daily_seaice_test.py
#
# LOCATION
#  integration_tests/cdr_daily_seaice_test.py
#
# PURPOSE
#  unit and validation tests for cdr_daily_seaice.py
#
# DESCRIPTION
#  unit and validation tests for cdr_daily_seaice.py
#
# AUTHOR
#  nsidc@nsidc.org
#  National Snow & Ice Data Center, University of Colorado, Boulder
#
# COPYRIGHT
#  Copyright 2011 Regents of the University of Colorado. The Regents of the
#  University of Colorado grants, and you accept, a personal, nonexclusive,
#  nontransferable license to use this software, at no charge, in accordance
#  with the terms herein, solely for internal research and development, which
#  is limited to testing, measuring, assessing, evaluating the software and/or
#  data that you generate using this software, and not for any commercial
#  purpose. You may not develop derivative works based on this software, or
#  redistribute this software.
#
# CREATION DATE
#  Aug 2011
#
# EXTERNALS
#   __future__.division
#   __future__.print_function
#   __future__.absolute_import
#   os
#   subprocess
#   sys
#   unittest2
#   cdr_lib
#   cdr_configuration.configuration_dictionary
#   cdr_daily_seaice
#
# SUBROUTINES
#   class SequenceFunctionsTest
#     setUp
#     test_north_fields
#     test_south_fields
#
# INPUTS
#  n/a
#
# OUTPUTS
#  n/a
#
# MODIFICATION HISTORY
#
#
#@*****
################################################################################

# import a few __future__ features to keep possible upgrades cleaner
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import


import os
import subprocess
import sys
import unittest2 as unittest
import cdr_lib as cl


from cdr_configuration import configuration_dictionary
import cdr_daily_seaice as cdr


class SequenceFunctionsTest(unittest.TestCase):

    def setUp(self):
        if (os.path.dirname(sys.argv[0]) != ""):
            self.root_dir = os.path.realpath(
                os.path.abspath(os.path.dirname(sys.argv[0])+"/.."))
        else:
            self.root_dir = os.path.realpath(os.path.abspath(".."))

        self.nconf = configuration_dictionary('n', '20070315')
        self.sconf = configuration_dictionary('s', '20070315')
        self.nconf["brightness_input"] = "%(CDR_root)s/test_data" % self.nconf
        self.sconf["brightness_input"] = "%(CDR_root)s/test_data" % self.sconf
        self.nconf["nt_final_dir"] = "%(CDR_root)s/test_data" % self.nconf
        self.sconf["nt_final_dir"] = "%(CDR_root)s/test_data" % self.sconf
        self.nconf["bt_final_dir"] = "%(CDR_root)s/test_data" % self.nconf
        self.sconf["bt_final_dir"] = "%(CDR_root)s/test_data" % self.sconf
        self.n_size = self.nconf["h_dims"][0] * self.nconf["h_dims"][1]
        self.s_size = self.sconf["h_dims"][0] * self.sconf["h_dims"][1]
        self.nconf["work"] = "%(CDR_root)s/test_data/work" % self.nconf
        self.sconf["work"] = "%(CDR_root)s/test_data/work" % self.sconf

    def test_north_fields(self):
        """Testing creation of combined sea-ice fields in northern hemisphere"""

        # Check northern hemisphere
        cdr.conf = self.nconf

        # Check Nasateam
        tb_root = cdr.get_rss_brightness_files_root(self.nconf)
        nt_command = cdr.make_raw_nasateam("%(CDR_root)s/test_data/test_n_nt"
                                           % self.nconf, tb_root)
        cl.run_and_check(nt_command)
        cdr.nasateam_grid_from_raw_grid("%(CDR_root)s/test_data/test_n_nt"
                                        % self.nconf)
        nt_test = cl.CDRVariable(
            "test_nt_var", self.nconf["h_dims"],
            nctype=cl.NC_BYTE, intype=cl.NC_BYTE,
            filters=[cdr.nasateam_value_filter, cl.recast]
        )
        nt_test.from_file("%(CDR_root)s/test_data/test_n_nt" % self.nconf)
        nt_test.to_file("%(CDR_root)s/test_data/test_n_nt_two" % self.nconf)
        test_call = (
            "cmp "
            "%(CDR_root)s/test_data/nt_for_cdr_20070315_north_f13 "
            "%(CDR_root)s/test_data/test_n_nt") % self.nconf
        print(test_call)
        self.assertEqual(subprocess.call(test_call, shell=True), 0,
                         ("testcall failed: %s" % test_call))

        # Check Bootstrap
        tb_root = cdr.get_rss_brightness_files_root(self.nconf)
        bt_command = cdr.make_raw_bootstrap("%(CDR_root)s/test_data/test_n_bt" %
                                            self.nconf, tb_root)
        cl.run_and_check(bt_command)
        test_call = ("cmp "
                     "%(CDR_root)s/test_data/bt_for_cdr_20070315_north_f13 "
                     "%(CDR_root)s/test_data/test_n_bt") % self.nconf
        print(test_call)
        self.assertEqual(subprocess.call(test_call, shell=True), 0,
                         ("testcall failed: %s" % test_call))

        # northern hemisphere, add coast
        coastFromNasateamMask = cl.cdr_mask_filter_factory(nt_test, use=[253])
        bt_test = cdr.bootstrap_grid_from_file(
            self.nconf["bt_var"],
            "%(CDR_root)s/test_data/test_n_bt" % self.nconf, coastFromNasateamMask
        )
        bt_test.tofile("%(CDR_root)s/test_data/test_n_bt_two" % self.nconf)

        # Check combo field
        cdr.create_combo_concentration(
            self.nconf["legacy_combo_var"],
            "%(CDR_root)s/test_data/test_n_combo" % self.nconf,
            "qa_of_seaice_conc_cdr", "%(CDR_root)s/test_data/test_n_qa" % self.nconf,
            nt_test, bt_test)
        # cmdstring loads field being tested, do not run here
        test_call = (
            "cmp "
            "%(CDR_root)s/test_data/combo_cdr_20070315_north_f13 "
            "%(CDR_root)s/test_data/test_n_combo"
        ) % self.nconf
        print(test_call)
        self.assertEqual(subprocess.call(test_call, shell=True), 0)
        # QA field is generated with combo field, test now
        test_call = (
            "cmp "
            "%(CDR_root)s/test_data/combo_cdr_qa_20070315_north_f13 "
            "%(CDR_root)s/test_data/test_n_qa"
        ) % self.nconf
        print(test_call)
        self.assertEqual(subprocess.call(test_call, shell=True), 0)

        # cleanup
        os.unlink("%(CDR_root)s/test_data/test_n_bt" % self.nconf)
        os.unlink("%(CDR_root)s/test_data/test_n_nt" % self.nconf)
        os.unlink("%(CDR_root)s/test_data/test_n_combo" % self.nconf)
        os.unlink("%(CDR_root)s/test_data/test_n_qa" % self.nconf)

    def test_south_fields(self):
        """Testing creation of combined sea-ice fields in southern hemisphere"""

        # Check southern hemisphere
        cdr.conf = self.sconf

        # Check Nasateam
        tb_root = cdr.get_rss_brightness_files_root(self.sconf)
        nt_command = cdr.make_raw_nasateam("%(CDR_root)s/test_data/test_nt" % self.sconf,
                                           tb_root)
        cl.run_and_check(nt_command)
        cdr.nasateam_grid_from_raw_grid("%(CDR_root)s/test_data/test_nt" % self.sconf)
        nt_test = cl.CDRVariable(
            "test_nt_var", self.sconf["h_dims"],
            nctype=cl.NC_BYTE, intype=cl.NC_BYTE,
            filters=[cdr.nasateam_value_filter,
                     cl.recast])
        nt_test.from_file("%(CDR_root)s/test_data/test_nt" % self.sconf)
        nt_test.to_file("%(CDR_root)s/test_data/test_nt_two" % self.sconf)
        test_call = (
            "cmp "
            "%(CDR_root)s/test_data/nt_for_cdr_20070315_south_f13 "
            "%(CDR_root)s/test_data/test_nt"
        ) % self.sconf
        print(test_call)
        self.assertEqual(subprocess.call(test_call, shell=True), 0)

        # Check Bootstrap
        tb_root = cdr.get_rss_brightness_files_root(self.sconf)
        bt_command = cdr.make_raw_bootstrap("%(CDR_root)s/test_data/test_bt"
                                            % self.sconf, tb_root)
        cl.run_and_check(bt_command)
        test_call = (
            "cmp "
            "%(CDR_root)s/test_data/bt_for_cdr_20070315_south_f13 "
            "%(CDR_root)s/test_data/test_bt"
        ) % self.sconf
        print(test_call)
        self.assertEqual(subprocess.call(test_call, shell=True), 0,
                         ("testcall failed: %s" % test_call))

        # southern hemisphere

        bt_test = cdr.bootstrap_grid_from_file(
            self.sconf["bt_var"],
            "%(CDR_root)s/test_data/test_bt" % self.sconf
        )
        bt_test.tofile("%(CDR_root)s/test_data/test_bt_two" % self.sconf)

        # Check combo field
        cdr.create_combo_concentration(
            self.sconf["legacy_combo_var"],
            "%(CDR_root)s/test_data/test_combo" % self.sconf,
            "qa_of_seaice_conc_cdr", "%(CDR_root)s/test_data/test_qa" % self.sconf,
            nt_test, bt_test)
        # cmdstring loads field being tested, do not run here
        test_call = ("cmp "
                     "%(CDR_root)s/test_data/combo_cdr_20070315_south_f13 "
                     "%(CDR_root)s/test_data/test_combo") % self.sconf
        print(test_call)
        self.assertEqual(subprocess.call(test_call, shell=True), 0)
        # QA field is generated with combo field, test now
        test_call = ("cmp "
                     "%(CDR_root)s/test_data/combo_cdr_qa_20070315_south_f13 "
                     "%(CDR_root)s/test_data/test_qa") % self.sconf
        print(test_call)
        self.assertEqual(subprocess.call(test_call, shell=True), 0)

        # cleanup
        os.unlink("%(CDR_root)s/test_data/test_bt" % self.sconf)
        os.unlink("%(CDR_root)s/test_data/test_nt" % self.sconf)
        os.unlink("%(CDR_root)s/test_data/test_combo" % self.sconf)


if __name__ == '__main__':
    unittest.main()
