################################################################################
#@***h* Sea_Ice_CDR/cdr_monthly_seaice_test.py
#
# NAME
#  cdr_monthly_seaice_test.py
#
# LOCATION
#  integration_tests/cdr_monthly_seaice_test.py
#
# PURPOSE
#  Integration tests for monthly CDR creation.
#
# DESCRIPTION
#  Integration tests for monthly CDR creation.
#
# AUTHOR
#  nsidc@nsidc.org
#  National Snow & Ice Data Center, University of Colorado, Boulder
#
# COPYRIGHT
#  Copyright 2012 Regents of the University of Colorado. The Regents of the
#  University of Colorado grants, and you accept, a personal, nonexclusive,
#  nontransferable license to use this software, at no charge, in accordance
#  with the terms herein, solely for internal research and development, which
#  is limited to testing, measuring, assessing, evaluating the software and/or
#  data that you generate using this software, and not for any commercial
#  purpose. You may not develop derivative works based on this software, or
#  redistribute this software.
#
# CREATION DATE
#  Jan 2012
#
# EXTERNALS
#  glob
#  numpy.testing
#  os
#  unittest2
#  netCDF4.Dataset
#  cdr_lib
#  cdr_configuration
#  cdr_add_melt
#  cdr_daily_seaice
#  cdr_monthly_seaice
#
#
# SUBROUTINES
#   class MonthlyCdrTest
#     generate_daily_data
#     setUpClass
#     test_QA_values
#     test_expected_month_files
#     test_file_has_correct_time_variable
#     test_month_of_cdr_daily_files
#     test_monthly_bt_field
#     test_monthly_nt_field
#     test_monthly_seaice_conc_monthly_cdr
#     test_monthly_standard_deviation_values
#     test_read_existing_cdr_file
#     test_time_coverage_fields
#     validate_monthly_file_has_proper_time_coverage_values
#     validate_ncdf_variable_matches_binary_file
#
#
# INPUTS
#  n/a
#
# OUTPUTS
#  n/a
#
# MODIFICATION HISTORY
#
#
#@*****
################################################################################

import glob
import numpy.testing as nptest
import os
import unittest2

from netCDF4 import Dataset

import cdr_lib as cl
import cdr_configuration as config
import cdr_add_melt as cam
import cdr_daily_seaice as cds
import cdr_monthly_seaice as cms


class MonthlyCdrTest(unittest2.TestCase):

    @classmethod
    def setUpClass(cls):
        """ Set member vars needed for tests. """

        cls.mon_test_dir = os.path.join(cl.cdr_root_dir(),
                                        "test_data",
                                        "monthly_cdr_test_data")

        test_cases = {'test1':  ["unused", 'north', '20060301'],
                      'test2':  ["unused", 'south', '20060301'],
                      'test3':  ["unused", 'north', '19800101']}

        cls.configuration = {}
        cls.cdrobj = {}
        cls.cdr_sample_outfile = {}

        cls.generate_daily_data()

        for tname, seed in test_cases.items():
            preset = {
                "cdr_version": cl.create_product_version_string("01", "00")
            }
            cls.configuration[tname] = config.create_configuration_dictionary(seed,
                                                                              preset)
            cls.cdrobj[tname] = cms.CDRMonthlySeaice(cls.configuration[tname])
            cls.cdr_sample_outfile[tname] = cls.cdrobj[tname].output_monthly_filename()
            cls.cdrobj[tname].process_monthly_cdr()

    @classmethod
    def generate_daily_data(cls):
        test_data = [(1980, 1, 32), (2006, 3, 62)]

        for year, month, number_of_files in test_data:
            files = glob.glob("output/seaice_conc_daily_*%04d%02d*.nc" % (year, month))
            if len(files) != number_of_files:
                for d in range(1, 32):
                    cds.main(['cdr_daily_seaice.py',
                              'b',
                              '%04d%02d%02d' % (year, month, d),
                              '01', '00'])
                if month >= 3:
                    cam.add_melt(year, '01', '00')

    def validate_ncdf_variable_matches_binary_file(self, params):
        """
        Read a field from a netcdf file and verify it is the same as the binary data.

        params => {
             'nc_filename': string,
             'nc_variable_name': string,
             'nc_file_data_type': string,
             'bin_filename': string,
             'bin_file_data_type': optional string (default 'uint8' [NC_BYTE])
             'hemi': string}
        """
        if 'bin_file_data_type' in params:
            bin_file_data_type = params['bin_file_data_type']
        else:
            bin_file_data_type = cl.NC_BYTE

        actual_array = cl.fetch_nc_grid(
            params['nc_filename'],
            params['nc_variable_name'],
            params['nc_file_data_type']
        )

        expected_array = cl.read_bin_grid(params['bin_filename'],
                                          params['hemi'],
                                          file_data_type=bin_file_data_type)

        nptest.assert_array_almost_equal(
            actual_array, expected_array,
            err_msg=("failed to match file %s" % params['bin_filename']))

    def validate_monthly_file_has_proper_time_coverage_values(self, filename):
        """
        Read a netCDF file and verify that the time_coverage_duration
        and time_coverage_resolution fields are both "P1M"

        file = the netCDF file to check
        """

        ncroot = Dataset(filename, 'r')
        duration = ncroot.getncattr('time_coverage_duration')
        resolution = ncroot.getncattr('time_coverage_resolution')

        self.assertTrue(duration == 'P1M', filename +
                        "as an incorrect time_coverage_duration value.")
        self.assertTrue(resolution == 'P1M', filename +
                        " has an incorrect time_coverage_resolution value.")
        ncroot.close()

    def test_expected_month_files(self):
        """
        Test that we are getting the expected monthly files from the configuration
        """

        conf = config.configuration_dictionary('n', '20070315')
        boot_final_base = os.path.normpath(conf['bootstrap_final_base'])
        nt_final_base = os.path.normpath(conf['nasateam_final_base'])
        hemi_params = {
            'test1': {'expected_bt_file': '%s/north/monthly/bt_200603_f13_v02_n.bin'
                      % boot_final_base,
                      'expected_nt_file': '%s/north/monthly/nt_200603_f13_v01_n.bin'
                      % nt_final_base},
            'test2': {'expected_bt_file': '%s/south/monthly/bt_200603_f13_v02_s.bin'
                      % boot_final_base,
                      'expected_nt_file': '%s/south/monthly/nt_200603_f13_v01_s.bin'
                      % nt_final_base}
        }

        for tname, params in hemi_params.items():
            expected_bt_file = params['expected_bt_file']
            expected_nt_file = params['expected_nt_file']

            self.assertEqual(expected_bt_file,
                             cl.find_matching_bt_month_file(self.configuration[tname]))
            self.assertEqual(expected_nt_file,
                             cl.find_matching_nt_month_file(self.configuration[tname]))

    def test_monthly_nt_field(self):
        """
        Compare the output NT monthly field matches the expected generated externally
        """

        hemi_params = {
            'test1': {
                'expected_base':  'expected_nt_field_N200603.byte',
                'hemi_initial': 'N'},
            'test2': {
                'expected_base':  'expected_nt_field_S200603.byte',
                'hemi_initial': 'S'}}

        for tname, hparams in hemi_params.items():
            nc_file = self.cdr_sample_outfile[tname]
            expected_file = os.path.join(self.mon_test_dir, hparams['expected_base'])
            params = {'nc_filename': nc_file,
                      'nc_variable_name': 'goddard_nt_seaice_conc_monthly',
                      'nc_file_data_type': cl.NC_BYTE,
                      'bin_filename': expected_file,
                      'bin_file_data_type': cl.NC_BYTE,
                      'hemi': hparams['hemi_initial']}
            self.validate_ncdf_variable_matches_binary_file(params)

    def test_monthly_bt_field(self):
        """
        Compare the output BT (bootstrap) monthly field matches the expected
        generated externally
        """

        hemi_params = {
            'test1': {
                'expected_base':  'expected_bt_field_N200603.byte',
                'hemi_initial': 'N'},
            'test2': {
                'expected_base':  'expected_bt_field_S200603.byte',
                'hemi_initial': 'S'}}

        for tname, hparam in hemi_params.items():
            nc_file = self.cdr_sample_outfile[tname]
            expected_file = os.path.join(self.mon_test_dir, hparam['expected_base'])
            params = {'nc_filename': nc_file,
                      'nc_variable_name': 'goddard_bt_seaice_conc_monthly',
                      'nc_file_data_type': cl.NC_BYTE,
                      'bin_filename': expected_file,
                      'bin_file_data_type': cl.NC_BYTE,
                      'hemi': hparam['hemi_initial']}
            self.validate_ncdf_variable_matches_binary_file(params)

    def test_monthly_standard_deviation_values(self):
        """
        Compare the Standard Deviation field to expected values.
        """

        hemi_params = {
            'test1': {
                'expected_base':  'cdr_std_200603N.float',
                'hemi_initial': 'N'},
            'test2': {
                'expected_base':  'cdr_std_200603S.float',
                'hemi_initial': 'S'},
            'test3': {
                'expected_base':  'cdr_std_198001N.float',
                'hemi_initial': 'N'}
        }

        for tname, hparam in hemi_params.items():

            nc_file = self.cdr_sample_outfile[tname]
            expected_file = os.path.join(self.mon_test_dir, hparam['expected_base'])
            params = {'nc_filename': nc_file,
                      'nc_variable_name': 'stdev_of_seaice_conc_monthly_cdr',
                      'nc_file_data_type': cl.NC_FLOAT,
                      'bin_filename': expected_file,
                      'bin_file_data_type': cl.NC_FLOAT,
                      'hemi': hparam['hemi_initial']}

            self.validate_ncdf_variable_matches_binary_file(params)

    def test_QA_values(self):
        """
        Compare QA calculated values to expected values.
        """

        hemi_params = {
            'test1': {
                'expected_base': 'expected_QA_field_N200603.byte',
                'hemi_initial': 'N'},
            'test2': {
                'expected_base': 'expected_QA_field_S200603.byte',
                'hemi_initial': 'S'},
            'test3': {
                'expected_base': 'expected_QA_field_N198001.byte',
                'hemi_initial': 'N'},
        }

        for tname, hparam in hemi_params.items():
            nc_file = self.cdr_sample_outfile[tname]
            expected_file = os.path.join(self.mon_test_dir, hparam['expected_base'])

            params = {'nc_filename': nc_file,
                      'nc_variable_name': 'qa_of_seaice_conc_monthly_cdr',
                      'nc_file_data_type': cl.NC_BYTE,
                      'bin_filename': expected_file,
                      'hemi': hparam['hemi_initial']}

            self.validate_ncdf_variable_matches_binary_file(params)

    def test_monthly_seaice_conc_monthly_cdr(self):
        """
        Compare computed Monthly Sea Ice values to expected values.
        """

        hemi_params = {
            'test3': {
                'expected_base': 'expected_monthly_concentration_N198001.byte',
                'hemi_initial': 'N'},
        }

        for tname, hparam in hemi_params.items():
            nc_file = self.cdr_sample_outfile[tname]
            expected_file = os.path.join(self.mon_test_dir, hparam['expected_base'])

            params = {'nc_filename': nc_file,
                      'nc_variable_name': 'seaice_conc_monthly_cdr',
                      'nc_file_data_type': cl.NC_BYTE,
                      'bin_filename': expected_file,
                      'hemi': hparam['hemi_initial']}

            self.validate_ncdf_variable_matches_binary_file(params)

    def test_file_has_correct_time_variable(self):
        """
        Testing to make sure we are writing the correct number
        of days since NOAA epoch time (1601 01 01).
        """

        nc_file = self.cdr_sample_outfile['test1']
        var_name = 'time'
        data_type = cl.NC_DOUBLE
        generated_value = cl.fetch_nc_grid(nc_file, var_name, data_type)
        expected_elapsed_days = 147982
        self.assertEqual(expected_elapsed_days, generated_value)

    def test_read_existing_cdr_file(self):
        """
        Read a netcdf CDR daily file and make sure it contains the same values as the
        binary equivalent created with IDL from the same file.

        All this tests is do I know how to use the python-netcdf4 library.
        """

        nc_file = os.path.join(self.mon_test_dir,
                               'seaice_conc_daily_nh_f13_20050917_v01r00.nc')
        bin_file = os.path.join(self.mon_test_dir,
                                'seaice_conc_daily_nh_f13_20050917.bin')
        nc_array = cl.fetch_nc_grid(nc_file, 'seaice_conc_cdr', cl.NC_BYTE)
        bin_array = cl.read_bin_grid(bin_file, 'N')
        nptest.assert_array_almost_equal(nc_array, bin_array)

    def test_month_of_cdr_daily_files(self):
        """
        Return an array of a month of cdr files.
        It will use a conf object that is initialized with a date. These files will be all
        of the files that match the Year and Month of the input YYYYMMDD.
        """

        results = []
        for d in range(1, 32, 2):
            results.append('seaice_conc_daily_nh_n07_198001%02d_v01r00.nc' % d)

        conf = config.configuration_dictionary('N', '19800101')
        files = cl.find_month_of_daily_cdr_netcdf_files(conf)
        file_bases = [os.path.basename(x) for x in files]
        self.assertEqual(results, file_bases)

    def test_time_coverage_fields(self):
        """
        Test to make sure the north and south files have the time coverage fields set
        correctly.
        """

        cases = ['test1', 'test2']

        for case in cases:
            nc_file = self.cdr_sample_outfile[case]
            self.validate_monthly_file_has_proper_time_coverage_values(nc_file)


if __name__ == '__main__':
    unittest2.main()
