################################################################################
#@***h* Sea_Ice_CDR/cdr_stddev_test.py
#
# NAME
#  cdr_stddev_test.py
#
# LOCATION
#  integration_tests/cdr_stddev_test.py
#
# PURPOSE
#  integration tests for the cdr standard deviation code
#
# DESCRIPTION
#  Test case for making sure standard deviation computations match the expected.
#
# AUTHOR
#  nsidc@nsidc.org
#  National Snow & Ice Data Center, University of Colorado, Boulder
#
# COPYRIGHT
#  Copyright 2011 Regents of the University of Colorado. The Regents of the
#  University of Colorado grants, and you accept, a personal, nonexclusive,
#  nontransferable license to use this software, at no charge, in accordance
#  with the terms herein, solely for internal research and development, which
#  is limited to testing, measuring, assessing, evaluating the software and/or
#  data that you generate using this software, and not for any commercial
#  purpose. You may not develop derivative works based on this software, or
#  redistribute this software.
#
# CREATION DATE
#  Aug 2011
#
# EXTERNALS
#    unittest2
#    os
#    numpy
#    numpy.testing
#    cdr_lib
#
# SUBROUTINES
#
#   class CDRStddevTests
#        setUp
#        test_sample_data
#
# INPUTS
#  n/a
#
# OUTPUTS
#  n/a
#
# MODIFICATION HISTORY
#
#
#@*****
################################################################################

import unittest2 as unittest
import os
import numpy as np
import numpy.testing as nptest

import cdr_lib as cl


class CDRStddevTests(unittest.TestCase):
    """ TestCase class for developing the CDR Standard Deviation Codes """

    def setUp(self):
        """ Set up some necessary variables """

        # Read the reference stddev data
        ref_file = os.path.join(cl.cdr_root_dir(),
                                "test_data", "stddev_ntbt_20070315_N.bin")
        self.nref_data = np.fromfile(ref_file, dtype=cl.NC_FLOAT, sep="")
        self.nref_data = np.reshape(self.nref_data, (448, 304))

        ref_file = os.path.join(cl.cdr_root_dir(),
                                "test_data", "stddev_ntbt_20070315_S.bin")
        self.sref_data = np.fromfile(ref_file, dtype=cl.NC_FLOAT, sep="")
        self.sref_data = np.reshape(self.sref_data, (332, 316))

    def test_sample_data(self):
        """ Read in sample data for 20070315 and compute the STD field and
        compare to the IDL computed version in the test_data directory """

        hemis = ['N', 'S']
        shapes = [(448, 304), (332, 316)]

        for hemi, shape in zip(hemis, shapes):
            nt_file = os.path.join(
                cl.cdr_root_dir(), "test_data", "nt_stdbase_20070315_%s.bin" % hemi)
            bt_file = os.path.join(
                cl.cdr_root_dir(), "test_data", "bt_stdbase_20070315_%s.bin" % hemi)

            bt_data = np.fromfile(bt_file, dtype=cl.NC_FLOAT, sep="")
            bt_data = np.reshape(bt_data, shape)
            nt_data = np.fromfile(nt_file, dtype=cl.NC_FLOAT, sep="")
            nt_data = np.reshape(nt_data, shape)
            stddev = cl.local_std_dev(bt_data, nt_data, -1)
            if hemi == 'N':
                ref = self.nref_data
            else:
                ref = self.sref_data

            nptest.assert_array_almost_equal(stddev, ref, decimal=5)

if __name__ == '__main__':
    unittest.main()
