################################################################################
#@***h* Sea_Ice_CDR/daily_metadata_tests.py
#
# NAME
#  daily_metadata_tests.py
#
# LOCATION
#  integration_tests/daily_metadata_tests.py
#
# PURPOSE
#  Daily CDR Integration tests.
#
# DESCRIPTION
#  simple integration tests
#
# AUTHOR
#  nsidc@nsidc.org
#  National Snow & Ice Data Center, University of Colorado, Boulder
#
# COPYRIGHT
#  Copyright 2013 Regents of the University of Colorado. The Regents of the
#  University of Colorado grants, and you accept, a personal, nonexclusive,
#  nontransferable license to use this software, at no charge, in accordance
#  with the terms herein, solely for internal research and development, which
#  is limited to testing, measuring, assessing, evaluating the software and/or
#  data that you generate using this software, and not for any commercial
#  purpose. You may not develop derivative works based on this software, or
#  redistribute this software.
#
# CREATION DATE
#  Feb 2013
#
# EXTERNALS
#  common_metadata_tests.CommonMetadataTests
#  netCDF4.Dataset
#  datetime
#  os
#  unittest2
#
# SUBROUTINES
#   class DailyMetadataTests
#     setUpClass
#     tearDownClass
#     test_source_contains_nsidc0001
#     test_source_contains_nsidc0051
#     test_source_contains_nsidc0079
#   class MissingVarsDailyMetadataTests
#     test_source_contains_nsidc0001
#     test_source_does_not_contain_nsidc0001
#   class NorthernDailyMetadataTests
#   class SouthernDailyMetadataTests
#
# INPUTS
#  n/a
#
# OUTPUTS
#  n/a
#
# MODIFICATION HISTORY
#
#
#@*****
################################################################################

from common_metadata_tests import CommonMetadataTests
from netCDF4 import Dataset
import datetime as dt
import os
import unittest2


HEMI = {'north': 'n',
        'south': 's'}


class DailyMetadataTests(object):

    @classmethod
    def setUpClass(cls):
        cmd = "python src/cdr_daily_seaice.py {hemisphere} {yyyymmdd} 01 00".format(
            hemisphere=cls.hemisphere,
            yyyymmdd=cls.date.strftime("%Y%m%d")
        )
        os.system(cmd)
        cls.ncobj = Dataset(cls.DAILY_FILENAME)

    @classmethod
    def tearDownClass(cls):
        os.remove(cls.DAILY_FILENAME)

    def test_source_contains_nsidc0001(self):
        base = "ftp://sidads.colorado.edu/DATASETS/nsidc0001_polar_stereo_tbs"\
               "/{hemisphere}/{year}/tb_{platform}_{yyyymmdd}_v2_{hemi}{channel}.bin"

        channels = ['19v', '19h', '22v', '37v', '37h']
        expected = ", ".join([base.format(hemisphere=self.hemisphere,
                                          year=self.date.strftime("%Y"),
                                          platform=self.platform,
                                          yyyymmdd=self.date.strftime("%Y%m%d"),
                                          hemi=HEMI[self.hemisphere],
                                          channel=channel) for channel in channels])

        actual = getattr(self.ncobj, 'source')

        self.assertTrue(expected in actual, "{0} is not in {1}".format(expected, actual))

    def test_source_contains_nsidc0051(self):
        base = "ftp://sidads.colorado.edu/DATASETS/"\
               "nsidc0051_gsfc_nasateam_seaice/final-gsfc"\
               "/{hemisphere}/daily/{year}/nt_{yyyymmdd}_{platform}_v01_{hemi}.bin"
        expected = base.format(
            hemisphere=self.hemisphere,
            year=self.date.strftime("%Y"),
            platform=self.platform,
            yyyymmdd=self.date.strftime("%Y%m%d"),
            hemi=HEMI[self.hemisphere]
        )

        actual = getattr(self.ncobj, 'source')

        self.assertTrue(expected in actual, "{0} is not in {1}".format(expected, actual))

    def test_source_contains_nsidc0079(self):
        base = "ftp://sidads.colorado.edu/DATASETS"\
               "/nsidc0079_gsfc_bootstrap_seaice/final-gsfc"\
               "/{hemisphere}/daily/{year}/bt_{yyyymmdd}_{platform}_v02_{hemi}.bin"
        expected = base.format(
            hemisphere=self.hemisphere,
            year=self.date.strftime("%Y"),
            platform=self.platform,
            yyyymmdd=self.date.strftime("%Y%m%d"),
            hemi=HEMI[self.hemisphere]
        )

        actual = getattr(self.ncobj, 'source')

        self.assertTrue(expected in actual, "{0} is not in {1}".format(expected, actual))


class SouthernDailyMetadataTests(DailyMetadataTests, CommonMetadataTests,
                                 unittest2.TestCase):
    date = dt.date(2007, 1, 1)
    platform = 'f13'
    hemisphere = 'south'
    DAILY_FILENAME = 'output/seaice_conc_daily_sh_f13_20070101_v01r00.nc'


class NorthernDailyMetadataTests(DailyMetadataTests, CommonMetadataTests,
                                 unittest2.TestCase):
    date = dt.date(2007, 1, 1)
    platform = 'f13'
    hemisphere = 'north'
    DAILY_FILENAME = 'output/seaice_conc_daily_nh_f13_20070101_v01r00.nc'


class MissingVarsDailyMetadataTests(DailyMetadataTests, CommonMetadataTests,
                                    unittest2.TestCase):
    date = dt.date(1979, 3, 1)
    platform = 'n07'
    hemisphere = 'north'
    DAILY_FILENAME = 'output/seaice_conc_daily_nh_{platform}_{yyyymmdd}_v01r00.nc'.format(
        platform=platform,
        yyyymmdd=date.strftime("%Y%m%d")
    )

    def test_source_contains_nsidc0001(self):
        pass

    def test_source_does_not_contain_nsidc0001(self):
        source = getattr(self.ncobj, 'source')
        self.assertTrue("nsidc0001" not in source)


if __name__ == '__main__':
    unittest2.main()
