################################################################################
#@***h* Sea_Ice_CDR/monthly_metadata_tests.py
#
# NAME
#  monthly_metadata_tests.py
#
# LOCATION
#  integration_tests/monthly_metadata_tests.py
#
# PURPOSE
#  Monthly CDR Integration tests.
#
# DESCRIPTION
#  simple integration tests
#
# AUTHOR
#  nsidc@nsidc.org
#  National Snow & Ice Data Center, University of Colorado, Boulder
#
# COPYRIGHT
#  Copyright 2013 Regents of the University of Colorado. The Regents of the
#  University of Colorado grants, and you accept, a personal, nonexclusive,
#  nontransferable license to use this software, at no charge, in accordance
#  with the terms herein, solely for internal research and development, which
#  is limited to testing, measuring, assessing, evaluating the software and/or
#  data that you generate using this software, and not for any commercial
#  purpose. You may not develop derivative works based on this software, or
#  redistribute this software.
#
# CREATION DATE
#  Feb 2013
#
# EXTERNALS
#   common_metadata_tests.CommonMetadataTests
#   netCDF4.Dataset
#   datetime
#   glob
#   os
#   unittest2
#
# SUBROUTINES
#   class MonthlyMetadataTests
#     setUpClass
#     test_source
#   class SouthernMonthlyMetadataTests
#     test_missing_comment_when_DataIsAvailable_in_southern_hemisphere
#   class NorthernMonthlyMetadataTests
#     test_missing_comment_when_DataIsAvailable_in_northern_hemisphere
#   class MissingVarsMonthlyMetadataTests
#     test_missing_comment_when_DataIsMissing
#     test_missing_melt_comment_when_DataIsMissing
#
# INPUTS
#  n/a
#
# OUTPUTS
#  n/a
#
# MODIFICATION HISTORY
#
#
#@*****
################################################################################

from common_metadata_tests import CommonMetadataTests
from netCDF4 import Dataset
import datetime as dt
import glob
import os
import unittest2


class MonthlyMetadataTests(object):

    @classmethod
    def setUpClass(cls):
        cls.ncobj = Dataset(cls.MONTHLY_FILENAME)

    def test_source(self):
        expected = ""
        base = "ftp://sidads.colorado.edu/pub/DATASETS/NOAA" \
               "/G02202/{0}/daily/{1}/".format(
                   self.hemisphere,
                   self.the_date.strftime("%Y")
               )

        base_dir = "/projects/DATASETS/NOAA/G02202/{hemisphere}/daily/{yyyy}".format(
            hemisphere=self.hemisphere,
            yyyy=self.the_date.strftime("%Y"))

        months_files = glob.glob(os.path.join(base_dir,
                                              "seaice_conc_daily_*_{date}*.nc".format(
                                                  date=self.the_date.strftime("%Y%m"))))

        correct_files = [os.path.join(base, os.path.basename(mfile)) for
                         mfile in sorted(months_files)]
        expected = ", ".join(correct_files)

        actual = getattr(self.ncobj, 'source')
        self.assertEqual(expected, actual)


class SouthernMetadataTests(MonthlyMetadataTests, CommonMetadataTests,
                            unittest2.TestCase):
    hemisphere = 'south'
    platform = 'f13'
    the_date = dt.date(2006, 3, 1)
    MONTHLY_FILENAME = "output/seaice_conc_monthly_sh_{platform}_{date}_v01r00.nc".format(
        platform=platform,
        date=the_date.strftime("%Y%m"))

    def test_missing_comment_when_DataIsAvailable_in_southern_hemisphere(self):
        expected = "variable melt_onset_day_seaice_conc_monthly_cdr is not available"
        actual = getattr(self.ncobj, 'comment')
        self.assertTrue(expected in actual, "[{expected}] is not in [{actual}]".format(
            expected=expected, actual=actual)
        )


class NorthernMetadataTests(MonthlyMetadataTests, CommonMetadataTests,
                            unittest2.TestCase):
    hemisphere = 'north'
    platform = 'f13'
    the_date = dt.date(2006, 3, 1)
    MONTHLY_FILENAME = "output/seaice_conc_monthly_nh_{platform}_{date}_v01r00.nc".format(
        platform=platform,
        date=the_date.strftime("%Y%m"))

    def test_missing_comment_when_DataIsAvailable_in_northern_hemisphere(self):
        expected = "none"
        actual = getattr(self.ncobj, 'comment')
        self.assertTrue(expected in actual, "[{expected}] is not in [{actual}]".format(
            expected=expected, actual=actual)
        )


class MissingVarsMonthlyMetadataTests(MonthlyMetadataTests, CommonMetadataTests,
                                      unittest2.TestCase):

    hemisphere = 'north'
    platform = 'n07'
    the_date = dt.date(1980, 1, 1)
    MONTHLY_FILENAME = "output/seaice_conc_monthly_nh_{platform}_{date}_v01r00.nc".format(
        platform=platform,
        date=the_date.strftime("%Y%m"))

    def test_missing_comment_when_DataIsMissing(self):
        expected = "not available for this month."
        actual = getattr(self.ncobj, 'comment')
        self.assertTrue(expected in actual, "[{expected}] is not in [{actual}]".format(
            expected=expected, actual=actual)
        )

    def test_missing_melt_comment_when_DataIsMissing(self):
        expected = "melt_onset_day_seaice_conc_monthly_cdr"
        actual = getattr(self.ncobj, 'comment')
        self.assertTrue(expected in actual, "[{expected}] is not in [{actual}]".format(
            expected=expected, actual=actual)
        )

if __name__ == "__main__":
    unittest2.main()
