################################################################################
#@***h* Sea_Ice_CDR/cdr_add_melt.py
#
# NAME
#  cdr_add_melt.py
#
# LOCATION
#  src/cdr_add_melt.py
#
# PURPOSE
#  Process daily files, calculate melt-related variables, and add them
#  to the CDR file.
#
# DESCRIPTION
#  This module processes all daily CDR files and uses the functions
#  provided in the cdr_melt module to calculate melt-related
#  variables. Specifically it calculates the melt onset day variable
#  and the melt start detected flag in the QA variable. Running the module
#  directly with the year, version, and revision flag will attempt to
#  process all daily files for the given year, e.g.:
#
#    python src/cdr_add_melt.py 2008 02 00
#
# AUTHOR
#  nsidc@nsidc.org
#  National Snow & Ice Data Center, University of Colorado, Boulder
#
# COPYRIGHT
#  Copyright 2011 Regents of the University of Colorado. The Regents of the
#  University of Colorado grants, and you accept, a personal, nonexclusive,
#  nontransferable license to use this software, at no charge, in accordance
#  with the terms herein, solely for internal research and development, which
#  is limited to testing, measuring, assessing, evaluating the software and/or
#  data that you generate using this software, and not for any commercial
#  purpose. You may not develop derivative works based on this software, or
#  redistribute this software.
#
# CREATION DATE
#  March 2013
#
# EXTERNALS
#   argparse
#   cdr_configuration
#   cdr_lib
#   cdr_melt
#   glob
#   numpy
#   os
#
# SUBROUTINES
#   add_melt
#   brightness_temperatures
#   concentrations
#   initial_melt_onset_day_field
#   initially_detectable
#   lcs_mask
#   parse_arguments
#   update_cdr
#
# INPUTS
#  1. Daily CDR files for the given year, including the CDR file for
#     the first day of the melt season (required to calculate the grid
#     cells at which melt is detectable)
#  2. Daily brightness temperature files
#  3. Coastal mask file (ancillary file)
#
# OUTPUTS
#  Updated daily CDR files
#
# MODIFICATION HISTORY
#
#
#@*****
################################################################################
import argparse as ap
import cdr_configuration as cfg
import cdr_lib as cl
import cdr_melt as melt
import glob
import numpy as np
import os


CONSTANTS = {}
cfg.cdr_constants(CONSTANTS)
FIRST_DAY_OF_SEASON = CONSTANTS['start_of_northern_melt_season']
cfg.qc_flag_values(CONSTANTS)
MELT_BIT = int(CONSTANTS['melt_started'])


def add_melt(year, version, revision):
    preset = {
        "cdr_version": cl.create_product_version_string(version, revision)
    }
    previous_mod = initial_melt_onset_day_field()
    detectable = initially_detectable(year, preset)

    for doy in range(1, cl.days_in_year(year) + 1):
        conf = cfg.configuration_dictionary('n', cl.yyyymmdd(year, doy), preset)

        try:
            tb19 = brightness_temperatures(conf, year, doy, '19h')
            tb37 = brightness_temperatures(conf, year, doy, '37h')

            with cl.cdr_dataset(conf) as cdr:
                conc = concentrations(conf, cdr)
                melting = melt.melting(conc, tb19, tb37)
                current_mod = melt.melt_onset_day(melting, doy, previous_mod, detectable)
                melt_start_detected = melt.melt_start_detected(doy, current_mod, conc)

                update_cdr(conf, cdr, current_mod, melt_start_detected)

                previous_mod = current_mod

        except RuntimeError:
            print "Couldn't find input files for {0} {1}".format(year, doy)
            continue


def initial_melt_onset_day_field():
    imod = np.zeros(cl.north_dims(), dtype=cl.NC_SHORT)
    imod[:] = melt.MELT_ONSET_DAY_FILL_VALUE
    return imod


def initially_detectable(year, preset):
    conf = cfg.configuration_dictionary('n', cl.yyyymmdd(year, FIRST_DAY_OF_SEASON),
                                        preset)
    first_day_concentrations = concentrations(conf, cl.cdr_dataset(conf))
    return melt.melt_detectable(first_day_concentrations, lcs_mask(conf))


def concentrations(conf, cdr):
    conc = cdr.variables[conf['cdr_combo_var']]
    conc.set_auto_maskandscale(False)
    conc = conc[0]
    conc[conc == -1] = 0
    return conc


def lcs_mask(conf):
    lcs_mask = np.fromfile(
        conf["coastal_mask"], dtype=cl.NC_BYTE,
        count=conf["h_dims"][1]*conf["h_dims"][0])
    lcs_mask.shape = tuple(conf["h_dims"])
    return lcs_mask


def brightness_temperatures(conf, year, doy, channel):

    tb_filename = os.path.join(conf['brightness_input'],
                               "{bt_root}{channel}.bin".format(bt_root=conf['bt_root'],
                                                               channel=channel))
    tb_files = glob.glob(tb_filename)
    if len(tb_files) == 0 or len(tb_files) > 1:
        raise RuntimeError
    tb_file = tb_files[0]

    tb_data = cl.CDRVariable(channel, conf["h_dims"],
                             nctype=cl.NC_SHORT, dtype=cl.NC_SHORT)
    tb_data.from_file(tb_file)

    return melt.scaled_brightness_temperatures(conf['sat'], channel, tb_data)


def update_cdr(conf, cdr, melt_onset_day, melt_start_detected):
    mod = cdr.variables['melt_onset_day_seaice_conc_cdr']
    mod[0, :, :] = melt_onset_day

    msd_bits = np.where(melt_start_detected, MELT_BIT, 0)
    qa = cdr.variables['qa_of_seaice_conc_cdr']
    final_qa = np.bitwise_or(qa[0], msd_bits)
    qa[0, :, :] = final_qa


def parse_arguments():
    parser = ap.ArgumentParser(description="Add melt onset day variable and add melt"
                               "start detected flag to daily CDR files")
    parser.add_argument('year', type=int, help='Year of daily CDR files to process')
    parser.add_argument('version', help='G02202 dataset version to process')
    parser.add_argument('revision', help='G02202 dataset revision to process')
    return parser.parse_args()

if __name__ == "__main__":
    args = parse_arguments()
    add_melt(args.year, args.version, args.revision)
