################################################################################
#@***h* Sea_Ice_CDR/cdr_configuration.py
#
# NAME
#  cdr_configuration.py
#
# LOCATION
#  src/cdr_configuration.py
#
# PURPOSE
#  Create and load contextual metadata for running sea ice algorithms.
#
# DESCRIPTION
#
#  Configuration is handled as a python dictionary.  The command
#  create_configuration_dictionary, first loads user defaults from an external
#  file. Then proceeds to set up standard constants and locations required to
#  produce the CDR.
#
# AUTHOR
#  nsidc@nsidc.org
#  National Snow & Ice Data Center, University of Colorado, Boulder
#
# COPYRIGHT
#  Copyright 2011 Regents of the University of Colorado. The Regents of the
#  University of Colorado grants, and you accept, a personal, nonexclusive,
#  nontransferable license to use this software, at no charge, in accordance
#  with the terms herein, solely for internal research and development, which
#  is limited to testing, measuring, assessing, evaluating the software and/or
#  data that you generate using this software, and not for any commercial
#  purpose. You may not develop derivative works based on this software, or
#  redistribute this software.
#
# CREATION DATE
#  July 2011
#
# EXTERNALS
#   __future__.division
#   __future__.absolute_import
#   copy
#   sys
#   os.path
#   ConfigParser
#   cdr_lib
#   datetime
#
# SUBROUTINES
#   _check_valid_platform_date_combination
#   brightness_file_glob
#   cdl_template_from_hemisphere
#   cdr_constants
#   configuration_dictionary
#   create_configuration_dictionary
#   create_file_names
#   executable_locations
#   expand_hemisphere
#   folder_locations
#   goddard_file_names
#   gpd_name
#   load_constant_values
#   masks_from_month
#   platform_data_from_date
#   qc_flag_values
#   set_hemisphere_info
#   set_test_years
#   class UserDefaults
#     convert_config_to_dict
#     load_user_defaults
#
#
# INPUTS
#  Optional preset dictionary, normally containing name/hemisphere/date.  If
#  no configuration file is present, this routine defaults to the one located
#  at "ROOT/config/cdr.config"
#
# OUTPUTS
#  configuration dictionary
#
# MODIFICATION HISTORY
#
#
#@*****
################################################################################

# import a few __future__ features to keep possible upgrades cleaner
from __future__ import division
from __future__ import absolute_import
# Default to unicode, for 8-bit strings prepend to the string declaration: b''
#from __future__ import unicode_literals

import copy
import sys
import os.path
import ConfigParser
import cdr_lib as cl
import datetime as dt


# "conf" is a  configuration dictionary throughout this file
#
# This file is imbued with the essence of cruft - I would like this to become a
# reader of proper configuration file, and preferably a configuration file that
# can be written out from a database, and easily read in multiple programming
# languages.


class UserDefaults:
    """ Simple helper class that uses the ConfigParser to to read and return
    configuration values from a user config file.  It also loads those values
    into an input dictionary and returns it."""

    def load_user_defaults(self, config_file, conf={}):
        """ read localization parameters from the configuraton file
        config_file - input config file in unix config format.
        """

        # Named parameters can and are used in the config file, and
        # will create errors if they are not populated when the file
        # is read. Currently, h_long, YEAR, and CDR_root are used.
        if ("h_long" in conf and "YYYY" in conf):
            cp = ConfigParser.SafeConfigParser(
                {'CDR_root': cl.cdr_root_dir(),
                 'h_long': conf["h_long"],
                 'YYYY': conf["YYYY"]
                 })
        else:
            cp = ConfigParser.SafeConfigParser({'CDR_root': cl.cdr_root_dir()})
        cp.read(config_file)
        UserDefaults.config = self.convert_config_to_dict(cp)

        # Load the input conf with the stuff just read from the file
        for key in UserDefaults.config:
            # Few values are set when this is called, but they are important
            if (key not in conf):
                conf[key] = UserDefaults.config[key]

        return conf

    def convert_config_to_dict(self, conParse):
        """ Take the configuration parser interpolate all values and return a
            dictionary. This is used to get a configuration dictionary so that
            %()s substitution can continue to work.
            Sections are ignored """

        conDict = {}
        sections = conParse.sections()
        for section in sections:
            opts = conParse.options(section)
            for opt in opts:
                conDict[opt] = conParse.get(section, opt)
        return conDict


def folder_locations(conf):
    """Set the paths to changable folders - input, output, work/temp.

    This may need to be updated during code porting or relocalization. The
    default is to assume that the code is being run from a subdirectory of
    where it is checked out or unpacked.
    """

    if ('brightness_input' not in conf):
        conf["brightness_input"] = os.path.normpath(
            "%(brightness_input_base)s/%(h_long)s/%(YYYY)s/" % conf)
    if ('nt_final_dir' not in conf):
        conf["nt_final_dir"] = os.path.normpath(
            "%(nasateam_final_base)s/%(h_long)s/daily/%(YYYY)s" % conf)
    if ('bt_final_dir' not in conf):
        conf["bt_final_dir"] = os.path.normpath(
            "%(bootstrap_final_base)s/%(h_long)s/daily/%(YYYY)s" % conf)
    if ('nt_monthly_final_dir' not in conf):
        conf["nt_monthly_final_dir"] = os.path.normpath(
            "%(nasateam_final_base)s/%(h_long)s/monthly" % conf)
    if ('bt_monthly_final_dir' not in conf):
        conf["bt_monthly_final_dir"] = os.path.normpath(
            "%(bootstrap_final_base)s/%(h_long)s/monthly" % conf)
    if ('work_base_dir' not in conf):
        conf["work_base_dir"] = os.path.normpath(os.path.join(conf["CDR_root"], "work"))
    conf["work"] = os.path.normpath(os.path.join(
        conf["work_base_dir"], "%(DATE)s%(hemisphere)s" % conf))
    if ('output' not in conf):
        conf["output"] = os.path.normpath(
            os.path.join(conf["CDR_root"], 'output'))
    if ('ancillary' not in conf):
        conf["ancillary"] = os.path.normpath(os.path.join(conf["CDR_root"], 'ancillary'))
    conf["gpd_dir"] = conf["ancillary"]
    conf["config_dir"] = os.path.join(conf["CDR_root"], "config")


def cdr_constants(conf):
    """Constants for the CDR definitions, including the cdl variable names to
    load.

    These constants are closely related to the cdl templates, and this
    function needs to be kept synchronized with the variables names in them.

    CDR variable name in "CDR variable and filenames" emails from Donna Scott
    and Walt Meier on July 14 and 15, 2011. Versioning from various
    conversations, The v00 versions are pre-release, released versions will
    start with 1.0.

    Uses hemisphere to pick land mask.
    """

    # These could be moved to the config file
    conf["cdr_combo_var"] = "seaice_conc_cdr"
    conf["nt_var"] = "goddard_nt_seaice_conc"
    conf["bt_var"] = "goddard_bt_seaice_conc"
    conf["cdr_stddev_var"] = "stdev_of_seaice_conc_cdr"
    conf["cdr_qa_var"] = "qa_of_seaice_conc_cdr"
    conf["legacy_combo_var"] = "goddard_merged_seaice_conc"
    # conf["cdr_version"] = "v01r00"
    conf["min_NT"] = 10
    conf["min_BT"] = 10
    conf["start_of_northern_melt_season"] = 60  # To match NSIDC-0105
    conf["end_of_northern_melt_season"] = 244  # To match NSIDC-0105
    # Tb cutoff per Mark Anderson, via email from Walt Meier 2011 Aug 24
    conf["melt_algo_delta_tb_cutoff_tenths_k"] = 20
    conf["melt_algo_min_ice_conc"] = 50


def load_constant_values(conf):
    """ Single location for magic numbers, particularly mask values """

    conf["smallest_flag_value"] = 251
    conf["smallest_land_mask_flag"] = 251
    conf["largest_land_mask_flag"] = 254
    conf["pole_flag"] = 251
    conf["lake_flag"] = 252
    conf["coast_flag"] = 253
    conf["land_flag"] = 254
    conf["missing_flag"] = 255
    conf["nsidc-0105-open-ocean"] = 0
    conf["nsidc-0105-no-melt-detected"] = 253
    conf["nsidc-0105-near-coast"] = 254
    conf["nsidc-0105-land"] = 255
    conf["combined_oceanmask_values"] = {
        'open_ocean': 0,
        'coast': 2,
        'land': 3,
        'lake': 4,
        'shore': 5,
        'near_shore': 6,
        'off_shore': 7
    }
    conf["nasateam_scale_factor"] = 2.5
    conf["nasateam_max_raw_data"] = 250
    conf["bootstrap_scale_factor"] = 10.0
    conf["bootstrap_max_raw_data"] = 1000
    conf["concentration_fill_value"] = 255
    conf["stdev_fill_value"] = -1.
    conf["qa_fill_value"] = 0


def set_test_years(conf):
    conf["test_year_n07"] = "1901"
    conf["test_year_f08"] = "1902"
    conf["test_year_f11"] = "1903"
    conf["test_year_f13"] = "1905"
    conf["test_year_f17"] = "1906"


def expand_hemisphere(conf):
    """Add the full hemisphere name as conf["h_long"]
    and the EASE grid dimensions as "h_dims"."""

    conf["HEMI"] = conf["hemisphere"].upper()
    if (conf["hemisphere"] == "s"):
        conf["h_long"] = "south"
        conf["h_dims"] = cl.south_dims()
    elif (conf["hemisphere"] == "n"):
        conf["h_long"] = "north"
        conf["h_dims"] = cl.north_dims()
    return conf


def executable_locations(conf):
    """Set the paths to important programs"""

    conf["CDR_bin"] = os.path.join(cl.cdr_root_dir(), "bin")
    conf["external_bin"] = os.path.join(cl.cdr_root_dir(),
                                        "externals/netcdf/bin")
    conf["ncgen"] = "%(external_bin)s/ncgen" % conf


def _check_valid_platform_date_combination(platform_string, conf):
    if (
            platform_string == 'pre-NIMBUS' and
            conf["iYY"] != int(conf["test_year_n07"]) and
            conf["iYY"] != int(conf["test_year_f08"]) and
            conf["iYY"] != int(conf["test_year_f11"]) and
            conf["iYY"] != int(conf["test_year_f13"]) and
            conf["iYY"] != int(conf["test_year_f17"])
    ):
        raise IOError("Date given is pre-NIMBUS (Oct 26, 1978) and "
                      "outside the operational range of this code.\n")


def platform_data_from_date(conf):
    """Add names for the satellite of reference on a particular date.

    Abbreviation for filenames in conf["sat"], GCMD name in
    conf["GCMD_platform"].

    The sat #s are from NT. BT is slightly different in some cases.  The
    original source notes "May need to think about this more in terms of the
    final CDR output."

    Pole holes are only specified for Arctic. In the Antarctic, the pole hole
    is not relevant to seaice concentration since it entirely on land.
    """

    run_date = dt.date(conf["iYY"], conf["iMM"], conf["iDD"])
    platform_string = cl.platform_from_date(run_date)
    conf["days_required_for_valid_month"] = 20

    _check_valid_platform_date_combination(platform_string, conf)

    if (platform_string == 'n07' or
            conf["iYY"] == int(conf["test_year_n07"])):
        # For monthly, update to use NIMBUS through August
        platform_string = 'n07'
        conf["GCMD_platform"] = 'NIMBUS-7'
        conf["GCMD_sensor"] = 'SMMR > ' \
                              'Scanning Multichannel Microwave Radiometer'
        conf["pole_hole_file"] = 'gsfc_pole_hole.N07'  # SMMR pole hole.
        conf["days_required_for_valid_month"] = 10
        # RSS input files not present, "tb_vsn" undefined
    if (
            platform_string == 'f08' or
            conf["iYY"] == int(conf["test_year_f08"])
    ):
        platform_string = 'f08'
        conf["GCMD_platform"] = 'DMSP 5D-2/F8 > ' \
                                'Defense Meteorological Satellite Program-F8'
        conf["GCMD_sensor"] = 'SSM/I > Special Sensor Microwave/Imager'
        conf["pole_hole_file"] = 'gsfc_pole_hole.N13'  # SSM/I pole hole.
        conf["tb_vsn"] = '2'
    if (
            platform_string == 'f11' or
            conf["iYY"] == int(conf["test_year_f11"])
    ):
        platform_string = 'f11'
        conf["GCMD_platform"] = 'DMSP 5D-2/F11 > ' \
            'Defense Meteorological Satellite Program-F11'
        conf["GCMD_sensor"] = 'SSM/I > Special Sensor Microwave/Imager'
        conf["pole_hole_file"] = 'gsfc_pole_hole.N13'  # SSM/I pole hole.
        conf["tb_vsn"] = '2'
    if (platform_string == 'f13' or
            conf["iYY"] == int(conf["test_year_f13"])):
        platform_string = 'f13'
        conf["GCMD_platform"] = 'DMSP 5D-2/F13 > ' \
            'Defense Meteorological Satellite Program-F13'
        conf["GCMD_sensor"] = 'SSM/I > Special Sensor Microwave/Imager'
        conf["pole_hole_file"] = 'gsfc_pole_hole.N13'  # SSM/I pole hole.
        conf["tb_vsn"] = '2'
    if (platform_string == 'f17' or
            conf["iYY"] == int(conf["test_year_f17"])):
        platform_string = 'f17'
        conf["GCMD_platform"] = 'DMSP 5D-3/F17 > ' \
            'Defense Meteorological Satellite Program-F17'
        conf["GCMD_sensor"] = 'SSMI/S > Special Sensor Microwave Imager/Sounder'
        conf["pole_hole_file"] = 'gsfc_pole_hole.N17'   # SSM/I pole hole.
        conf["tb_vsn"] = '4'

    conf["sat"] = platform_string
    conf["goddard_sat"] = platform_string
    conf["SAT"] = platform_string.upper()
    del(run_date)


def goddard_file_names(conf):
    """Set current version number for final BT and NT files from Goddard."""

    conf["nt_vsn"] = "01"
    conf["bt_vsn"] = "02"
    conf["goddard_nt_file"] = (
        "nt_%(DATE)s_%(goddard_sat)s_v%(nt_vsn)s_%(hemisphere)s.bin") % conf
    conf["goddard_bt_file"] = (
        "bt_%(DATE)s_%(goddard_sat)s_v%(bt_vsn)s_%(hemisphere)s.bin") % conf


def qc_flag_values(conf):
    """Set the flag values for the QC bit-flags."""

    conf["BT_value"] = "1"
    conf["NT_value"] = "2"
    conf["climate_masked"] = "4"
    conf["near_coast"] = "8"
    conf["low_concentration"] = "32"
    conf["low_conc_cutoff"] = "50"
    conf["melt_started"] = "128"


def gpd_name(conf):
    """Determine correct gpd file by hemisphere."""

    conf['gpd'] = "%(HEMI)s3B.gpd" % conf


def cdl_template_from_hemisphere(conf):
    """Determine correct template cdl file by hemisphere."""

    conf['cdr_template'] = "%(ancillary)s/sea_ice_CDR_%(h_long)s.cdl" % conf


def masks_from_month(conf):
    """Determine which ocean, weight (cmin), and climate masks to use by date
    and hemisphere.

    The ocean mask values are ocean(openwater):0, possible ice:1 coast:2,
    land:3, lake:4

    See "Land Masks" massages on NSIDC cdr mailing list, 7 July 2011.
    """

    # oceanmask
    if (conf['hemisphere'] == 'n'):
        conf["combined_oceanmask"] = (
            "%(ancillary)s/landmask/landCoastShoreMask.north.bin" % conf
        )
        if (conf["iMM"] < 11):
            conf["oceanmask"] = (
                "%(ancillary)s/oceanmask/oceanmask.%(MMM)s.1979-2003.n" % conf)
        elif (conf["iMM"] > 10):
            conf["oceanmask"] = (
                "%(ancillary)s/oceanmask/oceanmask.%(MMM)s.1978-2003.n" % conf)
    else:  # not north
        conf["oceanmask"] = (
            "%(ancillary)s/oceanmask/oceanmask.S%(MM)s_v2.bin" % conf)
        conf["combined_oceanmask"] = (
            "%(ancillary)s/landmask/landCoastShoreMask.south.bin" % conf)

    # cmin & landmask
    if (conf['hemisphere'] == 'n'):
        conf["cmin"] = "%(ancillary)s/cmin/pmat_north.byte" % conf
        conf["nt_landmask"] = "%(ancillary)s/landmask/landmask-n3b.byte" % conf
        conf["bt_landmask"] = (
            "%(ancillary)s/landmask/N3B.bootstrap_v02.landmask.dat" % conf)
    else:   # not north
        conf["cmin"] = "%(ancillary)s/cmin/pmat_south.byte" % conf
        conf["nt_landmask"] = "%(ancillary)s/landmask/landmask-s3b.byte" % conf
        conf["bt_landmask"] = (
            "%(ancillary)s/landmask/S3B.bootstrap_v02.landmask.dat" % conf)

    # nasateam climatology
    #     - climatology (-c)  iceclim.nasateam.[N|S][MM]_v[1|2].byte
    if (conf['hemisphere'] == 'n'):
        conf["nt_climatology"] = (
            "%(ancillary)s/climatology/icelakeclim.nasateam.N%(MM)s_v1.byte"
        ) % conf
    else:
        conf["nt_climatology"] = (
            "%(ancillary)s/climatology/icelakeclim.nasateam.S%(MM)s_v2.byte"
        ) % conf

    # bootstrap climatology
    #     - climatology (-c)  iceclim.boot.[N|S][MM]_v[1|2].byte
    if (conf['hemisphere'] == 'n'):
        conf["bt_climatology"] = (
            "%(ancillary)s/climatology/iceclim.bootv2.N%(MM)s_v1.byte"
        ) % conf
    else:
        conf["bt_climatology"] = (
            "%(ancillary)s/climatology/iceclim.bootv2.S%(MM)s_v2.byte"
        ) % conf

    # mask for near-shore quality flag
    conf["coastal_mask"] = (
        "%(ancillary)s/landmask/landCoastShoreMask.%(h_long)s.bin"
    ) % conf


def brightness_file_glob(conf):
    """Create a pattern to match the expected names of the brightness
    temperature files.
    """
    conf['bt_root'] = "tb_%(sat)s_%(DATE)s_v?_%(hemisphere)s" % conf


def create_file_names(conf):
    """Create the by-run names for working and output files.

    NetCDF4 filename in "CDR VAriable and filenames" emails from Donna Scott
    and Walt Meier on July 14 and 15, 2011.
    """

    conf["cdr_nc"] = (
        "seaice_conc_daily_%(hemisphere)sh_%(sat)s_%(DATE)s_%(cdr_version)s.nc"
    ) % conf
    conf["cdr_cdl"] = (
        "seaice_conc_daily_%(DATE)s_%(h_long)s_%(sat)s_%(cdr_version)s.cdl"
    ) % conf
    conf["auto_nt"] = "%(work)s/nt_for_cdr_%(DATE)s_%(h_long)s_%(sat)s" % conf
    conf["auto_bt"] = "%(work)s/bt_for_cdr_%(DATE)s_%(h_long)s_%(sat)s" % conf
    conf["auto_nt_scaled"] = conf["auto_nt"]+".rescaled"
    conf["auto_bt_scaled"] = conf["auto_bt"]+".rescaled"
    conf["scaled_nt"] = (
        "%(work)s/nt_rescaled_%(DATE)s_%(h_long)s_%(sat)s") % conf
    conf["scaled_bt"] = (
        "%(work)s/bt_rescaled_%(DATE)s_%(h_long)s_%(sat)s") % conf
    conf["stddev_file"] = (
        "%(work)s/stddev_for_cdr_%(DATE)s_%(h_long)s_%(sat)s") % conf
    conf["legacy_combo_file"] = (
        "%(work)s/combo_legacy_%(DATE)s_%(h_long)s_%(sat)s") % conf
    conf["cdr_combo_file"] = (
        "%(work)s/combo_cdr_%(DATE)s_%(h_long)s_%(sat)s") % conf
    conf["cdr_qa_file"] = (
        "%(work)s/combo_cdr_qa_%(DATE)s_%(h_long)s_%(sat)s") % conf
    conf["cdr_stddev_file"] = (
        "%(work)s/combo_cdr_stddev_%(DATE)s_%(h_long)s_%(sat)s") % conf
    conf["melt_summary_file"] = "%(output)s/melt_%(YYYY)s.bin" % conf
    conf["blank_monthly_nc_file"] = (
        "%(ancillary)s/%(HEMI)s_monthly_blank.nc") % conf


def configuration_dictionary(hemisphere, YYYYMMDD, preset={}):
    return create_configuration_dictionary(['configuration_dictionary',
                                            hemisphere, YYYYMMDD], preset=preset)


def set_hemisphere_info(hemi, conf):
    conf["hemisphere"] = hemi.lower().strip()[0]
    conf = expand_hemisphere(conf)
    return conf


def create_configuration_dictionary(argv, preset={}):
    """
    Use a list of arguments and a dictionary of defaults to return a
    configuration dictionary.

    This is the primary entry point for cdr_configuration.
    """

    # set presets
    conf = copy.deepcopy(preset)
    set_test_years(conf)

    # check for presence of hemisphere
    if ('hemisphere' not in conf):
        conf = set_hemisphere_info(argv[1].lower().strip()[0], conf)
    else:
        conf = expand_hemisphere(conf)

    # find the date for which we are running
    if (len(argv[2]) == 7):
        if argv[2].isdigit():
            sys.stderr.write(
                "\nDate field needs eight digit dates. Use YYYYMMDD.\n")
            exit(1)
        else:
            year_key = "test_year_" + argv[2][0:3]
            if (year_key not in conf or not argv[2][3:7].isdigit()):
                sys.stderr.write(
                    "\nDate field should be entirely digits. Use YYYYMMDD.\n")
                exit(1)
            conf["DATE"] = conf[year_key] + argv[2][3:7]
    elif not argv[2].isdigit():
        sys.stderr.write(
            "\nDate field should be entirely digits. Use YYYYMMDD.\n")
        exit(1)
    elif (len(argv[2]) == 8):
        # should be all number, so no case smashing
        conf["DATE"] = argv[2]
        conf["YYYY"] = argv[2][0:4]
        conf["iYY"] = int(argv[2][0:4])
        conf["iMM"] = int(argv[2][4:6])
        conf["MM"] = argv[2][4:6]
        conf["iDD"] = int(argv[2][6:8])
        conf["MMM"] = ["jan", "feb", "mar", "apr", "may", "jun",
                       "jul", "aug", "sep", "oct", "nov", "dec"][conf["iMM"]-1]
    else:
        # bad length
        sys.stderr.write(
            "\nDate field required after hemisphere. Use YYYYMMDD.\n")
        exit(1)
    conf["YYYY"] = conf["DATE"][0:4]
    conf["iYY"] = int(conf["DATE"][0:4])
    conf["iMM"] = int(conf["DATE"][4:6])
    conf["MM"] = conf["DATE"][4:6]
    conf["iDD"] = int(conf["DATE"][6:8])
    conf["MMM"] = ["jan", "feb", "mar", "apr", "may", "jun",
                   "jul", "aug", "sep", "oct", "nov", "dec"][conf["iMM"]-1]

    # Load the default configuration file and add the values to conf.
    if "config_filename" not in conf:
        conf["config_filename"] = cl.standard_config_filename()
    conf["CDR_root"] = cl.cdr_root_dir()
    conf = UserDefaults().load_user_defaults(conf["config_filename"], conf)

    # code environment setup
    load_constant_values(conf)
    cdr_constants(conf)  # needs hemisphere for land mask
    folder_locations(conf)  # before exceutableLocations()
    executable_locations(conf)

    # science metadata
    platform_data_from_date(conf)
    goddard_file_names(conf)
    qc_flag_values(conf)

    # metadata based configuration, including derived filenames
    gpd_name(conf)
    cdl_template_from_hemisphere(conf)
    masks_from_month(conf)
    brightness_file_glob(conf)  # needs date and platform
    create_file_names(conf)

    return conf
