#########################################################################################
#@***h* Sea_Ice_CDR/cdr_daily_seaice.py
#
# NAME
#  cdr_daily_seaice.py
#
# LOCATION
#  src/cdr_daily_seaice.py
#
# PURPOSE
#  Runs processing for creating the Climate Data Record of Average Daily Sea
#  Ice Concentration.
#
# DESCRIPTION
#  This code creates proper configuration scripts and then builds up a list of
#  commands that are then called out to the operating system in sequence to
#  process the daily sea ice data and create the CDR netcdf File
#
# AUTHOR
#  nsidc@nsidc.org
#  National Snow & Ice Data Center, University of Colorado, Boulder
#
# COPYRIGHT
#  Copyright 2011 Regents of the University of Colorado. The Regents of the
#  University of Colorado grants, and you accept, a personal, nonexclusive,
#  nontransferable license to use this software, at no charge, in accordance
#  with the terms herein, solely for internal research and development, which
#  is limited to testing, measuring, assessing, evaluating the software and/or
#  data that you generate using this software, and not for any commercial
#  purpose. You may not develop derivative works based on this software, or
#  redistribute this software.
#
# CREATION DATE
#  Aug 2011
#
# EXTERNALS
#    __future__.division
#    __future__.print_function
#    __future__.absolute_import
#    sys
#    glob
#    errno
#    re
#    datetime
#    cdr_lib
#    numpy
#    os.path
#    cdr_configuration
#
# SUBROUTINES
#   _dataset_ftp_url
#   bootstrap_grid_from_file
#   bootstrap_value_filter
#   cleanup_work
#   create_combined_field_qa
#   create_combo_concentration
#   create_working_cdl
#   get_rss_brightness_files_root
#   main
#   make_raw_bootstrap
#   make_raw_nasateam
#   nasateam_grid_from_raw_grid
#   nasateam_value_filter
#   pole_hole_filter_factory
#   required_environment
#   run_hemi
#   usage
#
#
# INPUTS
#  n/a
#
#
# OUTPUTS
#  n/a
#
# MODIFICATION HISTORY
#
#
#@*****
################################################################################


# import a few __future__ features to keep possible upgrades cleaner
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
# Default to unicode, for 8-bit strings prepend to the string declaration: b''
#from __future__ import unicode_literals

# Standard python libraries
import sys
import glob
import errno
import re
import datetime as dt
import cdr_lib as cl

# Numpy is not a python built-in
import numpy

# for checksums
#import hashlib
import os.path
if (os.path.dirname(sys.argv[0]) != ""):
    sys.path.append(os.path.abspath(os.path.dirname(sys.argv[0])))
else:
    sys.path.append(os.path.abspath("."))

import cdr_configuration as cfg

# "conf" is a global configuration dictionary - I expect to extend this later
# with argument parsing and configuration files.
# Some strings in the configuration will expect to be format strings using the
# conf dictionary itself to fill in the format. These contain "%(...)s" sub-
# strings, with the name of the conf{} member to insert in lieu of the ellipsis.
# The final "s" may be replaced with other format identifiers as needed.
conf = {}


def usage():
    print("""
    CDR Daily Sea Ice is generated by calling this script
    with the hemisphere, date, product version, and product
    revision you wish to process.

    > cdr_daily_seaice n 20071103 02 00
    """)
    exit(1)


################################################################################
#
# The following section is data filters for sea-ice sources
#
################################################################################


def nasateam_value_filter(rawarray):
    """
    A filter to remap values in the range 0-250 to the ranged 0-100, and
    preserves the flags in the range 251 to 255. This will be passed to the
    CDRVariable object for the nasateam grid as part of its filter list.
    """
    for i in numpy.ndindex(rawarray.shape):
        if (rawarray[i] < conf["smallest_flag_value"] and rawarray[i] > 0):
            # nt array is int, to round correctly adding .5 since this truncates
            # when it stores the int back into rawarray.
            rawarray[i] = (rawarray[i] * 100.0 / 250.0) + 0.5
        # other values do not change, so no else:


def bootstrap_value_filter(rawarray):
    """
    A filter to remap values in the range 0-1000 to the range 0-100, and
    map the flags above 1000. This will be passed to the CDRVariable
    object for the bootstrap grid as part of its filter list.
    """

    for i in numpy.ndindex(rawarray.shape):
        if (rawarray[i] < 1001):
            # the bt values are ints, so adding 5 creates correct rounding.
            rawarray[i] = (rawarray[i] + 5) / 10.0
        elif (rawarray[i] == 1100):  # missing, includes pole
            rawarray[i] = 255
        elif (rawarray[i] == 1200):  # land
            rawarray[i] = 254
        else:  # not part of BT definition
            rawarray[i] = 255  # "missing"


def pole_hole_filter_factory(conf):
    """
    Returns a filter for the correct pole hole, setting the pole hole to 251
    """

    print("Reading %(ancillary)s/%(pole_hole_file)s" % conf)
    ph_file = open(
        # by-sensor pole hole to match Goddard
        "%(ancillary)s/%(pole_hole_file)s" % conf,
        "rb")
    ph_file.read(300)
    pole_mask = numpy.fromfile(
        ph_file, dtype=cl.NC_BYTE, count=conf["h_dims"][0]*conf["h_dims"][1]
    )
    ph_file.close()
    del(ph_file)
    pole_mask.shape = tuple(conf["h_dims"])
    pole_mask[pole_mask == 1] = conf["pole_flag"]
    return cl.cdr_mask_filter_factory(pole_mask, use=[conf["pole_flag"]])


###############################################################################
#
# create_combo_concentration
#
################################################################################


def create_combo_concentration(combo_name, combo_file_name,
                               qa_name, qa_file_name, nt_var, bt_var):
    """Make a combined sea-ice concentration from nasateam and bootstrap
    sea-ice products."""

    #
    # Based on Walt Meier's example CDR field IDL code
    #
    combo_ice = cl.CDRVariable(combo_name, conf["h_dims"],
                               nctype=cl.NC_BYTE, dtype=cl.NC_BYTE)
    nt_copy = numpy.copy(nt_var)
    if (nt_copy.dtype != cl.NC_BYTE):
        print(">>> NT array for combo of wrong type!\n")
    bt_copy = numpy.copy(bt_var)
    if (bt_copy.dtype != cl.NC_BYTE):
        print(">>> BT array for combo of wrong type!\n")

    #
    # find union of missing
    #
    missing = numpy.copy(bt_var)
    missing[missing != conf["missing_flag"]] = 0
    missing[nt_var != conf["missing_flag"]] = 0
    missing_mask_fxn = cl.cdr_mask_filter_factory(
        missing, use=[conf["missing_flag"]]
    )

    #
    # Load masks needed for process
    #
    ocean_mask = numpy.fromfile(conf["oceanmask"], dtype=cl.NC_BYTE, count=combo_ice.size)
    ocean_mask.shape = tuple(conf["h_dims"])
    # lake mask value in oceanmask is 4
    ocean_mask[ocean_mask == 4] = conf["lake_flag"]
    lake_mask_fxn = cl.cdr_mask_filter_factory(ocean_mask, use=[conf["lake_flag"]])

    # Load the correct combination land mask
    land_mask = numpy.fromfile(conf["combined_oceanmask"], dtype=cl.NC_BYTE,
                               count=combo_ice.size)
    land_mask.shape = tuple(conf["h_dims"])
    # land_mask values in landmask are 2 coast and 3 land.
    land_mask[land_mask == 2] = conf["coast_flag"]
    land_mask[land_mask == 3] = conf["land_flag"]
    combo_land_mask_fxn = cl.cdr_mask_filter_factory(
        land_mask, use=[conf["land_flag"], conf["coast_flag"]]
    )

    #
    # Create combination
    #

    # Make lakes consistent
    lake_mask_fxn(nt_copy)
    lake_mask_fxn(bt_copy)

    # NT and BT are not used where they are less than their minimums
    nt_copy[nt_copy < conf["min_NT"]] = 0
    bt_copy[bt_copy < conf["min_BT"]] = 0

    # NT is not used where BT is less than BTs minimum
    # this is so that the ice edge is defined by the BT product
    # Do this before BT flags are zeroed so that NT is not masked by BT land/coast
    nt_copy[bt_copy < conf["min_BT"]] = 0

    # Zero out flags (so that maximum() picks up sea-ice instead of missing)
    nt_copy[nt_copy > 100] = 0
    bt_copy[bt_copy > 100] = 0

    # Make land-masks equal before taking maximum to avoid addities.
    combo_land_mask_fxn(nt_copy)
    combo_land_mask_fxn(bt_copy)

    # Get the maximum of the sea-ice values of the two concentration fields.
    combo = numpy.maximum(nt_copy, bt_copy)

    # apply mask values to newly created combined field
    missing_mask_fxn(combo)
    combo_land_mask_fxn(combo)
    lake_mask_fxn(combo)
    if (conf["hemisphere"] == "n"):
        pole_mask_fxn = pole_hole_filter_factory(conf)
        pole_mask_fxn(combo)
        del(pole_mask_fxn)

    combo_ice.data = numpy.array(combo, copy=True, dtype=cl.NC_BYTE).data
    combo_ice.to_file(combo_file_name)
    cmdstring = (
        "%(CDR_bin)s/ncload {0} {1}  %(work)s/%(cdr_nc)s raw\n" % conf
    ).format(combo_name, combo_file_name)

    if (qa_name != ""):
        combo_qa = create_combined_field_qa(qa_name, combo_ice, nt_var, bt_var)
        combo_qa.to_file(qa_file_name)
        cmdstring += "\n"
        cmdstring += (
            "%(CDR_bin)s/ncload {0} {1}  %(work)s/%(cdr_nc)s raw\n" % conf
        ).format(qa_name, qa_file_name)
        del(combo_qa)
    del(nt_copy, bt_copy)
    return cmdstring


def create_combined_field_qa(qa_name, combo_var, nt_var, bt_var):
    """
    Returns a QA array from a combined field, the base nt and bt fields, and a
    mask file.

    Exact flag values used are set in cdr_configuration.py. The mask file is
    used to provide a near-shore mask, and the combined field is checked to
    see which of nt and bt match the value used. If both are equal, they will
    both be flagged as being used. A flag is also set for sea ice
    concentration between 0% and 50%, where concentration can change quickly
    and algorithms tend to have larger errors.
    """

    combo_ice_qc = cl.CDRVariable(
        qa_name, conf["h_dims"], nctype=cl.NC_BYTE, dtype=cl.NC_BYTE)
    combo_ice_qc.data = numpy.zeros(conf["h_dims"], dtype=cl.NC_BYTE).data

    # use near-shore mask as base before setting other flags
    near_coast_mask = numpy.fromfile(
        conf["coastal_mask"], dtype=cl.NC_BYTE, count=combo_ice_qc.size)
    near_coast_mask.shape = tuple(conf["h_dims"])
    # do not accidentally use lake!
    near_coast_mask[near_coast_mask == int(conf["near_coast"])] += 128
    # near-coast flag value in cdr QA is , in near_coastmask it is 4
    near_coast_mask[near_coast_mask == 5] = conf["near_coast"]  # one pixel away
    near_coast_mask[near_coast_mask == 6] = conf["near_coast"]  # two pixels away
    near_coast_mask_fxn = cl.cdr_mask_filter_factory(
        near_coast_mask, use=[int(conf["near_coast"])])
    near_coast_mask_fxn(combo_ice_qc)

    # Flag climatology overlay
    #
    ocean_mask = numpy.fromfile(
        conf["oceanmask"], dtype=cl.NC_BYTE, count=combo_ice_qc.size)
    ocean_mask.shape = tuple(conf["h_dims"])
    # climate mask value in oceanmask is 0
    combo_ice_qc[ocean_mask == 0] += int(conf["climate_masked"])

    combo_ice_qc[nt_var == combo_var] += int(conf["NT_value"])
    combo_ice_qc[bt_var == combo_var] += int(conf["BT_value"])
    combo_ice_qc[
        (combo_var > 0) &
        (combo_var < int(conf["low_conc_cutoff"]))
    ] += int(conf["low_concentration"])
    return combo_ice_qc


##############################################################################
#
# create intermediate grids - pmalgos calls
#
################################################################################


def required_environment():
    """ set Environmental variables that must be set before performing pmalgos
    calls """

    command = (
        "export BOOTSTRAP_V2_ANCILLARY=%(cdr_root)s/ancillary/bootstrap_seaice/\n"
        "export MORPH_SE_PATH=%(cdr_root)s/ancillary/spillover_kernels/\n"
        "export PATHMPP=%(gpd_dir)s\n"
    ) % conf
    return command


def get_rss_brightness_files_root(conf):
    """
    Find the root file name for RSS brightness temperatures.

    Finds the root for the day indicated in the configuration dictionary passed.
    """

    tb_file = glob.glob("%(brightness_input)s/%(bt_root)s22v.bin" % conf)
    if len(tb_file) == 1:
        return tb_file[0][:-7]  # strip to the root that pmalgos takes
    elif len(tb_file) > 1:
        raise IOError("Multiple brightness files "
                      "found matching %(brightness_input)s/%(tb_glob)s\n" % conf)
    else:
        return None


def make_raw_bootstrap(raw_grid_name, brightness_filename):
    """
    Runs pmalgos to generate a fully reproducible BT grid.
    """

    #
    # Based on Matt's IDL wrapper calls to pmalgos
    #
    command = required_environment() + (
        "%(CDR_bin)s/pmalgos  "
        "-u 0 -b "
        "-d %(DATE)s  "
        "-l %(bt_landmask)s "
        "-p %(SAT)s "
        "-c %(bt_climatology)s "
        "%(gpd)s "
        # Product version, should be the same _everywhere_.
        "BOOTSTRAP_%(HEMI)s_V02_PUB_SEAICE  "
        + brightness_filename + " | "
        "%(CDR_bin)s/fdistill -s  >  "
        + raw_grid_name + "\n"
    ) % conf
    return command


def make_raw_nasateam(raw_grid_name, brightness_filename):
    """
    Runs pmalgos to generate a fully reproducible raw NT grid.
    """

    #
    # Based on Matt's IDL wrapper calls to pmalgos
    #

    command = required_environment() + (
        "%(CDR_bin)s/pmalgos "
        "-j -b -x 0 "
        # land mask
        "-m %(nt_landmask)s "
        # weightmask
        "-w %(cmin)s "
        "-f 255 "
        "-r "
        # hemisphere
        "-q %(HEMI)s "
        # platform
        "-p %(SAT)s "
        # climatology
        "-c %(nt_climatology)s "
        "%(gpd_dir)s/%(gpd)s "
        "GODDARD_PUB_SEAICE "
        + brightness_filename + " | "
        "%(CDR_bin)s/fdistill -u  > "
        + raw_grid_name + "\n"
    ) % conf
    return command


def bootstrap_grid_from_file(var_name, grid_file, missingValueMask=None):
    """
    Load bootstrap grid from Goddard-compatible bootstrap file.

    Applies requested filters for CDR combination sea ice.
    """

    #
    # Code based on Walt's IDL code.
    #
    if (conf["hemisphere"] == "n"):
        #     # Only apply the most recent pole mask, to avoid biasing the seaice
        #     # extent time series.
        poleOrCoastMask = pole_hole_filter_factory(conf)
    else:  # southern hemisphere
        # land and coast
        # Only apply the most recent mask, to avoid biasing the seaice extent time series
        land_mask = numpy.fromfile("%(combined_oceanmask)s" % conf, dtype=cl.NC_BYTE,
                                   count=conf["h_dims"][0]*conf["h_dims"][1])
        land_mask.shape = tuple(conf["h_dims"])
        # land mask land is 2, coast is 1
        land_mask[land_mask == 3] = conf["land_flag"]
        land_mask[land_mask == 2] = conf["coast_flag"]
        poleOrCoastMask = cl.cdr_mask_filter_factory(
            land_mask, use=[conf["land_flag"], conf["coast_flag"]])
    if (missingValueMask is None):
        filter_list = [bootstrap_value_filter,
                       poleOrCoastMask,
                       cl.recast]
    else:
        filter_list = [bootstrap_value_filter,
                       missingValueMask,
                       poleOrCoastMask,
                       cl.recast]

    bt_var = cl.CDRVariable(
        var_name, conf["h_dims"], intype=cl.NC_SHORT, nctype=cl.NC_BYTE,
        filters=filter_list
    )
    bt_var.from_file(grid_file)
    bt_var_recast = bt_var.astype(cl.NC_BYTE)
    return bt_var_recast


def nasateam_grid_from_raw_grid(gridFileName):
    """
    Applies necessary post-processing for fully reproducible NT grid.
    """

    #
    # Based on IDL post-processor for pmalgos output
    #
    ocean_mask = numpy.fromfile(
        conf["oceanmask"], dtype=cl.NC_BYTE,
        count=conf["h_dims"][0]*conf["h_dims"][1])
    ocean_mask.shape = tuple(conf["h_dims"])
    # lake mask value in cdr is 252, in oceanmask it is 4
    # not applied to nasateam, set to '1' - the "ice OK" value
    ocean_mask[ocean_mask == 4] = 1
    # land mask value in cdr is 254, in oceanmask it is 3
    ocean_mask[ocean_mask == 3] = conf["land_flag"]
    # coast mask value in cdr is 253, in oceanmask it is 2
    ocean_mask[ocean_mask == 2] = conf["coast_flag"]
    # "don't mask" value in oceanmask is 1!
    # climate mask value in oceanmask is 0, and should force zeros in cdr
    # ocean_mask[ocean_mask==0] = 0
    oceanMaskFilter = cl.cdr_mask_filter_factory(ocean_mask, ignore=[1])

    if (conf["hemisphere"] == "n"):
        poleMask = pole_hole_filter_factory(conf)
        nt_var = cl.CDRVariable(
            "auto_nt_var", conf["h_dims"], header_len=0,  # no header on temp
            nctype=cl.NC_BYTE, intype=cl.NC_BYTE,
            filters=[oceanMaskFilter, poleMask, cl.recast]
        )
    else:
        nt_var = cl.CDRVariable(
            "auto_nt_var", conf["h_dims"], header_len=0,  # no header on temp
            nctype=cl.NC_BYTE, intype=cl.NC_BYTE,
            filters=[oceanMaskFilter, cl.recast]
        )

    nt_var.from_file(gridFileName)
    os.unlink(gridFileName)
    nt_var.to_file(gridFileName)
    return


################################################################################
#
# create CDL for this run
#
################################################################################


def create_working_cdl(cdl_template, cdl_working, conf):
    """
    Copy template CDL file to working CDL file, filling in dynamic attributes.
    """

    input_sources = []
    re_source = re.compile(r'SOURCE_ATTRIBUTE')
    tb_file_root = get_rss_brightness_files_root(conf)
    # If there are no brightness temperatures, CDR-concentration will not be run and
    # the source field should not contain these files. This run is
    # just loading Goddard fields.
    # Per emails from Donna Scott & Walt Meier on Sept. 16 2011.
    # Updated for Pivotal story: https://www.pivotaltracker.com/story/show/42809093
    if tb_file_root is not None:
        input_sources.append(tb_file_root + "19v.bin")
        input_sources.append(tb_file_root + "19h.bin")
        input_sources.append(tb_file_root + "22v.bin")
        input_sources.append(tb_file_root + "37v.bin")
        input_sources.append(tb_file_root + "37h.bin")

    input_sources.append("%(nt_final_dir)s/%(goddard_nt_file)s" % conf)
    input_sources.append("%(bt_final_dir)s/%(goddard_bt_file)s" % conf)
    input_sources = [_dataset_ftp_url(s) for s in input_sources if os.path.exists(s)]
    source_value = ", ".join(input_sources)

    # 	:history = "HISTORY_ATTRIBUTE" ;
    re_history = re.compile(r'HISTORY_ATTRIBUTE')
    history_value = cl.history_attribute(conf['cdr_nc'])
    # //	GDS attributes

    # 	:platform = "PLATFORM_ATTRUBUTE" ;
    re_platform = re.compile(r'PLATFORM_ATTRUBUTE')
    platform_value = conf["GCMD_platform"]

    # 	:sensor = "SENSOR_ATTRIBUTE" ;
    re_sensor = re.compile(r'SENSOR_ATTRIBUTE')
    sensor_value = conf["GCMD_sensor"]

    # //	requirements table GA.2 "Unidata ACDD"
    # //	:date_created = "YYYY-mm-ddThh:mm:ssZ" ;
    re_date_created = re.compile(r'DATE_CREATED_ATTRIBUTE')
    date_created_value = dt.datetime.utcnow().strftime("%Y-%m-%dT%H:%M:%SZ")

    # //	:time_coverage_start = "YYYY-mm-ddT00:00:00Z" ;
    re_time_coverage_start = re.compile(r'TIME_COVERAGE_START_ATTRIBUTE')
    time_coverage_start_value = cl.time_coverage_start(conf)

    # //	:time_coverage_end = "YYYY-mm-ddT23:59:59Z" ;
    re_time_coverage_end = re.compile(r'TIME_COVERAGE_END_ATTRIBUTE')
    time_coverage_end_value = cl.time_coverage_end(conf)

    re_software_version_id = re.compile(r'SOFTWARE_VERSION_ID_ATTRIBUTE')
    software_version_id_value = cl.current_software_version()

    re_product_version = re.compile(r'PRODUCT_VERSION_ATTRIBUTE')
    product_version_value = conf["cdr_version"]

    cdl_in = open(cdl_template, 'r')
    cdl_out = open(cdl_working, 'w')

    for line in cdl_in:
        line = re_history.sub(history_value, line)
        line = re_platform.sub(platform_value, line)
        line = re_sensor.sub(sensor_value, line)
        line = re_date_created.sub(date_created_value, line)
        line = re_source.sub(source_value, line)
        line = re_time_coverage_start.sub(time_coverage_start_value, line)
        line = re_time_coverage_end.sub(time_coverage_end_value, line)
        line = re_product_version.sub(product_version_value, line)
        line = re_software_version_id.sub(software_version_id_value, line)
        cdl_out.write(line)


def _dataset_ftp_url(filename):
    return re.sub(conf['file_root_replaced_for_ftp'],
                  conf['root_ftp_url'],
                  filename)


def cleanup_work(conf):
    """ move final nc file to the desired output directory """

    if not os.path.exists("%(output)s" % conf):
        os.mkdir("%(output)s" % conf)

    # /ftp/pub/DATASETS/NOAA/G02202
    #   /hemisphere
    #       /daily
    #           /year
    #              --daily data files
    return "mv -f %(work)s/%(cdr_nc)s %(output)s/%(cdr_nc)s\n" % conf


################################################################################
#
# run_hemi - run the code to load the northern or southern hemisphere data fields
#
################################################################################
def run_hemi(hemisphere, date):
    """Runs a hemisphere for a day.
    """

    #
    # Write commands for creating empty NetCDF4 from template, and loading the
    # nasateam and bootstrap fields into it.
    #

    #
    # Check to see if any inputs exist. If not, return a no-op shell command
    #
    tb_file_root = get_rss_brightness_files_root(conf)
    if (tb_file_root is None
            and (not os.path.exists("%(bt_final_dir)s/%(goddard_bt_file)s" % conf))
            and (not os.path.exists("%(nt_final_dir)s/%(goddard_nt_file)s" % conf))):
        # no inputs
        return ":\n"

    #
    # Make certain the working directory exists
    #
    try:
        os.makedirs(conf["work"])
    except OSError as e:
        # ignore "already exists" errors
        if e.errno != errno.EEXIST:
            raise

    create_working_cdl(conf["cdr_template"], "%(work)s/%(cdr_cdl)s" % conf, conf)

    #
    # Start creating shell command
    #
    cmdstring = (
        "rm %(work)s/%(cdr_nc)s\n"
        "export NC_BIN=%(external_bin)s\n"
        "%(CDR_bin)s/make_empty.bash %(work)s/%(cdr_cdl)s %(work)s/%(cdr_nc)s\n"
    ) % conf

    # Process Goddard sea ice if it exists for the day.
    if (os.path.exists("%(nt_final_dir)s/%(goddard_nt_file)s" % conf)):
        #
        # Read nasateam sea ice grid and write scaled version to work dir.
        #
        nt_var = cl.CDRVariable("nt_sea_ice_conc", conf["h_dims"], header_len=300,
                                nctype=cl.NC_BYTE, intype=cl.NC_BYTE,
                                filters=[nasateam_value_filter, cl.recast])
        nt_var.from_file("%(nt_final_dir)s/%(goddard_nt_file)s" % conf)
        nt_var.to_file("%(scaled_nt)s" % conf)
        cmdstring += (
            "%(CDR_bin)s/ncload %(nt_var)s %(nt_final_dir)s/%(goddard_nt_file)s"
            " %(work)s/%(cdr_nc)s nt %(DATE)s\n"
        ) % conf

    #
    # Read bootstrap sea ice grid, if bootstrap is present
    #
    if (os.path.exists("%(bt_final_dir)s/%(goddard_bt_file)s" % conf)):
        if (conf["hemisphere"] == "n" and
                os.path.exists("%(nt_final_dir)s/%(goddard_nt_file)s" % conf)):
            # get coast from Nasateam
            flagsFromNasateamMask = cl.cdr_mask_filter_factory(nt_var, use=[253])
            bt_var = bootstrap_grid_from_file(
                conf["bt_var"], "%(bt_final_dir)s/%(goddard_bt_file)s" % conf,
                missingValueMask=flagsFromNasateamMask
            )
        else:  # southern hemisphere, or no nasateam to get coast from
            bt_var = bootstrap_grid_from_file(
                conf["bt_var"], "%(bt_final_dir)s/%(goddard_bt_file)s" % conf
            )
        bt_var.tofile("%(scaled_bt)s" % conf)
        cmdstring += (
            "%(CDR_bin)s/ncload %(bt_var)s %(scaled_bt)s %(work)s/%(cdr_nc)s raw\n"
        ) % conf

    #
    # Add Goddard combo loading to shell commands if both inputs exist
    #
    if (os.path.exists("%(bt_final_dir)s/%(goddard_bt_file)s" % conf) and
            os.path.exists("%(nt_final_dir)s/%(goddard_nt_file)s" % conf)):
        cmdstring += create_combo_concentration(
            conf["legacy_combo_var"], conf["legacy_combo_file"],
            "", "",  # blanks indicating no QA for this combo field
            nt_var, bt_var)

    #
    # If input brightness temperatures are available create CDR fields.
    #
    if tb_file_root is not None:
        # raw passes for 2007 03 15 in comparison to pmalgos
        runNowString = make_raw_nasateam(conf["auto_nt"], tb_file_root)
        # off by 44 bytes for 2007 03 15 in comparison to pmalgos
        runNowString += make_raw_bootstrap(conf["auto_bt"], tb_file_root)

        # wait here for the above pmalgos calls to complete, then read in
        # to variables and apply filters to it.
        cl.run_and_check(runNowString)

        nasateam_grid_from_raw_grid(conf["auto_nt"])
        auto_nt_var = cl.CDRVariable(conf["auto_nt"], conf["h_dims"],
                                     nctype=cl.NC_BYTE, intype=cl.NC_BYTE,
                                     filters=[nasateam_value_filter, cl.recast])
        auto_nt_var.from_file(conf["auto_nt"])
        auto_nt_var.to_file(conf["auto_nt_scaled"])
        if (conf["hemisphere"] == "n"):
            # coast - Pole and missing already "missing" in bootstrap
            flagsFromAutoNasateamMask = cl.cdr_mask_filter_factory(
                auto_nt_var, use=[conf["coast_flag"]]
            )
            auto_bt_var = bootstrap_grid_from_file(
                "auto_bt_var", conf["auto_bt"],
                missingValueMask=flagsFromAutoNasateamMask
            )
        else:  # southern hemisphere
            # missing values already in Bootstrap, has interpolation - don't mask!
            auto_bt_var = bootstrap_grid_from_file(
                "auto_bt_var", conf["auto_bt"]
            )
        auto_bt_var.tofile(conf["auto_bt_scaled"])

        #
        # Create CDR concentration and QA fields and to shell commands
        #
        cmdstring += create_combo_concentration(
            conf["cdr_combo_var"], conf["cdr_combo_file"],
            conf["cdr_qa_var"], conf["cdr_qa_file"],
            auto_nt_var, auto_bt_var)

        # need to remove lakes from NT and BT before creating stddev of CDR
        ocean_mask = numpy.fromfile(
            conf["oceanmask"], dtype=cl.NC_BYTE,
            count=conf["h_dims"][0]*conf["h_dims"][1]
        )
        ocean_mask.shape = tuple(conf["h_dims"])
        # lake mask value in oceanmask is 4
        auto_nt_var[ocean_mask == 4] = conf["lake_flag"]
        auto_bt_var[ocean_mask == 4] = conf["lake_flag"]
        # Compute the Standard Deviation Field.

        stddev_var = cl.local_std_dev(auto_nt_var, auto_bt_var, conf["stdev_fill_value"])
        stddev_var = cl.scale_from_percent_to_fraction(stddev_var,
                                                       conf["stdev_fill_value"])
        stddev_var.tofile(conf["stddev_file"])

        #
        # Add standard deviation to shell commands
        #
        cmdstring += (
            "%(CDR_bin)s/ncload %(cdr_stddev_var)s %(stddev_file)s"
            " %(work)s/%(cdr_nc)s\n"
        ) % conf

    return cmdstring


################################## BEGIN MAIN ##################################
def main(args):
    global conf
    # Areas is a list of recognized representations of what hemisphere should be
    # processed.
    areas = ["north", "south", "both", "n", "s", "b"]

    # Shell actions is a list of strings that will be executed by a shell at the
    # end of operations. This allows all of the metadata to be assembled before the
    # NetCDF file is created. The binary data for the CDR should be written to temp
    # files in subdirectories of the working directory
    shell_actions = []

    if (len(args) != 3 and len(args) != 5):
        usage()

    if ((len(args) == 5) and ((not args[3].isdigit()) or (not args[4].isdigit()))):
        usage()

    preset_config_dict = {}
    if (len(args) == 5):
        preset_config_dict = {"cdr_version":
                              cl.create_product_version_string(args[3], args[4])}
        del args[3]
        del args[3]

    # find which hemisphere to run, both is OK.
    try:
        # smash case to maintain correctness of lazy coding
        hemisphere = args[1].lower()
    except:
        print("\nHemisphere (and then date) required. "
              "Try one of north, south, or both.\n")
        exit(1)
    hemisphere = hemisphere.strip()
    if (hemisphere not in areas):
        print("\nHemisphere should be one of north, south, or both.\n")
        exit(1)
    else:
        hemisphere = hemisphere[0]

    if hemisphere == "n" or hemisphere == "b":
        # Create northen dictionary
        preset_config_dict['hemisphere'] = 'n'
        conf = cfg.create_configuration_dictionary(
            args,
            preset=preset_config_dict
        )
        shell_actions.extend(run_hemi("n", conf["DATE"]))
        shell_actions.extend(cleanup_work(conf))

    if hemisphere == "s" or hemisphere == "b":
        # Create southern dictionary
        preset_config_dict['hemisphere'] = 's'
        conf = cfg.create_configuration_dictionary(
            args,
            preset=preset_config_dict
        )
        shell_actions.extend(run_hemi("s", conf["DATE"]))
        shell_actions.extend(cleanup_work(conf))

    command_string = "".join(shell_actions)
    cl.run_and_check(command_string)


if __name__ == '__main__':
    main(sys.argv)
