#########################################################################################
#@***h* Sea_Ice_CDR/cdr_lib.py
#
# NAME
#  cdr_lib.py
#
# LOCATION
#  src/cdr_lib.py
#
# PURPOSE
#  Utility routines and classes for needed across the CDR project.
#
# DESCRIPTION
# This is just a general collection of routines that will be used in each part
# of the cdr process, things like standard directory locations, type
# conversions, masking functions, etc.
#
# AUTHOR
#  nsidc@nsidc.org
#  National Snow & Ice Data Center, University of Colorado, Boulder
#
# COPYRIGHT
#  Copyright 2011 Regents of the University of Colorado. The Regents of the
#  University of Colorado grants, and you accept, a personal, nonexclusive,
#  nontransferable license to use this software, at no charge, in accordance
#  with the terms herein, solely for internal research and development, which
#  is limited to testing, measuring, assessing, evaluating the software and/or
#  data that you generate using this software, and not for any commercial
#  purpose. You may not develop derivative works based on this software, or
#  redistribute this software.
#
# CREATION DATE
#  June 2011
#
# EXTERNALS
#  __future__.division
#  __future__.print_function
#  __future__.absolute_import
#  __future__.unicode_literals
#  calendar
#  counter
#  datetime
#  getpass
#  os
#  sys
#  numpy
#  re
#  subprocess
#  netCDF4.Dataset
#
# SUBROUTINES
#   _all_missing
#   _fetch_nasateam_raw_data
#   _fetch_raw_bootstrap_data
#   _formatted_comment
#   _missing_comment_generic
#   _svn_info
#   add_masked_border
#   apply_coastlines
#   apply_lakes
#   apply_land
#   apply_pole_hole
#   bootstrap_concentration_from_file
#   cdr_mask_filter_factory
#   cdr_root_dir
#   clip_masked_array
#   create_product_version_string
#   current_software_version
#   data_replacer
#   days_since_noaa_epoch
#   dims_from_hemisphere
#   fetch_combined_landmask
#   fetch_nc_grid
#   fetch_pole_hole
#   fetch_scaled_nasateam_from_file
#   find_all_matching_subfile
#   find_matching_bt_month_file
#   find_matching_nt_month_file
#   find_month_of_daily_cdr_netcdf_files
#   get_cdr_valid_year_range
#   history_attribute
#   is_north
#   is_south
#   local_std_dev
#   missing_daily_comment
#   missing_monthly_comment
#   north_dims
#   platform_from_date
#   read_bin_grid
#   recast
#   run_and_check
#   scale_from_percent_to_fraction
#   south_dims
#   standard_config_filename
#   time_coverage_end
#   time_coverage_start
#   work_dir_ncfilename
#   yyyymmdd
#   days_in_year
#   cdr_dataset
#   satellite_for_month
#   latest_satellite
#   class CDRVariable
#     __array_finalize__
#     __array_wrap__
#     __new__
#     from_file
#     to_file
#
#
# INPUTS
#  n/a
#
# OUTPUTS
#  n/a
#
# MODIFICATION HISTORY
#  Dates and Changes to the script
#
#@*****
################################################################################

# import a few __future__ features to keep possible upgrades cleaner
## division info: http://www.python.org/dev/peps/pep-0238/
##docs.python.org/py3k/howto/pyporting.html#try-to-support-python-2-6-and-newer-only
from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
## Default to unicode, for 8-bit strings prepend to the string declaration: b''
from __future__ import unicode_literals

import calendar
import counter
import datetime as dt
import getpass
import os
import sys
import numpy

import re
import subprocess
from netCDF4 import Dataset


################################################################################
#
# globals in lieu of #defines
#
# These are named as NetCDF types and defined as NumPy types. They only include
# numeric types - this does not suport NC_STRING, compound, or unknown types.
#
################################################################################
# In increasing memory size.
global NC_BYTE
global NC_CHAR
global NC_SHORT
global NC_INT
global NC_FLOAT
global NC_DOUBLE

NC_BYTE = b'uint8'
NC_UBYTE = b'uint8'  # Only listed in some lists, do not count on this type
NC_CHAR = b'int8'
NC_SHORT = b'int16'
NC_USHORT = b'uint16'
NC_INT = b'int32'
NC_LONG = NC_INT
NC_UINT = b'uint32'
NC_INT64 = b'int64'
NC_UINT64 = b'uint64'
NC_FLOAT = b'float32'
NC_DOUBLE = b'float64'
#NC_STRING
NC_BOOL = b'uint8'
DAILY_CDR_SATELLITE_DATE_REGEX = "seaice_conc_daily_[ns]h_(...)_([0-9]{8})"


def run_and_check(command_string):
    print("Running:\n%s\n" % (command_string))
    try:
        returns = subprocess.call(command_string, shell=True)
    except OSError as e:
        sys.stderr.write("\nError during system call:\n%s\n" % (str(e)))
        exit(1)
    return returns


def recast(array):
    array = array.astype(array.ncType)


def standard_config_filename():
    """ Default location of the configuration filename"""

    return os.path.join(cdr_root_dir(), "config", "cdr.config")


def cdr_root_dir():
    """
    Return the root directory for this project.
    Many file dependencies are computed relative to this location
    """

    if (os.path.dirname(sys.argv[0]) != ""):
        CDR_root = os.path.realpath(os.path.join(sys.path[0], '..'))
    else:
        CDR_root = os.path.realpath(os.path.abspath(".."))

    return CDR_root


def cdr_mask_filter_factory(mask, ignore=[0], use=[]):
    """Create a closure that applies a masking array to a data array.

    mask - a masking ndarray showing the values that should be overwritten
           onto a data array.

    ignore - a list of values in the mask to ignore (data at those locations
             is unchanged)

    The function returned will take a single argument of an ndarray of data,
    onto which the mask should be written, excepting those locations where the
    mask has values on the ignore list. Note that it will not check to see if
    the dimensions of the mask match the dimensions of the data.
    """
    if (use != []):
        def mask_function_by_use(cdrVar):
            for use_value in use:
                cdrVar[mask == use_value] = mask[mask == use_value]
        return mask_function_by_use
    else:
        def mask_function_by_ignore(cdrVar):
            original_cdrValues = cdrVar.copy()
            cdrVar[:] = mask.copy()
            for ignored_value in ignore:
                cdrVar[mask == ignored_value] = original_cdrValues[mask == ignored_value]
        return mask_function_by_ignore


# Set of functions used in computing the standard deviation field for the cdr
def add_masked_border(inArray):
    """ Create a 1 grid cell border around the input array that is masked as
    invalid and will be excluded in standard deviation calculations. This is
    just convenience to eliminate worrying about edge cases for the STDDEV
    computations.
    """

    borderArray = numpy.zeros(numpy.array(inArray.shape) + 2)
    borderArray = numpy.ma.masked_equal(borderArray, 0)
    borderArray[1:-1, 1:-1] = inArray
    return borderArray


def clip_masked_array(inArray):
    """ Given a recently masked array, snip out the center piece and return a
    normal ndarray
    """

    clippedArray = inArray.data[1:-1, 1:-1]
    return clippedArray


def local_std_dev(AArray, BArray, std_fill_value):
    """ Compute the local standard deviation of two sea ice concentration arrays

    AArray - first array of data
    BArray - second array of data
    std_fill_value - what to fill the value with if there are fewer
                            than 6 values to use to compute the stddev

    This code assumes we are working with SIConc data and masks input values
    <= 251 and ignores them during the computations.

    returns - standard deviation of the two arrays where each [I,J] in the
    output array represents the standard deviation of the 1 cell neighborhood
    around I, J in both AArray and BArray

    Easy enough to modify for different kernels in the future if you need.
    """

    # Based on field specification in emails from Walt,
    # some details elaborated in meeting on Aug 9, 2011
    #  - no standard deviation where no sea ice estimates
    #  - stddev may be calculated from just one of BT or NT
    #  - flags should be present for both "not applicable" and
    #    "not appropriate to calculate" (too much missing data)
    #  - a minimum of six non-missing cells needs to be present
    #    before calculating the stddev.
    #  - zeros from climate masks are considered valid
    #  - land, coast, and lake are all considered to be missing data

    AMasked = add_masked_border(AArray)
    BMasked = add_masked_border(BArray)
    assert(AMasked.shape == BMasked.shape)

    # This is the size of the bordered data.
    stddev = numpy.ma.empty_like(AMasked, dtype=NC_FLOAT)

    ## need to mask the input data by flags as well
    AMasked = numpy.ma.masked_greater_equal(AMasked, 251)
    BMasked = numpy.ma.masked_greater_equal(BMasked, 251)

    # Loop over the input array shape. Then by adding one to both indices, we
    # extract from the masked array neighborhoods without worrying about
    # overflowing the array indices.
    for (y, x), unused_idx in numpy.ndenumerate(AArray):
        ym = y+1
        xm = x+1
        AVals = AMasked[ym-1:ym+2, xm-1:xm+2]
        BVals = BMasked[ym-1:ym+2, xm-1:xm+2]
        comb = numpy.ma.concatenate((AVals, BVals))
        if comb.count() >= 6:
            stddev[ym, xm] = comb.std(ddof=1)
        else:
            stddev[ym, xm] = std_fill_value

    stddev = clip_masked_array(stddev)
    return stddev


def scale_from_percent_to_fraction(stddev_var, fill_value):
    """ Take valid percent values and scale them to a fraction [0,1] leaving
    any numbers less than the fill value as they are the standard fill value for
    stdev is -1
    """
    masked_stddev = numpy.ma.masked_less_equal(stddev_var, fill_value)
    masked_stddev = masked_stddev / 100.
    return masked_stddev.data


def time_coverage_end(conf):
    """
    return time coverage end string in CF format
    """

    return dt.datetime(
        conf["iYY"], conf["iMM"], conf["iDD"], 23, 59, 59
    ).strftime("%Y-%m-%dT%H:%M:%SZ")


def time_coverage_start(conf):
    """
    return time coverage start string in CF format
    """
    return dt.datetime(
        conf["iYY"], conf["iMM"], conf["iDD"], 0, 0, 0
    ).strftime("%Y-%m-%dT%H:%M:%SZ")


def days_since_noaa_epoch(conf):
    """
    return number of days since NOAA time epoch to time_coverage_start
    """
    return (dt.date(conf["iYY"], conf["iMM"], conf["iDD"]) - dt.date(1601, 1, 1)).days


def platform_from_date(date):
    """
    return preferred platform string for the date given
    """
    if date < dt.date(1978, 10, 26):
        platform_string = "pre-NIMBUS"
    elif date >= dt.date(1978, 10, 26) and date < dt.date(1987, 7, 8):
        platform_string = "n07"
    elif date >= dt.date(1987, 7, 8) and date < dt.date(1991, 12, 3):
        platform_string = "f08"
    elif date >= dt.date(1991, 12, 3) and date < dt.date(1995, 10, 1):
        platform_string = "f11"
    elif date >= dt.date(1995, 10, 1) and date < dt.date(2008, 1, 1):
        platform_string = "f13"
    elif date >= dt.date(2008, 1, 1):
        platform_string = "f17"
    return platform_string


def fetch_combined_landmask(conf):
    """
    fetch the current combined land mask and return as properly shaped ndarray
    """

    land_mask = numpy.fromfile(conf['combined_oceanmask'],
                               dtype=NC_BYTE).reshape(conf['h_dims'])
    return land_mask


def data_replacer(target_data, identifier_data, from_to):
    """
    Given an input target_data array, return a modified copy of that array
    where for each key in the from_to dictionary has been used to find every
    location in the identifier_data.  Those locations are used to replace the
    values in the target_data copy with the from_to['value']

    locs = identifier_data == from_to[key]
    target_data[locs] = from_to[value]

    Arguments:
    - `target_data`: input data
    - `identifier_data`: array to find key locations
    - `from_to`: key value pairs where key is into identifier_data and value is
    the replacement value
    """

    ret_data = target_data
    for loc_val, replace_val in from_to.items():
        ret_data[identifier_data == loc_val] = replace_val
    return ret_data


def apply_lakes(data, conf):
    """return a copy of the input data where combined oceanmask's lake areas
    have been replaced with the conf['lake_flag']

    Arguments:
    - `data`: data to return a modified copy of
    - `conf`: standard conf object.
    """

    lm_lake_value = conf['combined_oceanmask_values']['lake']
    land_mask = fetch_combined_landmask(conf)
    return data_replacer(data, land_mask, {lm_lake_value: conf['lake_flag']})


def apply_land(data, conf):
    """
    return a copy of the input data array replacing its values with land_flag
    in locations where the combined oceanmask has land values

    Arguments:
    - `data`: input concentration data
    - `conf`: standard config object
    """

    lm_land_value = conf["combined_oceanmask_values"]["land"]
    land_mask = fetch_combined_landmask(conf)
    out_data = data_replacer(data, land_mask, {lm_land_value: conf['land_flag']})
    return out_data


def apply_coastlines(data, conf):
    """
    return a copy of the input data array replacing its values with coast_flag
    in locations where the combined oceanmask has coast values

    Arguments:
    - `data`: input concentration data
    - `conf`: standard config object
    """

    lm_coast_value = conf["combined_oceanmask_values"]["coast"]
    land_mask = fetch_combined_landmask(conf)
    out_data = data_replacer(data, land_mask, {lm_coast_value: conf['coast_flag']})
    return out_data


def fetch_pole_hole(conf):
    """
    read the pole hole data file and return an 1/0 mask of the hole
    """
    pole_hole_file = "%(ancillary)s/%(pole_hole_file)s" % conf
    fp = open(pole_hole_file, 'rb')
    fp.read(300)  # discard header
    pole = numpy.fromfile(fp, dtype=NC_BYTE).reshape(conf['h_dims'])
    fp.close()
    return pole


def apply_pole_hole(data, conf):
    """
    return a copy of the input data where the pole hole has been applied

    Arguments:
    - `data`: input data
    - `conf`: standard config
    """

    if is_south(conf):
        return data

    pole = fetch_pole_hole(conf)
    out_data = numpy.where(pole != 0, conf['pole_flag'], data)
    return out_data


def _fetch_nasateam_raw_data(nt_filename, conf):
    """
    skip the header and read the data of a nasateam file, returning a
    properly shaped array

    Arguments:
    - `nt_filename`: full path to the nasateam input file.
    - `conf`: standard config object
    """

    assert os.path.exists(nt_filename)
    file_pointer = open(nt_filename, "rb")
    file_pointer.read(300)  # discard header
    nt_conc = numpy.fromfile(file_pointer, dtype=NC_BYTE).reshape(conf['h_dims'])
    file_pointer.close()
    return nt_conc


def fetch_scaled_nasateam_from_file(nt_filename, conf):
    """
    Given a standard nasateam file return an array with the data scaled properly 0-100
    and the flags standard.

    Arguments:
    - `nt_filename`: full path to the nasateam input file.
    - `config`: standard config object
    """

    nt_conc = _fetch_nasateam_raw_data(nt_filename, conf)
    nt_conc = numpy.ma.masked_greater(nt_conc, conf['nasateam_max_raw_data'])
    # add .5 and truncate for rounding when converting back to byte.
    nt_conc = numpy.asarray((((nt_conc) / conf["nasateam_scale_factor"]) + .5),
                            dtype=NC_BYTE)
    return nt_conc


def _fetch_raw_bootstrap_data(bt_filename, conf):
    """Read the raw bootstrap data and return an unscaled, unflagged,
    properly shaped array

    Arguments:
    - `bt_filename`: full path to the bootstrap file
    - `conf`: standard config object
    """

    assert os.path.exists(bt_filename)
    bt_conc = numpy.fromfile(bt_filename, dtype=NC_SHORT).reshape(conf['h_dims'])
    return bt_conc


def bootstrap_concentration_from_file(bt_filename, conf):
    """
    Given Bootstrap file in standard format standard scaling and offsets,
    return a numpy array to the caller that is properly scaled and flagged.

    Arguments:
    - `bt_filename`: full path to the published bootstrap data.
    - `conf`: standard config object
    """

    bt_conc = _fetch_raw_bootstrap_data(bt_filename, conf)
    bt_conc = numpy.ma.masked_greater(bt_conc, conf["bootstrap_max_raw_data"])

    # Walt has decided in email that we will use a more standard rounding,
    # which is to add .5 and truncate.
    bt_conc = numpy.floor((bt_conc / conf["bootstrap_scale_factor"]) + .5)
    bt_conc = numpy.asarray(bt_conc, dtype=NC_BYTE)

    # Add proper flags
    bt_conc = apply_pole_hole(bt_conc, conf)
    bt_conc = apply_land(bt_conc, conf)
    bt_conc = apply_coastlines(bt_conc, conf)
    return bt_conc


def fetch_nc_grid(nc_file, var_name, nc_var_type):
    """
    Return the 2d sea-ice grid as numpy array from the variable (var_name)
    contained in nc_file.

    This is used for reading the cdr variable out of at netcdf cdr file.
    it should theoretically work for several variables, but only now tested for the
    immediate need.

    nc_var_type should be set to NC_BYTE for netcdf variables with type byte,
    because we are using the classic model, we don't have unsigned types.  We
    will see values of -2 when we expect flags of 254

    >>> np.asarray([-2])
    array([-2])
    >>> np.asarray([-2],dtype=cl.NC_BYTE)
    array([254], dtype=uint8)
    """

    ncroot = Dataset(nc_file, 'r')
    cdr = ncroot.variables[var_name]
    cdr.set_auto_maskandscale(False)
    return_var = cdr[0]
    ncroot.close()
    return numpy.asarray(return_var, dtype=nc_var_type)


def _all_missing(nc_variable):
    return numpy.all(numpy.ma.getmask(nc_variable[:]))


def _formatted_comment(var_list, day_or_month):
    if not var_list:
        return "none"

    var = ['variable', 'variables'][len(var_list) > 1]
    tobe = ['is', 'are'][len(var_list) > 1]
    return " ".join(["The", var,  ", ".join(var_list),
                     tobe, "not available for this {id}.".format(id=day_or_month)])


def missing_daily_comment(nc_dataset):
    checked_variables = ['seaice_conc_cdr',
                         'stdev_of_seaice_conc_cdr',
                         'melt_onset_day_seaice_conc_cdr',
                         'qa_of_seaice_conc_cdr',
                         'goddard_merged_seaice_conc',
                         'goddard_nt_seaice_conc',
                         'goddard_bt_seaice_conc']
    day_or_month = "day"
    return _missing_comment_generic(nc_dataset, checked_variables, day_or_month)


def missing_monthly_comment(nc_dataset):
    checked_variables = ['seaice_conc_monthly_cdr',
                         'stdev_of_seaice_conc_monthly_cdr',
                         'melt_onset_day_seaice_conc_monthly_cdr',
                         'qa_of_seaice_conc_monthly_cdr',
                         'goddard_merged_seaice_conc_monthly',
                         'goddard_nt_seaice_conc_monthly',
                         'goddard_bt_seaice_conc_monthly']
    day_or_month = "month"
    return _missing_comment_generic(nc_dataset, checked_variables, day_or_month)


def _missing_comment_generic(nc_dataset, checked_variables, day_or_month):
    missing_variables = []
    for var in checked_variables:
        if var in nc_dataset.variables and _all_missing(nc_dataset.variables[var]):
            missing_variables.append(var)
    return _formatted_comment(missing_variables, day_or_month)


def history_attribute(filename):
    return "{filename} created: {date} by user: {user} using software: {version}".format(
        filename=filename,
        date=dt.datetime.utcnow().strftime("%Y-%m-%d %H:%m:%SZ"),
        user=getpass.getuser(),
        version=current_software_version()
    )


def work_dir_ncfilename(conf):
    return "%(work)s/%(cdr_nc)s" % conf


def read_bin_grid(bin_file, hemi, file_data_type=NC_BYTE):
    """
    Read a binary file from disk and return a numpy array
    """

    dims = dims_from_hemisphere(hemi)
    size = dims[0] * dims[1]
    outarray = numpy.fromfile(bin_file, dtype=file_data_type)
    assert outarray.size == size
    outarray.shape = tuple(dims)
    return outarray


def dims_from_hemisphere(hemi):
    """ Return the correct size of the grid for the input hemisphere.
    hemi is a case independent string (N, North, north, nor)
    """

    if hemi[0].upper() == 'N':
        return north_dims()
    else:
        return south_dims()


def is_south(conf):
    """
    Return True if this is a config for a Southern object

    Arguments:
    - `conf`: Standard config object
    """

    return conf["hemisphere"] == "s"


def is_north(conf):
    """
    Return True if this is a config for a Northen object

    Arguments:
    - `conf`: Standard config object
    """

    return conf["hemisphere"] == "n"


def south_dims():
    """ grid dimesions for south grids"""

    return [332, 316]


def north_dims():
    """ grid dimesions for north grids"""

    return [448, 304]


def find_all_matching_subfile(regex, top_dir):
    """
    given a regular expression and a top_level directory traverse the path returning all
    files that match the regular expression

    Arguments:
    - `regex`: compiled regular expression to match
    - `top_dir`: top level to start your search
    """

    all_files = []
    for root, dirs, files in os.walk(top_dir):
        files = filter(regex.search, files)
        for filename in files:
            all_files.append(os.path.join(root, filename))
    all_files.sort()
    return all_files


def get_cdr_valid_year_range(conf):
    """
    Use the existing CDR daily files to return the first and last year for
    which valid data exists
    """
    path_to_year_data = conf["cdr_netcdf_daily_file_dir"]
    (cdr_top_dir, year_dir) = os.path.split(path_to_year_data)
    years = list()

    for dir, subdir, files in os.walk(cdr_top_dir):
        for s in subdir:
            years.append(int(s))

    years.sort()
    return {'first_year': years[0], 'last_year': years[-1]}


def find_month_of_daily_cdr_netcdf_files(conf):
    """
    Given a config object return a list of daily CDR files for the year and month of the
    config object.
    """

    regex = re.compile("%1.1sh_..._%4.4s%2.2s[0-9]{2}_" % (conf['HEMI'].lower(),
                                                           conf["YYYY"], conf["MM"]))
    top_dir = conf["cdr_netcdf_daily_file_dir"]
    return find_all_matching_subfile(regex, top_dir)


def find_matching_bt_month_file(conf):
    """
    get the published BT Monthly file
    """

    top_dir = conf["bt_monthly_final_dir"]
    regex = re.compile("_%4.4s%2.2s_" % (conf["YYYY"], conf["MM"]))
    return find_all_matching_subfile(regex, top_dir)[0]


def find_matching_nt_month_file(conf):
    """
    get the published NT Monthly file
    """

    top_dir = conf["nt_monthly_final_dir"]
    regex = re.compile("_%4.4s%2.2s_" % (conf["YYYY"], conf["MM"]))
    return find_all_matching_subfile(regex, top_dir)[0]


def current_software_version():
    return "{0}@{1}".format(_svn_info("URL"), _svn_info("Revision"))


def create_product_version_string(product_version, product_revision):
    return "v%sr%s" % (product_version, product_revision)


def _svn_info(attribute):
    cmd = "svn info"
    svn_info = subprocess.Popen([cmd],
                                shell=True,
                                stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE).communicate()[0]

    pattern = "{0}: (.*)".format(attribute)
    m = re.search(pattern, svn_info)
    if m and m.groups():
        return m.group(1)
    else:
        raise RuntimeError("Unable to get svn attribute {0}".format(attribute))


def yyyymmdd(year, doy):
    date = dt.date(year, 1, 1) + dt.timedelta(days=doy-1)
    return date.strftime("%Y%m%d")


def days_in_year(year):
    return [365, 366][calendar.isleap(year)]


def cdr_dataset(conf):
    cdr_filename = os.path.join(conf['output'], conf['cdr_nc'])
    return Dataset(cdr_filename, mode='r+', format='NETCDF4_CLASSIC')


def satellite_for_month(daily_cdr_files):
    search_results = [re.search(DAILY_CDR_SATELLITE_DATE_REGEX, file)
                      for file in daily_cdr_files]
    satellites = [result.group(1) for result in search_results]
    ordered_counts = counter.Counter(satellites).most_common()

    if len(ordered_counts) > 1 and ordered_counts[0][1] == ordered_counts[1][1]:
        satellite_name = latest_satellite(daily_cdr_files)
    else:
        satellite_name = ordered_counts[0][0]

    return satellite_name


def latest_satellite(daily_cdr_files):
    search_results = [re.search(DAILY_CDR_SATELLITE_DATE_REGEX, file)
                      for file in daily_cdr_files]
    satellites = [result.group(1) for result in search_results]
    dates = [int(result.group(2)) for result in search_results]
    return satellites[dates.index(max(dates))]


class CDRVariable(numpy.ndarray):
    """CDR variable class

    This class inherits from numpy.ndarray so that it will have the ability to
    be used "naked" as a variable in by-element equations. The array functions that
    numpy provides, like the by-element maximum() function, can also be used on this
    class.

    It adds a set of attributes that ties it to its role in a NetCDF file, and allows
    NetCDF appropriate functions to be added.

    "Raw" binary files can be read and written with the .fromfile() and .tofile() methods.

    The filtering functions should be kept outside of this class for maximum flexibility.
    """
    # subtype and shape arguments, and optional arguments following

    def __new__(subtype, varname, shape, dtype=b"", nctype=b"",
                intype=b"", header_len=0, attributes={},
                filters=[], buffer=None, offset=0, strides=None, order=None):
        """  To instantiate new objects, numpy uses __new__() instead of __init__().
        This is done to allow the array to create new arrays in a more variable-like way.
        The arguments subtype, shape, dtype, buffer, offset, strides, and order are
        inherited and documented in the numpy documentation.

        new input arguments:
            varname - the variable name used in the NetCDF and/or cdl file
            dtype - in-memory data type (inherited from numpy)
            ncType - data type of NetCDF variable
            ncAttributes - a dictionary of NetCDF "attributes" of the variable
            inputType - data type to read from binary grid, defaults to float "f"
            inputFilters - a list of functions applied to data read from binary streams
            headerLength - length of header to skip when reading from binary streams
        """

        # If types are partially defined, then try to find consistent values that help
        # minimise the information lost during processing
        if (dtype == "" or intype == "" or nctype == ""):
            if (intype == "" and nctype == ""):
                nctype = NC_FLOAT
            if (nctype == ""):
                # if nctype is still undefined, then intype is defined
                nctype = intype
                # nctype now certain to be defined
            if (intype == ""):
                intype = nctype
                # intype now certain to be defined
            if (dtype == ""):
                inByteLen = numpy.dtype(intype).itemsize
                ncByteLen = numpy.dtype(nctype).itemsize
                if (inByteLen > ncByteLen):
                    dtype = intype
                else:
                    dtype = nctype
                # dtype now certain to be defined - can use to make numpy.ndarray

        # Create the ndarray instance of our type. This will call the standard
        # ndarray constructor, but return an object of our type.
        # It also triggers a call to CDRVariable.__array_finalize__
        obj = numpy.ndarray.__new__(subtype, shape, dtype, buffer, offset, strides, order)

        # information about variable for NetCDF
        obj.variableName = varname
        obj.ncType = nctype
        obj.ncAttributes = attributes

        # information about variable source
        obj.inType = intype
        obj.inputFilters = filters
        obj.headerLength = header_len

        # Finally, we must return the newly created object:
        return obj

    def __array_finalize__(self, obj):
        """Set the defaults of this subclass.

            We set the default value for 'info' here, rather than
            in the __new__ method, because this method sees all
            creation of default objects - with the CDRVariable.__new__
            constructor, but also with arr.view(CDRVariable).
        """

        # ``self`` is a new object resulting from
        # ndarray.__new__(CDRVariable, ...), therefore it only has
        # attributes that the ndarray.__new__ constructor gave it -
        # i.e. those of a standard ndarray.
        #
        # We could have got to the ndarray.__new__ call in 3 ways:
        # From an explicit constructor - e.g. cdrVariable():
        #    obj is None
        #    (we're in the middle of the CDRVariable.__new__ constructor)
        if obj is None:
            return
        # From view casting - e.g arr.view(CDRVariable):
        #    obj is arr
        #    (type(obj) can be CDRVariable)
        # From new-from-template - e.g infoarr[:3]
        #    type(obj) is CDRVariable
        #
        # Note that we set the default value for 'info' here, rather than
        # in the __new__ method, because this method sees all creation of
        # default objects - with the CDRVariable.__new__ constructor, but
        # also with arr.view(CDRVariable).
        self.variableName = getattr(obj, 'variableName', None)
        self.ncType = getattr(obj, 'ncType', NC_FLOAT)
        self.ncAttributes = getattr(obj, 'ncAttributes', {})
        self.inType = getattr(obj, 'inType', NC_FLOAT)
        self.inputFilters = getattr(obj, 'inputFilters', [])
        self.headerLength = getattr(obj, 'headerLength', [])
        # We do not need to return anything

    def __array_wrap__(self, out_arr, context=None):
        """Allow numpy ufuncs to return the correct type."""
        return numpy.ndarray.__array_wrap__(self, out_arr, context)

    def from_file(self, filename, sep=""):
        """Load the array from "file", using numpy.fromfile()

        input arguments:
            file - Open file object or filename.
            sep  - separator to use if text, if empty (default) then use a binary read

        effect:
            the CDRVariable becomes loaded with data
        """
        if ((sep is None or sep == "") and self.headerLength > 0):
            inFile = open(filename, "rb")
            self.header = inFile.read(self.headerLength)
            self.data = numpy.fromfile(inFile, dtype=self.inType).data
            inFile.close()
        else:
            self.data = numpy.fromfile(filename, dtype=self.inType, sep=sep).data
        ## May be advisable to recast after filters if nctype larger than intype
        ## consider testing for that condition.
        if (self.inType != self.ncType):
            if (recast not in self.inputFilters):
                recast(self)
        for fn in self.inputFilters:
            print(fn.__name__, end=" ")
            fn(self)
        if (len(self.inputFilters) > 0):
            print()

    def to_file(self, filename):
        """Write the array to "file", using numpy.tofile()

        input arguments:
            file - Open file object or filename.
        """
        temp = self.astype(self.ncType)
        numpy.ndarray.tofile(temp, filename)
        del(temp)
