################################################################################
#@***h* Sea_Ice_CDR/cdr_melt.py
#
# NAME
#  cdr_melt.py
#
# LOCATION
#  src/cdr_melt.py
#
# PURPOSE
#  Functions to calculate melt-related CDR variables.
#
# DESCRIPTION
#  This module contains functions to calculate grid cells where melt
#  is detectable, where melt is occurring, and the melt onset day. In
#  addition, it contains a function to scale F17 brightness
#  temperatures so they are consistent with earlier satellite
#  data, resulting in consistent melting measures.
#
# AUTHOR
#  nsidc@nsidc.org
#  National Snow & Ice Data Center, University of Colorado, Boulder
#
# COPYRIGHT
#  Copyright 2011 Regents of the University of Colorado. The Regents of the
#  University of Colorado grants, and you accept, a personal, nonexclusive,
#  nontransferable license to use this software, at no charge, in accordance
#  with the terms herein, solely for internal research and development, which
#  is limited to testing, measuring, assessing, evaluating the software and/or
#  data that you generate using this software, and not for any commercial
#  purpose. You may not develop derivative works based on this software, or
#  redistribute this software.
#
# CREATION DATE
#  March 2013
#
# EXTERNALS
#   cdr_configuration
#   numpy
#
# SUBROUTINES
#   melt_detectable
#   melt_onset_day
#   melt_start_detected
#   melting
#   out_of_season
#   out_of_season_value
#   scaled_brightness_temperatures
#
# INPUTS
#  n/a
#
# OUTPUTS
#  n/a
#
# MODIFICATION HISTORY
#
#
#@*****
################################################################################
import cdr_configuration as cfg
import numpy as np


CONSTANTS = {}
cfg.load_constant_values(CONSTANTS)
cfg.cdr_constants(CONSTANTS)

MELT_ONSET_DAY_FILL_VALUE = -1
CONCENTRATION_THRESHOLD = CONSTANTS['melt_algo_min_ice_conc']
TB_DELTA_THRESHOLD_TENTHS_K = CONSTANTS['melt_algo_delta_tb_cutoff_tenths_k']
TB_MISSING = 0
NO_ICE = 0

MELT_SEASON_START_DOY = CONSTANTS['start_of_northern_melt_season']
MELT_SEASON_FINISH_DOY = CONSTANTS['end_of_northern_melt_season']

COAST_MASK = CONSTANTS['combined_oceanmask_values']['coast']
LAND_MASK = CONSTANTS['combined_oceanmask_values']['land']
LAKE_MASK = CONSTANTS['combined_oceanmask_values']['lake']
SHORE_MASK = CONSTANTS['combined_oceanmask_values']['shore']
NEAR_SHORE_MASK = CONSTANTS['combined_oceanmask_values']['near_shore']
UNDETECTABLE_MASKS = [COAST_MASK, LAND_MASK, LAKE_MASK]


def melting(concentrations, tb19, tb37):
    melting = np.logical_and(concentrations >= CONCENTRATION_THRESHOLD,
                             tb19 - tb37 < TB_DELTA_THRESHOLD_TENTHS_K)
    melting[tb19 == TB_MISSING] = False
    melting[tb37 == TB_MISSING] = False

    return melting


def melt_detectable(concentrations, lcs_mask):
    detectable = concentrations >= CONCENTRATION_THRESHOLD

    for undetectable in UNDETECTABLE_MASKS:
        detectable[lcs_mask == undetectable] = False

    return detectable


def out_of_season(doy):
    return doy < MELT_SEASON_START_DOY or doy > MELT_SEASON_FINISH_DOY


def out_of_season_value(doy, melt_onset_day):
    out_of_season_variable = np.copy(melt_onset_day)

    if doy < MELT_SEASON_START_DOY:
        out_of_season_variable[:] = MELT_ONSET_DAY_FILL_VALUE
        return out_of_season_variable

    if doy > MELT_SEASON_FINISH_DOY:
        return out_of_season_variable

    return None


def melt_onset_day(melting, doy, previous_mod, detectable):
    today_mod = np.copy(previous_mod)

    if out_of_season(doy):
        return out_of_season_value(doy, previous_mod)

    new_melting_locations = np.logical_and(melting,
                                           previous_mod == MELT_ONSET_DAY_FILL_VALUE)
    new_melting_locations = np.logical_and(new_melting_locations, detectable)
    today_mod[new_melting_locations] = doy
    return today_mod


def melt_start_detected(doy, melt_onset_day, concentration):
    detected = np.copy(melt_onset_day)

    if doy > MELT_SEASON_FINISH_DOY:
        detected[:] = False
        return detected

    return np.logical_and(detected != MELT_ONSET_DAY_FILL_VALUE, concentration > NO_ICE)


def scaled_brightness_temperatures(satellite, channel, tbs):
    """
    This function scales brightness temperatures for the F17 satellite and
    leaves TBs for other satellites unchanged.

    The scaling is done via the following linear equations from Walt Meier:

      19h_new = 1.021 * 19h - 1.681
      37h_new = 1.001 * 37h - 0.650

    Since brightness temperatures are in tenths of Kelvin, the b value is
    scaled by a factor of ten, i.e.:

      19h_new = (1.021 * (19h / 10) - 1.681) * 10
              = 1.021 * 19h - 16.81
      37h_new = (1.001 * (37h / 10) - 0.650) * 10
              = 1.001 * 37h - 6.5

    Values which were originally missing (0) are reset to missing after scaling.
    """
    scaled_tbs = tbs
    missing = tbs == 0

    if (satellite == 'f17'):
        constants = {
            '19h': [1.021, -16.81],
            '37h': [1.001, -6.5]
        }

        a = constants[channel][0]
        b = constants[channel][1]
        scaled_tbs = a * tbs + b
        scaled_tbs[missing] = 0

    return scaled_tbs
