#########################################################################################
#@***h* Sea_Ice_CDR/cdr_monthly_seaice.py
#
# NAME
#  cdr_monthly_seaice.py
#
# LOCATION
#  src/cdr_monthly_seaice.py
#
# PURPOSE
#  monthly CDR creation.
#
# DESCRIPTION
#  Existing Daily CDR data are collected and monthly means standard deviation
#  and quality information is computed and written to a monthly NetCDF4 CDR file
#
# AUTHOR
#  nsidc@nsidc.org
#  National Snow & Ice Data Center, University of Colorado, Boulder
#
# COPYRIGHT
#  Copyright 2012 Regents of the University of Colorado. The Regents of the
#  University of Colorado grants, and you accept, a personal, nonexclusive,
#  nontransferable license to use this software, at no charge, in accordance
#  with the terms herein, solely for internal research and development, which
#  is limited to testing, measuring, assessing, evaluating the software and/or
#  data that you generate using this software, and not for any commercial
#  purpose. You may not develop derivative works based on this software, or
#  redistribute this software.
#
# CREATION DATE
#  Jan 2012
#
# EXTERNALS
#  datetime
#  netCDF4.Dataset
#  numpy
#  numpy.ma
#  os
#  re
#  shutil
#  sys
#  cdr_configuration
#  cdr_lib
#
#
# SUBROUTINES
#   class CDRMonthlySeaice
#     __add_dynamic_global_attributes
#     __add_qa_flags_to_results
#     __add_time_variable_to_a_netcdf_file
#     __apply_landmask_flags_to_data
#     __compute_mean_concentration
#     __compute_qa_bt_and_nt_source
#     __compute_qa_melt
#     __compute_qa_open_water
#     __compute_std_deviation
#     __fill_empty_variables
#     __generate_land_mask
#     __generate_missing_mask
#     __get_merged_monthly_data
#     __get_number_days_with_valid_conc_data_grid
#     __get_number_days_with_valid_qa_grid
#     __get_qa_ocean_climatology_and_near_coast
#     __is_valid_month
#     __load_cdr_daily_files
#     __mark_missing_data
#     __output_monthly_filename
#     __prepare_input_files
#     __prepare_zeroed_netcdf_file
#     __set_missing_qa_values
#     __transfer_melt_onset_from_last_file_of_month
#     __transfer_monthly_bt_data
#     __transfer_monthly_nt_data
#     __write_merged_monthly_data
#     __write_to_a_netcdf_file
#     __init__
#     _filename_parts
#     _ftp_url
#     _source_files
#     merge_monthly_fields
#     output_monthly_filename
#     process_monthly_cdr
#     usage
#
# INPUTS
#  Daily NOAA/NSIDC Climate Data Record of Passive Microwave Sea Ice Concentration
#  http://nsidc.org/data/g02202.html
#
# OUTPUTS
#  Monthly NOAA/NSIDC Climate Data Record of Passive Microwave Sea Ice Concentration
#  http://nsidc.org/data/g02202.html
#
# MODIFICATION HISTORY
#
#
#@*****
################################################################################

import datetime as dt
from netCDF4 import Dataset
import numpy as np
import numpy.ma as ma
import os
import re
import shutil
import sys

import cdr_configuration as config
import cdr_lib as cl


class CDRMonthlySeaice():

    @classmethod
    def usage(self):
        print """

        CDR Monthly Sea Ice is generated by calling this script
        with the hemisphere, month, product version, and product
        revision you wish to process.

        > cdr_monthly_seaice north 200711 02 00
        """
        exit(1)

    def __init__(self, config):
        """ Initialize Monthly CDR class """

        self.month_files = cl.find_month_of_daily_cdr_netcdf_files(config)
        self.config = config
        self.mean_concentration = None
        self.std_deviation = None
        #note: self.mean_concentration and self.std_deviation will be numpy arrays.
        self.qa_flags_results_cube = None
        #note: qa_flags_results_cube will be a numpy array

    def __is_valid_month(self):
        """
        This function returns False if either there are not enough data files
        available to create a valid cube (files < 20(smmr:10)
        *OR* if all data for every day for the month is invalid.
        If there is some valid data and enough files it returns True
        """
        number_available_days = self.stacked_conc_cube.shape[2]
        all_data_is_missing = (
            self.stacked_conc_cube == self.config["missing_flag"]
        ).all()

        if ((number_available_days < self.config["days_required_for_valid_month"]) or
                all_data_is_missing):
            return False
        else:
            return True

    def __fill_empty_variables(self):
        """
        If there is no data we fill the variables with standard values.
        """
        # Just use the shape of the first element of the stacked_conc_cube to fill these
        self.mean_concentration = (self.stacked_conc_cube[:, :, 0] * 0 +
                                   self.config["concentration_fill_value"])
        self.std_deviation = (self.stacked_conc_cube[:, :, 0] * 0 +
                              self.config["stdev_fill_value"])
        self.qa_flags_results_cube = (self.stacked_conc_cube[:, :, 0] * 0 +
                                      self.config["qa_fill_value"])

    def __prepare_input_files(self):
        """
        It loads the set of input DAILY netCDF CDR files
        into a numpy object (self.stacked_conc_cube)
         then it creates the missing and landmask flags masks.
        """
        self.__load_cdr_daily_files()
        self.__generate_missing_mask()
        self.__generate_land_mask()

    def __load_cdr_daily_files(self):
        """
        Load all of the cdr files into "depth" stacked masked array for each variable
        """

        self.stacked_qa_flags = cl.fetch_nc_grid(self.month_files[0],
                                                 'qa_of_seaice_conc_cdr', cl.NC_BYTE)
        self.stacked_conc_cube = cl.fetch_nc_grid(self.month_files[0],
                                                  'seaice_conc_cdr', cl.NC_BYTE)

        for ncfile in self.month_files[1:]:
            this_conc = cl.fetch_nc_grid(ncfile, 'seaice_conc_cdr', cl.NC_BYTE)
            this_qa = cl.fetch_nc_grid(ncfile, 'qa_of_seaice_conc_cdr', cl.NC_BYTE)
            self.stacked_conc_cube = ma.dstack((self.stacked_conc_cube, this_conc))
            self.stacked_qa_flags = np.dstack((self.stacked_qa_flags, this_qa))

    def __generate_land_mask(self):
        """
        Read the first file from the file list and create a mask of the
        standard flags where only flag values are masked. land, coast, ocean,
        pole are masked, but missing is not masked.  This mask is used to
        transfer over flag values to final std and conc arrays
        """

        concentration = cl.fetch_nc_grid(self.month_files[0],
                                         'seaice_conc_cdr', cl.NC_BYTE)
        self.land_mask = ma.masked_inside(concentration,
                                          self.config["smallest_land_mask_flag"],
                                          self.config["largest_land_mask_flag"])

        # To aply a land mask
        # make sure land_mask.mask.any() is true first and
        # data[land_mask.mask] = land_mask[land_mask.mask]
        # land_mask[land_mask.mask] is all values that are between smallest and largest
        # flag.

    def __generate_missing_mask(self):
        """
        Compute the set of cells that will have valid data after making computations.
        These are grid cells that have had valid data for 20 (10 for SMMR) or more days.
        This ignores standard flag values in the concentration and only looks at missing
        """
        assert (self.stacked_conc_cube is not None), "Must load the data " \
            "before computing valid cells"

        cube_masked_for_missing = ma.masked_equal(self.stacked_conc_cube,
                                                  self.config["missing_flag"])

        num_good_data_days = ma.count(cube_masked_for_missing, axis=2)
        # if num_good_data_days less than 20 (or 10) we're eventually going to mask
        # that data as missing
        self.missing_mask = ma.masked_less(num_good_data_days,
                                           self.config["days_required_for_valid_month"])

    def __mark_missing_data(self, array_data, fill_value):
        """
        Given the an array of month data, use the missing mask and set
        the values that are masked with the fill_value
        """
        return_data = array_data
        if self.missing_mask.mask.any():
            return_data[self.missing_mask.mask] = fill_value
        return return_data

    def __apply_landmask_flags_to_data(self, input, force_data_type=cl.NC_BYTE):
        """
        Take the properly missing masked input data and replace any
        flags with the proper values (e.g. pole, land, coast)
        """
        input_variable = input
        if self.land_mask.mask.any():
            input_variable[self.land_mask.mask] = self.land_mask[self.land_mask.mask]
        input_variable = np.asarray(input_variable, dtype=force_data_type)
        return input_variable

    def __set_missing_qa_values(self):
        """
        QA values where concentration was invalid, need to be reset to 0,
        and flags need to be reapplied
        """
        self.qa_flags_results_cube = self.__mark_missing_data(
            self.qa_flags_results_cube, self.config["qa_fill_value"]
        )

    def __compute_mean_concentration(self):
        """
        Computes the mean sea ice concentration for the current month
        and flags the missing data values.
        """
        masked_cube = ma.masked_equal(self.stacked_conc_cube,
                                      self.config["missing_flag"])
        mean_conc = ma.mean(masked_cube, axis=2)
        mean_conc = self.__mark_missing_data(mean_conc.data,
                                             self.config["concentration_fill_value"])
        self.mean_concentration = self.__apply_landmask_flags_to_data(mean_conc)

    def __compute_std_deviation(self, dof):
        """
        Computes the standard deviation for a given month with $dof degrees of freedom
        and flags the missing data values.
        """
        stddev_no_data_value = self.config["stdev_fill_value"]  # NOAA spec for this var
        masked_cube = ma.masked_greater_equal(self.stacked_conc_cube,
                                              self.config["smallest_flag_value"])

        std_deviation = ma.std(masked_cube, ddof=dof, axis=2)

        # Fill the places that were masked with the fill values so that scaling works in
        # the next step
        std_deviation.fill_value = stddev_no_data_value
        std_deviation = std_deviation.filled()
        std_deviation = self.__mark_missing_data(std_deviation,
                                                 stddev_no_data_value)
        self.std_deviation = cl.scale_from_percent_to_fraction(std_deviation,
                                                               stddev_no_data_value)

    def __add_qa_flags_to_results(self, new_qa_flags_cube):
        """
        Create a new qa flag cube, or add new flags to it if one exists.
        """

        if self.qa_flags_results_cube is None:
            self.qa_flags_results_cube = np.array(new_qa_flags_cube, dtype='uint8')
        else:
            self.qa_flags_results_cube |= new_qa_flags_cube

    def __get_number_days_with_valid_conc_data_grid(self):
        """
        Count the number of days that have data (ie, 0-100).
        Return this as an grid, where each cell in the grid is
        the number of valid days for that cell.
        """
        masked_cube = ma.masked_greater(self.stacked_conc_cube, 100)
        results_valid_data = ma.where(masked_cube > 100, 0, 1)
        results_num_valid_days = ma.sum(results_valid_data, axis=2)
        return results_num_valid_days

    def __get_number_days_with_valid_qa_grid(self):
        """
        Count the number of days that has valid QA data.
        From looking at the CDR file, if the qa value
        is 255 (all bits are set), the data is not valid.
        This probably would never be the case, but
        we should account gor it just in case.
        """
        results_valid_data = ma.where(self.stacked_qa_flags == 255, 0, 1)
        results_num_valid_days = ma.sum(results_valid_data, axis=2)

        return results_num_valid_days

    def __compute_qa_bt_and_nt_source(self):
        """
        QA mask = 8 bits
        bit 0 (value = 1) = BT algorithm
        bit 1 (value = 2) = NT algorithm
        This method counts which algorithm was more used for a given pixel
        i.e. if pixel m for 3 days has [00000001] day 2 [00000011] and day 3 [00000001]
        the count should be [[0,0,0,0,0,0,1,3] therefore the BT algorithm was used more
        and the monthly flag should reflect that [00000001]
        If both algorithms were used the same number of times we
        set both bits to 1 [00000011]
        """
        #self.stacked_qa_flags & 1 is a bitwise operation to mask out of
        #the bt flag that generates a true false kind of variable that
        #we count with np.sum over the time series a.k.a. axis=2 a.k.a. days

        results_bt_cube = np.sum((self.stacked_qa_flags & 1), axis=2)
        results_nt_cube = np.sum(((self.stacked_qa_flags & 2) / 2), axis=2)

        result_cube_nt_gt_bt = ma.where(results_nt_cube > results_bt_cube, 2, 0)
        result_cube_bt_gt_nt = ma.where(results_nt_cube < results_bt_cube, 1, 0)
        result_cube_bt_eq_nt = ma.where(results_nt_cube == results_bt_cube, 3, 0)

        self.__add_qa_flags_to_results(result_cube_nt_gt_bt |
                                       result_cube_bt_gt_nt | result_cube_bt_eq_nt)

    def __compute_qa_melt(self):
        """
        QA mask = 8 bits
        bit 7 (value = 64) = melt occurred > 0 days
        bit 8 (value = 128) = melt occurred > 1/2 days in month
        This method counts the number of melt days recorded for a given pixel.
        If the number of days recorded is greater than 0
        (ie, there was at least some melt recorded for the month), bit 7 is set.
        If at least half the days in the month have the melt set, bit 8 will be
        set.  Note that if bit 8 is set, bit 7 will also always be set.
        """

        #self.stacked_qa_flags & 128 is a bitwise operation to mask out the melt flag,
        #which we can then shift down to make a 0/1 variable tha twe can count with.
        results_melt_cube = np.sum((self.stacked_qa_flags & 128) / 128, axis=2)
        num_valid_qa_days_in_month_grid = self.__get_number_days_with_valid_qa_grid()

        result_cube_melt_gt_0 = ma.where(results_melt_cube > 0, 64, 0)
        result_cube_melt_gt_half = ma.where(
            results_melt_cube > (num_valid_qa_days_in_month_grid / 2.), 128, 0)

        self.__add_qa_flags_to_results(result_cube_melt_gt_0 | result_cube_melt_gt_half)

    def __get_qa_ocean_climatology_and_near_coast(self):
        """
        QA mask = 8 bits
        bit 3 (value = 4) Ocean Climatology Mask - No Ice Allowed
        bit 4 (value = 8) = Near to Coast mask
        This method sets the QA flags for determining whether a pixel
        should never have ice, and whether it is near the coast.
        """
        results_ocean_climatology_mask_cube = self.stacked_qa_flags[..., 0] & 4
        results_near_coast_cube = self.stacked_qa_flags[..., 0] & 8

        self.__add_qa_flags_to_results(results_ocean_climatology_mask_cube |
                                       results_near_coast_cube)

    def __compute_qa_open_water(self):
        """
        QA mask = 8 bits
        bit 6 (value = 32) = Open Water flag
        This method determines if the pixel in question was open water.
        This is calculated by determining if the pixel had no ice present
        for at least half the valid days in the month.  If so, it is flagged
        as open water.
        """

        valid_days_grid = self.__get_number_days_with_valid_conc_data_grid()
        masked_cube = ma.masked_equal(self.stacked_conc_cube,
                                      self.config["missing_flag"])
        results_had_some_ice_cube = ma.where(masked_cube == 0, 0, 1)
        results_num_days_with_ice_grid = ma.sum(results_had_some_ice_cube, axis=2)

        result_qa_open_water = ma.where(((results_num_days_with_ice_grid <
                                          (valid_days_grid / 2.))
                                        & (results_num_days_with_ice_grid > 0)), 32, 0)
        self.__add_qa_flags_to_results(result_qa_open_water)

    def merge_monthly_fields(self, scaled_bt_data, scaled_nt_data):
        """Given scaled Bootstrap Data and scaled Nasateam Data, return an array
        That contains the 'merged' product, this chooses the highest value between
        the two variables, while applying the cutoffs as described in the
        ATBD section 3.4.1.4


        Arguments:
        - `scaled_bt_data`: scaled array of bootstrap data
        - `scaled_nt_data`: scaled array of nasateam data
        """

        masked_bt = np.ma.masked_greater_equal(scaled_bt_data,
                                               self.config["smallest_flag_value"])
        masked_nt = np.ma.masked_greater_equal(scaled_nt_data,
                                               self.config["smallest_flag_value"])
        masked_bt[masked_bt < self.config["min_BT"]] = 0
        masked_nt[masked_nt < self.config["min_NT"]] = 0
        # Additional BT cutoff
        masked_nt[masked_bt < self.config["min_BT"]] = 0
        merged = np.maximum(masked_bt, masked_nt)
        return merged.data

    def __get_merged_monthly_data(self):
        """ returns the merged concentration product from the published
        monthly algorithms.
        """

        nt_file = cl.find_matching_nt_month_file(self.config)
        nt_conc = cl.fetch_scaled_nasateam_from_file(nt_file, self.config)

        bt_file = cl.find_matching_bt_month_file(self.config)
        bt_conc = cl.bootstrap_concentration_from_file(bt_file, self.config)
        merged_data = self.merge_monthly_fields(bt_conc, nt_conc)
        # After merge apply lakes, and coasts..
        merged_data = cl.apply_lakes(merged_data, self.config)
        merged_data = cl.apply_coastlines(merged_data, self.config)

        return merged_data

    def __write_merged_monthly_data(self):
        """Get the merged concentration data and write it to the netcdf file

        """

        try:
            merged_data = self.__get_merged_monthly_data()
            self.__write_to_a_netcdf_file(self.__output_monthly_filename(),
                                          'goddard_merged_seaice_conc_monthly',
                                          merged_data)
        except:
            print "Merged monthly data: Unable to get bootstrap and/or nasateam data"
            pass

    def __transfer_monthly_nt_data(self):
        """
        Take the data in the published NT monthly file and scale and apply flags and
        write into the netcdf_file.
        """

        try:
            nt_file = cl.find_matching_nt_month_file(self.config)
            nt_conc = cl.fetch_scaled_nasateam_from_file(nt_file, self.config)
            self.__write_to_a_netcdf_file(self.__output_monthly_filename(),
                                          'goddard_nt_seaice_conc_monthly',
                                          nt_conc)
        except:
            print "Monthly nasateam: Unable to get nasateam data"
            pass

    def __transfer_monthly_bt_data(self):
        """
        Take the data in a published BT file and scale, apply files, and write
        into the netcdf_file
        """

        try:
            bt_file = cl.find_matching_bt_month_file(self.config)
            bt_conc = cl.bootstrap_concentration_from_file(bt_file, self.config)
            self.__write_to_a_netcdf_file(self.__output_monthly_filename(),
                                          'goddard_bt_seaice_conc_monthly',
                                          bt_conc)
        except:
            print "Monthly bootstrap: Unable to get bootstrap data"
            pass

    def __transfer_melt_onset_from_last_file_of_month(self):
        last_file = self.month_files[-1]
        with Dataset(last_file, mode='r') as source:
            with Dataset(self.__output_monthly_filename(), mode='a') as target:
                target_var = target.variables['melt_onset_day_seaice_conc_monthly_cdr']
                source_var = source.variables['melt_onset_day_seaice_conc_cdr']
                target_var[0, :, :] = source_var[0, :, :]

    def output_monthly_filename(self):
        return self.__output_monthly_filename()

    def __output_monthly_filename(self):
        """
        Generate the proper output filename for this monthly CDR file
        """
        # Get most recent satellite sensor by fishing it off the last file in the list.
        self.config["sat"] = cl.satellite_for_month(self.month_files)
        outfile = "%(output)s/seaice_conc_monthly_%(hemisphere)" \
            "sh_%(sat)s_%(YYYY)s%(MM)s_%(cdr_version)s.nc" % self.config
        return outfile

    def __prepare_zeroed_netcdf_file(self):
        """
        Create a properly named netcdf file with blank elements in the output
        location

        In the Monthly data generation we are using pre-existing blank files
        created once at the beginning of the project, but the daily generation
        creates blank files from the ancillary .CDL files on the fly
        """

        out_ncfile = self.__output_monthly_filename()
        if not os.path.exists(os.path.dirname(out_ncfile)):
            os.makedirs(os.path.dirname(out_ncfile))
        shutil.copy(self.config["blank_monthly_nc_file"], out_ncfile)

    def __write_to_a_netcdf_file(self, netcdf_file, var_name, var_data):
        """
        Given an empty netcdf file, open it and replace the data stored in the
        variable: var_name with the data passed into var_data
        """

        ncroot = Dataset(netcdf_file, 'a')
        var = ncroot.variables[var_name]
        var.set_auto_maskandscale(False)

        assert var[0, :, :].shape == var_data.shape,\
            "Input variable data has incompatable shape"

        var[0, :, :] = var_data
        ncroot.close()

    def __add_time_variable_to_a_netcdf_file(self):
        """
        Fill the time variable in the NetCDF file with the number of days
        between the start of the month and the NOAA epoch time (1601 01 01).
        """

        ncroot = Dataset(self.output_monthly_filename(), 'a')
        var = ncroot.variables['time']
        var.set_auto_maskandscale(False)
        var_data = cl.days_since_noaa_epoch(self.config)
        assert int == type(var_data), "Input variable data has incompatable shape"
        var[0] = var_data
        ncroot.close()

    def __add_dynamic_global_attributes(self):
        """
        Add dynamic global attributes to the netcdf monthly file
        """

        result = re.search("seaice_conc_daily_[ns]h_..._[0-9]{6}([0-9]{2})",
                           self.month_files[-1])
        DD = result.group(1)
        YYYYMMDD = "%4s%2s%2s" % (self.config["YYYY"], self.config["MM"], DD)

        # We create a second configuration for the last element of the list of
        # files because the monthly data spans a longer time period than the
        # original design called for.  We use a set to get a list of the uniq
        # platforms and sensors.  We make the assumption that there are only
        # two platforms or sensors in a given month and only look at the first
        # and last files for the information in the metadata.

        fake_config = config.configuration_dictionary(self.config["HEMI"], YYYYMMDD)
        platforms = set()
        platforms.add(self.config["GCMD_platform"])
        platforms.add(fake_config["GCMD_platform"])
        sensors = set()
        sensors.add(self.config["GCMD_sensor"])
        sensors.add(fake_config["GCMD_sensor"])

        ncroot = Dataset(self.__output_monthly_filename(), 'a')
        ncroot.source = self._source_files()
        ncroot.platform = ", ".join(platforms)
        ncroot.sensor = ", ".join(sensors)
        ncroot.date_created = dt.datetime.utcnow().strftime("%Y-%m-%dT%H:%M:%SZ")
        ncroot.time_coverage_start = cl.time_coverage_start(self.config)
        ncroot.time_coverage_end = cl.time_coverage_end(fake_config)
        ncroot.product_version = self.config["cdr_version"]
        ncroot.software_version_id = cl.current_software_version()
        ncroot.history = cl.history_attribute(
            os.path.basename(self.__output_monthly_filename()))

        ncroot.comment = cl.missing_monthly_comment(ncroot)

        # Loop over all global attributes setting the value to it's own value,
        # but in the order you'd like to see it.

        ordered_values = ['Conventions', 'title', 'references', 'comment',
                          'cdr_program', 'cdr_variable',
                          'software_version_id', 'Metadata_Link',
                          'product_version', 'spatial_resolution',
                          'Metadata_Conventions', 'standard_name_vocabulary',
                          'id', 'naming_authority', 'license', 'summary',
                          'keywords', 'keywords_vocabulary', 'cdm_data_type',
                          'project', 'creator_url', 'creator_email',
                          'institution', 'processing_level',
                          'geospatial_lat_min', 'geospatial_lat_max',
                          'geospatial_lat_units', 'geospatial_lon_min',
                          'geospatial_lon_max', 'geospatial_lon_units',
                          'time_coverage_duration',
                          'time_coverage_resolution', 'contributor_name',
                          'contributor_role', 'acknowledgment', 'source',
                          'history', 'platform', 'sensor', 'date_created',
                          'time_coverage_start', 'time_coverage_end']

        [ncroot.setncattr(x, ncroot.getncattr(x)) for x in ordered_values]
        ncroot.close()

    def _source_files(self):
        return ", ".join([self._ftp_url(source) for source in self.month_files])

    def _ftp_url(self, source):
        HEMISPHERE = {
            'n': 'north',
            's': 'south'
        }
        filename = os.path.split(source)[1]
        hemi, year = self._filename_parts(filename)
        FTP_URL = self.config['g02202_root_ftp_url'] + "/{0}/daily/{1}/{2}"
        return FTP_URL.format(HEMISPHERE[hemi], year, filename)

    def _filename_parts(self, filename):
        pattern = "seaice_conc_daily_(n|s)h_[nf]\d\d_([\d]{4})\d{4}_v\d\dr\d\d.nc"
        m = re.search(pattern, filename)
        if m and m.groups():
            return m.groups()
        else:
            raise RuntimeError("Unable to parse filename parts: [{0}]".format(filename))

    def process_monthly_cdr(self):
        """
        Run the necessary steps to generate and write the monthly CDR file

        This is relatively procedural code encapsulated in an object.  This is
        the only routine that an external user should be calling as the
        precedence of the function calls is important.
        """
        self.__prepare_zeroed_netcdf_file()
        self.__prepare_input_files()

        # Generate computed fields
        if (self.__is_valid_month()):
            self.__compute_mean_concentration()
            self.__compute_std_deviation(1)  # 1 degree of freedom
            self.__compute_qa_bt_and_nt_source()
            self.__compute_qa_melt()
            self.__get_qa_ocean_climatology_and_near_coast()
            self.__compute_qa_open_water()
        else:
            self.__fill_empty_variables()

        # Populate the existing data fields.
        self.__transfer_monthly_bt_data()
        self.__transfer_monthly_nt_data()
        self.__write_merged_monthly_data()

        self.__set_missing_qa_values()

        # write concentration, standard_deviation, and qa
        self.__write_to_a_netcdf_file(self.__output_monthly_filename(),
                                      'seaice_conc_monthly_cdr',
                                      self.mean_concentration)
            #now we are adding the standard deviation to the nc file
        self.__write_to_a_netcdf_file(self.__output_monthly_filename(),
                                      'stdev_of_seaice_conc_monthly_cdr',
                                      self.std_deviation)
        self.__write_to_a_netcdf_file(self.__output_monthly_filename(),
                                      'qa_of_seaice_conc_monthly_cdr',
                                      self.qa_flags_results_cube)

        self.__transfer_melt_onset_from_last_file_of_month()
        self.__add_time_variable_to_a_netcdf_file()

        self.__add_dynamic_global_attributes()


if __name__ == '__main__':

    sys_args_len = len(sys.argv)
    if ((sys_args_len != 3) and (sys_args_len != 5)):
        CDRMonthlySeaice.usage()

    if (not sys.argv[2].isdigit()):
        CDRMonthlySeaice.usage()

    if (sys_args_len == 5
        and ((not sys.argv[3].isdigit())
             or (not sys.argv[4].isdigit()))):
        CDRMonthlySeaice.usage()

    preset_config_dict = {}
    if (sys_args_len == 5):
        preset_config_dict = {"cdr_version":
                              cl.create_product_version_string(sys.argv[3], sys.argv[4])}
        del sys.argv[3]
        del sys.argv[3]

    sys.argv[2] = "%6.6s01" % sys.argv[2]
    conf = config.create_configuration_dictionary(sys.argv, preset=preset_config_dict)

    cdr_month = CDRMonthlySeaice(conf)
    cdr_month.process_monthly_cdr()
