/*
 *@***h* Sea_Ice_CDR/ncload.c
 *
 * NAME
 *  ncload.c 
 *
 * LOCATION
 *  .../src/ncload.c 
 * 
 * PURPOSE
 *  Used to load the data into an empty NetCDF4
 *
 * DESCRIPTION
 *  Uses the NetCDF4 library from unidata to overwrite the data in 
 *  variable <variable> in the NetCDF file <cf file> with data from the
 *  binary file <datasource file>.
 *      
 * AUTHOR
 *  Sean Mallory, sean.mallory@nsidc.org
 *  National Snow & Ice Data Center, University of Colorado, Boulder
 *
 * CREATION DATE
 *  April 2011
 * 
 * COPYRIGHT
 *  Copyright 2011 Regents of the University of Colorado. The Regents of the
 *  University of Colorado grants, and you accept, a personal, nonexclusive,
 *  nontransferable license to use this software, at no charge, in accordance
 *  with the terms herein, solely for internal research and development, which
 *  is limited to testing, measuring, assessing, evaluating the software
 *  and/or data that you generate using this software, and not for any
 *  commercial purpose. You may not develop derivative works based on this
 *  software, or redistribute this software.
 * 
 * EXTERNALS
 *  n/a
 *
 * SUBROUTINES
 *   usage
 *   ncErrorCheck
 *   mallocOrABEND
 *   inPlaceArrayToLower
 *   struct_tm_Init
 *   parseCmdDate
 *   nasateam_header_DateCheck
 *   daysFromEpoch
 *   cdr_variable_info_Init
 *   checkNCVar
 *   overwriteData
 *   getData
 *   btProcessing
 *   ntProcessing
 *
 * INPUTS
 *  ncload.c <variable> <data source file> <cf file>
 *  variable - the variable name in the NetCDF file to overwrite.
 *  data source file - the binary source file to load to the variable.
 *  cf file - the NetCDF file that the variable to be updated resides in.
 *
 * OUTPUTS
 *  The NetCDF file <cf file> is updated.
 *      
 * MODIFICATION HISTORY
 *  n/a
 *
 *@*****
 */
/*--+---10----+---20----+---30----+---40----+---50----+---60----+---70----+---80----+---*/
#include "ncload.h"
/* The following three constants are automatically populated by the SVN version
   control sytem for provenance and binary identification. Line length will
   vary depending on the details of provenance. If using this code in another
   version control system, please update to use the correct attributes for
   that system. */
/*@ unused @*/ static const char NCLOAD_C_SVN_DATE[]=
  "$Date: 2011-09-16 20:39:12 -0600 (Fri, 16 Sep 2011) $";
/*@ unused @*/ static const char NCLOAD_C_SVN_URL[]=
  "$HeadURL: svn+ssh://newice.colorado.edu/SubversionRepositories/programmers/seaice_cdr/tags/PROD_20130520/src/ncload.c $";
/*@ unused @*/ static const char NCLOAD_C_SVN_REV[]=
  "$Revision: 19366 $";

static void usage() {
  printf("Usage:\n"
         "ncload varname grid-file cf-file [mode] [date]\n"
         "\t  cf-file - NetCDF4 to overwrite data in\n"
         "\tgrid-file - source sea-ice grid.\n"
         "\t     mode - (optional) enable a set of special processing\n"
         "\t     date - (optional) YYYYMMDD or YYYYdoy, the expected date of "
	 "the data for QC.\n"
         "Modes:\n"
         "nt (for nasateam sea-ice)\n"
         "\tupdates 'time' dimension of NetCDF file from the year and day-of-"
	 "year in the sea-ice header\n"
         "\tScales the sea-ice concentration range from 0-250 to 0-100, while "
	 "251 to 255 remain unchanged.\n"
         "bt (for bootstrap sea-ice)\n"
         "\tScales the sea-ice concentration range from 0-1000 to 0-100, and "
	 "maps flags to match nasateam codes.\n"
         "For the users convenience(!), this program:\n"
         "\tassumes the array is of type BYTE,\n"
         "\n");
}


/*::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
 * Utility routines.
 *::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::*/

/*------------------------------------------------------------------------
 * check_nc_err - check for NetCDF library errors
 *
 *      input : stat - status code returned from an libnc.h function.
 *              line - current line of source, always use __LINE__
 *              file - current source file, always use __FILE__
 *
 *      output: no output parameters.
 *
 *      result: (void) returns if no error, exits otherwise
 *
 *      effect: program exits on NetCDF library error, after printing message
 *
 *------------------------------------------------------------------------*/
static void ncErrorCheck(const int stat, const int line, const char *file) {
  const char * err_messg;
  if (stat != NC_NOERR) {
    err_messg = nc_strerror(stat);
    (void)fprintf(stderr,
		  "NetCDF library error: line %d of %s: %s\n",
		  line, file, err_messg);
    (void)fflush(stderr);
    exit(EXIT_FAILURE);
  }
}

/*------------------------------------------------------------------------
 * mallocOrABEND - try to re-allocate memory, exit failing if not allocated
 *
 *      input : size - how much memory to allocate
 *              purpose - human readable what is being allocated
 *
 *      output: no output parameters.
 *
 *      result: pointer to allocated memory
 *
 *      effect: memory allocated or program exits on allocation error after
 *              printing message
 *
 *------------------------------------------------------------------------*/
static void * mallocOrABEND(const size_t size, const char * purpose) {
  void * point;
  if (size == 0) {
    point = NULL;
  } else {
    point = malloc(size);
    if (NULL == point) {
      (void)fprintf(stderr,
		    "Could not allocate memory for %d byte %s.\n",
		    (int)size, purpose);
      (void)fflush(stderr);
      exit(EXIT_FAILURE);
    }
  }
  /*@ -nullret @*/
  return point;
  /*@ +nullret @*/
}


/*------------------------------------------------------------------------
 * inPlaceArrayToLower - smash case of str_length caracters to lower case
 *
 *      input : string - string to convert
 *              str_length - string length
 *
 *      output: string - converted string
 *
 *      result: void
 *
 *      effect: string is converted to lower case in-place
 *
 *------------------------------------------------------------------------*/
static void inPlaceArrayToLower(char * string, size_t str_length) {
  char * temp_char_p = NULL;
  char * end_guard = string + str_length;
  for (temp_char_p = string ; temp_char_p < end_guard; temp_char_p++){
    *temp_char_p = tolower(*temp_char_p);
  }
}

/*::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
 * Date handling routines.
 *::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::*/

/*------------------------------------------------------------------------
 * struct_tm_Init - initialize a POSIX date structure
 *
 *      input : ddate - pointer to uninitialized struct tm
 *
 *      output: ddate - pointer to initialized struct tm
 *
 *      result: returns pointer to allocated memory if no error,
 *              prints error to stderr and exits otherwise.
 *
 *      effect: sets the struct tm pointed to by ddate to a known state.
 *
 *------------------------------------------------------------------------*/
static void struct_tm_Init(/*@ out @*/ struct tm * const ddate) {
  /* init date */
  ddate->tm_sec= 0;
  ddate->tm_min= 0;
  ddate->tm_hour= 12; /* default timestamp to midpoint of day */
  ddate->tm_mday= 1;      /* zero not valid mday in struct tm */
  ddate->tm_mon= 0;
  ddate->tm_year= 0;
  ddate->tm_wday= 0;
  ddate->tm_yday= 0;
  ddate->tm_isdst= 0;
}

/*------------------------------------------------------------------------
 * parseCmdDate - create a timestamp and struct tm from a YYYYMMDD or
 *                YYYYdoy string
 *
 *      input : datestr - a pointer to the start of a char array with a 
 *                        date in YYYYMMDD or YYYYdoy format. (does not
 *                        need terminator)
 *              datelen - the length of the date in the datestr char array.
 *              verbose - 0 for quiet operation, higher numbers for more
 *                         messages
 *
 *      output: day_of_data - time structure of result
 *              day_of_data_time - time stamp of result
 *
 *      result: void
 *
 *      effect: assign date to day_of_data and day_of_data_time.
 *
 *------------------------------------------------------------------------*/
static void parseCmdDate (const char * datestr, const size_t datelen, 
			  struct tm * day_of_data, time_t * day_of_data_time,
			  const int verbose) {
  /* If this needs anything more, replace it with a library.
   *  Date parsing is a bottomless hole.
   *
   * Maybe http://pds-rings.seti.org/toolkits/julian_133_html/ ?
   * Or maybe getdate()/strptime() will become more widely available in libc?
   */
  char tempstr[256];
  int temp;
  if (datelen == (size_t)8) {                                 /* YYYYMMDD */
    (void)sscanf(datestr, "%4d%2d%2d",
                 &(day_of_data->tm_year), 
		 &(day_of_data->tm_mon),
		 &(day_of_data->tm_mday));
    day_of_data->tm_mon -= 1;              /* January = 0 in POSIX months */
  } else if (datelen == (size_t)7) {                           /* YYYYdoy */
    (void)sscanf(datestr, "%4d%3d",
		 &(day_of_data->tm_year),
		 &(day_of_data->tm_mday));
    /* mktime uses mday not yday, so set DOY as day-of-January */
    day_of_data->tm_mon = 0;
                                 
  }
  day_of_data->tm_year -= EPOCH_YEAR_FOR_STRUCT_TM;
  /* Use side effect of mktime() to correct fields in day_of_data */
  *day_of_data_time = mktime(day_of_data);
  if (verbose > 0) {
    temp = (int)strftime(tempstr, 255, "%Y %B %d (DOY %j)", day_of_data);
    if (temp != 0) {
      printf("  Expecting to process data for %s.\n", tempstr);
    } else {
      (void)fprintf(stderr, "Something went horribly wrong processing the "
		    "QC date.\nThis should NOT happen!\n");
    }
  }
  return;
}

/*------------------------------------------------------------------------
 * nasateam_header_DateCheck -
 *
 *      input : nasateam_header - a 300 byte header read from a Nasateam 
 *                                Sea-Ice file
 *              validdate - a struct tm holding the date which is supposed
                            to be read
 *
 *      output: none
 *
 *      result: returns "true" if date is OK.
 *
 *      effect: exit if date in nasateam header does not match date in validdate
 *
 *      This routine is designed to help prevent putting data for an incorrect
 *      day into a file.
 *
 *------------------------------------------------------------------------*/
static bool nasateam_header_DateCheck(const nasateam_header * h,
				      const struct tm * valid_date) {
  /* Check for date equivalence, disregarding time of day */
  struct tm reference_day;
  char * end_pointer ;
  time_t header_time;
  const int width = 6;

  struct_tm_Init(&reference_day);

  end_pointer = (char *)(h->grid_data_year + width);
  reference_day.tm_year = (int)( strtol(
					h->grid_data_year,
					&end_pointer,
					USER_NUMBER_SYSTEM_BASE)
                          - EPOCH_YEAR_FOR_STRUCT_TM);
  /* Input data is NOT all Y2K compliant, adjust 2 digit years to twentieth 
     century */
  if ((reference_day.tm_year + EPOCH_YEAR_FOR_STRUCT_TM) < 100) {
    reference_day.tm_year += 1900;
  }
  end_pointer = (char *)(h->grid_data_day_of_year + width);
  reference_day.tm_mday = (int)strtol(h->grid_data_day_of_year,
				      &end_pointer,
				      USER_NUMBER_SYSTEM_BASE);
  /* make splint happy, overwritten by mktime */
  reference_day.tm_yday = reference_day.tm_mday;
  /* day of January equals day of year */
  reference_day.tm_mon = 0;                     
  /* Used for side effect: corrects fields in reference_day */
  header_time = mktime(&reference_day);

  /* mktime defined TM_yday despite lint warning */
  return (reference_day.tm_year == valid_date->tm_year &&
          reference_day.tm_yday == valid_date->tm_yday);
}

/*------------------------------------------------------------------------
 * daysFromEpoch - calculate days since a given Gregorian date
 *
 *      input : t - a POSIX time stamp to calculate the number of days for
 *              epochYear - the Gregorian year of epoch
 *              epochMonth - the Gregorian Month of epochYear, Jan = 0
 *              epochDay - the Gregorian Day of epochMonth
 *
 *      output: none
 *
 *      result: a long value representing the number of says since epoch
 *
 *------------------------------------------------------------------------*/
long daysFromEpoch(const time_t * t, int epochYear, int epochMonth,
		   int epochDay) {
  double seconds;
  long days; /* Life may be short, but days are long. */
  time_t ref_t;
  struct tm refday;

  struct_tm_Init(&refday);
  refday.tm_year = epochYear - EPOCH_YEAR_FOR_STRUCT_TM;
  refday.tm_mday = epochDay;
  refday.tm_mon = epochMonth;

  ref_t = mktime(&refday);
  seconds = difftime(*t, ref_t);
  /* ignoring leap seconds */
  days = (long)floor(seconds/(24*60*60));

  return days;
}

/*::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
 * NetCDF data handling routines.
 *::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::*/

/*------------------------------------------------------------------------
 * cdr_variable_info_Init - initialize a cdr_variable_info structure
 *
 *      input : cdr_var - a pointer to the structure to initialize
 *
 *      output: cdr_var - a pointer to the initialized structure
 *
 *      result: void
 *
 *      effect: set members to predictable defaults in a cdr_variable_info
 *
 *------------------------------------------------------------------------*/
void  cdr_variable_info_Init(/*@ out @*/cdr_variable_info *cdr_var) {
  /*
   * Variables used during data processing
   *
   * Does not allocate or initialize pointers to data and header arrays.
   */
  cdr_var->variable_id = -1;
  cdr_var->var_type = 0;
  cdr_var->rank = -1;
  memset(cdr_var->dimension_ids, 0, NC_MAX_VAR_DIMS);
  memset(cdr_var->dimension_lengths, 0 ,NC_MAX_VAR_DIMS);
  cdr_var->attribute_count = -1;
  cdr_var->header = NULL;
  cdr_var->grid_size = (size_t)0;
  cdr_var->data_array_data = NULL;
}

/*------------------------------------------------------------------------
 * cdr_file_info_Init - initialize a cdr_file_info structure
 *
 *      input : cdr_file - a pointer to the structure to initialize
 *
 *      output: cdr_file - a pointer to the initialized structure
 *
 *      result: void
 *
 *      effect: set members to predictable defaults in a cdr_file_info
 *
 *------------------------------------------------------------------------*/
void  cdr_file_info_Init(/*@ out @*/ cdr_file_info *cdr_file) {
  struct_tm_Init(&(cdr_file->day_of_data));
  /* day_of_ddata_time is different than day_of_data and days_since_1601 !*/
  cdr_file->day_of_data_time = (time_t)0;
  cdr_file->days_since_1601 = 0;
  cdr_file->qc_date_flag = false;
}

/*------------------------------------------------------------------------
 * checkNCVar - verify presence and read attributes of a NetCDF variable
 *
 *      input : cf_file - the name of the NetCDF file the variable is in
 *              cdr_var->bin_file_name - what file to read data from
 *
 *      output: cdr_var->variable_id - libnc variable id
 *              cdr_var->var_type - NetCDF type for variable
 *              cdr_var->rank - dimensional rank of variable
 *              cdr_var->dimension_ids - NetCDF dimension ids for variable
 *              cdr_var->attribute count - count of NetCDF attributes for 
 *                                         variable
 *              cdr_var->grid_size - the size of the data grid needed for
 *                                   variable
 *
 *      result: 0 indicates success
 *
 *      effect: cdr_var structure updated to reflect attributes in NetCDF file
 *
 *      Function aborts program if the name returned from the NetCDF does not
 *      match the name requested. This would indicate a file being concurrently
 *      rewritten in place, or corruption of the filehandle table.
 *
 *      It also aborts on libnc errors.
 *
 *------------------------------------------------------------------------*/
int checkNCVar(const char* cf_file, cdr_variable_info *cdr_var) {
  int stat = 0;
  int ncid = 0;
  int i = 0;
  int return_val = 0;
  char var_name_check[NC_MAX_NAME] = "";

  printf("Opening file %s to check %s.\n",cf_file, cdr_var->variable_name);
  stat = nc_open(cf_file, NC_NOWRITE, &ncid);
  ncErrorCheck(stat,__LINE__,__FILE__);

  stat = nc_inq_varid(ncid, cdr_var->variable_name, &(cdr_var->variable_id));
  ncErrorCheck(stat,__LINE__,__FILE__);

  stat = nc_inq_var(ncid, cdr_var->variable_id,
		    var_name_check, 
		    &(cdr_var->var_type),
                    &(cdr_var->rank),
		    cdr_var->dimension_ids,
		    (&cdr_var->attribute_count));
  ncErrorCheck(stat,__LINE__,__FILE__);

  if (strncmp(cdr_var->variable_name, var_name_check, NC_MAX_NAME) != 0) {
    (void)fprintf(stderr,
		  "Inconsistency in NetCDF %s, variable name changed from "
		  "%s to %s.\n",
                  cf_file, cdr_var->variable_name, var_name_check);
    (void)fflush(stderr);
    exit(EXIT_FAILURE);
  }

  /* check the dimension sizes and calculate total size */
  cdr_var->grid_size = 1;
  for (i = 0; i < cdr_var->rank; i++) {
    stat = nc_inq_dimlen(ncid,
			 cdr_var->dimension_ids[i],
			 &cdr_var->dimension_lengths[i]);
    ncErrorCheck(stat,__LINE__,__FILE__);
    cdr_var->grid_size *= cdr_var->dimension_lengths[i];
  }

  /* adjust for data size */
  switch ( cdr_var->var_type ) {
    case NC_BYTE:                         /* signed 1 byte integer */
    case NC_CHAR:                               /* ASCII character */
    case NC_UBYTE:                          /* unsigned 1 byte int */
      /* already correct length */
      break;
    case NC_SHORT:                        /* signed 2 byte integer */
    case NC_USHORT:                         /* unsigned 2-byte int */
      cdr_var->grid_size *= (size_t)2;
      break;
    case NC_INT:                          /* signed 4 byte integer */
    case NC_FLOAT:       /* single precision floating point number */
    case NC_UINT:                           /* unsigned 4-byte int */
      cdr_var->grid_size *= (size_t)4;
      break;
    case NC_DOUBLE:      /* double precision floating point number */
    case NC_INT64:                            /* signed 8-byte int */
    case NC_UINT64:                         /* unsigned 8-byte int */
      cdr_var->grid_size *= (size_t)8;
      break;
    case NC_STRING:                                      /* string */
      (void)fprintf(stderr,
                    "Variable %s is of type string, length to load is"
		    " indeterminate.\nBailing.\n",
                    cdr_var->variable_name);
      (void)fflush(stderr);
      exit(EXIT_FAILURE);
      /*@  -unreachable  @*/ /* for splint checking */
      break;
      /*@  +unreachable  @*/ /* for splint checking */
    case NC_NAT:                  /* NAT = 'Not A Type' (c.f. NaN) */
    default:
      fprintf(stderr,
              "Variable %s returns 'Not A Type', data length cannot be"
	      " calculated.\nBailing.\n",
              cdr_var->variable_name);
      (void)fflush(stderr);
      return_val = 1;
      /*@  -unreachable  @*/ /* for splint checking */
      break;
      /*@  +unreachable  @*/ /* for splint checking */
  }

  stat = nc_close(ncid);
  ncErrorCheck(stat,__LINE__,__FILE__);
  return return_val;
}

/*------------------------------------------------------------------------
 * overwriteData - overwrite the data array of a variable in a NetCDF file
 *
 *      input : cf_file - the name of the NetCDF file the variable is in
 *              data_array_data - the array of data to overwrite with
 *              data_array_dim_len - the array of dimension sizes of the 
 *                                   array of data
 *              variable_name - the variable to overwrite the data of
 *
 *      output: none
 *
 *      result: 0 indicates success
 *
 *      effect: data for variable_name in cf_file is overwritten with
 *              data_array_data
 *
 *      This function aborts on libnc errors.
 *
 *------------------------------------------------------------------------*/
int overwriteData(const char* cf_file, const void* data_array_data,
                  const size_t * data_array_dim_len,
		  const char* variable_name) {
  int  stat;  /* return status */
  int  ncid = -1;  /* NetCDF id */

  /* variable ids and shapes */
  int data_array_id = -1;
  /*@ -initallelements @*/ /* for splint checking */
  /* The zero index, See C99 std, 6.7.8 for the additional zeros. */
  size_t data_array_start[NC_MAX_VAR_DIMS] = {0};                             
                                                    
  /*@ +initallelements @*/ /* for splint checking */

  /* open NetCDF4 to update data */
  printf("Opening file %s to update %s.\n",cf_file, variable_name);
  stat = nc_open(cf_file, NC_WRITE, &ncid);
  ncErrorCheck(stat,__LINE__ - 1, __FILE__);

  stat = nc_inq_varid(ncid, variable_name, &data_array_id);
  ncErrorCheck(stat,__LINE__ - 1, __FILE__);

  stat = nc_put_vara(ncid,
		     data_array_id,
		     data_array_start,
		     data_array_dim_len,
		     data_array_data);
  if (stat != 0) printf("Error %d.\n", stat);
  ncErrorCheck(stat,__LINE__ - 2, __FILE__);

  stat = nc_close(ncid);
  ncErrorCheck(stat,__LINE__ - 1, __FILE__);
  return 0;
}

/*------------------------------------------------------------------------
 * getData - read a data array from a binary file
 *
 *      input : ice_file - the name of the binary file to read data from
 *              header_length - the header size in the binary file (0 for none)
 *              header - a pointer to space allocated for the header
 *              data_length - the size of the data array in the binary file
 *              data - a pointer to space allocated for the date
 *
 *      output: header - the data from the file header
 *              data - the data grid from the file
 *
 *      result: the number of bytes remaining in the file after the data
 *              array is returned
 *
 *      effect: data and header arrays are overwritten from the binary file
 *
 *------------------------------------------------------------------------*/
long int getData(const char* ice_file, const size_t header_length, 
		 void * header, const size_t data_length, void * data) {
  FILE *fp;
  size_t count = 0;
  long int remainder = 0;;

  fp = fopen(ice_file, "rb");
  if (errno != 0) {
    fprintf(stderr, "Error reading input data file %s: %s",
	    ice_file, strerror(errno));
    exit(EXIT_FAILURE);
  }
  if (fp == NULL) {
    fprintf(stderr,"Error opening input data file %s.",ice_file);
    exit(EXIT_FAILURE);
  }

  if (header_length > 0) {
    count = fread(header, sizeof(byte), header_length, fp);
    if (count != header_length) {
      fprintf(stderr,
	      "error reading data file header of %s, read %d bytes.\n",
              ice_file, (int)count);
      exit(EXIT_FAILURE);
    }
  }

  count = fread(data, sizeof(byte), data_length, fp);
  if (errno != 0) {
    fprintf(stderr,"Error reading input data file %s: %s",
	    ice_file, strerror(errno));
    exit(EXIT_FAILURE);
  }
  if (count != data_length) {
    fprintf(stderr,"error reading data file grid of %s, found %d bytes "
	    "for %d byte grid.\n", ice_file, (int)count,
	    (int)data_length);
    exit(EXIT_FAILURE);
  }

  /* Return bytes remaining in file so that caller can check their 
     expectations. */
  remainder -= ftell(fp);
  (void)fseek(fp, 0, SEEK_END);
  remainder += ftell(fp);

  (void)fclose(fp);
  return remainder;
}

/*::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
 * Mode specific processing routines.
 *
 * These are all essentially part of main(), but put in functions to
 * improve readability and maintainability by isolating large conditionally
 * executed blocks.
 *::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::*/

/*------------------------------------------------------------------------
 * btProcessing - special processing for bootstrap seaice concentration
 *
 *      input : cdr_var->grid_size - the size of the data grid to read
 *              cdr_var->bin_file_name - what file to read data from
 *              verbose - 0 for quiet operation, higher numbers for more 
 *                        messages
 *
 *      output: cdr_var->data_array_data - rescaled data from 
 *                                         cdr_var->bin_file_name
 *
 *      result: void
 *
 *      effect: cdr_var.data_array_data is loaded with data from the file
 *              indicated in cdr_var->bin_file_name.
 *
 *   Bootstrap sea-ice grids are originally produced with a data range of
 *   0-1000 (percent times ten, stored as a little-endian 16 bit int) that
 *   this scales to 0-100, and the values greater than 1000 are flags that
 *   are mapped to flags matching the nasateam flags. Thus 1200 (land)
 *   becomes 254, and 1100 (pole) becomes 251. I have not found a value
 *   for "missing".
 *
 *   The established format lacks a header.
 *
 *------------------------------------------------------------------------*/
static void btProcessing(cdr_variable_info *cdr_var, int verbose) {
  long int i;
  uint16_t *in_buffer = NULL;
  size_t in_buff_size;

  /* bt_ice_conc is of type byte */
  in_buff_size = sizeof(uint16_t) * cdr_var->grid_size;           
  in_buffer = mallocOrABEND(in_buff_size,
			    "temporary buffer for reading bootstrap sea-ice");
  if (verbose > 0) {
    printf("  Allocated %ld bytes for input.\n", (long int)in_buff_size);
  }

  i = getData( cdr_var->bin_file_name, (size_t)0, cdr_var->header,
	       in_buff_size, in_buffer);
  if (i != 0) {
    (void)fprintf(stderr, "Error reading binary file %s, %ld extra"
		  " bytes after reading data.\n", cdr_var->bin_file_name, i);
    (void)fflush(stderr);
    exit(EXIT_FAILURE);
  }

  if (verbose > 0) {
    printf("  Read data, preparing to normalize data.\n");
  }

  /* Normalize data from bootstrap-style 0-1000 to percentages, 
     recode flag values */
  for (i = 0 ; i < (int)(cdr_var->grid_size) ; i++ ) {
    if (in_buffer[i] <= BT_MAX_UNSCALED ) {
      cdr_var->data_array_data[i] =
	(unsigned char)floor(( (float)in_buffer[i] * BT_SCALE_FACTOR ) + 0.5 ) ;
    } else if (in_buffer[i] == BT_POLE_MASK ) {
      cdr_var->data_array_data[i] = CDR_POLE_MASK;
    } else if (in_buffer[i] == BT_LAND_MASK ) {
      cdr_var->data_array_data[i] = CDR_LAND_MASK;
    } else { /* undocumented value reached */
      cdr_var->data_array_data[i] = CDR_UNUSED_FILL;
    }
  }

  FREE(in_buffer);
}

/*------------------------------------------------------------------------
 * ntProcessing - special processing for nasateam seaice concentration
 *
 *      input : cdr_var->grid_size - the size of the data grid to read
 *              cdr_var->bin_file_name - what file to read data from
 *              cdr_file->day_of_data - what date the output NetCDF is for
 *              cdr_file->qc_date_flag - if true, verify that date in 
 *                                       nasateam header matches
 *                                       cdr_file->day_of_data
 *              verbose - 0 for quiet operation, higher numbers for more 
 *                        messages
 *
 *      output: cdr_var->data_array_data - rescaled data from cdr_var->bin_file_name
 *              cdr_var->header - nasateam_header from cdr_var->bin_file_name
 *
 *      result: void
 *
 *      effect: cdr_var.data_array_data is loaded with data from the file
 *              indicated in cdr_var->bin_file_name
 *
 *   Nasateam sea-ice grids are originally produced with a data range of
 *   0-250 that should be scaled to 0-100, and the values 251 through 255
 *   are flags that (at this time) should be preserved as-is in the 
 *   resulting data array.
 *
 *   The established format also has a 300-byte header designed to hold a
 *   considerable amount of metadata, of which only one date is currently
 *   being used by this code.
 *------------------------------------------------------------------------*/
static void ntProcessing(cdr_variable_info *cdr_var,
			 const cdr_file_info *cdr_file,
			 int verbose) {
  long int i;
  char tempstr[256];
  cdr_variable_info * cdr_time_var;

  i = getData( cdr_var->bin_file_name, sizeof(nasateam_header), cdr_var->header,
               cdr_var->grid_size, cdr_var->data_array_data);
  if (i != 0) {
    (void)fprintf(stderr,
                  "Error reading binary file %s, %ld extra bytes after reading"
		  " header and data.\n", cdr_var->bin_file_name, i);
    (void)fflush(stderr);
    exit(EXIT_FAILURE);
  }

  if (cdr_file->qc_date_flag) {
    if( !nasateam_header_DateCheck((nasateam_header *)cdr_var->header,
				   &(cdr_file->day_of_data))) {
      (void)strftime(tempstr, 255, "day %j of %Y", &cdr_file->day_of_data);
      (void)fprintf(stderr,
		    "Date mismatch: header of %s has day %s of %s, and "
		    "QC date is %s.\n",
                    cdr_var->bin_file_name,
                    ((nasateam_header*)cdr_var->header)->grid_data_day_of_year,
                    ((nasateam_header*)cdr_var->header)->grid_data_year,
                    tempstr );
      (void)fflush(stderr);
      exit(EXIT_FAILURE);
    } else {
      if (verbose > 0) {
        (void)printf("Date check passed.\n");
      }
    }
    /* since we verified the date, we will set the date */
    cdr_time_var = mallocOrABEND(sizeof(cdr_variable_info),
				 "NetCDF time variable");
    cdr_time_var->variable_name = CDR_TIME_VARIABLE;
    i = checkNCVar(cdr_file->cf_file_name, cdr_time_var);
    if (i == 0) {
      i = overwriteData(cdr_file->cf_file_name,
			&(cdr_file->days_since_1601),
                        cdr_time_var->dimension_lengths,
			cdr_time_var->variable_name );
    } else {
      (void)fprintf(stderr,
		    "Error getting information about variable %s "
		    "from NetCDF file %s.\n",
                    cdr_time_var->variable_name, cdr_file->cf_file_name);
      (void)fflush(stderr);
    }
    FREE(cdr_time_var);
  }

  if (verbose > 0) {
    printf("  Read data, preparing to normalize data.\n");
  }

  /* Normalize data from nasateam-style 0-250 to percentages, leave 
     flag values alone */
  for (i = 0 ; i < (int)(cdr_var->grid_size) ; i++ ) {
    if (cdr_var->data_array_data[i] <= (unsigned char)NT_MAX_UNSCALED ) {
      cdr_var->data_array_data[i] =
        (unsigned char)floor(( (float)cdr_var->data_array_data[i] * 
			       NT_SCALE_FACTOR) + 0.5 ) ;
    }
  }
}


/*::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
 * main() - see usage() for interface documentation.
 *::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::*/

int main(int argc, char *argv[]) {
  /* utility vars */
  int temp;
  long int i;

  /* mode vars */
  int verbose = VERBOSITY;
  enum OP_MODE opmode = GENERIC;

  /* Structs for NOAA CDR */
  cdr_file_info cdr_file;
  cdr_variable_info cdr_var;

  /*
   * Argument Processing
   */
  if (argc < 4) {
    usage();
    return 1;
  };
  cdr_var.variable_name = argv[1];
  cdr_var.bin_file_name = argv[2];
  cdr_file.cf_file_name = argv[3];

  cdr_file_info_Init(&cdr_file);
  cdr_variable_info_Init(&cdr_var);

  /*
   * Parse optional arguments if present
   */
  if (argc > 4) {
    inPlaceArrayToLower(argv[4], strlen(argv[4]));
    if (strncmp("nt", argv[4], 3) == 0 || 
	strncmp("nasateam", argv[4], 3) == 0 ) {
      fprintf(stderr,"Running in nasateam sea-ice mode.\n");
      opmode = NT_MODE;
    } else if (strncmp("bt", argv[4], 3) == 0 ||
	       strncmp("bootstrap", argv[4], 3) == 0) {
      fprintf(stderr,"Running in bootstrap sea-ice mode.\n");
      opmode = BT_MODE;
    } else if (strncmp("none", argv[4], 3) == 0 ||
	       strncmp("raw", argv[4], 3) == 0
	       || strncmp("bin", argv[4], 3) == 0 ) {
      fprintf(stderr,"Running in raw grid mode.\n");
      opmode = GENERIC;
    } else {
      (void)fprintf(stderr,"Unrecognised mode \"%s\" requested.\n", argv[4]);
      (void)fflush(stderr);
      exit(EXIT_FAILURE);
    }
  }
  if (argc > 5) {
    cdr_file.qc_date_flag = true;
    parseCmdDate(argv[5], strlen(argv[5]), &cdr_file.day_of_data,
		 &cdr_file.day_of_data_time, verbose);
    cdr_file.days_since_1601 =
      (double)daysFromEpoch(&cdr_file.day_of_data_time,
                            DEFAULT_EPOCH_YEAR, DEFAULT_EPOCH_MONTH,
			    DEFAULT_EPOCH_DAY) ;
    /* 148727 for 2007 Mar 15 */
    printf("Calculated time field is: %f\n", cdr_file.days_since_1601);
  }

  /*
   * Gather variable attributes and allocate appropriate data array.
   */
  i = checkNCVar(cdr_file.cf_file_name, &cdr_var);
  if (i != 0) {
    (void)fprintf(stderr,
		  "Error getting information about variable %s "
		  "from NetCDF file %s.\n",
                  cdr_var.variable_name, cdr_file.cf_file_name);
    (void)fflush(stderr);
    exit(EXIT_FAILURE);
  }

  if (verbose > 0) {
    printf("  ncload var=%s cdr_var.rank=%d ", cdr_var.variable_name,
	   cdr_var.rank);
    if (cdr_var.rank > 0) {
      printf("  dimension lengths=( %li", (long)cdr_var.dimension_lengths[0]);
      for ( i = 1; i < cdr_var.rank; i++) {
        printf(", %ld", (long)cdr_var.dimension_lengths[i]);
      }
      printf(") ");
    }
    printf("bin-file=%s cf-file=%s\n  Total data size to be read "
	   "is %ld bytes of type %d.\n",
           cdr_file.cf_file_name, cdr_var.bin_file_name,
	   (long)cdr_var.grid_size,
           cdr_var.var_type);
  }

  free(cdr_var.data_array_data);
  cdr_var.data_array_data = mallocOrABEND(cdr_var.grid_size,
					  "variable data array");

  if (verbose > 0) {
    printf("  Allocated %ld bytes for data.\n"
	   "  Preparing to read data from %s.\n",
           (long)cdr_var.grid_size, cdr_var.bin_file_name);
  }

  /*
   * Mode-specific operations and output to NetCDF
   */
  switch (opmode) {
    case BT_MODE:
      btProcessing(&cdr_var, verbose);
      break;

    case NT_MODE:
      /* cdr_var was an "out" variable, so splint thinks the header 
	 array is allocated */
      /*@ -mustfreeonly @*/
      cdr_var.header = mallocOrABEND(sizeof(nasateam_header), "header");
      /*@ +mustfreeonly @*/
      ntProcessing(&cdr_var, &cdr_file, verbose);
      FREE(cdr_var.header);
      break;

    case GENERIC:
      /* same as default, fall through */
    default:
      i = getData( cdr_var.bin_file_name, 0, NULL,
		   cdr_var.grid_size,
		   cdr_var.data_array_data);
      if (i != 0) {
        (void)fprintf(stderr,
                      "Error reading binary file %s, %ld extra "
		      "bytes after reading data.\n",
                      cdr_var.bin_file_name, i);
        (void)fflush(stderr);
        exit(EXIT_FAILURE);
      }
      break;

  } /* End of switch for mode-specific processing */

  if (verbose > 0) {
    printf("  Data normalized, preparing to add to %s.\n",
	   cdr_file.cf_file_name);
  }

  temp = overwriteData(cdr_file.cf_file_name, cdr_var.data_array_data,
                       cdr_var.dimension_lengths, cdr_var.variable_name);

  if (verbose > 0) {
    printf("  Data loaded preparing to free memory and exit.\n\n");
  }

  FREE(cdr_var.data_array_data);
  exit(temp);
}
