/*
 *@***h* Sea_Ice_CDR/ncload.h
 *
 * NAME
 *  ncload.h
 *
 * LOCATION
 *  .../src/ncload.h
 * 
 * PURPOSE
 *  To provide a header file for constants used by the code to load the data into an
 *  empty NetCDF4
 *
 * DESCRIPTION
 *  Sets compile time constants for ncload.c.
 *      
 * AUTHOR
 *  Sean Mallory, sean.mallory@nsidc.org
 *  National Snow & Ice Data Center, University of Colorado, Boulder
 *
 * CREATION DATE
 *  April 2011
 * 
 * COPYRIGHT
 *  Copyright 2011 Regents of the University of Colorado. The Regents of the
 *  University of Colorado grants, and you accept, a personal, nonexclusive,
 *  nontransferable license to use this software, at no charge, in accordance
 *  with the terms herein, solely for internal research and development, which
 *  is limited to testing, measuring, assessing, evaluating the software
 *  and/or data that you generate using this software, and not for any
 *  commercial purpose. You may not develop derivative works based on this
 *  software, or redistribute this software.
 *
 * EXTERNALS
 *   n/a
 *
 * SUBROUTINES
 *   n/a
 *
 * INPUTS
 *  n/a
 *
 * OUTPUTS
 *  n/a
 *      
 * MODIFICATION HISTORY
 *  n/a
 *
 *@*****
 */
#ifndef ncload_header
#define ncload_header

#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <stdint.h>
#include <string.h>
#include <math.h>
#include <errno.h>
#include <time.h>
#include <ctype.h>

/* Includes outside the standard libraries. */
#include <netcdf.h>
#include "sea_ice.h"
/* #include "/home/smallory/seaice/mapx/matrix_io.h" */

typedef uint8_t byte;

/* this is safer than free(), since x will not point to freed memory */
#define FREE(POINTERX) free(POINTERX); POINTERX = NULL

/*
 * Constants
 */
/* The following three constants are automatically populated by the SVN version
   control sytem for provenance and binary identification. Line length will
   vary depending on the details of provenance. If using this code in another
   version control system, please update to use the correct attributes for
   that system. */
/*@ unused @*/ static const char NCLOAD_H_SVN_DATE[]=
  "$Date: 2011-09-01 14:41:01 -0600 (Thu, 01 Sep 2011) $";
/*@ unused @*/ static const char NCLOAD_H_SVN_URL[]=
  "$HeadURL: svn+ssh://newice.colorado.edu/SubversionRepositories/programmers/seaice_cdr/tags/PROD_20130520/src/ncload.h $";
/*@ unused @*/ static const char NCLOAD_H_SVN_REV[]=
  "$Revision: 19231 $";
#define VERBOSITY 1
#define USER_NUMBER_SYSTEM_BASE 10
#define EPOCH_YEAR_FOR_STRUCT_TM 1900

/* The following MUST match the epoch used throughout the NetCDF cdl templates
   used. */
#define DEFAULT_EPOCH_YEAR 1601
#define DEFAULT_EPOCH_MONTH 0
#define DEFAULT_EPOCH_DAY 1
static const char CDR_TIME_VARIABLE[]="time";

/*::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
 *  Scaling factors and masking values
 *:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::*/ 
/*--+---10----+---20----+---30----+---40----+---50----+---60----+---70----+---80----+---*/
/* NT goes into unsigned char - ranging past 255 will require more code changes
   than just here. */
#define NT_MAX_UNSCALED 250
#define NT_MAX_SCALED 100
#define NT_SCALE_FACTOR ( (double)NT_MAX_SCALED / (double)NT_MAX_UNSCALED )

/* BT goes into int */
#define BT_MAX_UNSCALED 1000
#define BT_MAX_SCALED 100
#define BT_SCALE_FACTOR ( (double)BT_MAX_SCALED / (double)BT_MAX_UNSCALED )
#define BT_POLE_MASK 1100
#define BT_LAND_MASK 1200

/* Target flag values */
#define CDR_POLE_MASK (unsigned char)251
#define CDR_UNUSED_FILL (unsigned char)252
#define CDR_COAST_MASK (unsigned char)253
#define CDR_LAND_MASK (unsigned char)254
#define CDR_MISSING_DATA (unsigned char)255

/*
 * Special operating mode list
 *
 */
enum OP_MODE {
  BT_MODE,  /* bootstrap seaice */
  NT_MODE,  /* nasateam seaice */
  GENERIC
};

/*::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
 * The following structures are used to hold data related to NetCDF files
 * or NetCDF variables (depending on which struct). They are typedef'd to
 * a short name to eliminate the need to use the "struct" label where ever
 * they are used.
 * This typedef is not done in a single step with the struct declaration
 * because the static code analysis program being used (splint) does not
 * permit fine grained variable-use annotation at the lowest level of a
 * compound reference. These structures use this annotation to indicate that
 * the pointers to * strings from the command line that are allocated and
 * freed as part of argv are "dependent". To splint, "dependent" means it is
 * not a memory leak if the dependent variable is not freed, because the
 * memory is via a different pointer. These are inherently const because of
 * the dependence, and so declared.
 *::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::*/

/*
 * Structure to hold dynamic data for create/load/update of NOAA NetCDF4
 * files for CDR's
 */
struct cdr_file_struct {
  /*@ dependent @*/ const char * cf_file_name;
  /* the following three forms of date all represent the same date,
   * which is the date of the observations in the file.
   */
  struct tm day_of_data;
  time_t day_of_data_time;
  /* The type of the following days variable currently MUST match 
     the type of the "time" field in the NetCDF file, or the field in the 
     NetCDF will be incorrect with no error generated. */
  double days_since_1601;
  /* only update the NetCDF date and check the input file header dates if a
     date is given. */
  bool qc_date_flag;

};
typedef struct cdr_file_struct cdr_file_info ;

struct cdr_variable_struct {
  /* Variable in the NetCDF file */
  /*@ dependent @*/ const char * variable_name;
  int variable_id;
  nc_type var_type;
  int rank;
  int dimension_ids[NC_MAX_VAR_DIMS];
  size_t dimension_lengths[NC_MAX_VAR_DIMS];
  int attribute_count;

  /* Variable in the binary source file */
  char * header;
  size_t grid_size;
  /*@ dependent @*/ const char * bin_file_name;

  /* The location of the variable data. */
  unsigned char * data_array_data;

} ;
typedef struct cdr_variable_struct cdr_variable_info;

/*::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
 * exported functions
 *
 * input-only variables are declared "const" so that they can receive both
 * const and non-const values.
 *
 * Parameters annotated as "out" are assumed by splint to be fully initialized
 * following a call to the function. All-non pointers not annotated as dependent
 * within an "out" parameter structure are assumed to be pointing to memory
 * allocated for them.
 *
 * Functions with used for only a particular structure start with the structure
 * name, and thus have slightly different formatting than general functions.
 *::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::*/

void  cdr_variable_info_Init(/*@ out @*/cdr_variable_info *cdr_var);
void  cdr_file_info_Init(/*@ out @*/ cdr_file_info *cdr_var);
long daysFromEpoch(const time_t * t, int epochYear, int epochMonth,
		   int epochDay);
int checkNCVar(const char* cf_file, cdr_variable_info *cdr_var);
int overwriteData(const char* cf_file, const void* data_array_data,
		  const size_t * data_array_dim_len,
		  const char* variable_name);
long int getData(const char* ice_file, const size_t header_length,
		 void *header, const size_t data_length, void *data) ;
int main(int argc, char *argv[]) ;

#endif
