################################################################################
#@***h* Sea_Ice_CDR/melting_tests.py
#
# NAME
#  melting_tests.py
#
# LOCATION
#  tests/melting_tests.py
#
# PURPOSE
#  Unit tests for melt related CDR computations.
#
# DESCRIPTION
#  Unit tests for melt related CDR computations.
#
# AUTHOR
#  nsidc@nsidc.org
#  National Snow & Ice Data Center, University of Colorado, Boulder
#
# COPYRIGHT
#  Copyright 2011 Regents of the University of Colorado. The Regents of the
#  University of Colorado grants, and you accept, a personal, nonexclusive,
#  nontransferable license to use this software, at no charge, in accordance
#  with the terms herein, solely for internal research and development, which
#  is limited to testing, measuring, assessing, evaluating the software and/or
#  data that you generate using this software, and not for any commercial
#  purpose. You may not develop derivative works based on this software, or
#  redistribute this software.
#
# CREATION DATE
#  March 2013
#
# EXTERNALS
#  numpy
#  numpy.testing
#  unittest2
#  cdr_melt
#
# SUBROUTINES
#   class MeltDetectableTests
#     test_that_50_percent_concentration_is_detectable
#     test_that_land_coast_shore_near_shore_and_lakes_are_undetectable
#     test_with_all_valid_concentrations
#     test_with_some_valid_concentrations
#   class MeltOnsetDayTests
#     test_with_doy_after_end_of_melt_season
#     test_with_doy_before_beginning_of_melt_season
#     test_with_melting_but_grid_cell_initially_undetectable
#     test_with_new_melting_and_nothing_previously
#     test_with_new_melting_and_some_previously
#     test_with_no_melting
#     test_with_prior_melt_and_concentrations_below_threshold
#   class MeltStartDetectedTests
#     test_with_doy_after_end_of_melt_season
#     test_with_doy_before_beginning_of_melt_season
#     test_with_no_melting
#     test_with_previous_melting_and_ice_disappears_in_one_cell
#     test_with_some_melting
#   class MeltingTests
#     test_that_50_percent_concentration_is_in_range
#     test_that_tb_threshold_is_2_kelvin
#     test_with_both_missing_values
#     test_with_melting_everywhere
#     test_with_missing_tb19_values
#     test_with_missing_tb37_values
#     test_with_some_non_melting_due_to_large_tb_deltas
#     test_with_some_non_melting_due_to_low_concentrations
#   class ScaledBrightnessTempTests
#     test_f08_tbs_are_not_scaled
#     test_f11_tbs_are_not_scaled
#     test_f13_tbs_are_not_scaled
#     test_f17_19h_tbs_are_scaled
#     test_f17_37h_tbs_are_scaled
#     test_f17_missing_tbs_are_not_scaled
#
# INPUTS
#  n/a
#
# OUTPUTS
#  n/a
#
# MODIFICATION HISTORY
#
#
#@*****
################################################################################
import numpy as np
import numpy.testing as npt
import unittest2
import cdr_melt as melt


class MeltingTests(unittest2.TestCase):

    def test_with_melting_everywhere(self):
        concentrations = np.array([80, 80, 80, 90])
        tb19 = np.array([200, 200, 200, 200])
        tb37 = np.array([200, 200, 200, 200])
        expected = np.array([True, True, True, True])

        actual = melt.melting(concentrations, tb19, tb37)

        npt.assert_array_equal(expected, actual)

    def test_with_some_non_melting_due_to_large_tb_deltas(self):
        concentrations = np.array([80, 80, 80, 90])
        tb19 = np.array([200, 200, 200, 200])
        tb37 = np.array([150, 200, 200, 200])
        expected = np.array([False, True, True, True])

        actual = melt.melting(concentrations, tb19, tb37)

        npt.assert_array_equal(expected, actual)

    def test_that_tb_threshold_is_2_kelvin(self):
        # Brightness temperatures are in *tenths* of kelvin
        concentrations = np.array([80, 80, 80, 90])
        tb19 = np.array([200, 200, 200, 200])
        tb37 = np.array([150, 180, 181, 200])
        expected = np.array([False, False, True, True])

        actual = melt.melting(concentrations, tb19, tb37)

        npt.assert_array_equal(expected, actual)

    def test_with_some_non_melting_due_to_low_concentrations(self):
        concentrations = np.array([10, 40, 80, 90])
        tb19 = np.array([200, 200, 200, 200])
        tb37 = np.array([200, 200, 200, 200])
        expected = np.array([False, False, True, True])

        actual = melt.melting(concentrations, tb19, tb37)

        npt.assert_array_equal(expected, actual)

    def test_that_50_percent_concentration_is_in_range(self):
        concentrations = np.array([49, 50, 80, 90])
        tb19 = np.array([200, 200, 200, 200])
        tb37 = np.array([200, 200, 200, 200])
        expected = np.array([False, True, True, True])

        actual = melt.melting(concentrations, tb19, tb37)

        npt.assert_array_equal(expected, actual)

    def test_with_missing_tb19_values(self):
        concentrations = np.array([80, 80, 80, 90])
        tb19 = np.array([melt.TB_MISSING, melt.TB_MISSING, 200, 200])
        tb37 = np.array([200, 200, 200, 200])
        expected = np.array([False, False, True, True])

        actual = melt.melting(concentrations, tb19, tb37)

        npt.assert_array_equal(expected, actual)

    def test_with_missing_tb37_values(self):
        concentrations = np.array([80, 80, 80, 90])
        tb19 = np.array([200, melt.TB_MISSING + 1, melt.TB_MISSING + 1, 200])
        tb37 = np.array([200, melt.TB_MISSING, melt.TB_MISSING, 200])
        expected = np.array([True, False, False, True])

        actual = melt.melting(concentrations, tb19, tb37)

        npt.assert_array_equal(expected, actual)

    def test_with_both_missing_values(self):
        concentrations = np.array([80, 80, 80, 90])
        tb19 = np.array([200, melt.TB_MISSING, melt.TB_MISSING, 200])
        tb37 = np.array([200, melt.TB_MISSING, melt.TB_MISSING, 200])
        expected = np.array([True, False, False, True])

        actual = melt.melting(concentrations, tb19, tb37)

        npt.assert_array_equal(expected, actual)


class MeltDetectableTests(unittest2.TestCase):

    def test_with_all_valid_concentrations(self):
        concentrations = np.array([80, 80, 80, 90])
        expected = np.array([True, True, True, True])
        lcs_mask = np.array([0, 0, 0, 0])

        actual = melt.melt_detectable(concentrations, lcs_mask)
        npt.assert_array_equal(expected, actual)

    def test_with_some_valid_concentrations(self):
        concentrations = np.array([80, 80, 20, 90])
        expected = np.array([True, True, False, True])
        lcs_mask = np.array([0, 0, 0, 0])

        actual = melt.melt_detectable(concentrations, lcs_mask)
        npt.assert_array_equal(expected, actual)

    def test_that_50_percent_concentration_is_detectable(self):
        concentrations = np.array([80, 50, 49, 90])
        expected = np.array([True, True, False, True])
        lcs_mask = np.array([0, 0, 0, 0])

        actual = melt.melt_detectable(concentrations, lcs_mask)
        npt.assert_array_equal(expected, actual)

    def test_that_land_coast_shore_near_shore_and_lakes_are_undetectable(self):
        undetectable_masks = [melt.LAND_MASK, melt.COAST_MASK, melt.LAKE_MASK]
        concentrations = np.array([80, 60, 50, 90])

        for mask in undetectable_masks:
            lcs_mask = np.array([mask, 0, 0, 0])
            expected = np.array([False, True, True, True])
            actual = melt.melt_detectable(concentrations, lcs_mask)
            npt.assert_array_equal(expected, actual)


class MeltOnsetDayTests(unittest2.TestCase):

    def test_with_no_melting(self):
        previous_mod = np.array([melt.MELT_ONSET_DAY_FILL_VALUE,
                                 melt.MELT_ONSET_DAY_FILL_VALUE,
                                 melt.MELT_ONSET_DAY_FILL_VALUE,
                                 melt.MELT_ONSET_DAY_FILL_VALUE])
        detectable = np.array([True, True, True, True])
        melting = np.array([False, False, False, False])
        expected = np.array([melt.MELT_ONSET_DAY_FILL_VALUE,
                             melt.MELT_ONSET_DAY_FILL_VALUE,
                             melt.MELT_ONSET_DAY_FILL_VALUE,
                             melt.MELT_ONSET_DAY_FILL_VALUE])

        actual = melt.melt_onset_day(melting, 60, previous_mod, detectable)

        npt.assert_array_equal(expected, actual)

    def test_with_new_melting_and_nothing_previously(self):
        doy = 60
        previous_mod = np.array([melt.MELT_ONSET_DAY_FILL_VALUE,
                                 melt.MELT_ONSET_DAY_FILL_VALUE,
                                 melt.MELT_ONSET_DAY_FILL_VALUE,
                                 melt.MELT_ONSET_DAY_FILL_VALUE])
        detectable = np.array([True, True, True, True])
        melting = np.array([True, False, False, False])
        expected = np.array([doy,
                             melt.MELT_ONSET_DAY_FILL_VALUE,
                             melt.MELT_ONSET_DAY_FILL_VALUE,
                             melt.MELT_ONSET_DAY_FILL_VALUE])

        actual = melt.melt_onset_day(melting, doy, previous_mod, detectable)

        npt.assert_array_equal(expected, actual)

    def test_with_new_melting_and_some_previously(self):
        doy = 61
        previous_mod = np.array([60,
                                 melt.MELT_ONSET_DAY_FILL_VALUE,
                                 melt.MELT_ONSET_DAY_FILL_VALUE,
                                 melt.MELT_ONSET_DAY_FILL_VALUE])
        detectable = np.array([True, True, True, True])
        melting = np.array([True, True, False, False])
        expected = np.array([60,
                             doy,
                             melt.MELT_ONSET_DAY_FILL_VALUE,
                             melt.MELT_ONSET_DAY_FILL_VALUE])

        actual = melt.melt_onset_day(melting, doy, previous_mod, detectable)

        npt.assert_array_equal(expected, actual)

    def test_with_prior_melt_and_concentrations_below_threshold(self):
        doy = 62
        previous_mod = np.array([60,
                                 61,
                                 melt.MELT_ONSET_DAY_FILL_VALUE,
                                 melt.MELT_ONSET_DAY_FILL_VALUE])
        detectable = np.array([True, True, True, True])
        melting = np.array([False, False, False, False])
        expected = np.array([60,
                             61,
                             melt.MELT_ONSET_DAY_FILL_VALUE,
                             melt.MELT_ONSET_DAY_FILL_VALUE])

        actual = melt.melt_onset_day(melting, doy, previous_mod, detectable)

        npt.assert_array_equal(expected, actual)

    def test_with_melting_but_grid_cell_initially_undetectable(self):
        doy = 63
        previous_mod = np.array([melt.MELT_ONSET_DAY_FILL_VALUE, 61, 62])
        detectable = np.array([False, True, True])
        melting = np.array([True, True, True])
        expected = np.array([melt.MELT_ONSET_DAY_FILL_VALUE, 61, 62])

        actual = melt.melt_onset_day(melting, doy, previous_mod, detectable)

        npt.assert_array_equal(expected, actual)

    def test_with_doy_before_beginning_of_melt_season(self):
        doy = 59
        previous_mod = np.array([melt.MELT_ONSET_DAY_FILL_VALUE,
                                 melt.MELT_ONSET_DAY_FILL_VALUE])
        detectable = np.array([True, True])
        melting = np.array([True, False])
        expected = np.array([melt.MELT_ONSET_DAY_FILL_VALUE,
                             melt.MELT_ONSET_DAY_FILL_VALUE])

        actual = melt.melt_onset_day(melting, doy, previous_mod, detectable)

        npt.assert_array_equal(expected, actual)

    def test_with_doy_after_end_of_melt_season(self):
        doy = 250
        previous_mod = np.array([70, 100, melt.MELT_ONSET_DAY_FILL_VALUE])
        detectable = np.array([True, True, True])
        melting = np.array([True, True, True])
        expected = np.array([70, 100, melt.MELT_ONSET_DAY_FILL_VALUE])

        actual = melt.melt_onset_day(melting, doy, previous_mod, detectable)

        npt.assert_array_equal(expected, actual)


class MeltStartDetectedTests(unittest2.TestCase):

    def test_with_no_melting(self):
        doy = 60
        melt_onset_day = np.array([melt.MELT_ONSET_DAY_FILL_VALUE,
                                   melt.MELT_ONSET_DAY_FILL_VALUE,
                                   melt.MELT_ONSET_DAY_FILL_VALUE])
        concentration = np.array([80, 80, 80])
        expected = np.array([False, False, False])

        actual = melt.melt_start_detected(doy, melt_onset_day, concentration)

        npt.assert_array_equal(expected, actual)

    def test_with_some_melting(self):
        doy = 60
        melt_onset_day = np.array([melt.MELT_ONSET_DAY_FILL_VALUE,
                                   60,
                                   melt.MELT_ONSET_DAY_FILL_VALUE])
        concentration = np.array([80, 80, 80])
        expected = np.array([False, True, False])

        actual = melt.melt_start_detected(doy, melt_onset_day, concentration)

        npt.assert_array_equal(expected, actual)

    def test_with_previous_melting_and_ice_disappears_in_one_cell(self):
        doy = 61
        melt_onset_day = np.array([60, 60, melt.MELT_ONSET_DAY_FILL_VALUE])
        concentration = np.array([8, 0, 8])
        expected = np.array([True, False, False])

        actual = melt.melt_start_detected(doy, melt_onset_day, concentration)

        npt.assert_array_equal(expected, actual)

    def test_with_doy_before_beginning_of_melt_season(self):
        doy = 59
        melt_onset_day = np.array([melt.MELT_ONSET_DAY_FILL_VALUE,
                                   melt.MELT_ONSET_DAY_FILL_VALUE])
        concentration = np.array([100, 100])
        expected = np.array([False, False])

        actual = melt.melt_start_detected(doy, melt_onset_day, concentration)

        npt. assert_array_equal(expected, actual)

    def test_with_doy_after_end_of_melt_season(self):
        doy = 255
        melt_onset_day = np.array([101, 200])
        concentration = np.array([100, 100])
        expected = np.array([False, False])

        actual = melt.melt_start_detected(doy, melt_onset_day, concentration)

        npt.assert_array_equal(expected, actual)


class ScaledBrightnessTempTests(unittest2.TestCase):

    def test_f08_tbs_are_not_scaled(self):
        satellite = 'f08'
        channel = '19h'
        tbs = np.array([100, 200, 300, 400])

        actual = melt.scaled_brightness_temperatures(satellite, channel, tbs)

        npt.assert_array_almost_equal(tbs, actual)

    def test_f11_tbs_are_not_scaled(self):
        satellite = 'f11'
        channel = '19h'
        tbs = np.array([100, 200, 300, 400])

        actual = melt.scaled_brightness_temperatures(satellite, channel, tbs)

        npt.assert_array_almost_equal(tbs, actual)

    def test_f13_tbs_are_not_scaled(self):
        satellite = 'f13'
        channel = '19h'
        tbs = np.array([100, 200, 300, 400])

        actual = melt.scaled_brightness_temperatures(satellite, channel, tbs)

        npt.assert_array_almost_equal(tbs, actual)

    def test_f17_missing_tbs_are_not_scaled(self):
        satellite = 'f17'
        channels = ['19h', '37h']
        tbs = np.array([0, 0])

        for channel in channels:
            actual = melt.scaled_brightness_temperatures(satellite, channel, tbs)
            npt.assert_array_almost_equal(tbs, actual)

    def test_f17_19h_tbs_are_scaled(self):
        satellite = 'f17'
        channel = '19h'
        tbs = np.array([0, 16.46425, 1000, 2000])
        expected = np.array([0, 0, 1004.19, 2025.19])

        actual = melt.scaled_brightness_temperatures(satellite, channel, tbs)

        npt.assert_array_almost_equal(expected, actual)

    def test_f17_37h_tbs_are_scaled(self):
        satellite = 'f17'
        channel = '37h'
        tbs = np.array([0, 6.493506, 1000, 2000])
        expected = np.array([0, 0, 994.5, 1995.5])

        actual = melt.scaled_brightness_temperatures(satellite, channel, tbs)

        npt.assert_array_almost_equal(expected, actual)


if __name__ == "__main__":
    unittest2.main()
