################################################################################
#@***h* Sea_Ice_CDR/test_cdr_configuration.py
#
# NAME
#  test_cdr_configuration.py
#
# LOCATION
#  tests/test_cdr_configuration.py
#
# PURPOSE
#  unittests for testing the cdr configuration code
#
# DESCRIPTION
#  simple unit tests
#
# AUTHOR
#  nsidc@nsidc.org
#  National Snow & Ice Data Center, University of Colorado, Boulder
#
# COPYRIGHT
#  Copyright 2011 Regents of the University of Colorado. The Regents of the
#  University of Colorado grants, and you accept, a personal, nonexclusive,
#  nontransferable license to use this software, at no charge, in accordance
#  with the terms herein, solely for internal research and development, which
#  is limited to testing, measuring, assessing, evaluating the software and/or
#  data that you generate using this software, and not for any commercial
#  purpose. You may not develop derivative works based on this software, or
#  redistribute this software.
#
# CREATION DATE
#  Aug 2011
#
# EXTERNALS
#    unittest
#    cdr_configuration.configuration_dictionary
#    cdr_configuration.platform_data_from_date
#    cdr_configuration.set_test_years
#    cdr_lib
#
# SUBROUTINES
# class TestCdrConfiguration
#    setUp
#    test_simple_creation
#    test_rootdir
#    test_invalid_input_date
#    test_platform_from_date_with_pre_nimbus_date
#    test_platform_from_date_with_n07_start_date
#    test_platform_from_date_with_f08_start_date
#    test_platform_from_date_with_f11_start_date
#    test_platform_from_date_with_f13_start_date
#    test_platform_from_date_with_f13_start_date
#
#
# INPUTS
#  n/a
#
# OUTPUTS
#  n/a
#
# MODIFICATION HISTORY
#
#
#@*****
################################################################################

import unittest


from cdr_configuration import configuration_dictionary
from cdr_configuration import platform_data_from_date
from cdr_configuration import set_test_years

import cdr_lib as cl


class TestCdrConfiguration(unittest.TestCase):

    def setUp(self):
        self.root = cl.cdr_root_dir()

    def test_simple_creation(self):
        """
        Test that I can create a configuration dictionary
        and some expected values exist
        """
        config = configuration_dictionary('n', '20070315')
        self.assertEqual(config["MMM"], 'mar')
        self.assertEqual(config["nt_var"],
                         "goddard_nt_seaice_conc")

    def test_rootdir(self):
        """ Test that I can get the root directory of my project via lib call """

        config = configuration_dictionary('n', '20070315')
        base = cl.cdr_root_dir()
        self.assertEqual(base, config["CDR_root"])

    def test_invalid_input_date(self):
        """
        Make sure we can throw exception when bad date is used.
        """
        try:
            config = configuration_dictionary('n', '19000101')
            config
        except IOError as error:
            # Just make sure we get here.
            self.assertTrue(True)
            error

    def test_platform_from_date_with_pre_nimbus_date(self):
        config = {'iYY': 1900,
                  'iMM': 10,
                  'iDD': 26}
        set_test_years(config)
        try:
            platform_data_from_date(config)
        except IOError as io_error:
            self.assertEqual(str(io_error), "Date given is pre-NIMBUS (Oct 26, 1978) and "
                             "outside the operational range of this code.\n")

    def test_platform_from_date_with_n07_start_date(self):
        config = {'iYY': 1978,
                  'iMM': 10,
                  'iDD': 26}
        set_test_years(config)
        platform_data_from_date(config)
        self.assertEqual(config['sat'], 'n07')
        self.assertEqual(config['goddard_sat'], 'n07')
        self.assertEqual(config['SAT'], 'N07')

    def test_platform_from_date_with_f08_start_date(self):
        config = {'iYY': 1987,
                  'iMM': 7,
                  'iDD': 8}
        set_test_years(config)
        platform_data_from_date(config)
        self.assertEqual(config['sat'], 'f08')
        self.assertEqual(config['goddard_sat'], 'f08')
        self.assertEqual(config['SAT'], 'F08')

    def test_platform_from_date_with_f11_start_date(self):
        config = {'iYY': 1991,
                  'iMM': 12,
                  'iDD': 3}
        set_test_years(config)
        platform_data_from_date(config)
        self.assertEqual(config['sat'], 'f11')
        self.assertEqual(config['goddard_sat'], 'f11')
        self.assertEqual(config['SAT'], 'F11')

    def test_platform_from_date_with_f13_start_date(self):
        config = {'iYY': 1995,
                  'iMM': 10,
                  'iDD': 1}
        set_test_years(config)
        platform_data_from_date(config)
        self.assertEqual(config['sat'], 'f13')
        self.assertEqual(config['goddard_sat'], 'f13')
        self.assertEqual(config['SAT'], 'F13')

    def test_platform_from_date_with_f17_start_date(self):
        config = {'iYY': 2008,
                  'iMM': 1,
                  'iDD': 1}
        set_test_years(config)
        platform_data_from_date(config)
        self.assertEqual(config['sat'], 'f17')
        self.assertEqual(config['goddard_sat'], 'f17')
        self.assertEqual(config['SAT'], 'F17')


if __name__ == '__main__':
    unittest.main()
