################################################################################
#@***h* Sea_Ice_CDR/test_cdr_lib.py
#
# NAME
#  test_cdr_lib.py
#
# LOCATION
#  tests/test_cdr_lib.py
#
# PURPOSE
#  unit tests for cdr_lib.py
#
# DESCRIPTION
#  simple unit tests
#
# AUTHOR
#  nsidc@nsidc.org
#  National Snow & Ice Data Center, University of Colorado, Boulder
#
# COPYRIGHT
#  Copyright 2013 Regents of the University of Colorado. The Regents of the
#  University of Colorado grants, and you accept, a personal, nonexclusive,
#  nontransferable license to use this software, at no charge, in accordance
#  with the terms herein, solely for internal research and development, which
#  is limited to testing, measuring, assessing, evaluating the software and/or
#  data that you generate using this software, and not for any commercial
#  purpose. You may not develop derivative works based on this software, or
#  redistribute this software.
#
# CREATION DATE
#  Feb 2013
#
# EXTERNALS
#   cdr_lib
#   datetime
#   mock
#   numpy
#   os
#   unittest2
#   netCDF4.Dataset
#
# SUBROUTINES
#   class MockedMissingDataset
#     __enter__
#     __exit__
#     __init__
#   class TestCDRLib
#     test__all_missing_when_allDataIsMissing
#     test__all_missing_when_someDataIsAvailable
#     test__formatted_comment_when_TwoMissing
#     test__formatted_comment_when_multipleMissing
#     test__formatted_comment_when_none_missing
#     test__formatted_comment_when_oneMissing
#     test_latest_satellite
#     test_mask_filters_applying_single_value
#     test_mask_filters_applying_values
#     test_mask_filters_ignoring_multiple_values
#     test_mask_filters_ignoring_value
#     test_missing_comment_when_no_data_is_missing
#     test_missing_comment_when_one_variable_is_missing
#     test_missing_comment_when_someVariablesAreMissing
#     test_missing_comment_with_multiple_vars_missing
#     test_platform_from_date
#     test_satellite_for_month
#     test_satellite_for_month_equal_numbers_picks_latest_date
#     test_satellite_for_month_equal_numbers_picks_latest_date_reversed_order
#     test_satellite_for_month_should_be_n07
#
#
# INPUTS
#  n/a
#
# OUTPUTS
#  n/a
#
# MODIFICATION HISTORY
#
#
#@*****
################################################################################

import cdr_lib as cl
import datetime as dt
import mock
import numpy as np
import os
import unittest2
from netCDF4 import Dataset


class MockedMissingDataset(object):

    vars = ['seaice_conc_cdr',
            'stdev_of_seaice_conc_cdr',
            'melt_onset_day_seaice_conc_cdr',
            'qa_of_seaice_conc_cdr',
            'goddard_merged_seaice_conc',
            'goddard_nt_seaice_conc',
            'goddard_bt_seaice_conc']

    def __init__(self, missing_vars):
        self.filename = 'fake_fn'
        self.ncDataset = Dataset(self.filename, 'w', diskless=True)
        self.ncDataset.createDimension('length', 3)
        for v in self.vars:
            self.ncDataset.createVariable(v, 'int8', ('length',), fill_value=1)
        cl._all_missing = mock.MagicMock(
            side_effect=[v in missing_vars for v in self.vars])

    def __enter__(self):
        return self.ncDataset

    def __exit__(self, type, value, traceback):
        self.ncDataset.close()
        # diskless doesn't currently work... so you need to clean up.
        os.remove(self.filename)


class TestCDRLib(unittest2.TestCase):

    def test_mask_filters_applying_values(self):
        arr = np.array([0, 0, 0, 0, 0])
        expected = np.array([1, 0, 3, 0, 0])
        the_mask = np.array([1, 2, 3, 4, 5])
        mask_closure = cl.cdr_mask_filter_factory(the_mask, use=[1, 3])
        mask_closure(arr)
        np.testing.assert_array_equal(expected, arr)

    def test_mask_filters_applying_single_value(self):
        arr = np.array([0, 0, 0, 0, 0])
        expected = np.array([1, 0, 1, 0, 1])
        the_mask = np.ma.array([1, 2, 1, 4, 1])
        mask_closure = cl.cdr_mask_filter_factory(the_mask, use=[1, 3])
        mask_closure(arr)
        np.testing.assert_array_equal(expected, arr)

    def test_mask_filters_ignoring_value(self):
        arr = np.array([0, 1, 2, 1, 0])
        the_mask = np.array([10, 20, 30, 40, 50])
        expected = np.array([10, 20, 2, 40, 50])
        mask_closure = cl.cdr_mask_filter_factory(the_mask, ignore=[30])
        mask_closure(arr)
        np.testing.assert_array_equal(expected, arr)

    def test_mask_filters_ignoring_multiple_values(self):
        arr = np.array([0, 1, 2, 3, 4])
        the_mask = np.array([50, 40, 30, 40, 50])
        expected = np.array([50, 1, 2, 3, 50])
        mask_closure = cl.cdr_mask_filter_factory(the_mask, ignore=[30, 40])
        mask_closure(arr)
        np.testing.assert_array_equal(expected, arr)

    def test_platform_from_date(self):
        self.assertEqual(cl.platform_from_date(dt.date(1900, 10, 26)), 'pre-NIMBUS')
        self.assertEqual(cl.platform_from_date(dt.date(1978, 10, 26)), 'n07')
        self.assertEqual(cl.platform_from_date(dt.date(1981, 3, 4)), 'n07')
        self.assertEqual(cl.platform_from_date(dt.date(1987, 7, 8)), 'f08')
        self.assertEqual(cl.platform_from_date(dt.date(1989, 12, 12)), 'f08')
        self.assertEqual(cl.platform_from_date(dt.date(1991, 12, 3)), 'f11')
        self.assertEqual(cl.platform_from_date(dt.date(1993, 6, 15)), 'f11')
        self.assertEqual(cl.platform_from_date(dt.date(1995, 10, 1)), 'f13')
        self.assertEqual(cl.platform_from_date(dt.date(2000, 11, 1)), 'f13')
        self.assertEqual(cl.platform_from_date(dt.date(2008, 1, 1)), 'f17')
        self.assertEqual(cl.platform_from_date(dt.date(2013, 2, 13)), 'f17')

    def test__all_missing_when_allDataIsMissing(self):
        FakeDataset = Dataset('fake_fn', 'w', diskless=True)
        FakeDataset.createDimension('length', 5)
        ncVariable = FakeDataset.createVariable('var', 'int8', ('length',), fill_value=-1)
        ncVariable[:] = np.array([-1, -1, -1, -1, -1])
        try:
            self.assertTrue(cl._all_missing(ncVariable))
        finally:
            FakeDataset.close()
            os.remove('fake_fn')

    def test__all_missing_when_someDataIsAvailable(self):
        FakeDataset = Dataset('fake_fn', 'w', diskless=True)
        FakeDataset.createDimension('length', 5)
        ncVariable = FakeDataset.createVariable('var', 'int8', ('length',), fill_value=-1)
        ncVariable[:] = np.array([-1, -1, -1, -1, 0])
        try:
            self.assertFalse(cl._all_missing(ncVariable))
        finally:
            FakeDataset.close()
            os.remove('fake_fn')

    def test_missing_comment_when_no_data_is_missing(self):
        missing = []
        expected = cl._formatted_comment(missing, 'day')
        with MockedMissingDataset(missing) as mmd:
            self.assertEqual(expected, cl.missing_daily_comment(mmd))

    def test_missing_comment_when_someVariablesAreMissing(self):
        missing = ['seaice_conc_cdr']
        expected = cl._formatted_comment(missing, 'day')
        with MockedMissingDataset(missing) as mmd:
            self.assertEqual(expected, cl.missing_daily_comment(mmd))

    def test_missing_comment_when_one_variable_is_missing(self):
        the_vars = ['seaice_conc_cdr',
                    'stdev_of_seaice_conc_cdr',
                    'melt_onset_day_seaice_conc_cdr',
                    'qa_of_seaice_conc_cdr',
                    'goddard_merged_seaice_conc',
                    'goddard_nt_seaice_conc',
                    'goddard_bt_seaice_conc']

        for var in the_vars:
            with MockedMissingDataset([var]) as mmd:
                expected = cl._formatted_comment([var], 'day')
                self.assertEqual(expected, cl.missing_daily_comment(mmd))

    def test_missing_comment_with_multiple_vars_missing(self):
        missing = ['melt_onset_day_seaice_conc_cdr',
                   'qa_of_seaice_conc_cdr']
        expected = cl._formatted_comment(missing, 'day')
        with MockedMissingDataset(missing) as mmd:
            self.assertEqual(expected, cl.missing_daily_comment(mmd))

    def test__formatted_comment_when_none_missing(self):
        missing = []
        expected = "none"
        self.assertEqual(expected, cl._formatted_comment(missing, "irrelevant"))

    def test__formatted_comment_when_oneMissing(self):
        missing = ['variablea']
        expected = "The variable variablea is not available for this day."
        self.assertEqual(expected, cl._formatted_comment(missing, "day"))

    def test__formatted_comment_when_TwoMissing(self):
        missing = ['variablea', 'variableb']
        expected = "The variables variablea, variableb are not available for this day."
        self.assertEqual(expected, cl._formatted_comment(missing, "day"))

    def test__formatted_comment_when_multipleMissing(self):
        missing = ['vara', 'varb', 'var3']
        expected = "The variables vara, varb, var3 are not available for this day."
        self.assertEqual(expected, cl._formatted_comment(missing, "day"))

    def test_satellite_for_month(self):
        files = ['seaice_conc_daily_nh_f08_19870710_v01r00.nc',
                 'seaice_conc_daily_nh_f08_19870711_v01r00.nc',
                 'seaice_conc_daily_nh_f08_19870712_v01r00.nc',
                 'seaice_conc_daily_nh_n07_19870709_v01r00.nc', ]

        satellite = cl.satellite_for_month(files)
        self.assertEqual('f08', satellite)

    def test_satellite_for_month_should_be_n07(self):
        files = ['seaice_conc_daily_nh_f08_19870712_v01r00.nc',
                 'seaice_conc_daily_nh_n07_19870709_v01r00.nc',
                 'seaice_conc_daily_nh_n07_19870710_v01r00.nc',
                 'seaice_conc_daily_nh_n07_19870711_v01r00.nc', ]

        satellite = cl.satellite_for_month(files)
        self.assertEqual('n07', satellite)

    def test_satellite_for_month_equal_numbers_picks_latest_date(self):
        files = ['seaice_conc_daily_nh_f08_19870712_v01r00.nc',
                 'seaice_conc_daily_nh_f08_19870711_v01r00.nc',
                 'seaice_conc_daily_nh_n07_19870709_v01r00.nc',
                 'seaice_conc_daily_nh_n07_19870710_v01r00.nc', ]

        satellite = cl.satellite_for_month(files)
        self.assertEqual('f08', satellite)

    def test_satellite_for_month_equal_numbers_picks_latest_date_reversed_order(self):
        files = ['seaice_conc_daily_nh_f08_19870709_v01r00.nc',
                 'seaice_conc_daily_nh_f08_19870710_v01r00.nc',
                 'seaice_conc_daily_nh_f11_19870712_v01r00.nc',
                 'seaice_conc_daily_nh_f11_19870711_v01r00.nc', ]

        satellite = cl.satellite_for_month(files)
        self.assertEqual('f11', satellite)

    def test_latest_satellite(self):
        files = ['seaice_conc_daily_nh_f08_19870709_v01r00.nc',
                 'seaice_conc_daily_nh_f13_19870712_v01r00.nc',
                 'seaice_conc_daily_nh_f08_19870710_v01r00.nc',
                 'seaice_conc_daily_nh_f11_19870711_v01r00.nc', ]

        satellite = cl.latest_satellite(files)
        self.assertEqual('f13', satellite)


if __name__ == '__main__':
    unittest2.main()
