################################################################################
#@***h* Sea_Ice_CDR/test_cdr_monthly_seaice.py
#
# NAME
#  test_cdr_monthly_seaice.py
#
# LOCATION
#  tests/test_cdr_monthly_seaice.py
#
# PURPOSE
#  Unit tests for monthly CDR creation.
#
# DESCRIPTION
#  Unit tests for monthly CDR creation.
#
# AUTHOR
#  nsidc@nsidc.org
#  National Snow & Ice Data Center, University of Colorado, Boulder
#
# COPYRIGHT
#  Copyright 2012 Regents of the University of Colorado. The Regents of the
#  University of Colorado grants, and you accept, a personal, nonexclusive,
#  nontransferable license to use this software, at no charge, in accordance
#  with the terms herein, solely for internal research and development, which
#  is limited to testing, measuring, assessing, evaluating the software and/or
#  data that you generate using this software, and not for any commercial
#  purpose. You may not develop derivative works based on this software, or
#  redistribute this software.
#
# CREATION DATE
#  Jan 2012
#
# EXTERNALS
#  numpy
#  numpy.ma
#  numpy.testing
#  unittest2
#  cdr_lib
#  cdr_configuration
#  cdr_monthly_seaice
#
#
# SUBROUTINES
#   class TestMonthlyCdr
#     setUpClass
#     test_compound_mask
#     test_data_replacer
#     test_masked_count
#     test_monthly_merged_data_fields
#     test_numpy_concat
#
#
# INPUTS
#  n/a
#
# OUTPUTS
#  n/a
#
# MODIFICATION HISTORY
#
#
#@*****
################################################################################

import numpy as np
import numpy.ma as ma
import numpy.testing as nptest
import unittest2 as unittest

import cdr_lib as cl
import cdr_configuration as config
import cdr_monthly_seaice as cms


class TestMonthlyCdr(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        """ Set member vars needed for tests. """

        test_cases = {'test1':  ["unused", 'north', '20060301'],
                      'test2':  ["unused", 'south', '20060301'],
                      'test3':  ["unused", 'north', '19800101']}

        cls.configuration = {}
        cls.cdrobj = {}

        for tname, seed in test_cases.items():
            preset = {
                "cdr_version": cl.create_product_version_string("01", "00")
            }
            cls.configuration[tname] = config.create_configuration_dictionary(seed,
                                                                              preset)
            cls.cdrobj[tname] = cms.CDRMonthlySeaice(cls.configuration[tname])

    def test_data_replacer(self):
        """Test that the data_replacer function works as expected.

        Arguments:
        - `self`:
        """
        in_data = np.asarray([0, 10, 22, 33, 44, 55, 66], dtype=cl.NC_BYTE)
        loc_data = np.asarray([0,  1,  2,  2,  4,  5,  5], dtype=cl.NC_BYTE)
        from_to = {1: 100, 2: 60, 5: 5}
        expected_out = np.asarray([0, 100, 60, 60, 44, 5, 5], dtype=cl.NC_BYTE)
        out_data = cl.data_replacer(in_data, loc_data, from_to)
        np.testing.assert_array_equal(expected_out, out_data)

    def test_monthly_merged_data_fields(self):
        """Test that merge_monthly_fields works as expected

        Arguments:
        - `self`:
        """
        bt_array = np.asarray([0, 33, 9, 9, 10, 15, 16, 90, 99, 40, 253, 252],
                              dtype=cl.NC_BYTE)
        nt_array = np.asarray([0, 3, 9, 40, 10,  1, 60, 90, 40, 99, 253, 252],
                              dtype=cl.NC_BYTE)
        merged = self.cdrobj['test1'].merge_monthly_fields(bt_array, nt_array)
        expected_merged = np.asarray([0, 33, 0, 0,  10, 15, 60, 90, 99, 99, 253, 252],
                                     dtype=cl.NC_BYTE)
        np.testing.assert_array_equal(expected_merged, merged)

    def test_masked_count(self):
        """
        Just an example to show the correct axis for counting masked elements.
        This is used in the determiniation of the number of missing days.
        """

        #      reshape(depth, row, column)
        a = np.arange(18).reshape(3, 3, 2)
        # Create a 3x3x2 array
        #  array([[[0,  1],
        #        [2,  3],
        #        [4,  5]],

        #        [[6,  7],
        #         [8,  9],
        #         [10, 11]],

        #        [[12, 13],
        #         [14, 15],
        #         [16, 17]]])
        a = ma.array(a)
        a[0, 0, 0] = ma.masked    # 0
        a[1, 0, 0] = ma.masked    # 6
        a[2, 2, 1] = ma.masked    # 17
        self.assertEqual(ma.count_masked(a), 3)
        nptest.assert_array_equal(ma.count_masked(a, axis=0), [[2, 0], [0, 0], [0, 1]])

    def test_numpy_concat(self):
        a = np.array([[1, 2], [3, 4]])
        y = np.dstack((a, a, a, a, a, a))
        nptest.assert_array_almost_equal(y.mean(axis=2), a)
        a = ma.array(a)
        b = a.copy()
        b[0, 0] = ma.masked
        y = ma.dstack((a, b, a))
        y = ma.dstack((y, b))
        # y = masked_array(data =
        # [[[1 -- 1 --]
        #  [2 2 2 2]]

        # [[3 3 3 3]
        #  [4 4 4 4]]],
        #             mask =
        # [[[False  True False  True]
        #  [False False False False]]

        # [[False False False False]
        #  [False False False False]]],
        #       fill_value = 999999)
        nptest.assert_array_equal(y.mean(axis=2), a,
                                  err_msg="Masked array average is not equal")

    def test_compound_mask(self):
        """
        Want to make sure that you can separate 255 from 250-254 and return missing
        This test was to check how masked arrays can be manipulated to use as a way
        to add static flagged values (pole, land, coast, etc) back into a masked array
        """

        a = ma.array([0, 1, 50, 100, 250, 251, 252, 253, 254, 255])
        b = a.data[:] * 5000
        bad = ma.masked_greater(a, 100)
        missing = ma.masked_equal(a, 255)
        mask_flags_only = ma.array(bad.mask * ~missing.mask)
        nptest.assert_array_equal(b, [0, 5000, 250000, 500000,
                                      1250000, 1255000, 1260000,
                                      1265000, 1270000, 1275000])
        with_flags = ma.where(mask_flags_only, a, b)
        nptest.assert_array_equal(with_flags, [0, 5000, 250000, 500000, 250,
                                               251, 252, 253, 254, 1275000])


if __name__ == '__main__':
    unittest.main()
