################################################################################
#@***h* Sea_Ice_CDR/test_cdr_stddev.py
#
# NAME
#  test_cdr_stddev.py
#
# LOCATION
#  tests/test_cdr_stddev.py
#
# PURPOSE
#  unittests for testing the cdr standard deviation code
#
# DESCRIPTION
#  Test case for making sure standard deviation computations match the expected.
#
# AUTHOR
#  nsidc@nsidc.org
#  National Snow & Ice Data Center, University of Colorado, Boulder
#
# COPYRIGHT
#  Copyright 2011 Regents of the University of Colorado. The Regents of the
#  University of Colorado grants, and you accept, a personal, nonexclusive,
#  nontransferable license to use this software, at no charge, in accordance
#  with the terms herein, solely for internal research and development, which
#  is limited to testing, measuring, assessing, evaluating the software and/or
#  data that you generate using this software, and not for any commercial
#  purpose. You may not develop derivative works based on this software, or
#  redistribute this software.
#
# CREATION DATE
#  Aug 2011
#
# EXTERNALS
#    unittest2
#    numpy
#    numpy.ma
#    numpy.testing
#    cdr_lib
#
# SUBROUTINES
#   class CDRStddevTests
#     setUp
#     test__small_arrays_on_local_std_dev
#     test_clip_masked_array
#     test_masked_border
#     test_masked_std
#     test_num_valid_items
#     test_scale_from_percent_to_fraction
#     test_sub_array
#
# INPUTS
#  n/a
#
# OUTPUTS
#  n/a
#
# MODIFICATION HISTORY
#
#
#@*****
################################################################################

import unittest2 as unittest
import numpy as np
import numpy.ma as ma
import numpy.testing as nptest

import cdr_lib as cl


class CDRStddevTests(unittest.TestCase):
    """ TestCase class for developing the CDR Standard Deviation Codes """

    def setUp(self):
        """ Set up some necessary variables """

        # Set up Small test array
        self.x = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])

    def test_sub_array(self):
        """ Test that you can nest an array inside an container array.  This
        will be used to add the border around the initial grids so that we can
        ignore edge cases.
        """

        border = np.zeros([4, 4])
        border[1:, 1:] = self.x
        sub = np.array([[0, 0, 0, 0],
                        [0, 1, 2, 3],
                        [0, 4, 5, 6],
                        [0, 7, 8, 9]], dtype=b'float32')
        self.assertTrue(np.equal(border, sub).all())

    def test_masked_border(self):
        """
        Check that you can create a masked array that contains a sample array
        and that STDDEV computations are correctly ignored at the borders.
        """

        x = self.x
        border = cl.add_masked_border(x)
        sub = np.array([[0, 0, 0, 0, 0],
                        [0, 1, 2, 3, 0],
                        [0, 4, 5, 6, 0],
                        [0, 7, 8, 9, 0],
                        [0, 0, 0, 0, 0]], dtype=b'float32')

        self.assertTrue(np.equal(sub, ma.filled(border, fill_value=0)).all())

    def test_masked_std(self):
        """ Test that the masking works for computing standard deviation.
        """

        masked = cl.add_masked_border(self.x)
        maskedStd = np.std(masked[0:3, 0:3], ddof=1)
        normStd = np.std([1, 2, 4, 5], ddof=1)
        nptest.assert_array_almost_equal(maskedStd, normStd)
        maskedStd = np.std(masked[1:4, 1:4], ddof=1)
        normStd = np.std([1, 2, 3, 4, 5, 6, 7, 8, 9], ddof=1)
        nptest.assert_array_almost_equal(maskedStd, normStd)

    def test_clip_masked_array(self):
        """ Make sure you can return a clipped array """

        masked = cl.add_masked_border(self.x)
        clipped = cl.clip_masked_array(masked)
        self.assertTrue(np.equal(clipped, self.x).all())

    def test_num_valid_items(self):
        """ Test to make sure the std code will require
        the correct number of points."""

        masked = cl.add_masked_border(self.x)
        sub = masked[0:3, 0:3]
        count = 4
        self.assertEqual(count, sub.count())

    def test__small_arrays_on_local_std_dev(self):
        """ Test that the stddev is what we expect for small datasets """

        # Test simple small array
        x = np.array([[1, 2], [3, 4]])
        y = np.array([[5, 6], [7, 8]])
        allStd = np.std([1, 2, 3, 4, 5, 6, 7, 8], ddof=1)
        presumed = np.ones([2, 2]) * allStd
        stddev = cl.local_std_dev(x, y, -1)
        nptest.assert_array_almost_equal(stddev, presumed)

        # Test for masked arrays by Masking out values 7 & 8 in y
        y = ma.masked_greater(y, 6.5)
        newStd = np.std([1, 2, 3, 4, 5, 6], ddof=1)
        presumed = np.ones([2, 2]) * newStd
        stddev = cl.local_std_dev(x, y, -1)
        nptest.assert_array_almost_equal(stddev, presumed)

        # Test If not enough valid elements.
        y = ma.masked_greater(y, 5.5)
        newStd = np.std([1, 2, 3, 4, 5], ddof=1)
        # Presumed is missing
        presumed = np.ones([2, 2]) * -1.
        stddev = cl.local_std_dev(x, y, -1)
        nptest.assert_array_almost_equal(stddev, presumed)

    def test_scale_from_percent_to_fraction(self):
        test_in = np.array([[30., 40., 50.],
                            [30., 40., -1.],
                            [30., 40., 80.]])

        test_out_master = np.array([[.3, .4, .5],
                                    [.3, .4, -1.],
                                    [.3, .4, .8]])

        test_out = cl.scale_from_percent_to_fraction(test_in, -1)
        nptest.assert_array_almost_equal(test_out_master, test_out)


if __name__ == '__main__':
    unittest.main()
