################################################################################
#@***h* Sea_Ice_CDR/validate.py
#
# NAME
#  validate.py
#
# LOCATION
#  validate/validate.py
#
# PURPOSE
#  The validate module compares locally generated CDR files with those
#  that have been published to the NSIDC FTP site (via the filesystem,
#  not via FTP).
#
# DESCRIPTION
#  Compare locally generated CDR files with those published on the NSIDC
#  FTP directory via:
#
#    python validate/validate.py year type
#
#  where:
#    year: a four-digit year to compare
#    type: either 'daily' or 'monthly' (without quotes)
#
#  This module requires the nccmp program which is not included with
#  the CDR distribution. It can be obtained at:
#
#    http://nccmp.sourceforge.net
#
# AUTHOR
#  nsidc@nsidc.org
#  National Snow & Ice Data Center, University of Colorado, Boulder
#
# COPYRIGHT
#  Copyright 2011 Regents of the University of Colorado. The Regents of the
#  University of Colorado grants, and you accept, a personal, nonexclusive,
#  nontransferable license to use this software, at no charge, in accordance
#  with the terms herein, solely for internal research and development, which
#  is limited to testing, measuring, assessing, evaluating the software and/or
#  data that you generate using this software, and not for any commercial
#  purpose. You may not develop derivative works based on this software, or
#  redistribute this software.
#
# CREATION DATE
#  February 2013
#
# EXTERNALS
#   argparse
#   calendar
#   datetime
#   os
#   subprocess
#   sys
#   cdr_lib.platform_from_date
#
# SUBROUTINES
#   compare
#   dataset_filename
#   filename
#   local_filename
#   main
#   platform_from_date
#
# INPUTS
#  Local CDR files (generally in the ./output directory)
#
# OUTPUTS
#  Textual comparison between the local and published CDR files (stdout)
#
# MODIFICATION HISTORY
#
#
#@*****
################################################################################
import argparse
import calendar
import datetime as dt
import os
import subprocess
import sys

from cdr_lib import platform_from_date


LOCAL_DIR = "output"
DATASET_DIR = "/projects/DATASETS/NOAA/G02202"


def filename(product, hemisphere, date, platform):
    if product.lower() == 'daily':
        filename = "seaice_conc_{0}_{1}h_{2}_{3}_v01r00.nc"\
            .format(product.lower(),
                    hemisphere[0].lower(),
                    platform,
                    date.strftime("%Y%m%d"))
    else:
        filename = "seaice_conc_{0}_{1}h_{2}_{3}_v01r00.nc".format(product.lower(),
                                                                   hemisphere[0].lower(),
                                                                   platform,
                                                                   date.strftime("%Y%m"))
    return filename


def local_filename(product, hemisphere, date, platform):
    name = filename(product, hemisphere, date, platform)
    return os.path.join(LOCAL_DIR, name)


def dataset_filename(product, hemisphere, date, platform):
    name = filename(product, hemisphere, date, platform)
    if product.lower() == "daily":
        directory_path = os.path.join(DATASET_DIR, hemisphere, product.lower(),
                                      str(date.year), name)
    else:
        directory_path = os.path.join(DATASET_DIR, hemisphere, product.lower(), name)
    return directory_path


def compare(product, date, platform):
    for hemisphere in ["north", "south"]:
        local = local_filename(product, hemisphere, date, platform)
        dataset = dataset_filename(product, hemisphere, date, platform)
        print "Comparing {0} {1}  Files: {2}  {3}".format(date.strftime("%Y%m%d"),
                                                          hemisphere, local, dataset)
        sys.stdout.flush()
        for comparison in ["-fmg -G date_created,software_version_id", "-fd"]:
            cmd = "nccmp {0} {1} {2}".format(comparison, local, dataset)
            subprocess.call(cmd, stdout=sys.stdout, stderr=sys.stdout, shell=True)


def main(year, product):
    if product == "daily":
        begin_range = 1
        end_range = 365
        if (calendar.isleap(int(year))):
            end_range = 366
        date_format_string = "%Y%j"
    else:
        begin_range = 1
        end_range = 12
        date_format_string = "%Y%m"
    for timestep in xrange(begin_range, end_range + 1):
        date = dt.datetime.strptime("%s%d" % (year, timestep), date_format_string).date()
        platform = platform_from_date(date)
        compare(product, date, platform)


if __name__ == "__main__":
    parser = argparse.ArgumentParser("Compare the current monthly or daily CDRs to the "
                                     "published CDRs.")
    parser.add_argument("year", help="Year of data you would like to compare.")
    parser.add_argument("product", help="Either 'daily' or 'monthly'.")
    args = parser.parse_args()
    main(args.year, args.product)
