      subroutine seaice_goddard(tb19h,tb19v,tb22v,tb37v,ihem,iplatform,
     1     scale,totout,myrout)
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c
c                       SEAICE_GODDARD.F  
c
c PURPOSE:  This program runs routines to calculate multi-year and
c           total sea ice concentrations for a single pixel using the
c           NASA TEAM algorithm with modified tie points, used for
c           so-called "Goddard Time Series" data set.
c
c COMMENTS: (This code was adapted from the code RFK created to run the
c           seaice concentrations for N. Hemisphere EASE-Grid which
c           was code adapted from Nick's code that RFK got from 
c           him around April 5, 1994 (and includes weather filter)
c
c AUTHOR:   Raymond Kokaly   15 Nov 1994
c
c $Id: seaice_goddard.f 22320 2013-01-18 21:48:52Z kbeam $
c
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc        
c          
c   ALGORITHM DESCRIPTION:
c     The ice concentration algorithm is an adaptation of the     
c     algorithm designed by the NODS Software Group (refer to the     
c     NODS User Handbook, Appendix CC, pp 22-26). The algorithm     
c     uses the latest SSM/I tiepoints which were determined by Don     
c     Cavalieri and Koni Steffen.     
c
c     A weather filter is applied to the 25km TB cell to correct for
c     false concentrations computed in regions of open water.
c     Weather filter thresholds obtained from:
c     Cavalieri, D. J., C. L. Parkinson, N. DiGirolamo, and A. Ivanoff.
c     2012. Intersensor Calibration Between F13 SSMI and F17 SSMIS for
c     Global Sea Ice Data Records. IEEE Geoscience and Remote Sensing
c     Letters, Vol. 9, No. 2, March 2012.
c
c   INPUTS:
c     tb19h  -  Brightness temperature for 19GHz H (integer*2)
c     tb19v  -  Brightness temperature for 19GHz V (integer*2)
c     tb22v  -  Brightness temperature for 22GHz V (integer*2)
c     tb37v  -  Brightness temperature for 37GHz V (integer*2)
c     ihem   -  Defines the appropriate hemisphere to apply this
c                algorithm (1=North, 2=South).  Used for tie points.
c     iplatform -  Defines the SSM/I platform (1=F08, 2=F11, 3=F13,
c                4=F15_BRIDGE, 5=F15 (not implemented), 6=F17, 7=F17_FINAL)
c                Used for tie points.
c     scale = 1=output scaled to 0-100%, default: output is scaled to
c             0-250.  SUSPECT_VALUE is 251 in either case.
c      
c
c   INTERNALS
c      c     -  Array of coefficients used to compute ice concentrations.
c                 Refer to NODS Handbook Volue II. Appendix CC pp 23-24.
c      fyr   -  First-year ice concentration (real)
c      gr1   -  Gradient ratio (37v,19v)
c      gr2   -  Gradient ratio (22v,19v)
c      PR    -  Polarization ratio
c          
c   OUTPUTS          
c      totout   -  Output total ice concentration (integer*2)
c      myrout   -  Multi-year ice concentration (integer*2)
c      
c              Sea ice percentages range from 0 - 100%.  Default output is
c              scaled to 0 - 250; if scale=1, output is 0-100(%).  
c              Cells for which concentrations were suspect or           
c              not computed are flagged at 251.
c
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc       

      implicit     none 

      integer*2    ihem, iplatform
      integer*2    tb19h,tb19v,tb22v,tb37v
      real         fyr, myr, tot
      integer*2    scale
      integer*2    myrout, totout
      real         BT19V,BT19H,BT37V,BT22V
      real         c(12),ANF,ANM,DD                                        
      double precision  gr1,gr2,PR,PRGR
      integer*4    verbose
      integer*2    force_rov
      real         f15_bridge_22v_regression
      real         scale_factor, maxconc
      real         gr1_threshold

      integer*2    SUSPECT_VALUE
      real         PERCENT, MINCONC
      integer*2    F15_BRIDGE
      PARAMETER    (SUSPECT_VALUE = 251)
      PARAMETER    (PERCENT       = 100.0)
      PARAMETER    (MINCONC       = 0.0)
      PARAMETER    (F15_BRIDGE    = 4)

      verbose = 0
      if (1 .lt. verbose) then
        WRITE(0,*)
     1       '>> $RCSfile: seaice_goddard.f,v $ $Revision: 1.2.8.1 $'
      endif

c+++
c     Decide which scaling factor to use for output
c---
      scale_factor = 2.5
      if (1 .eq. scale) then
        scale_factor = 1.0
      endif
      maxconc = scale_factor * PERCENT
        
c+++
c     GET ICE ALGORITHM COEFFICIENTS DERIVED FROM TIEPOINTS.
c---
      call gtstie(c, ihem, iplatform)
c+++
c     CONVERT INTS TO REALS
c---
      BT19H = float(tb19h)
      BT19V = float(tb19v)
      BT22V = float(tb22v)
      BT37V = float(tb37v)

c+++  
c     FOR F15_BRIDGE, DO A REGRESSION ON 22V TBS ONLY.
c     ADJUSTS F15 22V TBS THAT HAVE BEEN CONTAMINATED BY
c     RADCON BEACON BACK DOWN TO PSEUDO-F13 TBS
c---
      if ( F15_BRIDGE .eq. iplatform ) then
        BT22V = f15_bridge_22v_regression( BT22V, ihem )
      endif
      
c+++
c     COMPUTE OPEN WATER WEATHER FILTER.
c---
      gr1 = (BT37V - BT19V)/(BT37V + BT19V)            
      gr2 = (BT22V - BT19V)/(BT22V + BT19V)
c+++
c     WEATHER FILTER: SET ICE CELL TO 0% CONCENTRATION IF
c     F17_FINAL platform in Southern Hemisphere:
c         gr1  > 0.053      indicates weather over open water
c         gr2  > 0.045      indicates weather over open water
c     All other platforms / hemispheres:
c         gr1  > 0.05       indicates weather over open water
c         gr2  > 0.045      indicates weather over open water
c---
      if (iplatform .eq. 7 .and. ihem .eq. 2) then
        gr1_threshold = 0.053
      else
        gr1_threshold = 0.05
      endif
      
      if (gr1 .gt. gr1_threshold .or. gr2 .gt. .045) then

        myrout = int( MINCONC )
        totout = myrout

      else                                    
c+++
c       CHECK IF 19V < 19H, IF SO THEN A PROBLEM WITH TBs
c---
        if (BT19V.LT.BT19H)  then           

          myrout = SUSPECT_VALUE
          totout = myrout

        else                             
c+++
c       IF PREVIOUS TWO CONDITIONS NOT MET CONTINUE WITH CALC
c       COMPUTE POLARIZATION RATIO PR
c---
          PR   = (BT19V - BT19H)/(BT19V + BT19H)               
          PRGR = PR*GR1                             
          ANF  = c(1) + c(2)*PR + c(3)*GR1 + c(4)*PRGR           
          ANM  = c(9) + c(10)*PR + c(11)*GR1 + c(12)*PRGR
          DD   = c(5) + c(6)*PR + c(7)*GR1 + c(8)*PRGR  
c+++
c         COMPUTE FIRST-YEAR, MULTI-YEAR, AND TOTAL ICE CONCENTRATIONS
c         DEFAULT IS SCALE 0 - 100% CONCENTRATIONS TO THE AVAILABLE INTEGER
c         RANGE OF 0 - 250.  IF SCALE INPUT ARGUMENT IS SET, SCALE
c         OUTPUT TO 0-100
c---
          fyr = scale_factor * PERCENT * (ANF/DD)
          myr = scale_factor * PERCENT * (ANM/DD)
          tot = fyr + myr    
c+++
c	  FORCE OUTPUT RANGE OF VALUES, ROUND TO NEAREST INTEGER
c---
          totout = force_rov( tot, MINCONC, maxconc )
          myrout = force_rov( myr, MINCONC, maxconc )
          if ( myrout .gt. totout ) myrout = totout

        endif                             
      endif                             
      return                               
      end

      subroutine gtstie (C, ihem, iplatform)             
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c
c GTSTIE - This subroutine computes coefficients for the ice algorithm
c          using the Goddard Time Series tie points.
c          
c   Inputs           
c          c     : Real array of coefficients. Refer to NODS User 
c                  handbook for description of coefficients A-L.
c          ihem    : Integer denoting hemisphere (1=north, 2=south). Used
c                    with iplatform to determine which set of coefficients to use.
c     iplatform : Integer denoting platform (1=F08, 2=F11, 3=F13,
c                 4=F15_BRIDGE), 5=F15 (not implemented), 6=F17, 7=F17_FINAL.
c     Used with ihem to determine which set of coefficients to use
c
c   Internals
c          TB...W: Open water tiepoints. One for each of the three channels
c                  used in the ice concentration algorithm (19V, 19H, 37V)
c          TB...F: Firstyear tiepoints. One for each of the three channels
c                  used in the ice concentration algorithm (19V, 19H, 37V)
c          TB...M: Multiyear tiepoints. One for each of the three channels
c                  used in the ice concentration algorithm (19V, 19H, 37V)
c          DW    : Difference between component tiepoints for open water.
c          SW    : Sum of the component tiepoints for open water.
c          DF    : Difference between component tiepoints for first year ice.
c          SF    : Sum of the component tiepoints for first year ice.
c          DM    : Difference between component tiepoints for multiyear ice.
c          SM    : Sum of the component tiepoints for multiyear ice.
c
c   Outputs          
c          C     : Array of coefficients which have been computed from 
c                  tiepoints.
c
c Platform-specific tiepoints in
c     Cavalieri, D. J., C. L. Parkinson, P. Gloersen, J. C. Comiso and 
c     H. J. Zwally. 1999.  Deriving long-term time series of sea ice
c     cover from satellite passive-microwave multisensor data sets.  
c     Journal of Geophysical Research, 104(C7), pp. 15,803-15,814.
c
c     F15_BRIDGE tiepoints were calculated by Walt Meier at NSIDC, May
c     2008.  Late this winter, F13 started failing, and we decided
c     to start using F15 data.  Walt determined that F13 tiepoints
c     were not acceptable for F15 data, so he calculated new ones.
c     I expect these to eventually be replaced by a slightly different
c     set from Cavalieri et al, so I'm designating these F15_BRIDGE
c     in order to distinguish them in the future.  See also
c     regression of F15 22V data.
c
c     F17 tiepoints wer calculated by Walt Meier at NSIDC, April, 2009.
c     Walt will likely publish this, so I will have a citation eventually.
c
c     F17_FINAL tiepoints from (Table 1):
c     Cavalieri, D. J., C. L. Parkinson, N. DiGirolamo, and A. Ivanoff.
c     2012. Intersensor Calibration Between F13 SSMI and F17 SSMIS for
c     Global Sea Ice Data Records. IEEE Geoscience and Remote Sensing
c     Letters, Vol. 9, No. 2, March 2012.
c
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc       

      PARAMETER (NHEMS = 2, NPLATFORMS = 7)
      real C(12)
      integer*2 ihem, iplatform       

c
c     Open water tiepoints 
c      
      real TB19VW(NHEMS, NPLATFORMS), TB19HW(NHEMS, NPLATFORMS),
     1     TB37VW(NHEMS, NPLATFORMS)
c
c     First-year tiepoints 
c
      real TB19VF(NHEMS, NPLATFORMS), TB19HF(NHEMS, NPLATFORMS),
     1     TB37VF(NHEMS, NPLATFORMS)
c
c     Multi-year tiepoints
c
      real TB19VM(NHEMS, NPLATFORMS), TB19HM(NHEMS, NPLATFORMS),
     1     TB37VM(NHEMS, NPLATFORMS)
c
      real DW(2) , DF(2) , DM(2)
      real SW(2) , SF(2) , SM(2)
c
c     Initialize tiepoints, (Cavalieri, et al.)
c     These arrays are populated like this:
c     data ARRAY  / N-F08  S-F08  N-F11  S-F11  N-F13  S-F13
c                   N-F15_BRIDGE, S-F15_BRIDGE,
c                   N-F15, S-F15,
c                   N-F17, S-F17,
c                   N-F17_FINAL, S-F17_FINAL/
      data TB19VW / 183.4, 185.3, 185.1, 186.2, 185.2, 186.0,
     1     184.849, 185.465, 0.0, 0.0, 182.2, 187.7, 184.9, 184.9 /
      data TB19VF / 251.5, 256.6, 251.4, 255.5, 251.2, 256.0,
     1     251.047, 255.325, 0.0, 0.0, 251.7, 256.2, 248.4, 253.1 /
      data TB19VM / 222.1, 246.9, 222.5, 246.2, 222.4, 246.6,
     1     222.160, 245.944, 0.0, 0.0, 223.4, 246.9, 220.7, 244.0 /

      data TB19HW / 113.2, 117.0, 113.6, 115.7, 114.4, 117.0,
     1     114.813, 117.464, 0.0, 0.0, 116.5, 118.4, 113.4, 113.4 /
      data TB19HF / 235.5, 242.6, 235.3, 241.2, 235.4, 241.4,
     1     235.571, 241.242, 0.0, 0.0, 235.4, 241.1, 232.0, 237.8 /
      data TB19HM / 198.5, 215.7, 198.3, 214.6, 198.6, 214.9,
     1     198.845, 214.874, 0.0, 0.0, 199.0, 214.8, 196.0, 211.9 /

      data TB37VW / 204.0, 207.1, 204.8, 207.1, 205.2, 206.9,
     1     205.029, 206.577, 0.0, 0.0, 206.5, 208.9, 207.1, 207.1 /
      data TB37VF / 242.0, 248.1, 242.0, 245.6, 241.1, 245.6,
     1     240.749, 244.426, 0.0, 0.0, 242.7, 246.4, 242.3, 246.6 /
      data TB37VM / 184.2, 212.4, 185.1, 211.3, 186.2, 211.1,
     1     186.124, 210.685, 0.0, 0.0, 188.1, 212.6, 188.5, 212.6 /
       
      DW(1) = TB19VW(ihem, iplatform) - TB19HW(ihem, iplatform)                                
      DF(1) = TB19VF(ihem, iplatform) - TB19HF(ihem, iplatform)                                 
      DM(1) = TB19VM(ihem, iplatform) - TB19HM(ihem, iplatform)                                   
      SW(1) = TB19VW(ihem, iplatform) + TB19HW(ihem, iplatform)                                   
      SF(1) = TB19VF(ihem, iplatform) + TB19HF(ihem, iplatform)                                   
      SM(1) = TB19VM(ihem, iplatform) + TB19HM(ihem, iplatform)                                   
      DW(2) = TB37VW(ihem, iplatform) - TB19VW(ihem, iplatform)                                   
      DF(2) = TB37VF(ihem, iplatform) - TB19VF(ihem, iplatform)                                   
      DM(2) = TB37VM(ihem, iplatform) - TB19VM(ihem, iplatform)                                   
      SW(2) = TB37VW(ihem, iplatform) + TB19VW(ihem, iplatform)                                   
      SF(2) = TB37VF(ihem, iplatform) + TB19VF(ihem, iplatform)                                   
      SM(2) = TB37VM(ihem, iplatform) + TB19VM(ihem, iplatform)                                   
c+
c     Derive coefficients from tiepoints   
c-
      C(1) = DM(1)*DW(2) - DM(2)*DW(1)                                        
      C(2) = DM(2)*SW(1) - DW(2)*SM(1)                                        
      C(3) = DW(1)*SM(2) - DM(1)*SW(2)                                        
      C(4) = SM(1)*SW(2) - SM(2)*SW(1)           
      C(5) = DF(1)*(DM(2) - DW(2)) + DW(1)*(DF(2) - DM(2))            
     1+    DM(1)*(DW(2) - DF(2))                                        
      C(6) = DF(2)*(SM(1) - SW(1)) + DW(2)*(SF(1) - SM(1))           
     1+    DM(2)*(SW(1) - SF(1))                                        
      C(7) = DF(1)*(SW(2) - SM(2)) + DW(1)*(SM(2) - SF(2))           
     1+    DM(1)*(SF(2) - SW(2))                 
      C(8) = SF(2)*(SW(1) - SM(1)) + SW(2)*(SM(1) - SF(1))               
     1+    SM(2)*(SF(1) - SW(1))                                  
      C(9) = DF(2)*DW(1) - DF(1)*DW(2)                                        
      C(10) = DW(2)*SF(1) - DF(2)*SW(1)                                        
      C(11) = SW(2)*DF(1) - DW(1)*SF(2)                                        
      C(12) = SF(2)*SW(1) - SF(1)*SW(2)     
c
      return
      end

      integer*2 function force_rov (val, minval, maxval)
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c
c     FORCE_ROV - forces the range of values and converts to integer*2
c                 for output
c   Inputs           
c          value : scaled ice concentration
c          min_value : the minimum allowed value
c          max_value : the maximum allowed value
c      
c   Returns integer*2 value, forced to the range [min_value, max_value]
c   suitable for program output
c
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc       

      real val, minval, maxval
      real tmp

      tmp = val

      if ( tmp .lt. minval ) then
        tmp = minval
      end if

      if ( tmp .gt. maxval ) then
        tmp = maxval
      end if

      force_rov = nint( tmp )

      return
      
      end

      
      real function f15_bridge_22v_regression ( tb22v, ihem )
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c
c     f15_bridge_22v_regression : does a regression on 22v TBs to adjust
c                 F15 to F13.  This is for F15_bridge processing only.
c                 Input/output TB is assumed to be scaled.
c
c   Inputs           
c          tb22v : scaled TB
c          ihem  : hemisphere index (1=North, 2=South)
c      
c   Returns adjusted 22V scaled brightness temperature.
c      
c
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc       

      real         TB_SCALE_FACTOR
      integer*2    NHEMS
      PARAMETER    (TB_SCALE_FACTOR = 0.1)
      PARAMETER    (NHEMS = 2)
      real tb22v
      integer*2 ihem
      real tmp

c
c     F15->F13 22V regression coefficients for F15_BRIDGE
c     version of algorithm, derived by W. Meier, NSIDC, May 2008
c
      real f15_bridge_22v_intercept(NHEMS)
      real f15_bridge_22v_slope(NHEMS)

      data f15_bridge_22v_intercept / -13.2, -7.915 /
      data f15_bridge_22v_slope / 1.007, 0.993 /
      
      tmp = tb22v * TB_SCALE_FACTOR
      tmp = f15_bridge_22v_intercept( ihem ) +
     1     ( f15_bridge_22v_slope( ihem ) * tmp )
      tmp = tmp / TB_SCALE_FACTOR
      
      f15_bridge_22v_regression = tmp

      return
      
      end

      
