      subroutine seaice_nasateam(tb19h,tb19v,tb22v,tb37v,ihem,tot,myr)
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c
c                       SEAICE_NASATEAM.F  
c
c PURPOSE:  This program runs routines to calculate multi-year and
c           total sea ice concentrations for a single pixel using the
c           NASA TEAM algorithm.
c
c COMMENTS: (This code was adapted from the code RFK created to run the
c           seaice concentrations for N. Hemisphere EASE-Grid which
c           was code adapted from Nick's code that RFK got from 
c           him around April 5, 1994 (and includes weather filter)
c
c AUTHOR:   Raymond Kokaly   15 Nov 1994
c
c $Revision: 1.6.8.1 $
c
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc        
c          
c   ALGORITHM DESCRIPTION:
c     The ice concentration algorithm is an adaptation of the     
c     algorithm designed by the NODS Software Group (refer to the     
c     NODS User Handbook, Appendix CC, pp 22-26). The algorithm     
c     uses the latest SSM/I tiepoints which were determined by Don     
c     Cavalieri and Koni Steffen.     
c     A weather filter (GR1 > .05 and GR2 > .045) is applied to the 
c     25km TB cell to correct for false concentrations computed in regions 
c     of open water.     
c
c   INPUTS:
c     tb19h  -  Brightness temperature for 19GHz H (integer*2)
c     tb19v  -  Brightness temperature for 19GHz V (integer*2)
c     tb22v  -  Brightness temperature for 22GHz V (integer*2)
c     tb37v  -  Brightness temperature for 37GHz V (integer*2)
c     ihem   -  Defines the appropriate hemisphere to apply this
c                algorithm (1=North, 2=South).  Used for tie points.
c
c   INTERNALS
c      c     -  Array of coefficients used to compute ice concentrations.
c                 Refer to NODS Handbook Volue II. Appendix CC pp 23-24.
c      fyr   -  First-year ice concentration (integer)
c      gr1   -  Gradient ratio (37v,19v)
c      gr2   -  Gradient ratio (22v,19v)
c      PR    -  Polarization ratio
c          
c   OUTPUTS          
c      tot   -  Total ice concentration (integer*1)
c      myr   -  Multi-year ice concentration (integer*1)
c      
c              Percentages range from 0 - 100%.
c              Cells for which concentrations were suspect or           
c              not computed are flagged with -99(157).
c
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc       

      implicit     none 

      integer*2    ihem
      integer*2    tb19h,tb19v,tb22v,tb37v
      integer*1    fyr,myr,tot
      real         BT19V,BT19H,BT37V,BT22V
      real         c(12),ANF,ANM,DD                                        
      double precision  gr1,gr2,PR,PRGR
      integer*4    verbose

      verbose = 0
      if (1 .lt. verbose) then
        WRITE(0,*)
     1       '>> $Header: /tmp_mnt/FILES/pmalgos/seaice_nasateam.f,v 1.6.8.1 2008-06-10 20:36:35 brodzik Exp $'
      endif
        
c+++
c     GET ICE ALGORITHM COEFFICIENTS DERIVED FROM TIEPOINTS.
c---
      call usetie(c,ihem)
c+++
c     CONVERT INTS TO REALS
c---
      BT19H = float(tb19h)
      BT19V = float(tb19v)
      BT22V = float(tb22v)
      BT37V = float(tb37v)
c+++
c     COMPUTE OPEN WATER WEATHER FILTER.
c---
      gr1 = (BT37V - BT19V)/(BT37V + BT19V)            
      gr2 = (BT22V - BT19V)/(BT22V + BT19V)
c+++
c     WEATHER FILTER: SET ICE CELL TO 0% CONCENTRATION IF 
c     gr1  > 0.05       indicates weather over open water
c     gr2  > 0.045      indicates weather over open water
c---
      if (gr1 .gt. .05 .or. gr2 .gt. .045) then
        tot      = 0
        fyr      = 0
        myr      = 0
      else                                    
c+++
c       CHECK IF 19V < 19H, IF SO THEN A PROBLEM WITH TBs
c---
        if (BT19V.LT.BT19H)  then           
          fyr = -99                             
          myr = -99                             
          tot = -99                             
        else                             
c+++
c       IF PREVIOUS TWO CONDITIONS NOT MET CONTINUE WITH CALC
c       COMPUTE POLARIZATION RATIO PR
c---
          PR   = (BT19V - BT19H)/(BT19V + BT19H)               
          PRGR = PR*GR1                             
          ANF  = c(1) + c(2)*PR + c(3)*GR1 + c(4)*PRGR           
          ANM  = c(9) + c(10)*PR + c(11)*GR1 + c(12)*PRGR
          DD   = c(5) + c(6)*PR + c(7)*GR1 + c(8)*PRGR  
c+++
c         COMPUTE FIRST-YEAR, MULTI-YEAR, AND TOTAL ICE CONCENTRATIONS.
c---
          fyr = nint(100.0*(ANF/DD))
          myr = nint(100.0*(ANM/DD))
          tot = fyr + myr    
c+++
c	  CHECK FOR VALUES OVER 100, AND LESS THAN 0.
c---
          if (tot.lt.  0) tot=  0
          if (myr.lt.  0) myr=  0
          if (tot.gt.100) tot=100
          if (myr.gt.100) myr=100
          if (myr.gt.tot) myr=tot

        endif                             
      endif                             
      return                               
      end

      subroutine usetie (C, ihem )             
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc
c
c USETIE - This subroutine computes coefficients for the ice algorithm
c          using the NASA-approved tiepoints.
c          
c   Inputs           
c          c     : Real array of coefficients. Refer to NODS User 
c                  handbook for description of coefficients A-L.
c          ihem  : Integer denoting hemisphere (1=north, 2=south). Used
c                  in determining which set of coefficients to use.
c   Internals
c          TB...W: Open water tiepoints. One for each of the three channels
c                  used in the ice concentration algorithm (19V, 19H, 37V)
c          TB...F: Firstyear tiepoints. One for each of the three channels
c                  used in the ice concentration algorithm (19V, 19H, 37V)
c          TB...M: Multiyear tiepoints. One for each of the three channels
c                  used in the ice concentration algorithm (19V, 19H, 37V)
c          DW    : Difference between component tiepoints for open water.
c          SW    : Sum of the component tiepoints for open water.
c          DF    : Difference between component tiepoints for first year ice.
c          SF    : Sum of the component tiepoints for first year ice.
c          DM    : Difference between component tiepoints for multiyear ice.
c          SM    : Sum of the component tiepoints for multiyear ice.
c
c   Outputs          
c          C     : Array of coefficients which have been computed from 
c                  tiepoints.
ccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccccc       
      real C(12)
      integer*2 ihem       

c Tiepoints for open water, first-year ice and multi-year ice           
c The first element of each tiepoint array represents the tiepoints
c for the North Polar region; the second element is applied to the
c South Polar region.
c                                
c     declare the two-element open water tiepoints 
c      
      real TB19VW(2), TB19HW(2), TB37VW(2)
c
c     declare the two-element first-year tiepoints 
c
      real TB19VF(2), TB19HF(2), TB37VF(2)
c
c     declare the two-element multi-year tiepoints
c
      real TB19VM(2), TB19HM(2), TB37VM(2)
c
      real DW(2) , DF(2) , DM(2)
      real SW(2) , SF(2) , SM(2)
c
c First component of each tiepoint pair is for Northern Hemisphere
c Second component of each tiepoint pair is of Southern Hemisphere
c These tiepoints approved for use by the NASA Sea Ice Algorithm
c Working Group (1/90)
c

      data TB19VW / 177.1, 176.6/
      data TB19VF / 258.2, 249.8/        
      data TB19VM / 223.2, 221.6/                                        
                                        
      data TB19HW / 100.8, 100.3/                                        
      data TB19HF / 242.8, 237.8/                                        
      data TB19HM / 203.9, 193.7/                                        

      data TB37VW / 201.7, 200.5/                                        
      data TB37VF / 252.8, 243.3/                                    
      data TB37VM / 186.3, 190.3/              
c       
      DW(1) = TB19VW(ihem) - TB19HW(ihem)                                
      DF(1) = TB19VF(ihem) - TB19HF(ihem)                                 
      DM(1) = TB19VM(ihem) - TB19HM(ihem)                                   
      SW(1) = TB19VW(ihem) + TB19HW(ihem)                                   
      SF(1) = TB19VF(ihem) + TB19HF(ihem)                                   
      SM(1) = TB19VM(ihem) + TB19HM(ihem)                                   
      DW(2) = TB37VW(ihem) - TB19VW(ihem)                                   
      DF(2) = TB37VF(ihem) - TB19VF(ihem)                                   
      DM(2) = TB37VM(ihem) - TB19VM(ihem)                                   
      SW(2) = TB37VW(ihem) + TB19VW(ihem)                                   
      SF(2) = TB37VF(ihem) + TB19VF(ihem)                                   
      SM(2) = TB37VM(ihem) + TB19VM(ihem)                                   
c+
c     Derive coefficients from tiepoints   
c-
      C(1) = DM(1)*DW(2) - DM(2)*DW(1)                                        
      C(2) = DM(2)*SW(1) - DW(2)*SM(1)                                        
      C(3) = DW(1)*SM(2) - DM(1)*SW(2)                                        
      C(4) = SM(1)*SW(2) - SM(2)*SW(1)           
      C(5) = DF(1)*(DM(2) - DW(2)) + DW(1)*(DF(2) - DM(2))            
     1+    DM(1)*(DW(2) - DF(2))                                        
      C(6) = DF(2)*(SM(1) - SW(1)) + DW(2)*(SF(1) - SM(1))           
     1+    DM(2)*(SW(1) - SF(1))                                        
      C(7) = DF(1)*(SW(2) - SM(2)) + DW(1)*(SM(2) - SF(2))           
     1+    DM(1)*(SF(2) - SW(2))                 
      C(8) = SF(2)*(SW(1) - SM(1)) + SW(2)*(SM(1) - SF(1))               
     1+    SM(2)*(SF(1) - SW(1))                                  
      C(9) = DF(2)*DW(1) - DF(1)*DW(2)                                        
      C(10) = DW(2)*SF(1) - DF(2)*SW(1)                                        
      C(11) = SW(2)*DF(1) - DW(1)*SF(2)                                        
      C(12) = SF(2)*SW(1) - SF(1)*SW(2)     
c
      return
      end
